﻿' Module: Crypt_rc4.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings. 

Module Crypt_rc4
  Public Const cCrypt_rc4_version As Date = #4/15/2021# ' M/d/y.

  Public Function Crypt_rc4_fn(ByVal tPlaintext As String, ByVal tPrivate_key As String, ByVal tEncrypt As Boolean) As String
    ' Revised 2021.04.15.
    Dim tLocals As udtLocals = Nothing
    Dim tX As Integer
    tLocals.Subr = "Crypt rc4"
    tLocals.Status = ""
    Crypt_rc4_fn = ""

    If tPrivate_key = "" Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    If tEncrypt Then
      tPlaintext = RC4_processor_fn(tPlaintext, tPrivate_key)
      For tX = 1 To Len(tPlaintext)
        Crypt_rc4_fn = Crypt_rc4_fn & Right("0" & Hex(Asc(Mid(tPlaintext, tX))), 2)
      Next tX
    Else
      For tX = 1 To Len(tPlaintext) Step 2
        Mid(tPlaintext, (tX + 1) / 2, 1) = Chr(Val("&H" & Mid(tPlaintext, tX, 2)))
      Next tX
      Crypt_rc4_fn = RC4_processor_fn(Left(tPlaintext, Len(tPlaintext) / 2), tPrivate_key)
    End If

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

    If tLocals.Status <> udtuStandard_texts.Failed Then uOp.Success = True Else uOp.Success = False : Stop
  End Function

  Public Function RC4_processor_fn(tPlaintext As String, tPrivate_key As String) As String
    ' RC4 encrypts tText using tKey as the key.
    ' Revised 2014.05.25.
    Dim tKey_len As Integer
    Dim tTemp As Integer
    Dim tX1 As Integer
    Dim tX2 As Integer
    Dim tX3 As Integer
    Dim aintB(256) As Integer
    Dim aintK(256) As Integer
    RC4_processor_fn = ""

    ' Initialise buffer and keystream:
    tKey_len = Len(tPrivate_key)
    For tX1 = 0 To 255
      aintB(tX1) = tX1
      aintK(tX1) = Asc(Mid(tPrivate_key, (tX1 Mod tKey_len) + 1))
    Next

    ' Scramble up the data in the buffer a bit:
    tX2 = 0
    For tX1 = 0 To 255
      tX2 = (tX2 + aintB(tX1) + aintK(tX1)) Mod 255
      tTemp = aintB(tX1)
      aintB(tX1) = aintB(tX2)
      aintB(tX2) = tTemp
    Next

    ' Encode, decode (but process n bytes through the stream before starting):
    tX2 = 0
    tX3 = 0
    For tX1 = 1 To 3072 + Len(tPlaintext)
      tX2 = (tX2 + 1) Mod 255
      tX3 = (tX3 + aintB(tX2)) Mod 255
      tTemp = aintB(tX2)
      aintB(tX2) = aintB(tX3)
      aintB(tX3) = tTemp
      If tX1 > 3072 Then
        RC4_processor_fn = RC4_processor_fn & Chr(Asc(Mid(tPlaintext, tX1 - 3072)) Xor aintB((aintB(tX2) + aintB(tX3)) Mod 255))
      End If
    Next
  End Function

End Module
