' Module: Database.
' Copyright  LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings.

Module Database_
  Public Const cDatabase_version As Date = #4/7/2021# ' M/d/y.

#Region "Create, instantiate, set-up"

  Sub Database_instantiate_local()
    ' Separated 2017.05.06. Revised 2021.04.04.
    ' Use for uMemory_uncommon, or a database that does not use udtDatabase*.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_instantiate_local
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1

    tLocals.Action = udtuStandard_texts.Main_start
    ' Example:
    ' If uMemory_ProgFiles.Database <> "" Then
    ' Call Database_instantiate(tDb)
    ' Or instatiate manually:
    ' uMemory_ProgFiles.Cmd = New OleDb.OleDbCommand
    ' End If

    If gMemory.Database <> "" Then
      Call Database_instantiate(gMemory)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Continue_ = False
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

    tLocals.Action = "Instatiate uncommon database(s)"

    tLocals.Action = "Check the connection of data sources"
    ' See Financier for example.

    tLocals.Action = "Set database connection dependants"
    ' See Financier for example.

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If uOp.Event_ <> "" And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    Call frmMain.ToolStripProgressBar1_set()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Memory_define_Preference()
    ' Revised 2021.04.04.
    With uDef_Preference
      .Table = udtTables.Preference
      uOp.Subr = "Memory define " & .Table
      .Test_field = "App_name"
      .Length = 36
      ReDim .Name_(.Length)
      ReDim .Type_(.Length)
      ReDim .Value(.Length)
      ReDim .Condition_operator(.Length)
      ReDim .Condition_value(.Length)

      .Name_(1) = "Id"
      .Type_(1) = "Autoincrement"
      .Name_(2) = "App_name"
      .Type_(2) = "Text"
      .Name_(3) = "App_build"
      .Type_(3) = "Single"
      .Name_(4) = "App_revision"
      .Type_(4) = "Date"
      .Name_(5) = "Branch"
      .Type_(5) = "Text"
      .Name_(6) = "Country"
      .Type_(6) = "Text"
      .Name_(7) = "Corporation"
      .Type_(7) = "Text"
      .Name_(8) = "Currency_"
      .Type_(8) = "Text"
      .Name_(9) = "Currency_format"
      .Type_(9) = "Text"
      .Name_(10) = "Currency_symbol"
      .Type_(10) = "Text"
      .Name_(11) = "Environment_dump_emailed"
      .Type_(11) = "Bit"
      .Name_(12) = "Environment_dump_completed_date"
      .Type_(12) = "Date"
      .Name_(13) = "First_run_completed_date"
      .Type_(13) = "Date"
      .Name_(14) = "First_run_wizard"
      .Type_(14) = "Bit"
      .Name_(15) = "First_run_wizard_step"
      .Type_(15) = "Single"
      .Name_(16) = "First_run_wizard_completed"
      .Type_(16) = "Bit"
      .Name_(17) = "Main_height"
      .Type_(17) = "Single"
      .Name_(18) = "Main_left"
      .Type_(18) = "Single"
      .Name_(19) = "Main_top"
      .Type_(19) = "Single"
      .Name_(20) = "Main_width"
      .Type_(20) = "Single"
      .Name_(21) = "Memory_check_completed_date"
      .Type_(21) = "Date"
      .Name_(22) = "Min_sys_tray"
      .Type_(22) = "Bit"
      .Name_(23) = "Reg_email"
      .Type_(23) = "Text"
      .Name_(24) = "Reg_30_day_trial"
      .Type_(24) = "Bit"
      .Name_(25) = "Reg_30_day_trial_from_date"
      .Type_(25) = "Date"
      .Name_(26) = "Reg_30_day_trial_20_days_notified"
      .Type_(26) = "Bit"
      .Name_(27) = "Reg_code"
      .Type_(27) = "Text"
      .Name_(28) = "Reg_code_sent"
      .Type_(28) = "Bit"
      .Name_(29) = "Registered"
      .Type_(29) = "Bit"
      .Name_(30) = "Speech"
      .Type_(30) = "Single"
      .Name_(31) = "Start_home_tab"
      .Type_(31) = "Single"
      .Name_(32) = "Start_maximised"
      .Type_(32) = "Bit"
      .Name_(33) = "Tab_last"
      .Type_(33) = "Text"
      .Name_(34) = "Theme"
      .Type_(34) = "Single"
      .Name_(35) = "User_first_name"
      .Type_(35) = "Text"
      .Name_(36) = "Updated"
      .Type_(36) = "Date"

Uncommon:
    End With

    uOp.Subr = ""
  End Sub ' Sub Memory_define_preference

  Sub Memory_define_International()
    ' Revised 2020.12.30. Demoted 2021.04.04.
    With uDef_International
      .Table = udtTables_local.International
      uOp.Subr = "Memory define " & .Table
      .Test_field = "Country_name"
      .Length = 9
      ReDim .Name_(.Length)
      ReDim .Type_(.Length)
      ReDim .Value(.Length)
      ReDim .Condition_operator(.Length)
      ReDim .Condition_value(.Length)

      .Name_(1) = "Id"
      .Type_(1) = "Autoincrement"
      .Name_(2) = "Two_letter"
      .Type_(2) = "Text"
      .Name_(3) = "Country_name"
      .Type_(3) = "Text"
      .Name_(4) = "Currency"
      .Type_(4) = "Text"
      .Name_(5) = "Currency_format"
      .Type_(5) = "Text"
      .Name_(6) = "Currency_symbol"
      .Type_(6) = "Text"
      .Name_(7) = "Currency_ISO_code"
      .Type_(7) = "Text"
      .Name_(8) = "Telephone_code"
      .Type_(8) = "Text"
      .Name_(9) = "Note"
      .Type_(9) = "Text"
    End With

    uOp.Subr = ""
  End Sub

  Sub Memory_define_Preference_common()
    ' Written 2018.10.28. Revised 2021.01.24.
    With uDef_Preference_common
      .Table = udtTables_local.Preference_common
      uOp.Subr = "Memory define " & .Table
      .Test_field = "Test_field"
      .Length = 4
      ReDim .Name_(.Length)
      ReDim .Type_(.Length)
      ReDim .Value(.Length)
      ReDim .Condition_operator(.Length)
      ReDim .Condition_value(.Length)

      .Name_(1) = "Id"
      .Type_(1) = "Autoincrement"
      .Name_(2) = "Kernel_distribute_date"
      .Type_(2) = "Date"
      .Name_(3) = "Kernel_distribute_time"
      .Type_(3) = "Time"
      .Name_(4) = "Test_field"
      .Type_(4) = "Text"
    End With

    uOp.Subr = ""
  End Sub

#End Region ' End region, Create, instantiate, set-up.

#Region "Database diagnostic local"

  Sub Database_diagnostic_local(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.01.
    Dim tLocals As udtLocals = Nothing
    Dim tChanged As Boolean
    If tDb.Upgrade_table = "" Then Exit Sub
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_local
    tLocals.Continue_ = True
    tLocals.Error_count = 0
    tLocals.Status = ""
    tChanged = False

    tLocals.Action = "Test if actions needed"
    If uMachine.Name = uMachine.Development_machine_name Then
      If tDb.Database = uMemory_ProgFiles.Database Then

        If tDb.Upgrade_table = udtTables_local.International Or tDb.Upgrade_table = "All" Then
          tDb.Table_name = udtTables_local.International
          tLocals.Action = "Checking table " & tDb.Table_name
          Call frmMain.txtDiagnostic_result_append(tLocals.Action & "...")
          If uDef_International.Length = 0 Then Call Memory_define_International()
          Call Database_diagnostic_level_05_part_01(tDb, uDef_International)
          uOp.Subr = tLocals.Subr
          If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            If Not tLocals.Continue_ Then GoTo Error_process
          End If
          uOp.Success = False
        End If

        If tDb.Upgrade_table = udtTables_local.Preference_common Or tDb.Upgrade_table = "All" Then
          tDb.Table_name = udtTables_local.Preference_common
          tLocals.Action = "Checking table " & tDb.Table_name
          Call frmMain.txtDiagnostic_result_append(tLocals.Action & "...")
          If uDef_Preference_common.Length = 0 Then Call Memory_define_Preference_common()
          Call Database_diagnostic_level_05_part_01(tDb, uDef_Preference_common)
          If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            If Not tLocals.Continue_ Then GoTo Error_process
          End If
          uOp.Success = False
        End If

      Else
        tLocals.Success = udtBoolean_extended.True_
      End If ' If tDb.Database = uMemory_ProgFiles.Database Then
    End If ' If uMachine.Name = uMachine.Development_machine_name Then

    ' Check any other databases:

    ' Reference 1.
Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check cancelled." & vbCrLf)
    Else
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If
    ' Reference 2.

    ' Reference 3.
    If tLocals.Status <> udtuStandard_texts.Failed Then uOp.Success = True Else uOp.Success = False : Stop
    ' Reference 4.
  End Sub

  Sub Test_sql_date_format()
    ' Written MJBH 2019.10.19. Demoted 2021.04.04.
    ' Stop ' To do, to develop.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = udtgApplication_subroutine_names.Test_sql_date_format

    uOp.Success = False
    uMemory.Table_name = uTables.Log
    uMemory.Cmd_str = "SELECT [Date] FROM " & uMemory.Table_name & " WHERE [Date]=#" & Date.Today & "#" ' Works for M/d/yyyy.
    ' Call Database_general(uMemory)

    uOp.Success = False
    uMemory.Cmd_str = "SELECT [Date] FROM " & uMemory.Table_name & " WHERE [Date]=Format(#" & Date.Today & "#, " & Chr(34) & uMachine.Date_format & Chr(34) & ")" ' Works for M/d/yyyy.
    Call Database_general(uMemory)

    If Not uOp.Success Then
      uOp.Success = False
      uMemory.Cmd_str = "SELECT [Date] FROM " & uMemory.Table_name & " WHERE [Date]=Format(#" & Date.Today & "#, " & Chr(34) & "yyyy.mm.dd" & Chr(34) & ")" ' Fail.
      ' Call Database_general(uMemory)

      If Not uOp.Success Then
        uOp.Success = False
        uMemory.Cmd_str = "SELECT [Date] FROM " & uMemory.Table_name & " WHERE [Date]=Format(#" & DateAndTime.DateValue(Date.Today) & "#, " & Chr(34) & uMachine.Date_format & Chr(34) & ")" ' Works for M/d/yyyy.
        ' Call Database_general(uMemory)
      End If
    End If

    If uOp.Success Then
      Call Date_detect_system_format(True)
      If uMachine.Date_format <> "" Then Call Test_sql_date_format()
    Else
      Stop ' To know which one works.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    End If

    uOp.Subr = ""
  End Sub

  ' Reference 5.
#End Region ' End region, Database diagnostic local.

#Region "Preferences"
  ' Reference 6.

  Sub Preferences_load_local(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.07.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Preferences_load_local
    If uMachine.Development_machine Then uReg.Registered = True

    tLocals.Action = "Shared partial begin"

    'tSearch.Keyword = "Corporation"
    'If IsDBNull(uMemory.Dataset.Tables( uMemory.Table_name ).Rows(0).item(tSearch.Keyword)) Then
    '	uUser.Preferences_changed = True
    'Else
    ' If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uUser.Corporation = uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)
    'End If

    'tSearch.Keyword = "Branch"
    'Select Case tValueType
    '   Case "System.DBNull"
    '      uUser.Preferences_changed = True
    '   Case Else
    ' If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uUser.Branch = uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)
    'End Select

    tLocals.Action = "Shared partial end"

    tSearch.Keyword = uApp.Short_name & "_bank_account_last"
    'If IsDBNull(tDB.Dataset.Tables( tDb.Table_name ).Rows(0).item(tSearch.Keyword)) Then
    '   uUser.Preferences_changed = True
    'Else
    ' If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uUser_uncommon.Bank_account_last = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword).ToString
    'End If

    uOp.Subr = ""
  End Sub

  Sub Preferences_write_local(ByRef tCmd_str As String)
    ' Revised 2020.02.06.
    ' Application specific preferences.
    uOp.Subr = udtApplication_subroutine_names.Preferences_write_local

    ' tCmd_str = tCmd_str & "Installer_path                   ='" & tFile.Installer_path & "', " & vbCr
    ' tCmd_str = tCmd_str & "Data_drive='" & uUser_uncommon.Data_drive & "', " & vbcr 
    ' tCmd_str = tCmd_str & "Post_path='" & uUser_uncommon.Post_path & "', "
    ' tCmd_str = tCmd_str & "Temp_path='" & uUser_uncommon.Temp_path & "', "

    uOp.Subr = ""
    ' Reference 7.
  End Sub

  Sub Preferences_common_write()
    ' Revised 2021.04.12.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtgApplication_subroutine_names.Preferences_common_write
    uOp.Subr = tLocals.Subr

    If uMemory_ProgFiles.Conn Is Nothing Then
      tLocals.Action = "Database instantiate"
      Call Database_instantiate(uMemory_ProgFiles)
      uOp.Subr = tLocals.Subr
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

    tLocals.Action = "Set uMemory_ProgFiles"
    With uMemory_ProgFiles
      .Table_name = udtTables_local.Preference_common
      With uDef_Preference_common
        If .Length = 0 Then Call Memory_define_Preference_common()
        uOp.Subr = tLocals.Subr
        ReDim .Value(.Length)
        .Value(2) = Date.Today
        .Value(3) = Trim(DateAndTime.TimeOfDay)
      End With
      Call Sql_builder_update(uMemory_ProgFiles, uDef_Preference_common)
      uOp.Subr = tLocals.Subr
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      tLocals.Success = udtBoolean_extended.True_

    End With
    Call Database_clear_partial(uMemory_ProgFiles, False)
    uOp.Subr = tLocals.Subr

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  ' Reference 8.
#End Region ' End region, Preferences.

#Region "Database operations general"
  ' Reference 9.

  ' Reference 10.
#End Region ' End region, Database operations general.

#Region "Shared optional"
  ' Reference 11.

  Sub Insert_sample_data(ByRef tChanged As Boolean)
    ' Revised 2021.03.16.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Insert_sample_data
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1
    tLocals.Success = udtBoolean_extended.Null
    tLocals.Status = ""

    tLocals.Action = udtApplication_subroutine_names.Insert_sample_data
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": " & LCase(tLocals.Action) & ". ")

    tLocals.Action = udtuStandard_texts.Sql_build_insert
    ' uMemory.Table_name  = uTables.Iweb
    ' uMemory.Cmd_str = "INSERT INTO " & uMemory.Table_name & " ([Parent], [Name], [Keyword], [Description], [Path], [Date_create], [Counter]) VALUES (" & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'" & gStd.Iweb & "'" & ", " & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'Communication'" & ", " & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'', " & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'Communication is a process of transferring an idea or information from a source point to a destination point.', " & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'', " & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "'" & Date.Today & "'," & vbCr
    ' uMemory.Cmd_str = uMemory.Cmd_str & "0)"
    'Call Database_general(uMemory)
    'uOp.Subr =  tLocals.Subr 
    'If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
    '  ' Stop ' Permanent.
    '  tLocals.Continue_ = False
    '        tLocals.Error_count   =       tLocals.Error_count   + 1
    '  tLocals.Status = udtuStandard_texts.Failed
    '  tLocals.Success = udtBoolean_extended.False_
    '  ' GoTo Error_process
    'Else
    ' tChanged = True
    'End If
    'Call frmMain.Statuslabel1_set
    tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If

    Call Log_write(tLocals, 0)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
    End If
    uOp.Success = False

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  ' Reference 12.
#End Region ' End region, Shared optional.

#Region "Unshared, this application only"
  ' Reference 13.

  Sub Country_detect_from_uMemory_ProgFiles_2(ByRef tCountry As String, ByRef tName_two_letter As String, ByRef tISOCurrencySymbol As String, ByRef tCurrency_format As String, ByRef tCurrencySymbol As String, ByRef tTelephone_code As String, ByRef tNote As String)
    ' Separated 2019.02.03. Depreciated ?. Revised 2021.04.07.
    Stop ' Temporary, to check:
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tSearch As udtSearch = Nothing

    tLocals.Subr = udtApplication_subroutine_names.Country_detect_from_uMemory_ProgFiles
    tLocals.Action = udtuStandard_texts.Initiation
    uOp.Subr = tLocals.Subr
    On Error GoTo Error_captured
    tCurrency_format = ""
    tCurrencySymbol = ""
    tISOCurrencySymbol = ""
    tName_two_letter = ""
    tTelephone_code = ""
    tNote = ""
    tDb.Database = uMemory_ProgFiles.Database
    Call Database_instantiate(tDb)
    uOp.Subr = udtApplication_subroutine_names.Database_instantiate_local
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

    tLocals.Action = "Detect values stored in uMemory_ProgFiles"
    tLocals.Action = udtuStandard_texts.Sql_build_select
    tDb.Table_name = udtTables_local.International
    tDb.Cmd_str = "SELECT * FROM " & tDb.Table_name & " WHERE "
    If Len(tCountry) = 2 Then
      tDb.Cmd_str = tDb.Cmd_str & "Two_letter='" & tCountry & "'"
    Else
      tDb.Cmd_str = tDb.Cmd_str & "Country_name='" & tCountry & "'"
    End If
    Call Database_general(tDb)
    uOp.Subr = tLocals.Subr
    If tDb.Rows_affected = 0 Then
      Call Country_settings_scrape()
      uOp.Subr = tLocals.Subr
    End If

    tLocals.Action = udtuStandard_texts.Dataset_read
    tSearch.Keyword = "Two_letter"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tName_two_letter = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Country_name"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tCountry = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Currency"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tISOCurrencySymbol = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Currency_format"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tCurrency_format = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Currency_symbol"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tCurrencySymbol = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Currency_ISO_code"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tCurrencySymbol = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Telephone_code"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tTelephone_code = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If
    tSearch.Keyword = "Note"
    If Not IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
      tNote = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
    End If

    Call Database_clear_partial(tDb, False)

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0 ' This is needed in case Infinite_loop is bypassed.
          Case Else
            'If tDb.Cmd_str = "" Then
            '   Stop ' Permanent.
            '               Call Error_display(tLocals) 
            'Else
            Call Database_error_handler(tLocals, tDb)
            If tLocals.Action <> "" Or tDb.Upgrade_table <> "" Then
              If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
              Call Database_diagnostic_00(tDb)
              uOp.Called_from = uOp.Subr
              Call Country_detect_from_uMemory_ProgFiles(tCountry, tName_two_letter, tISOCurrencySymbol, tCurrency_format, tCurrencySymbol, tTelephone_code, tNote)
            End If
            ' End If
        End Select
      End If
    End If

    If uOp.Called_from = uOp.Subr Then uOp.Called_from = ""
    uOp.Subr = ""
  End Sub

#Region "Data fixes"
  ' Reference 14.

  ' Reference 15.
#End Region ' End region, "Data fixes."

  ' Reference 16.
#End Region ' End region, Unshared, this application only.

End Module
