﻿' Module: Database_Adox.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson, 2019.12.28.
' Adox version.

' Internal references:
' Microsoft ADO Ext. for DDL and Security.

' External reference:
' www.connectionstrings.com.

Structure udtDatabase_adox
  ' Revised 2021.03.14.
  Public Database As String ' Database name, such as on a server; or if in a file system, including a path.
  Public Table_name As String ' Revised 2020.04.15.

  Public Conn_str As String
  Public Cmd_str As String
  Public Conn As ADODB.Connection
  ' Public Cmd As OleDb.OleDbCommand
  ' Public Data_adapter As OleDb.OleDbDataAdapter
  Public Dataset As DataSet
  Public Catalogue As ADOX.Catalog
  Public Table As ADOX.Table

  Public Attempt As Integer ' 2019.10.16. Counter. If less than 3 or maybe 5, try again. If >, stop with error message.
  Public Changed As Boolean
  Public Indent As String
  Public Rows_affected As Short
  Public Success As Boolean
  Public Upgrade_table As String
  Public Upgrade_counter As Integer
  Public Upgrade_in_progress As Boolean
  Public Use_password As Boolean
End Structure

Module Database_Adox
  Public Const cDatabase_common_Adox_version As Date = #4/4/2021# ' M/d/y.

  Sub List_Access_tables_Adox(strDBPath As String)
    ' Imported from https://docs.microsoft.com/en-us/previous-versions/office/developer/office2000/aa165325(v=office.10)?redirectedfrom=MSDN, 2019.12.28.
    ' Revised 2021.04.04.
    ' Works.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    ' Dim catDb As ADOX.Catalog
    Dim tblList As ADOX.Table
    Dim tDb As udtDatabase_adox
    On Error GoTo Error_captured

    ' Original:
    ' catDb = New ADOX.Catalog
    ' Open the catalogue:
    ' catDB.ActiveConnection = "Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" & strDBPath

    ' Revision:
    With tDb
      .Conn.Open("Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" & strDBPath)
      .Catalogue.ActiveConnection = tDb.Conn

      ' Loop through all the tables, but not queries, and print their names and types:
      For Each tblList In .Catalogue.Tables
        If tblList.Type <> "VIEW" Then
          ' Debug.Print(tblList.Name & vbTab & tblList.Type)
          tLocals.Msg = tblList.Name & vbTab & tblList.Type
        End If
      Next

Ender_full:
      .Catalogue = Nothing
    End With

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

    If uOp.Called_from = uOp.Subr Then uOp.Called_from = ""
    uOp.Subr = ""
    Call frmMain.Statuslabel1_set()
  End Sub

End Module
