﻿' Module: Database_DAO.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson, 2019.07.19. Revised 2020.12.16.
' DAO library.

' Internal references:
' Microsoft DAO 3.6 Object Library.
' Microsoft.Office.Interop.Access?
' Microsoft Office 16.0 Access database engine Object Library?

' External reference:
' www.connectionstrings.com.

' Imports DAO

Structure udtDatabase_DAO
  ' Separated 202.04.15.
  Public Server As String
  Public Port As String
  Public Database As String ' Database name, such as on a server; or if in a file system, including a path.
  Public User_name As String
  Public Password As Boolean
  Public Table_name As String

  Public Conn_str As String
  Public Cmd_str As String
  Public Conn As DAO.Connection
  ' Public Data_adapter As 
  Public Dataset As DataSet

  Public Changed As Boolean
  Public Indent As String
  Public Rows_affected As Short
  Public Success As Boolean
  Public Upgrade_table As String
  Public Upgrade_counter As Integer
  Public Upgrade_in_progress As Boolean
End Structure

Module Database_DAO

  Function DoesTblFieldExist(sTableName As String, sFieldName As String) As Boolean
    ' Author    : CARDA Consultants Inc.
    ' Website   : www.cardaconsultants.com.
    ' Purpose   : Determine if a field exists with the specified table.
    ' Copyright : The following may be altered and reused as you wish so long as the copyright notice is left unchanged (including author, website and
    '             copyright). It may not be sold resold or reposted on other sites (links back to this site are allowed).

    ' Input Variables:
    ' sTableName: Name of the table to check the existance of the field in.
    ' sFieldName: Name of the field to check the existance of.

    ' Usage Example:
    ' DoesTblFieldExist("Table1","Field1")

    ' Revision History:
    ' Rev       Date(yyyy/mm/dd)        Description
    ' **************************************************************************************
    ' 1         2010-Feb-02             Initial release.
    ' 2         2019.07.19              LivingInfo Systems syntax corrections.

    Dim db As DAO.Database
    Dim tdf As DAO.TableDef
    Dim I As String
    Dim bTableExists As Boolean

    On Error GoTo Error_handler_exit
    ' Initialise our variables:
    DoesTblFieldExist = False
    bTableExists = False

    'db = CurrentDb()
    'tdf = db.TableDefs(sTableName)

    bTableExists = True  'If we made it to here without triggering an error
    'the table exists

    I = tdf.Fields(sFieldName).Name

    DoesTblFieldExist = True   'If we made it to here without triggering an
    'error the table field exists

Error_handler_exit:
    On Error Resume Next
    tdf = Nothing
    db = Nothing
    Exit Function

Error_process:
    If Err.Number = 3265 And bTableExists = False Then
      'Table not found in the current database
      MsgBox("The specified Table could not be found in the current database", vbCritical)
    ElseIf Err.Number = 3265 And bTableExists = True Then
      'Field not found in the specified table

    Else
      MsgBox("MS Access has generated the following error" & vbCrLf & vbCrLf & "Error number: " & _
      Err.Number & vbCrLf & "Error source: DoesTblFieldExist" & vbCrLf & "Error description: " & _
      Err.Description, vbCritical, "An error has occured.")
    End If
    Resume Error_handler_exit
  End Function

End Module
