﻿' Module: Database_MsSql.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Separated 2021.01.24.

Structure udtDatabase_MsSql
  ' Separated 2017.08.27. Revised 2020.04.15.
  Public Server As String
  Public Port As String
  Public Database As String ' Database name, such as on a server; or if in a file system, including a path.
  Public Table_name As String

  Public Conn_str As String
  Public Cmd_str As String
  Public Conn As SqlClient.SqlConnection
  Public Cmd As SqlClient.SqlCommand
  Public Data_adapter As SqlClient.SqlDataAdapter
  Public Dataset As DataSet

  Public Changed As Boolean
  Public Indent As String
  Public Rows_affected As Short
  Public Success As Boolean
  Public Upgrade_table As String
  Public Upgrade_counter As Integer
  Public Upgrade_in_progress As Boolean
  Public Use_password As Boolean
End Structure

Module Database_MsSql
  Public Const cDatabase_common_MsSql_version As Date = #4/4/2021# ' M/d/y.

  Sub Database_create_MsSql(ByRef tDb As udtDatabase_MsSql)
    ' Separated 2021.02.17.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    tLocals.Subr = "Database create"
    uOp.Subr = tLocals.Subr

    With tDb
      If .Database <> "" Then
        If System.IO.File.Exists(.Database) Then
          uApp.First_run_in_progress = False
        Else
          tLocals.Action = "Create a memory database"
          If uOp.Called_from <> "First run" Then
            If tDb.Conn_str = uMemory.Conn_str Then uOp.Called_from = "First run"
          End If
          If System.IO.File.Exists(tDb.Database) Then
            uOp.Success = True
            Call Log_write(tLocals, 0)
            uOp.Subr = tLocals.Subr
            uOp.Success = True
          Else
            .Cmd = Nothing
            .Cmd_str = ""
            .Conn = Nothing
            .Conn_str = ""
            .Database = ""
          End If
        End If
      End If
    End With

    uOp.Subr = ""
  End Sub

  Private Sub Database_instantiate_MsSql(ByRef tDb As udtDatabase_MsSql)
    ' Separated 2017.08.27. Revised 2021.01.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Database instantiate MsSql"

    With tDb
      ' .Conn_str = "Provider=SQLOLEDB.1;Integrated Security=SSPI;Persist Security Info=False;"
      ' .Conn_str = .Conn_str & "Initial catalog=" & .Database & "; "
      ' .Conn_str = .Conn_str & "Data Source=" & .SQL_server & "; "
      .Conn_str = "Server=" & .Server & "; "
      .Conn_str = .Conn_str & "Database=" & .Database
      .Conn_str = .Conn_str & "Integrated Security=SSPI;" ' Uses Windows login credentials.
      ' If .Username <> "" Then .Conn_str = .Conn_str & "; User Id=" & .Username & "; "
      ' If .Password <> "" Then .Conn_str = .Conn_str & "Password=" & .Password & "; "
      ' If .Password <> "" Then .Conn_str = .Conn_str & "; Pwd=" & .Password & "; "
      If Not uLIS.Dev Then .Conn_str = .Conn_str & ";Jet OLEDB:Database Password=" & Encrypt_LIS_fn()
      If Dir(.Database, FileAttribute.Normal) = "" Then
        Stop
        tDb.Data_adapter = New SqlClient.SqlDataAdapter(tDb.Cmd_str, tDb.Conn_str)
        uOp.Subr = tLocals.Subr
      End If
      .Conn = New SqlClient.SqlConnection
      .Cmd = New SqlClient.SqlCommand
      .Cmd.Connection = .Conn
      .Cmd.CommandType = CommandType.Text
    End With
  End Sub

  Private Sub Sql_conditioning_MsSql(ByRef tDb As udtDatabase_MsSql)
    ' Separated 2017.08.27. Revised 2018.04.09.
    ' Sql conditioning:  
    ' Modify the sql command string depending on the sqlserver being used. 
    ' References: 
    ' Having www.w3schools.com/sql/sql_having.asp. 
    ' Top    www.w3schools.com/sql/sql_top.asp. 
    If tDb.Cmd_str = "" Then
      Call Sound_alert()
      Stop
    Else
      Dim tStatement As String = UCase(Left(tDb.Cmd_str, InStr(tDb.Cmd_str, " ") - 1))
      ' To restrict the number of rows returned, use: 
      ' SELECT TOP number column_name(s)
      ' FROM table_name 
      ' or
      ' SELECT TOP 50 PERCENT * FROM Persons
      If InStr(1, tDb.Cmd_str, Chr(34)) > 0 Then
        ' MS sql uses the apostrophe, not speech marks.
        Stop
      End If
      tDb.Cmd_str = Replace(tDb.Cmd_str, "Autoincrement", "Int IDENTITY(1,1)")
      tDb.Cmd_str = Replace(tDb.Cmd_str, "Single", "Int")
    End If
  End Sub

End Module
