﻿' Module: Database_SqLite.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Separated 2019.12.24.

' References:
' Devart.Data. C:\LivingInfo Systems\Dept 11A-2, Application development\Template\Template\SqLite\Devart.Data.dll.
' Devart.Data.Sqlite. C:\LivingInfo Systems\Dept 11A-2, Application development\Template\Template\SqLite\Devart.Data.SQLite.dll.

' To do:
' SqLite db encryption.

Imports Devart.Data.SQLite

Structure udtDatabase_SqLite
  ' Separated 2017.07.23. Revised 2020.04.15.
  Public Server As String
  Public Port As String
  Public Database As String ' Database name, such as on a server; or if in a file system, including a path.
  Public Table_name As String

  Public Conn_str As String
  Public Cmd_str As String
  Public Conn As SQLiteConnection
  Public Cmd As SQLiteCommand
  Public Data_adapter As SQLiteDataAdapter
  Public Dataset As DataSet

  Public Changed As Boolean
  Public Indent As String
  Public Rows_affected As Short
  Public Success As Boolean
  Public Upgrade_table As String
  Public Upgrade_counter As Integer
  Public Upgrade_in_progress As Boolean
  Public Use_password As Boolean
End Structure

Module Database_SqLite
  Public Const cDatabase_common_SqLite_version As Date = #5/2/2021# ' M/d/y.

#Region "Create, instantiate, set-up"

  Sub Memory_create_SqLite(ByRef tChanged As Boolean)

  End Sub

  Sub Database_create_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.02.28.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    tLocals.Subr = "Database create"
    uOp.Subr = tLocals.Subr

    With tDb
      If .Database <> "" Then
        If System.IO.File.Exists(.Database) Then
          uApp.First_run_in_progress = False ' ?
        Else
          tLocals.Action = "Create a memory database"
          SQLiteConnection.CreateFile(.Database)
          If uOp.Called_from <> "First run" Then
            If tDb.Conn_str = uMemory.Conn_str Then uOp.Called_from = "First run"
          End If
          If System.IO.File.Exists(tDb.Database) Then
            uOp.Success = True
            Call Log_write(tLocals, 0)
            uOp.Subr = tLocals.Subr
            uOp.Success = True
          Else
            .Cmd = Nothing
            .Cmd_str = ""
            .Conn = Nothing
            .Conn_str = ""
            .Database = ""
          End If
        End If
      End If
    End With

    uOp.Subr = ""
  End Sub

  Sub Database_table_create_SqLite(ByRef tDb As udtDatabase_SqLite, ByVal tTable_def As udtMemory_Fields)
    ' Separated 2018.04.08. Revised 2021.03.04.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim i As Integer
    tLocals.Subr = "Create table " & tDb.Table_name
    tDb.Table_name = tTable_def.Table
    uOp.Subr = tLocals.Subr
    tLocals.Action = udtuStandard_texts.Main_start
    If uOp.Called_from <> "First run" Then uOp.Called_from = uOp.Subr

    ' Watch for Allow Null, etc. 
    tDb.Cmd_str = "CREATE TABLE " & tDb.Table_name & " (" & vbCr
    For i = 1 To tTable_def.Length
      tDb.Cmd_str = tDb.Cmd_str & "[" & tTable_def.Name_(i) & "] " & tTable_def.Type_(i)
      If tTable_def.Name_(i) = "Id" Then
        tDb.Cmd_str = "CREATE TABLE " & tDb.Table_name & " (" & vbCr
        tDb.Cmd_str = tDb.Cmd_str & "[" & tTable_def.Name_(i) & "] "
        tDb.Cmd_str = tDb.Cmd_str & "INTEGER NOT NULL PRIMARY KEY"
      End If
      If i = tTable_def.Length Then
        tDb.Cmd_str = tDb.Cmd_str & ")"
      Else
        tDb.Cmd_str = tDb.Cmd_str & ", " & vbCr
      End If
    Next
    Call Database_general_SqLite(tDb)
    uOp.Subr = tLocals.Subr

    If Not tDb.Success Then
      If InStr(tLocals.Msg, "already exists.") = 0 Then
        Stop
      End If
    End If
    tDb.Upgrade_counter = tDb.Upgrade_counter + 1
    uOp.Success = tDb.Success
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

    uOp.Called_from = ""
    uOp.Subr = ""
  End Sub

  Private Sub Database_instantiate_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.01.28
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    tLocals.Subr = "Database instantiate SqLite"

    With tDb
      ' .Conn_str = "Data Source=" & .Database
      .Conn_str = "DataSource=" & .Database & "; Version=3; New=False; Compress=True;"
      ' If Not uLIS.Dev Then .Conn_str = .Conn_str & ";Jet OLEDB:Database Password=" & Encrypt_LIS_fn()
      If Dir(.Database, FileAttribute.Normal) = "" Then
        Stop
        Call Database_create_SqLite(tDb)
        uOp.Subr = tLocals.Subr
      End If
      .Conn = New SQLiteConnection
      .Cmd = New SQLiteCommand
      .Cmd.Connection = .Conn
      .Cmd.CommandType = CommandType.Text
    End With
  End Sub

  Sub DataAdapter_instantiate(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.02.28.
    tDb.Data_adapter = New SQLiteDataAdapter(tDb.Cmd_str, tDb.Conn_str)

  End Sub

#End Region ' End Create, instantiate, set-up.

#Region "Database diagnostic"

  Sub Database_diagnostic_00_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2018.08.21. Revised 2021.04.04.

    ' A Level 1 diagnostic: A thorough troubleshooting test series, which requires taking the associated system off-line.
    ' Requires human input and can take hours. A level 1 diagnostic requires a team of crew members to physically
    ' verify operation of system mechanisms and to system readings, rather than depending on the automated programmes,
    ' thereby guarding against possible malfunctions in self-testing hardware and software.
    ' A level 2 diagnostic: A thorough troubleshooting test series, which does not require taking the associated system off-line.
    ' Like a level 1, involves extensive automated routines, but requires crew verification of fewer operational elements.
    ' This yields a somewhat less reliable system analysis, but is a procedure that can be conducted in less than half the
    ' time of the more complex tests.
    ' A level 3 diagnostic: Involves crew verification of only key mechanics and systems readings. Level 3 diagnostics are
    ' intended to be performed in ten minutes or less.
    ' A Level 4 diagnostic: Involves more sophisticated batteries of automated diagnostics. For most systems, Level 4
    ' diagnostics can be performed in under 30 seconds.
    ' A Level 5 diagnostic: Aa fully automated programme, fully run by a computer and only takes seconds;

    ' http://scifi.stackexchange.com/questions/31189/meaning-of-levels-of-diagnostic-tests-in-star-trek:
    ' 1    Take system out of service, full shutdown, manual inspection, observed re-initialisation, manual testing.
    ' 2    Take system out of service, limited manual inspection, crew monitored automated tests.
    ' 3    Take system out of service, minimal manual inspection, automated tests.
    ' 4    Take system out of service, automated tests.
    ' 5    Automated tests.

    ' Warning - SPECULATION (although it is supported by good engineering princples):
    ' The higher (or lower depending on how you do it) number means that you get more detailed information.
    ' Some people ask why you would not put it on the highest level at all times, the reason you do not and you
    ' try to pick your level carefully is that you risk too much information being spewed out and causing you to
    ' miss something important or to have to spend a lot of time sifting through information to find what you need.
    ' Also, more output means more time spent producing output - not much of a consideration normally but on a
    ' real time machine (like a nuclear power plant monitor) a couple of microseconds writing output and not
    ' monitoring the systems can be serious. There is also the fact that more diagnostics means bigger log files
    ' which can cause an issue on some systems where users do not delete them and you end up with, literally,
    ' years of log files eating up disc space.

    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tTable As String = ""
    Dim tCmd_str As String = ""
    tLocals.Subr = "Database diagnostic 00 SqLite"
    uOp.Subr = "Database diagnostic 00"
    tLocals.Action = udtuStandard_texts.Main_start

    Call Error_write_mega(tLocals, "Command string:" & vbCr & tDb.Cmd_str)
    uError.Number = 0
    uError.Description = ""
    uError.Source = ""

    tTable = tDb.Table_name
    tCmd_str = tDb.Cmd_str
    tDb.Upgrade_table = tDb.Table_name
    Call Database_diagnostic_level_05_SqLite(tDb)

    If tTable <> "" And tCmd_str <> "" Then
      tDb.Table_name = tTable
      tDb.Cmd_str = tCmd_str
      Call Database_general_SqLite(tDb)
      tDb.Upgrade_table = ""
    End If

    uOp.Subr = ""
  End Sub

  Sub Database_diagnostic_level_05_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.09.12. Revised 2021.05.02.
    ' A Level 5 diagnostic is a fully automated programme, fully run by a computer and only takes seconds.
    ' Top level. Knows each kernel table, and calls the first upgrade step if necessary.
    If tDb.Upgrade_table = "" Then Exit Sub
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tTable As String
    tLocals.Subr = "Database diagnostic level 05"
    uOp.Subr = tLocals.Subr
    On Error GoTo Error_captured
    tLocals.Msg = "Performing database level five diagnostic..."
    Call frmMain.Statuslabel1_set(tLocals.Msg)
    uOp.Subr = tLocals.Subr

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    If Not (System.IO.File.Exists(tDb.Database)) Then
      Call Memory_create_SqLite(tLocals.Changed)
      uOp.Subr = tLocals.Subr
    End If

    If InStr(tDb.Database, udtLivingInfoSystems.LIS_memory) > 0 Or InStr(tDb.Database, uApp.Internal_name & " Memory") > 0 Then
      If tDb.Upgrade_table = uTables.Log Or tDb.Upgrade_table = "All" Then
        tDb.Table_name = uTables.Log
        If uDef_Log.Length = 0 Then Call Memory_define_Log()
        Call Database_SqLite.Database_diagnostic_level_05_part_01_SqLite(tDb, uDef_Log)
        uOp.Subr = tLocals.Subr
        uOp.Success = False
      End If

      If tDb.Upgrade_table = udtTables.Preference Or tDb.Upgrade_table = "All" Then
        tDb.Table_name = udtTables.Preference
        If uDef_Preference.Length = 0 Then Call Memory_define_Preference()
        Call Database_diagnostic_level_05_part_01_SqLite(tDb, uDef_Preference)
        uOp.Subr = tLocals.Subr
        uOp.Success = False
      End If
    End If

    If tDb.Upgrade_table <> "" Then
      Stop
      ' Call Database_upgrade_local(tDb)
      uOp.Subr = tLocals.Subr
      uOp.Success = False
    End If

    tDb.Upgrade_in_progress = False
    tDb.Upgrade_table = ""
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then uOp.Success = True

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 5 ' Description varies. 
            If InStr(uError.Description, "is already opened exclusively by another user") > 0 Then
              Stop ' To do, to re-develop:
              '     uOp.Result  = Msgbox_LIS_top ("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel,  tLocals.Subr )
              ' uDialogue.Dialogue_result = Msgbox_LIS_top ("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel,  tLocals.Subr )
              Call Msgbox_LIS_top("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel, tLocals.Subr)
              'If     uOp.Result  = vbCancel Then
              '  Stop
              '  frmMain.Close()
              '  Call End_game()
              'End If
            ElseIf InStr(uError.Description, "The Microsoft Jet database engine cannot find the input table or query") > 0 Then
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = "Create"
              Resume Next
            ElseIf InStr(uError.Description, "The INSERT INTO statement contains the following unknown field name") > 0 Then
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = udtuStandard_texts.Upgrade
              tDb.Conn.Close()
              Resume Next
            ElseIf InStr(uError.Description, "No value given for one or more required parameters.") > 0 Then
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = "Alter"
              Resume Next
            Else
              Stop ' Permanent.
              Call Error_display(tLocals)
            End If
          Case Else
            Call Database_error_handler_SqLite(tLocals.Action, tDb)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set(uOp.Subr & " completion.")

    uOp.Result = Nothing
    uOp.Subr = ""
  End Sub

  Private Sub Database_diagnostic_level_05_part_01_SqLite(ByRef tDb As udtDatabase_SqLite, ByRef tTable_def As udtMemory_Fields)
    ' Separated 2019.03.17. Revised 2021.05.02.
    ' Knows and co-ordinates each step of the diagnostic process.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tSearch As udtSearch = Nothing

    Dim tField As String = ""
    tLocals.Subr = "Database diagnostic level 05, part 01"
    On Error GoTo Error_captured
    uOp.Subr = tLocals.Subr
    tDb.Upgrade_in_progress = True

    tDb.Table_name = tTable_def.Table
    Call Database_diagnostic_level_05_part_02_SqLite(tDb) ' Check if the table exists.
    uOp.Subr = tLocals.Subr

    If tDb.Success Or uOp.Success Then
      tDb.Success = False
      uOp.Success = False
      ' If the table exists, test each field:
      Call Database_diagnostic_level_05_part_04_SqLite(tDb, tTable_def, tTable_def.Test_field)
      uOp.Subr = tLocals.Subr

      If tDb.Upgrade_counter < 2 Then
        tField = Replace(tField, "The INSERT INTO statement contains the following unknown field name: '", "")
        tField = Replace(tField, "'.  Make sure you have typed the name correctly, and try the operation again.", "")
      Else
        tField = ""
      End If
    ElseIf tLocals.Action = "Create" Then
      Call Database_table_create_SqLite(tDb, tTable_def)
      uOp.Subr = tLocals.Subr
      If tDb.Success Then
        If tTable_def.Table = udtTables.Preference Then
          tLocals.Success = uOp.Success
          Call Preference_insert_first_Sqlite(tDb)
          uOp.Subr = tLocals.Subr
        End If

        ' To do: 1. Action not ended with a full stop. 2. Doesn't return success = true: 
        uOp.Success = True
        tLocals.Action = "Table " & tTable_def.Table & " added to " & tDb.Database
        Call Log_write(tLocals, 0)
        uOp.Subr = tLocals.Subr
      End If
    End If

    If tLocals.Action = "Create db" Then
      Call Database_create_SqLite(tDb)
      uOp.Subr = tLocals.Subr
    End If

    If tLocals.Action = udtuStandard_texts.Upgrade Then
      If tField = "" Then
        Call Database_upgrade_insert_SqLite(tDb, tTable_def)
        tSearch.Keyword = Left$(tSearch.Keyword, Len(tSearch.Keyword) - 2)
        tLocals.Action = "Fields added to " & tTable_def.Table & " table; " & Trim(tSearch.Keyword)
      Else
        Call Database_upgrade_insert_SqLite(tDb, tTable_def, tField)
        If uOp.Success Then
          tLocals.Action = "Fields added to " & tTable_def.Table & " table; " & tField
        Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
        End If
      End If
      uOp.Subr = tLocals.Subr

      uOp.Success = True
      tLocals.Success = uOp.Success
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
    End If

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
    If tDb.Upgrade_table <> "All" Then tDb.Upgrade_table = ""
    tDb.Success = False

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 5 ' Description varies.
            If InStr(uError.Description, "is already opened exclusively by another user") > 0 Then
              Stop ' To do, to re-develop:
              '     uOp.Result  = Msgbox_LIS_top ("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel, uOp.Subr)
              ' uDialogue.Dialogue_result = Msgbox_LIS_top ("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel, uOp.Subr)
              Call Msgbox_LIS_top("Please close the memory database. Clicking Cancel will close " & uApp.App_name & ".", vbOKCancel, udtDialogue_types.Okay_cancel, uOp.Subr)
              If uOp.Result = vbCancel Then
                Stop ' to do, to test:
                frmMain.Close()
                Call End_game()
              End If
            ElseIf InStr(uError.Description, "The Microsoft Jet database engine cannot find the input table or query") > 0 Or _
               InStr(uError.Description, "Unable to open database file") > 0 Or _
               (InStr(uError.Description, "SQLite error") > 0 And _
               InStr(uError.Description, ": syntax error") > 0 Or _
               InStr(uError.Description, "no such table") > 0) Then
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = "Create"
              Resume Next
            ElseIf InStr(uError.Description, "The INSERT INTO statement contains the following unknown field name") > 0 Or _
             InStr(uError.Description, "Invalid column name") > 0 Then
              tField = uError.Description
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = udtuStandard_texts.Upgrade
              tDb.Conn.Close()
              Resume Next
            ElseIf InStr(uError.Description, "No value given for one or more required parameters.") > 0 Then
              tLocals.Error_count = 0
              uError.Number = 0
              uError.Description = ""
              uError.Source = ""
              tLocals.Action = "Alter"
              Resume Next
            Else
              Stop ' Permanent.
              Call Error_display(tLocals)
            End If
          Case Else
            Call Database_error_handler_SqLite(tLocals.Action, tDb)
        End Select
      End If
    End If
    uOp.Result = Nothing
    uOp.Subr = ""
  End Sub

  Private Sub Database_diagnostic_level_05_part_02_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.03.03.
    ' Check if the table exists.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Database diagnostic level 05, part 02"

    With tDb
      .Rows_affected = 0
      tLocals.Action = "Check table " & .Table_name

      .Cmd_str = "SELECT * FROM " & .Table_name & " LIMIT 1"
      If .Table_name = udtTables.Preference Then
        .Cmd_str = .Cmd_str & " WHERE " & uDef_Preference.Test_field & "='" & uApp.Internal_name & "'"
      End If
      tDb.Data_adapter = New SQLiteDataAdapter(tDb.Cmd_str, tDb.Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, tDb.Table_name)
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count

      .Cmd_str = ""
      .Success = True

      tLocals.Action = udtuStandard_texts.Dataset_clear
      .Dataset.Clear()
      .Dataset.Dispose()
      .Dataset = Nothing
      .Data_adapter.Dispose()
      .Data_adapter = Nothing
    End With

    uOp.Success = True
    uOp.Subr = ""
  End Sub

  Private Sub Database_diagnostic_level_05_part_03_SqLite(ByRef tDb As udtDatabase_SqLite, ByVal tField As String)
    ' Separated 2017.08.27. Revised 2021.03.03.
    ' Check if a record exists.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Database diagnostic level 05, part 03"

    With tDb
      .Rows_affected = 0
      tLocals.Action = "Check " & .Table_name
      .Cmd_str = "SELECT TOP 1 " & tField & " FROM " & .Table_name & " WHERE App_name='" & uApp.Internal_name & "'"

      .Data_adapter = New SQLiteDataAdapter(.Cmd_str, .Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, .Table_name)
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count

      tLocals.Action = udtuStandard_texts.Dataset_clear
      .Dataset.Clear()
      .Dataset.Dispose()
      .Dataset = Nothing
      .Data_adapter.Dispose()
      .Data_adapter = Nothing
      .Cmd_str = ""
    End With

    uOp.Subr = ""
  End Sub

  Private Sub Database_diagnostic_level_05_part_04_SqLite(ByRef tDb As udtDatabase_SqLite, ByRef tTable_var As udtMemory_Fields, ByVal tTest_field As String)
    ' Separated 2017.08.27. Revised 2021.03.04.
    ' Test write, to check that all fields are there:
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim i As Integer
    Dim tTest_type As String = ""
    Dim tInsert As String
    Dim tDelete As String
    tLocals.Subr = "Database diagnostic level 05, part 04"
    uOp.Subr = tLocals.Subr

    tLocals.Action = udtuStandard_texts.Sql_build_insert
    tInsert = "INSERT INTO [" & tDb.Table_name & "] (" & vbCr
    For i = 1 To tTable_var.Length
      tLocals.Msg = tTable_var.Name_(i)
      If tTable_var.Type_(i) = "Autoincrement" Then
        ' Skip.
      Else
        tInsert = tInsert & tDb.Indent & "[" & tTable_var.Name_(i) & "]"
      End If
      If i <> tTable_var.Length And tTable_var.Type_(i) <> "Autoincrement" Then
        tInsert = tInsert & ", " & vbCr
      End If
      If tTable_var.Name_(i) = tTest_field Then tTest_type = tTable_var.Type_(i)
    Next
    tInsert = tInsert & vbCr & ") VALUES (" & vbCr
    For i = 1 To tTable_var.Length
      tLocals.Msg = tTable_var.Type_(i)
      Select Case tTable_var.Type_(i)
        Case "Autoincrement"
          ' Do nothing. 
        Case "Bit"
          tInsert = tInsert & tDb.Indent & "True"
        Case "Decimal", "Integer", "Single", "Short"
          tInsert = tInsert & tDb.Indent & "1"
        Case "Text", "VARCHAR(MAX)"
          tInsert = tInsert & tDb.Indent & "'Test'"
        Case "Date"
          tInsert = tInsert & tDb.Indent & "'" & Date.Today & "'"
        Case "Time"
          tInsert = tInsert & tDb.Indent & "'" & TimeString & "'"
        Case Else
          tLocals.Msg = tTable_var.Type_(i)
          ' Check that the gDef_*.length is the correct value.
          Stop
      End Select
      If i = tTable_var.Length Then
        tInsert = tInsert & ")"
      Else
        If tTable_var.Type_(i) <> "Autoincrement" Then
          tInsert = tInsert & ", " & vbCr
        End If
      End If
    Next

    tLocals.Action = udtuStandard_texts.Sql_build_delete
    tDelete = "DELETE FROM " & tDb.Table_name & " WHERE " & tTest_field & "="
    If uApp.Debug Then tLocals.Msg = tTest_field.GetTypeCode()
    Select Case tTest_type
      ' Case "Autoincrement"
      ' Do nothing. 
      'Case "Bit"
      '   stop
      '   tDelete = tDelete & tDb.Indent & "True"
      'Case "Decimal", "Integer", "Single", "Short"
      '   stop
      '   tDelete = tDelete & "'1'"
      Case "Text", "VARCHAR(MAX)"
        tDelete = tDelete & "'Test'"
        'Case "Date"
        '   tDelete = tDelete & "'" & Date.Today & "'"
        'Case "Time"
        '   tDelete = tDelete & "'" & TimeString & "'"
      Case Else
        Stop
    End Select

    With tDb
      .Rows_affected = 0
      .Success = False

      If uApp.Debug Then tLocals.Msg = .Conn.State
      If .Conn.State <> ConnectionState.Open Then .Conn.Open()

      tLocals.Action = "Check " & .Table_name
      .Cmd_str = tInsert
      Call Sql_conditioning_SqLite(tDb)
      Call Error_write_micron(tInsert & vbCr & tDelete)
      Stop
      uOp.Subr = tLocals.Subr
      .Cmd.CommandText = .Cmd_str
      .Cmd.ExecuteNonQuery()
      .Success = True ' INSERT INTO table with all fields works. 

      .Cmd_str = tDelete
      .Cmd.CommandText = .Cmd_str
      .Cmd.ExecuteNonQuery()
      .Success = True ' DELETE FROM into table with all fields works. 

      .Conn.Close()
      .Cmd_str = ""
      .Cmd.CommandText = .Cmd_str
    End With
    tLocals.Action = ""
    uOp.Subr = tLocals.Subr

    uOp.Subr = ""
  End Sub

  Private Sub Database_upgrade_insert_SqLite(ByRef tDb As udtDatabase_SqLite, ByVal tTable_var As udtMemory_Fields, Optional ByVal tField As String = "")
    ' Separated 2017.08.27. Revised 2021.04.07.
    ' Insert any omitted fields.
    ' Try to insert each field; for any that do, the Insert will fail gracefully; for any that don't exist,
    ' the Insert will create it.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tSearch As udtSearch = Nothing

    Dim i As Integer
    Dim tTable As String
    tLocals.Subr = "Database upgrade insert"
    uOp.Subr = tLocals.Subr
    tSearch.Keyword = ""
    tTable = tDb.Table_name

    If tField = "" Then
      For i = 1 To tTable_var.Length
        tLocals.Msg = ""
        tDb.Table_name = tTable
        tLocals.Action = Replace(Replace("Insert field " & tTable_var.Name_(i), "[", ""), "]", "")

        tLocals.Action = udtuStandard_texts.Sql_build_alter
        tDb.Cmd_str = "ALTER TABLE " & tDb.Table_name & " ADD [" & tTable_var.Name_(i) & "] " & tTable_var.Type_(i)
        If tTable_var.Name_(i) = "Id" Or tTable_var.Name_(i) = "[Id]" Then
          tDb.Cmd_str = tDb.Cmd_str & " NOT NULL PRIMARY KEY"
        End If
        Call Database_general_SqLite(tDb)
        uOp.Subr = tLocals.Subr
        If tDb.Success Then
          tSearch.Keyword = tSearch.Keyword & tTable_var.Name_(i) & ", "
        End If
      Next
    Else
      Dim tType As String = ""
      tType = tTable_var.Type_(Array.IndexOf(tTable_var.Name_, tField))
      Call Database_upgrade_insert_one_SqLite(tDb, tField, tType)
      uOp.Subr = tLocals.Subr
    End If

    uOp.Subr = ""
  End Sub

  Private Sub Database_upgrade_insert_one_SqLite(ByRef tDb As udtDatabase_SqLite, tField As String, tType As String)
    ' Separated 2017.08.27. Revised 2021.04.04.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Database upgrade insert one"
    tLocals.Action = "Insert field " & tField & " into " & tDb.Upgrade_table

    tDb.Cmd_str = "ALTER TABLE " & tDb.Table_name & " ADD [" & tField & "] " & tType
    Call Database_general_SqLite(tDb)
    If uOp.Success Then
      tDb.Upgrade_counter = tDb.Upgrade_counter + 1
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    End If

    uOp.Subr = ""
  End Sub

#End Region ' End Database diagnostic.

#Region "Preferences"

  Sub Preference_insert_first_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.02.15.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Stop ' Indefinite.
    Dim tCmd_str As String
    tLocals.Subr = "Preference insert first"
    tLocals.Action = "Insert first Preference"
    uApp.First_run_in_progress = True

    Stop ' Call Sql_builder_insert()
    With tDb
      .Table_name = udtTables.Preference
      tCmd_str = "INSERT INTO " & .Table_name & " (" & vbCr
      tCmd_str = tCmd_str & .Indent & "App_name " & vbCr
      tCmd_str = tCmd_str & ") VALUES (" & vbCr
      tCmd_str = tCmd_str & .Indent & "'" & uApp.Internal_name & "' " & vbCr
      tCmd_str = tCmd_str & ")"
      .Cmd_str = tCmd_str
    End With
    Call Database_general_SqLite(tDb)
    tLocals.Success = tDb.Success
    Call Log_write(tLocals, 0)

    uOp.Subr = ""
  End Sub

#End Region

#Region "Database operations general"

  Sub Database_general_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.27. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    Dim tgOpSubr As String

    Dim tTable_loc As Short ' The position of the table name in the command string.
    Dim tPara_loc As Short  ' The position of the first open parenthesis in the command string.
    Dim tTable_name As String = ""
    Dim tCmd_str As String
    Dim tStatement As String = ""

    tLocals.Action = ""

    On Error GoTo Error_captured
    tLocals.Subr = "Database general"
    tgOpSubr = uOp.Subr
    uError.Message = ""

    tLocals.Action = "Check database and sql basics"
    If tDb.Table_name = "" Then
      Call Sound_alert()
      uError.Message = "Table not specified in  tDb.Table_name ."
      GoTo Ender_full
    End If
    If tDb.Cmd_str = "" Then
      uError.Message = "No s.q.l. command."
      GoTo Ender_full
    Else
      tDb.Cmd_str = Trim(tDb.Cmd_str)
      tCmd_str = tDb.Cmd_str
    End If
    If tDb.Database = "" Then
      If tDb.Conn Is Nothing Then
        tDb.Table_name = ""
        GoTo Ender_full
      Else
        Call Sound_alert()
        tLocals.Subr = uOp.Subr
        uOp.Called_from = uOp.Subr
        tCmd_str = tDb.Cmd_str
        Call Environment_determine()
        If uOp.Subr = "" Then uOp.Subr = tgOpSubr
        tDb.Cmd_str = tCmd_str
      End If
    End If
    If tDb.Conn_str = "" Or tDb.Conn Is Nothing Or tDb.Cmd Is Nothing Then ' 2014.03.02. 
      tLocals.Subr = uOp.Subr
      tTable_name = tDb.Table_name
      tCmd_str = tDb.Cmd_str

      Call Database_instantiate_SqLite(tDb)
      If uOp.Subr = "" Then uOp.Subr = tgOpSubr
      If Not tDb.Success Then
        uError.Message = "Database not instantiated."
        GoTo Ender_full
      End If

      tDb.Table_name = tTable_name
      tDb.Cmd_str = tCmd_str
    End If
    If InStr(1, tDb.Database, ":") > 0 And Not (System.IO.File.Exists(tDb.Database)) Then
      Call Sound_alert()
      tLocals.Subr = uOp.Subr
      tCmd_str = tDb.Cmd_str
      Call Memory_create_SqLite(Nothing)
      If uOp.Subr = "" Then uOp.Subr = tgOpSubr
      tDb.Cmd_str = tCmd_str
    End If

    tLocals.Msg = ""
    uOp.Success = False
    uError.Message = ""

    tLocals.Action = "Check sql syntax"
    tTable_loc = InStr(tDb.Cmd_str, " ")
    tStatement = UCase(Left(tDb.Cmd_str, tTable_loc - 1))
    Call Sql_conditioning_SqLite(tDb)
    If uOp.Subr = "" Then uOp.Subr = tgOpSubr
    If uError.Message <> "" Then GoTo Ender_full
    tLocals.Msg = ""

    Call Error_write_micron(tDb.Cmd_str)
    If uOp.Subr = "" Then uOp.Subr = tgOpSubr
    tLocals.Msg = ""

    tLocals.Action = "Determine table name"
    tTable_loc = 0
    If UCase(tStatement) = udtuStandard_texts.Sql_insert Then
      tTable_loc = InStr(tDb.Cmd_str, "INTO", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "into ", "")
      tTable_name = Replace(tTable_name, "INTO ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_select Then
      tTable_loc = InStr(tDb.Cmd_str, "FROM", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "from ", "")
      tTable_name = Replace(tTable_name, "FROM ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_create Then
      tTable_loc = InStr(tDb.Cmd_str, "CREATE TABLE", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "create table ", "")
      tTable_name = Replace(tTable_name, "CREATE TABLE ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_update Then
      tTable_loc = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_update, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "update ", "")
      tTable_name = Replace(tTable_name, "UPDATE ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_delete Then
      tTable_loc = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_delete, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "delete from ", "")
      tTable_name = Replace(tTable_name, "DELETE FROM ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_alter Then
      tTable_loc = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_alter, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_loc)
      tTable_name = Replace(tTable_name, "alter table ", "")
      tTable_name = Replace(tTable_name, "ALTER TABLE ", "")
    ElseIf UCase(tStatement) = "PROCEDURE" Then
      Stop
    End If
    If tTable_name = "" Then
      Stop
    End If
    tTable_loc = InStr(1, tTable_name, " ", CompareMethod.Text)
    If UCase(tStatement) = udtuStandard_texts.Sql_insert And tTable_loc < tPara_loc Then
      tPara_loc = InStr(tTable_name, "(")
      If tTable_loc < tPara_loc Then
        Stop
        uError.Message = "Alert! The table is not specified in the command string."
        GoTo Ender_full
      End If
    End If
    If tTable_loc > 0 Then
      tTable_name = Mid(tTable_name, 1, tTable_loc - 1)
    End If
    tTable_loc = InStr(tTable_name, vbCr, CompareMethod.Text)
    If tTable_loc > 0 Then
      tTable_name = Mid(tTable_name, 1, tTable_loc - 1)
    End If
    tTable_name = Replace(tTable_name, "[", "")
    tTable_name = Replace(tTable_name, "]", "")

    With tDb
      .Rows_affected = -1
      .Success = False
      ' Check table name:
      If UCase(.Table_name) <> UCase(tTable_name) Then
        .Table_name = tTable_name
      End If

      Select Case tStatement
        Case udtuStandard_texts.Sql_create, udtuStandard_texts.Sql_insert, udtuStandard_texts.Sql_update, udtuStandard_texts.Sql_delete, udtuStandard_texts.Sql_alter
          tLocals.Action = tStatement & " " & tDb.Table_name
          If .Conn.State <> ConnectionState.Open Then .Conn.Open()
          .Cmd.CommandText = tDb.Cmd_str
          .Cmd.ExecuteNonQuery()
          .Conn.Close()
          .Table_name = ""

        Case udtuStandard_texts.Sql_select
          .Dataset = New DataSet
          .Data_adapter = New SQLiteDataAdapter(.Cmd_str, .Conn_str)
          tLocals.Action = udtuStandard_texts.Dataset_fill
          .Rows_affected = .Data_adapter.Fill(.Dataset, .Table_name) ' A momentary connection. 
          tLocals.Msg = "records returned: " & .Rows_affected & "."
          If .Rows_affected = 0 Then
            ' Check server case sensitivity. 
            Stop
          End If
        Case Else
          Stop
      End Select

      .Success = True
      .Cmd_str = ""
      .Cmd.CommandText = ""
    End With

Upgrade_section:
    Select Case tLocals.Action
      Case udtuStandard_texts.Upgrade
        If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
        Call Database_diagnostic_00_SqLite(tDb)
        If uOp.Subr = "" Then uOp.Subr = tgOpSubr
        tLocals.Action = ""
      Case "Password"
        tDb.Use_password = True
        Call Database_instantiate_SqLite(tDb)
        uOp.Subr = tLocals.Subr
        Call Database_general_SqLite(tDb)
        uOp.Subr = tLocals.Subr
    End Select

Ender_full:
    Stop
    Select Case uError.Message
      Case ""
        uOp.Success = tDb.Success
        If tDb.Success Then
          If tLocals.Msg = "" Then
            tLocals.Msg = "Success."
          Else
            tLocals.Msg = "Success; " & tLocals.Msg
          End If
          uOp.Result = tLocals.Msg
        Else
          Stop
        End If
      Case "Database not instantiated."
        ' Exit gracefully.
      Case "Exclusive access denied."
        Call Sound_alert()
        Stop ' To do, to test:
        uError.Message = uApp.App_name & " does not have exclusive access to this database" & vbCr & "(" & tDb.Database & ")." & vbCr & "Please contact your database administrator."
        Stop ' To do, to re-develop:
        '     uOp.Result  = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        ' uDialogue.Dialogue_result = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
      Case "No command."
        Call Sound_alert()
        Stop ' To do, to re-develop:
        '     uOp.Result  = Msgbox_LIS_top ("The s.q.l. command was empty.", udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        ' uDialogue.Dialogue_result = Msgbox_LIS_top ("The s.q.l. command was empty.", udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        Call Msgbox_LIS_top("The s.q.l. command was empty.", udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
      Case Else
        Stop ' To do, to re-develop:
        '     uOp.Result  = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        ' uDialogue.Dialogue_result = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        If InStr(uError.Description, "The process cannot access the file") > 0 Then
          Stop ' To do, to test:
          frmMain.Close()
          Call End_game()
        End If
    End Select
    uError.Message = ""

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    uError.Message = tDb.Cmd_str
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      If Not tDb.Upgrade_in_progress Then uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
      If tLocals.Action = "Fill" Then tLocals.Action = ""

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 75 ' "The process cannot access the file 'C:\Users\Michael\AppData\Roaming\LivingInfo Systems\Financier\Financier error log.txt' because it is being used by another process."
            If InStr(uError.Description, "The process cannot access the file") > 0 Then
              If InStr(UCase(uError.Description), ".TXT") > 0 Then
                uError.Message = uError.Description & vbCr
                uError.Message = uError.Message & "To resolve this, please restart your computer."
                Resume Next
              Else
                uError.Message = "Exclusive access denied."
                Resume Ender_full
              End If
            Else
              Stop ' Permanent.
              Call Error_display(tLocals)
            End If
          Case Else
            If tDb.Database = "" Then
              Stop ' Permanent.
              Call Error_display(tLocals)
            Else
              Call Database_error_handler_SqLite(tLocals.Action, tDb)
              If (tLocals.Action <> "" Or tDb.Upgrade_table <> "") And Not tDb.Upgrade_in_progress Then
                Resume Upgrade_section
              Else
                Resume Error_captured
              End If
            End If
        End Select
      End If
    End If
    If tStatement <> udtuStandard_texts.Sql_select Then
      If tDb.Conn IsNot Nothing Then
        If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
      End If
    End If
    tDb.Cmd_str = ""
    If tDb.Cmd IsNot Nothing Then tDb.Cmd.CommandText = ""
    uOp.Result = Nothing
    ' uOp.Subr = ""
  End Sub

  Private Sub Sql_conditioning_SqLite(ByRef tDb As udtDatabase_SqLite)
    ' Separated 2017.08.13. Revised 2021.02.28.
    ' Sql conditioning:  
    ' Modify the sql command string depending on the sqlserver being used. 
    ' References: 
    ' Having www.w3schools.com/sql/sql_having.asp. 
    ' Top    www.w3schools.com/sql/sql_top.asp. 
    Dim i As Integer
    Stop
    If tDb.Cmd_str = "" Then
      Call Sound_alert()
      Stop
    Else
      Dim tStatement As String = UCase(Left(tDb.Cmd_str, InStr(tDb.Cmd_str, " ") - 1))
      If InStr(tDb.Cmd_str, "CREATE TABLE") > 0 Then
        tDb.Cmd_str = Replace(tDb.Cmd_str, "SINGLE", "INTEGER")
      ElseIf InStr(tDb.Cmd_str, udtuStandard_texts.Sql_select) <> 0 Then
        If InStr(tDb.Cmd_str, " TOP ") <> 0 Then
          Dim tStr As String
          Dim tSql_command As String
          i = 0
          tStr = tDb.Cmd_str
          tSql_command = Left(tStr, InStr(tStr, " ") - 1)
          tStr = Mid(tStr, InStr(1, tStr, " TOP ") + 5)
          i = Val(Left(tStr, InStr(tStr, " ")))
          tStr = Mid(tStr, InStr(tStr, " ") + 1)
          ' And the number afterwards.
          tStr = tSql_command & " " & tStr & vbCr & "LIMIT" & Str(i)
          tDb.Cmd_str = tStr
        End If
      End If
      If InStr(tDb.Cmd_str, "True,") > 0 Then
        tDb.Cmd_str = Replace(tDb.Cmd_str, "True,", "1,")
      End If
      If InStr(tDb.Cmd_str, "=True") > 0 Then
        tDb.Cmd_str = Replace(tDb.Cmd_str, "=True", "=1")
      End If
      If InStr(tDb.Cmd_str, "False,") > 0 Then
        tDb.Cmd_str = Replace(tDb.Cmd_str, "False,", "0,")
      End If
      If InStr(tDb.Cmd_str, "=False") > 0 Then
        tDb.Cmd_str = Replace(tDb.Cmd_str, "=False", "=0")
      End If
    End If
  End Sub

  Sub Database_error_handler_SqLite(ByRef tAction As String, ByRef tDb As udtDatabase_SqLite)
    ' Separated by Michael John Boyack Hodgson, 2017.08.27. Revised 2021.04.01.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    Select Case Err.Number
      Case 0
        Stop

      Case 5 ' Description varies: 
        If InStr(uError.Description, "already exists") > 0 Then
          ' Skip and continue.
          tLocals.Msg = uError.Description
          uError.Message = ""
          uError.Number = 0
          uError.Description = ""
          uError.Source = ""
          tLocals.Action = ""

        ElseIf InStr(uError.Description, "Could not find output table") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        ElseIf uError.Description = "Cannot find table or constraint." Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        ElseIf InStr(uError.Description, "does not belong to table") > 0 Then
          tDb.Upgrade_table = tDb.Table_name

        ElseIf InStr(uError.Description, "ERROR [08001] Client unable to establish connection") > 0 Then
          Stop
          Call Error_display(tLocals)

        ElseIf InStr(uError.Description, "is exclusively locked by user ") > 0 Then
          uError.Message = "Exclusive access denied."

        ElseIf InStr(uError.Description, "No value given for one or more required parameters.") > 0 Then
          ' Try a memory upgrade; but that may not be the solution. 
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        ElseIf InStr(uError.Description, "Syntax error (missing operator) in query expression ") > 0 Then
          ' Look for a missing comma. 
          Stop
          Call Error_display(tLocals)

        ElseIf InStr(uError.Description, "The INSERT INTO statement contains the following unknown field name") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        ElseIf InStr(uError.Description, "The Microsoft Jet database engine cannot find the input table or query") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        ElseIf InStr(uError.Description, "Unrecognized database format") > 0 Then
          Stop
          ' Kill it.
          ' If not resolved, use uUser.Ap_data_path & "\" & ulis.Name & "\" & uApp.App_name & "\" & uApp.internal & " Memory.mdb"

        ElseIf InStr(uError.Description, "SQLite error") > 0 And InStr(uError.Description, "no such table") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals.Action = udtuStandard_texts.Upgrade

        Else
          Stop
          Call Error_display(tLocals)

        End If

      Case Else
        Stop
        Call Error_display(tLocals)
    End Select

  End Sub

  Function Last_use_SqLite_fn() As Boolean
    ' Separated 2021.02.17. Revised 2021.04.07.
    Stop ' Temporary.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tSearch As udtSearch = Nothing

    Dim tToday As Date
    Dim tToday_str As String
    Dim tLast_date As Date
    tLocals.Subr = "fnLast use SqLite"
    uOp.Subr = tLocals.Subr

    Stop ' To do, use tDb.
    tToday = DateAndTime.Today ' M/d/y.
    tToday_str = CStr(tToday) ' Y/m/d.
    'uMemory.Table_name = uTables.Log
    'uMemory.Cmd_str = "SELECT TOP 1 Date FROM " & uMemory.Table_name & " " & vbCr
    'tToday_str = Replace(tToday_str, ".", "-")
    'uMemory.Cmd_str = uMemory.Cmd_str & "WHERE Date='" & tToday_str & "'" & vbCr
    'uMemory.Cmd_str = uMemory.Cmd_str & "AND   Subr='End game'" & vbCr
    'uMemory.Cmd_str = uMemory.Cmd_str & "ORDER BY Id DESC"
    'Call Database_SqLite.Database_general(uMemory)
    uOp.Subr = tLocals.Subr
    If uMemory.Rows_affected <> 1 Then
      'uMemory.Table_name = uTables.Log
      'uMemory.Cmd_str = "SELECT TOP 1 Date FROM " & uMemory.Table_name & " " & vbCr
      'uMemory.Cmd_str = uMemory.Cmd_str & "WHERE Date='" & tToday_str & "'" & vbCr
      'uMemory.Cmd_str = uMemory.Cmd_str & "ORDER BY Id DESC"
      'Call Database_general(uMemory)
      uOp.Subr = tLocals.Subr
      tSearch.Keyword = "Date"
      If uMemory.Rows_affected > 0 Then
        tLocals.Action = udtuStandard_texts.Dataset_read
        If IsDBNull(uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          tLast_date = CDate(uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)) ' Mm/dd/yyyy.
          If DateAdd(DateInterval.Day, 7, tLast_date) < DateAndTime.Today Then
            Last_use_SqLite_fn = True
          End If
        End If
      End If
    End If

    uOp.Called_from = ""
  End Function

#End Region ' End Database operations general.

End Module
