﻿' Module: Declarations_common.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

Option Explicit On
Imports System.Net

Public Module Declarations_common
  Public Const cDeclarations_common_version As Date = #5/9/2021# ' M/d/y.
  Declare Function Get_this_user_fn Lib "advapi32.dll" Alias "GetUserNameA" (ByVal lpBuffer As String, ByRef nSize As Integer) As Integer
  Declare Function Get_system_metrics_fn Lib "user32" Alias "GetSystemMetrics" (ByVal nIndex As Integer) As Integer

  ' 2018.10.28.
  ' U for universal; used in all applications.
  ' G will now be used for global, to mean global to this application.

  Structure udtuStandard_texts
    ' Added 2018.10.28. Revised 2021.04.27.
    Const Alphabetical As String = "Alphabetical" ' 2019.11.10.
    Const Chronological As String = "Chronological"

    Const Continue_ As String = "Continue" ' 2020.12.01.
    Const Diagnostic As String = "Diagnostic"
    Const Date_null As Date = #1/1/1904#

    ' Status:
    Const Initiation As String = "Initiation"
    Const Cancelled As String = "Cancelled" ' Usually by the user.
    Const Failed As String = "Failed"       ' 2021.01.15. Means procedure failure; it didn't perform as expected.
    Const Completion As String = "Completion"
    ' Const Dialogue_active As String = "Dialogue active" ' 2021.05.02; to consider.

    Const Loading As String = "Loading"
    Const Main_start As String = "Main start" ' Revised 2021.03.03.
    Const Password_upgrade As String = "Password upgrade"
    Const Retry As String = "Retry" ' 2020.12.01.
    Const Session_begin As String = " session begin" ' 2021.04.27.
    Const Time_null As Date = #12:00:00 AM#
    Const Upgrading As String = "Upgrading..."
    Const Upgrade As String = "Upgrade" ' 2020.12.01.

    Const Sql_alter As String = "ALTER TABLE"
    Const Sql_create As String = "CREATE"
    Const Sql_delete As String = "DELETE"
    Const Sql_insert As String = "INSERT INTO"
    Const Sql_modify As String = "MODIFY COLUMN" ' 2021.03.05. Added.
    Const Sql_select As String = "SELECT"
    Const Sql_show As String = "SHOW" ' 2021.03.05. Added.
    Const Sql_update As String = "UPDATE"

    Const Sql_build_alter As String = "Sql build alter"
    Const Sql_build_delete As String = "Sql build delete"
    Const Sql_build_insert As String = "Sql build insert"
    Const Sql_build_select As String = "Sql build select"
    Const Sql_build_update As String = "Sql build update"
    Const Sql_build_upsert As String = "Sql build upsert"

    Const Sql_fill As String = "Sql fill"
    Const Sql_write As String = "Sql write"
    Const Dataset_clear As String = "Dataset clear"
    Const Dataset_fill As String = "Dataset fill"
    Const Dataset_read As String = "Dataset read"
  End Structure

  Enum udtStandard_off_on ' 2019.03.14. Revised 2020.03.13.
    Null ' Because default value is zero.
    Off
    On_
  End Enum
  Enum udtBoolean_extended ' 2020.03.13.
    Null
    False_
    True_
  End Enum

  Structure udtColours
    ' Promoted to universal, 2020.04.14. Revised 2020.11.08.
    Public Highlight As Color
    Public On_ As Color
    Public Off_ As Color
    Public Okay As Color
    Public Cancel As Color

    Public Initialised As Boolean

    Sub Initialise()
      ' Revised 2020.11.08.
      If Not Initialised Then
        Highlight = Color.LemonChiffon
        On_ = Color.DarkGreen
        Off_ = Color.Red
        Okay = Color.LightGreen
        Cancel = Color.OrangeRed
        Initialised = True
      End If
    End Sub

  End Structure

  Enum Taskbar_Location
    Top
    Bottom
    Left
    Right
  End Enum

  Enum udtTheme ' 2019.03.26.
    Null ' Because defult value is zero.
    Light
    Dark
    System
  End Enum

  Structure udtLivingInfoSystems
    ' Revised 2021.04.08.
    Public Config_loc As String
    Const Name_short As String = "LivingInfo Systems" ' Renamed 2020.10.23.
    Const Name_long As String = "LivingInfo Systems Limited" ' 2020.10.23.
    Public Dev As Boolean

    Public Development_path As String
    Const Domain As String = "livinginfosystems.com"
    Const LIS_memory As String = "LIS_memory.mdb"
    Const LIS_memory_populated_password As String = "LIS_memory_populated_password.mdb"
    Const LIS_memory_populated_no_password As String = "LIS_memory_populated_no_password.mdb"
    Const LIS_memory_common As String = "LIS_memory_common.mdb" ' 2021.03.03.

    Public LLC As String
    Public Prog_file_path As String ' C:\Program Files (x86)\LivingInfo Systems.
    Public Profile_path As String ' C:\Users\(User)\AppData\Roaming\LivingInfo Systems.
    Public Resources_path As String ' Typically C:\Program Files (x86)\LivingInfo Systems, but if that fails, C:\Users\(User)\AppData\Roaming\LivingInfo Systems\Resources.
    Public Support_email As String

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.03. Revised 2021.04.08.
      If Not Initialised Then
        Development_path = "C:\" & udtLivingInfoSystems.Name_short
        LLC = udtLivingInfoSystems.Name_long
        Profile_path = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) & "\" & udtLivingInfoSystems.Name_short
        ' Prog_file_path = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles) & "\" & udtLivingInfoSystems.Name_short
        Prog_file_path = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFilesX86) & "\" & udtLivingInfoSystems.Name_short
        Resources_path = Prog_file_path & "\Resources"
        Config_loc = Resources_path & "\Config.xml"
        Try
          Dev = CBool(XML_read_fn(Config_loc, "Dev")) ' !
        Catch ex As Exception

        End Try
        uOp.Success = False
        Support_email = "support@" & udtLivingInfoSystems.Domain

        Initialised = True
      End If
    End Sub

  End Structure
  Public uLIS As udtLivingInfoSystems

  Structure udtMachine
    ' Revised 2021.04.15.
    Public Name As String
    Public Environment_known As Boolean
    Public Os_major As Single
    Public Os_minor As Short
    Public Os_name As String
    Public Development_machine As Boolean
    Public Development_machine_name As String
    Public Motherboard_id As String
    Public Testing_machine As Boolean
    ' Public Installer_path As String ' 2014.11.12 Temp. 
    Public Environment_dump As String
    Public Date_format As String      ' 2019.10.19.

    Const Mjbh_Surface_01 As String = "MJBH-SURFACE-01" ' All caps.
    Const Lis_Vm_Test_01 As String = "LIS-VM-TEST-01"
    Const Lis_Vm_Test_02 As String = "LIS-VM-TEST-02"

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.04. Revised 2020.11.07.
      If Not Initialised Then
        Development_machine_name = udtMachine.Mjbh_Surface_01
        Name = Environment.MachineName
        If Name = Development_machine_name Or uLIS.Dev Or _
           Left$(Name, 7) = "LIS_DEV" Or _
           Left$(Name, 7) = "LIS-DEV" Or _
           Left$(Name, 8) = "MJBH_DEV" Or _
           Left$(Name, 8) = "MJBH-DEV" Then
          Development_machine = True
        ElseIf Left$(Name, 8) = "LIS_TEST" Or Left$(Name, 8) = "LIS-TEST" Or Left$(Name, 11) = "LIS-VM-TEST" Then
          Testing_machine = True
        End If
        Environment_dump = uLIS.Profile_path & "\Environment dump.txt"

        Initialised = True
      End If
    End Sub

    Sub Motherboard_id_determine()
      ' References:
      ' System.Management.
      ' Moved to Declarations_common 2021.04.14. Revised 2021.04.15.
      Dim tLocals As udtLocals = Nothing
      tLocals.Subr = udtApplication_subroutine_names.Motherboard_id_fn
      tLocals.Action = udtuStandard_texts.Main_start
      If Not Initialised Then Call Initialise()
      If uMachine.Development_machine Or uMachine.Testing_machine Or uMachine.Name = udtMachine.Lis_Vm_Test_01 Or uMachine.Name = udtMachine.Lis_Vm_Test_02 Then
      Else
        Stop ' Indefinite.
      End If

      If Motherboard_id = "" Then
        Motherboard_id = ""
        ' tLocals.Action = "Dim tMotherboard_id"
        ' Dim tMotherboard_id As String = String.Empty
        tLocals.Action = "Dim tQuery"
        Dim tQuery As New System.Management.SelectQuery("Win32_BaseBoard")
        tLocals.Action = "Dim tSearch"
        Dim tSearch As New System.Management.ManagementObjectSearcher(tQuery)
        tLocals.Action = "Dim tInfo"
        Dim tInfo As System.Management.ManagementObject
        Try
          For Each tInfo In tSearch.Get()
            Motherboard_id = tInfo("SerialNumber").ToString
            Motherboard_id = Trim(Motherboard_id)
          Next
          ' Motherboard_id = tMotherboard_id
        Catch ex As Exception
          Stop ' Permanent.
          tLocals.Error_description = ex.Message
          tLocals.Success = udtBoolean_extended.False_
          ' GoTo Error_process
        End Try
      End If

      uOp.Subr = ""
      If tLocals.Success <> udtBoolean_extended.False_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
    End Sub

  End Structure
  Public uMachine As udtMachine

  ' Internal environment:

  Structure udtApplication
    ' Revised 2021.04.30.
    Public App_name As String          ' Full name. Don't change me after publication.
    Public Internal_name As String     ' Short name. Used for Preferences, etc.
    Public Short_name As String        ' 2020.05.16.
    Public Name_for_log As String      ' 2020.06.28. Typically either Internal_name, or Short_name.
    Public Name_for_preferences As String ' 2020.12.11.
    Public English_name As String      ' 2021.02.17. Without underscore.

    Public App_build_number As Decimal ' 2019.10.13.
    Public App_revision_date As Date   ' 2019.10.13.
    Public Appdata_path As String      ' To the folder within the user's appdata path, used by this application.
    Public Debug As Boolean
    Public End_game As Boolean ' u.

    Const App_Template As String = "Template"
    Const App_Contacts_merge As String = "Contacts Merge"
    Const App_Financier_Personal_Edition As String = "Financier Personal Edition"
    Const App_Financier_Commercial_Edition As String = "Financier Commercial Edition"
    Const App_LivingInfo_Systems_Server As String = "LivingInfo Systems Server"
    Const App_Seer As String = "Seer"

    Public Colours As udtColours ' 2020.04.14.
    Public Default_left As Short
    Const Default_date_format As String = "yyyy.MM.dd" ' 2019.11.25.
    Const Default_date_long_format As String = "yyyy MMMM d" ' 2019.11.25. Revised 2020.12.04.
    Public Default_top As Short
    Public Default_tab As String
    ' Public Email_smtp_working As Boolean ' 2020.11.19.
    Public Email_udt As udtEmail ' 2020.11.19.
    Public Environment_dump_emailed As Boolean
    Public Environment_dump_completed_date As Date ' 2021.02.16. aa.
    Public Error_log_loc As String
    ' Public First_run_completed As Boolean ' 2021.02.16 renamed for consistency. 2021.02.25 Commented.
    Public First_run_completed_date As Date ' 2021.02.16. ab.
    Public First_run_in_progress As Boolean ' 2021.02.14. In progress. ac.
    Public Initialising As Boolean
    Public Log_loc As String
    Public Me_ As String
    Public Memory_check_completed_date As Date ' 2021.02.16. ad. Revised 2021.02.17.
    Public Prog_file_path As String
    Public Resources() As String ' 2018.12.31.
    ' Public Run_count As Short    ' 2018.09.30.
    Public Run_count As udtRun_count_wizard ' 2021.04.28.
    Public Startup_sequence As Integer ' 2019.11.10. ae.
    Public Shutdown As Boolean ' af. 2021.03.24 Redefined from string to boolean.
    Public Sql_last_loc As String ' ag. Path to file showing last sql write; e.g. C:\Users\micha\AppData\Roaming\LivingInfo Systems\Seer\Seer sql last.txt. 2021.03.24 Renamed from Sql_last to Sql_last_loc.
    Public Progressbar_show As Boolean             ' 2019.07.31.
    Public Progressbar_step As Integer             ' 2019.06.02.
    Public Progressbar_steps_difference As Integer ' 2019.06.13.
    Public Progressbar_steps_total As Integer      ' 2019.06.02.
    Public Terminating As Short                    ' For premature termination under an error condition.
    Public Upgraded As Boolean
    ' Public Use_dialogue_tab As Boolean ' 2018.12.29.

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.05. Revised 2021.04.30.
      If Not Initialised Then
        Call Colours.Initialise()
        If Not uLIS.Initialised Then Call uLIS.Initialise()
        If Name_for_log = "" Then
          Call Kernel_local_pre()
        End If
        Call frmMain.Statuslabel1_set()

        Default_tab = frmMain.tabHome.Name
        Default_top = 0
        If (My.Computer.Screen.Bounds.Height - My.Computer.Screen.WorkingArea.Height) > Default_top Then
          Default_top = CShort(My.Computer.Screen.Bounds.Height - My.Computer.Screen.WorkingArea.Height)
        End If
        ' .Default_left = 0
        Default_tab = frmMain.tabHome.Name
        Email_udt.Smtp_working = True
        Environment_dump_completed_date = udtuStandard_texts.Date_null
        First_run_completed_date = udtuStandard_texts.Date_null
        Memory_check_completed_date = udtuStandard_texts.Date_null
        ReDim Resources(10)
        Resources(1) = "Alert.wav"
        Resources(2) = "Critical.png"
        Resources(3) = "Exclamation.jpg"
        Resources(4) = "Gear.png"
        Resources(5) = "Info_tree.jpg"
        Resources(6) = "Information.jpg"
        Resources(7) = "LIS.gif"
        Resources(8) = "Ready.wav"
        Resources(9) = udtLivingInfoSystems.LIS_memory_populated_password
        Resources(10) = "Vandiana_platin_lite.ttf"
        If uMachine.Development_machine Then
          ReDim Preserve Resources(Resources.GetLength(0))
          Resources(Resources.GetLength(0) - 1) = udtLivingInfoSystems.LIS_memory_populated_no_password
        End If

        Me_ = uLIS.Prog_file_path & "\" & Short_name
        Appdata_path = Environ("APPDATA") & "\" & udtLivingInfoSystems.Name_short & "\" & uApp.Short_name
        ' Appdata_path = Environ("APPDATA") & "\" & udtLivingInfoSystems.Name_short & "\" & uApp.Internal_name
        Error_log_loc = Appdata_path & "\" & Short_name & " error log.txt"
        Prog_file_path = uLIS.Prog_file_path & "\" & Short_name
        Sql_last_loc = Appdata_path & "\" & Short_name & " sql last.txt"

        With Run_count
          If .Iteration Is Nothing Then
            ReDim .Iteration(2)
            ReDim .Iteration_message(2)
          End If
          .Iteration(1) = 3
          .Iteration_message(1) = "Looks like you've run " & uApp.App_name & " a few times. Do you want to start at the Home tab (or the last tab you used)?" & vbCrLf
          .Iteration_message(1) = .Iteration_message(1) & "You can set this in Edit menu: " & frmMain.tabSettings.Text & "."

          .Iteration(2) = 7
          .Iteration_message(2) = "Looks like you've run " & uApp.App_name & " a few times. Do you want to set the amount of speech " & uApp.App_name & " uses?" & vbCrLf
          .Iteration_message(2) = .Iteration_message(2) & "You can set this in Edit menu: " & frmMain.tabSettings.Text & "."
        End With

        If Progressbar_steps_total < 25 Then Progressbar_steps_total = 26 ' Initial value.

        Initialised = True
      End If

      If Appdata_path = "" Or Name_for_log = "" Then
        Stop ' Indefinite.
        Log_loc = ""
      Else
        Log_loc = Appdata_path & "\" & Name_for_log & " log " & Year(DateAndTime.Now) & ".txt"
      End If

    End Sub

  End Structure
  Public uApp As udtApplication

  Structure udtRun_count_wizard
    ' Separated 2021.04.28. To track subsequent application run wizard messages.
    Public Count As Short
    Public Iteration() As Short ' One based. Ensure that none of the values are duplicate.
    Public Iteration_message() As String ' One based.
  End Structure

  Structure udtApplication_subroutine_names
    ' Started 2019.11.08. Revised 2021.05.02.

    ' Common modules:

    ' Module Kernel common:
    ' Region Kernel:
    Const Environment_determine As String = "Detect environment"
    Const Operating_system_determine As String = "Detect operating system"
    Const Detect_country As String = "Detect country"
    Const First_run As String = "First run"
    Const First_run_make_folders As String = "First run make folders" ' Revised 2021.02.18.
    Const First_run_check_resources As String = "First run check resources" ' 2021.03.28.
    Const Installed_check As String = "Installed check"
    Const Check_run_count As String = "Check run count"
    Const Check_run_count_after_dialogue_tab As String = "Check run count after dialogue tab"
    Const Speak_it As String = "Speak it"
    Const Speak_greeting As String = "Speak greeting"
    Const English_convert_fn As String = "Convert to English"
    Const Title_case_fn As String = "Convert to Title Case"
    Const Clipboard_copy_to As String = "Clipboard copy to"
    Const Sound_alert As String = "Sound alert"
    Const Ready As String = "Ready"
    Const Get_taskbar_location As String = "Get taskbar location"
    ' Region Dialogue:
    Const Msgbox_LIS_top As String = "Msgbox LIS top"
    Const Msgbox_LIS As String = "Msgbox LIS"
    Const Inputbox_LIS_top As String = "Inputbox LIS top"
    Const Inputbox_LIS As String = "Inputbox LIS"
    Const Input_chooser_top_LIS As String = "Input chooser top LIS"
    Const Dialogue_clear As String = "Dialogue clear"
    ' Region External:
    Const XML_read_fn As String = "XML read"
    Const Environment_dump As String = "Environment dump"
    Const Environment_dump_email As String = "Environment dump email"
    Const Log_write As String = "Log write"
    Const Sql_write As String = "Sql write"
    Const Config_xml_create As String = "Config.xml create"
    Const File_properties_determine_fn As String = "File properties determine" ' 2021.04.14 renamed.
    Const Me_update As String = "Me update"
    Const Motherboard_id_fn As String = "Motherboard id"
    Const Registry_get_value_fn As String = "Registry get value"
    ' Region, "Registration":
    Const Registration As String = "Registration"
    Const Registration_request As String = "Registration request"

    Const Registration_request_email As String = "Registration request email"

    Const Registration_request_detail As String = "Registration request detail"
    Const Registration_code_submission_fn As String = "Registration code submission"
    Const Registration_code_fn As String = "Registration code"
    Const Registration_validate_email_fn As String = "Registration validate email"
    Const Register As String = "Register"
    ' Region, "Error handling":
    Const Error_clear As String = "Error clear"
    Const Error_display As String = "Error display"
    Const Error_email As String = "Email the error file"
    Const Error_write_micron As String = "Error write micron"
    Const Error_write_mega As String = "Error write mega"
    Const Infinite_loop As String = "Infinite loop"
    ' Region, "Version maintenance":
    Const Version_check As String = "Version check"
    Const Version_check_after_dialogue_tab As String = "Version check after dialogue tab"
    Const Version_check_sub_01 As String = "Version check sub 01"
    Const Version_check_sub_02 As String = "Version check sub 02"
    Const Extract_transform_load_Kernel_vb As String = "Extract transform load Kernel_vb"
    Const Module_list_load As String = "Module list load"
    Const App_revision_upgrade As String = "App_revision upgrade"

    Const End_game As String = "End game"

    ' Module: Database_common:
    ' Region, "Create, instantiate, set-up":
    Const Memory_create As String = "Memory create"
    Const uMemory_copy_from_progfiles As String = "uMemory copy from progfiles"
    Const Database_password_upgrade As String = "Database password update"
    Const Database_table_create As String = "Database table create"
    Const Database_instantiate As String = "Database instantiate"
    Const Memory_instantiate As String = "Memory instantiate"
    Const tDb_point_to_uMemory As String = "tDb point to uMemory"
    Const Memory_Define_log As String = "(Specific to that sub-routine.)"
    Const Memory_define_International As String = "(Specific to that sub-routine.)"

    ' Region, "Database diagnostic":
    Const Database_diagnostic_00 As String = "Database diagnostic 00"
    Const Database_diagnostic_level_05 As String = "Database diagnostic level 05"
    Const Database_diagnostic_level_05_part_01 As String = "Database diagnostic level 05, part 01"
    Const Database_diagnostic_level_05_part_02 As String = "Database diagnostic level 05, part 02"
    Const Database_diagnostic_level_05_part_03 As String = "Database diagnostic level 05, part 03"
    Const Database_diagnostic_level_05_part_04 As String = "Database diagnostic level 05, part 04"
    Const Database_upgrade_insert As String = "Database upgrade insert"
    Const Database_upgrade_insert_one As String = "Database upgrade insert one"
    Const Diagnostic_all_tables As String = "Diagnostic all tables" ' 2021.02.07.

    ' Region, "Data fixes":
    Const Data_fix_currency_sign_to_currency_symbol As String = "Data fix currency_sign to currency_symbol"
    Const Data_fix_currency_to_currency_ As String = "Data fix Currency to Currency_"
    Const Data_fix_field_rename As String = "Data fix field rename"
    Const Preference_insert_first As String = "Preference insert first"
    Const Preferences_load As String = "Preferences load"
    Const Preferences_write As String = "Preferences write"
    Const Preferences_write_one As String = "Preferences write one"
    Const Preferences_duplicate_fix As String = "Preferences duplicate fix"
    Const Log_write_memory As String = "Log write memory"
    Const Database_general As String = "Database general"
    Const Database_general_detect_table As String = "Database general detect table"
    Const Sql_conditioning As String = "Sql conditioning"
    Const Database_error_handler As String = "Database error handler"
    Const Database_query_one_fn As String = "Database query one"
    Const Memory_query_one_fn As String = "Memory query one"
    Const Last_use_fn As String = "Last use"
    Const Country_detect_from_uMemory_ProgFiles As String = "Country detect from uMemory_ProgFiles"
    Const Database_clear_partial As String = "Database clear partial"

    ' Module: Ftp common:
    Const Ftp_download_prep As String = "Ftp download preparation"
    Const Ftp_download As String = "Ftp download"

    ' Module: Web_common:
    Const Internet_connection_check As String = "Internet connection check"
    Const Get_default_browser_fn As String = "Get default browser"
    Const Url_validate_fn As String = "Url validate"
    Const Email_validate_fn As String = "Email validate"
    Const Me_web_update As String = "Me web update"
    Const Country_settings_scrape As String = "Country settings scrape"

    ' Module: Kernel:
    ' Region "Shared partial":
    Const Check_run_count_local As String = "Check run count local"
    Const Insert_sample_data As String = "Insert sample data"

    ' Module: clsEncrypt:
    Const Encrypt_or_decrypt_fn As String = "Encrypt or decrypt"
    Const Encrypt_array_fn As String = "Encrypt array"
    Const Encrypt_byte_fn As String = "Encrypt byte"
    Const String_to_byte_array_fn As String = "String to byte array"
    Const Unicode_bytes_to_string_fn As String = "Unicode bytes to string"
    ' Region "Encryption":
    Const Encrypt_LIS_fn As String = "Encrypt LIS"
    Const Encrypt_LIS_two_fn As String = "Encrypt LIS two"
    Const Encrypt_it_fn As String = "Encrypt it"

    ' Module: Crypt_rc4:
    Const Crypt_rc4_fn As String = "Crypt rc4"
    Const RC4_processor_fn As String = "RC4 processor"

    ' Module: Email_smtp:
    Const Email_smtp_test As String = "Email smtp test"
    Const Email_smtp_send_fn As String = "Email smtp send"

    ' Module: Database:
    ' Region "Create, instantiate, set-up:
    Const Memory_create_local As String = "Memory create local"
    Const Database_instantiate_local As String = "Database instantiate local"
    ' Const Memory_define_preference As String=
    ' Const Memory_define_preference_common As String=
    ' Region "Database diagnostic local:
    Const Database_diagnostic_local As String = "Database diagnostic local"
    ' Region "Preferences":
    Const Preferences_load_local As String = "Preferences load local"
    Const Preferences_write_local As String = "Preferences write local"

    ' Module: General:
    Const Credit_card_number_check As String = "Credit card number check"
    Const Luhn_check As String = "Luhn check"
    Const Nested_if_01 As String = "Nested if 01"
    Const Nested_if_02 As String = "Nested if 02"

    ' Module: Geo_location:
    Const Geo_location_get_location_info_fn As String = "Geo location get location info"
    Const Geo_location_detect As String = "Geo location detect"
    Const Geo_location_get_location_property As String = "Get location property"

    ' Module: Mp3:
    Const Read_id3_v2_tag As String = "Read id3 v2 tag"
    Const Read_id3_v2_tag_v2 As String = "Read id3 v2 tag v2"

    ' Form: Main:
    Const Form_load As String = "Form load"
    Const frmMain_FormClosed As String = "frmMain_FormClosed"
    Const Form_refresh As String = "Form refresh"
    Const btnContext_more_Click As String = "btnContext_more_Click"
    Const Enter_registration_code As String = "Enter registration code"
    Const Key_handle As String = "Key handle"

    ' tabcontrol top
    Const tabControl_top_tabs_show As String = "TabControl_top tabs show"
    Const tabControl_top_SelectedIndexChanged As String = "TabControl_top_SelectedIndexChanged"
    Const Top_dialogue_preparation As String = "Top dialogue preparation" ' 2020.11.05.
    Const top_dialogue_okay_button_Click_handler As String = "Top dialogue_okay_button_Click_handler"
    Const top_dialogue_cancel_button_Click As String = "Top dialogue_cancel_button_Click"
    Const First_run_wizard_set As String = "First run wizard set"
    Const btnFirst_run_next_Click As String = "btnFirst_run_next_Click"

    ' Region, "tabControl_main":
    Const tabControl_main_show_tabs As String = "tabControl_main tabs show"
    Const tabControl_main_SelectedIndexChanged As String = "tabControl_main_SelectedIndexChanged" ' As of 2020.03.13.
    Const Variable_update As String = "Variable update"
    Const Form_fill As String = "Form fill"

    Const btnDiagnostic_environment_dump_Click As String = "btnDiagnostic_environment_dump_Click" ' 2021.02.07.
    Const btnDiagnostic_first_run_Click As String = "btnDiagnostic_first_run_Click"               ' 2021.02.07.
    Const btnDiagnostic_check_memory_structure_Click As String = "btnDiagnostic_check_memory_structure_Click" ' 2021.02.07.
    Const btnDiagnostic_check_all_Click As String = "btnDiagnostic_check_all_Click"                           ' 2021.02.07.

    Const tabMemory_query_preparation As String = "tabMemory_query preparation"

    Const tabClient_registration_preparation As String = "tabClient registration preparation" ' LivingInfo Systems Server only.
    Const txtReg_rcv_code_LostFocus As String = "txtReg_rcv_code_LostFocus" ' LivingInfo Systems Server only.
  End Structure

  Structure udtTables
    ' Promoted 2019.07.19. Revised 2020.12.30.
    Public Log As String
    Const Preference As String = "Preference"
    Const Information_Schema_Columns As String = "Information_Schema.Columns" ' 2019.07.19.

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.08.
      If Not Initialised Then
        Log = uApp.Name_for_log & "_" & "log"
        Initialised = True
      End If
    End Sub

  End Structure
  Public uTables As udtTables

  Structure udtError
    ' Revised 2021.03.13.
    Public Number As Decimal     ' 2021.03.18: To do, consider deleting.
    Public Source As String      ' 2021.03.18: To do, consider deleting.
    Public Description As String ' 2021.03.18: To do, consider deleting.
    Public Count As Integer      ' Revised 2021.03.13. Counts the total number of errors this session, by the number of visits to Error_display. ' 2021.03.18: To do, consider deleting.

    Public Extra As String
    Const Infinite_loop_limit As Integer = 20
    Public Last_error_sub As String ' 2019.11.09. If that sub generated an un-handled error, don't run it again; to prevent infinite looping.
    Public Message As String ' 2021.03.14: To do, consider deleting.
    ' Tracking:
    Public Reported As Boolean ' 2019.11.09; means that an error message has been shown to the user after an unhandled error has occured.
    Public Speech_error_reported As Boolean
    ' Public Suppress_error As Short ' 2019.11.09. Usage change; 0 = not set. 1 = yes. 2 = no. 2020.10.30 turned off.
  End Structure
  Public uError As udtError ' 2021.04.27: Use for captured errors (Error_captured), rather than process errors (Error_process).

  Structure udtFile
    ' Revised 2021.04.06.
    Public File_attributes As String ' 2018.11.03. Read-write.
    Public File_creation_time As Date ' 2018.11.03. Read-write.
    Public File_exists As Boolean ' 2018.11.03.
    Public File_extension As String
    Public File_full_name As String
    Public File_is_read_only As Boolean ' 2018.11.03. Read-write.
    Public File_last_access_time As Date ' 2018.11.03. Read-write.
    Public File_last_write_time As Date ' 2018.11.03. Read-write.
    Public File_length As Long ' 2018.11.03. Revised from integer to long 2019.08.31.
    Public File_name As String
    Public File_path As String
    Public File_short_name As String ' Without extension.
    Public Path_delimiter As String ' Re-instated 2018.11.03. Slash / or back-slash \.
    Public Free_file As Short ' 2021.04.06 revised.
    ' Public Extended As udtFile_extended_properties ' 2019.11.30.
    Public Extended As Object ' 2019.11.30.
    ' Fso As Object ' File system object.
    ' Fso_file As Object

    Public Copy_source As String ' l. 2021.03.17; Updated and expanded.
    Public Copy_source_path As String
    Public Copy_destination As String ' m.
    Public Copy_destination_path As String

    Public Open_file As String ' 2018.08.07.
    Public Text_read_line As String ' 2018.08.08.
  End Structure
  ' Public uFile As udtFile ' 2021.04.06 deleted.

  Structure udtDialogue
    ' Added 2016.04.30. Revised 2020.10.25.
    Public Dialogue_type As udtDialogue_types
    Public Alert As udtAlert_types
    Public Combobox_last As udtDialogue_combobox_items
    Public Parent As Object
    Public Prompt As String
    Public Dialogue_result As DialogResult
    Public Position As udtDialogue_position ' 2019.11.27.
    Public Response_num As Integer ' cboDialogue_response.SelectedIndex.
    Public Response_str As String  '  cboDialogue_response.Text.

    ' Flow control:
    Public Active As Boolean     ' 2018.12.29.
    Public Called_from As String ' 2018.12.29.

    ' Section 2019.06.03, revised 2020.05.03.
    Const txtPrompt_MouseHover As String = "This is asking you a question."
    Const txtResponse_MouseHover As String = "This is a suggested response."
    Const cboResponse_MouseHover As String = "Select from one of the available options."
    Const Okay_button_MouseHover As String = "Click the Okay button when you are happy with your response."
    Const Cancel_button_MouseHover As String = "Click the Cancel button to cancel this choice."
  End Structure
  Public uDialogue As udtDialogue

  Enum udtAlert_types
    ' Created 2015.04.03. Revised 2018.12.29.
    Null ' 2018.12.29, because defult value is zero.
    Critical
    Exclamation
    Information
  End Enum

  Public Enum udtDialogue_types
    ' Created 2015.04.03. Revised 2018.12.29.
    Null ' 2018.12.29, because defult value is zero.
    Chooser
    Okay_only
    Okay_cancel
    Yes_no ' 2018.05.09.
    Question
  End Enum

  Public Enum udtDialogue_position
    ' Created 2015.04.03. Revised 2020.11.25.
    Null ' 2018.12.29, because default value is zero.
    Full ' 2020.11.25.
    Top_left
    Top_right
    Bottom_left
    Bottom_right
  End Enum

  'Structure udtThreading
  ' Globalised 2020.11.18. Demoted 2021.01.28.
  '  Public Pool_thread As Boolean
  '  Public Thread As System.Threading.Thread
  '  Public Name As String

  '  Sub Start(tSub As System.Threading.ParameterizedThreadStart)
  '    Pool_thread = System.Threading.Thread.CurrentThread.IsThreadPoolThread
  '    Thread = New System.Threading.Thread(tSub)
  '    Thread.Priority = Threading.ThreadPriority.Highest
  '    If Not Thread.IsAlive Then Thread.Start()
  '    Pool_thread = System.Threading.Thread.CurrentThread.IsThreadPoolThread
  '  End Sub

  'End Structure
  'Public uThreading As udtThreading ' 2020.11.19.

  ' General:
  Structure udtOperation
    ' Revised 2021.04.04.
    ' Shows the status or results of user activity, and returning status from a sub-routing to it's calling sub-routine.
    Public Subr As String ' a.
    ' Public Action As String ' b. 2021.04.04 deleted.
    Public Called_from As String ' c.
    Public Event_ As String ' p. 2020.04.24. The event that triggered the current programme flow.
    ' Public Status As String ' q. 2021.01.15. 2021.04.04 deleted.
    Public Success As Boolean ' l. Definition: that sub-routine completed everything it was supposed to, with no un-handled errors.

    Public Administer As Boolean ' r.
    Public Diagnostics As udtDiagnostic ' s. 2021.01.07.
    Public Form_initialising As Boolean ' w.
    Public Form_updating As Boolean   ' x. Updating some data in a form. Do not confuse with Form_refreshing, which is a full refresh of the form or tab.
    Public Form_refreshing As Boolean ' y. 2019.08.14.
    Public Form_loaded As Boolean     ' z. 2021.02.08.
    ' Public Msg As String 'd. 2021.03.04 Deleted.
    ' Public Num_err As Integer ' aa. 2021.03.05 Deleted.
    Public Preferences_loaded As Boolean ' ab. 2019.04.21.
    Public Result As String ' i. What the result of a sub-routine was, besides just success, or not. 2021.03.07 renamed.

    Sub Initialise()
      ' Written 2021.01.15. Revised 2021.04.15.
      Subr = ""
      ' Action = ""
      Called_from = ""
      ' Event_ = ""
      ' Status = ""
      Success = False

      ' Administer = False
      ' Msg = ""
      ' Num_err = 0
      Result = ""
    End Sub

  End Structure
  Public uOp As udtOperation

  Structure udtLocals
    ' Separated 2021.03.14.
    ' Shows the status or results of sub-routine activity.
    Public Subr As String
    Public Action As String
    Public Changed As Boolean ' 2021.03.14.
    Public Status As String ' 2021.01.14.
    Public Success As udtBoolean_extended ' Definition: that sub-section completed everything it was supposed to, with no un-handled errors.

    Public Continue_ As Boolean
    Public Msg As String
    Public Response As String ' 2021.03.07; to do, consider deleting, in favour of uDialogue.Response_num or uDialogue.Response_str.
    Public Steps_completed As Short ' 2021.03.14.
    Public Steps_to_complete As Short ' 2021.03.14.

    ' Error section added 2021.03.13. Revised 2021.03.14.
    Public Error_number As Decimal
    Public Error_source As String
    Public Error_description As String
    Public Error_count As Integer ' 2021.03.14.
    ' Public Error_structure As udtError ' 2021.03.18; To do, consider.

  End Structure

  Structure udtDiagnostic
    ' Separated 2021.01.07.
    Public In_progress As Boolean
    Public All_tables_in_progress As Boolean ' 2020.12.10.
    Public All_tables_done As Boolean ' 2020.12.10.
  End Structure

  Structure udtSearch
    Public Keyword As String
    Public Sub_keyword As String
  End Structure
  ' Public uSearch As udtSearch ' 2021.04.07 deleted.

  ' Memory:
  Structure udtMemory_Fields
    ' Revised 2020.12.27.
    Public Table As String
    Public Length As Integer
    Public Name_() As String
    Public Type_() As String
    Public Value() As Object
    Public Condition_operator() As String ' 2020.12.26.
    Public Condition_value() As Object    ' 2020.12.27.
    Public Test_field As String
  End Structure
  Public uDef_Log As udtMemory_Fields
  Public uDef_Preference As udtMemory_Fields
  Public uDef_International As udtMemory_Fields
  Public uDef_Preference_common As udtMemory_Fields

  Structure udtDatabase_oledb
    ' Revised 2020.11.18.
    Public Database As String          ' Database name, such as on a server; or if in a file system, including a path.
    Public Database_version As Integer ' 2020.07.05.
    Public Table_name As String        ' Revised 2020.04.15.

    Public Conn_str As String
    Public Cmd_str As String
    Public Conn As OleDb.OleDbConnection
    Public Cmd As OleDb.OleDbCommand
    Public Data_adapter As OleDb.OleDbDataAdapter
    Public Dataset As DataSet

    Public Attempt As Integer    ' 2019.10.16. Counter. If less than 3 or maybe 5, try again. If >, stop with error message.
    Public Changed As Boolean
    Public Error_ As String      ' 2020.06.10. Error status.
    Public Error_table As String ' 2020.11.18. Indicates what table generated an error, so don't do that again.
    Public Indent As String
    Public Rows_affected As Short
    Public Success As Boolean
    Public Upgrade_table As String
    Public Upgrade_counter As Integer
    Public Upgrade_in_progress As Boolean
    Public Use_password As Boolean

    Public Initialised As Boolean

    Sub Initialise(ByRef tDatabase As String)
      ' Internalised 2020.11.04. Revised 2020.11.07.
      If Not Initialised Then
        '...
        Initialised = True
      End If
      If tDatabase <> "" Then Database = tDatabase
    End Sub

  End Structure
  Public uMemory As udtDatabase_oledb
  Public uMemory_ProgFiles As udtDatabase_oledb

  Structure udtUser_Parameters
    ' Revised 2021.04.15.
    ' Data about the user, his activities or preferences.
    Public Appdata_path As String      ' n. To c:\user\(user)\appdata.
    Public Cancel As Boolean           ' o.
    ' Country                            a.
    ' Currency                           b.
    ' Currency_format                    c.
    ' Currency_sign                      d.
    ' Environment dump emailed           e.
    Public First_name As String        ' v.

    Public First_run_wizard As Boolean           ' w. Wizard is in progress.
    Public First_run_wizard_max As Short         ' x. Maximum number of steps in the wizard.
    Public First_run_wizard_step As Short        ' y. Current step.
    Public First_run_wizard_loading As Boolean   ' 2020.04.13. Data and form elements are in progress being set. z.
    Public First_run_wizard_completed As Boolean ' 2020.12.08.
    ' Public First_run_wizard_completed_date As Date  ' 2021.03.07.

    Public Main_top As Short ' aa.
    Public Main_left As Short ' ab.
    Public Main_width As Short ' ac.
    Public Main_height As Short ' ad.
    Public Min_sys_tray As Boolean ' ae.
    Public My_dox As String ' p
    ' Public Option_num As Integer q.
    ' Public Option_str As String r.
    Public Preferences_changed As Boolean ' s.
    Public Profile_path As String ' t.
    Public Resources_path As String
    ' User_first_name m.
    ' User_name u.

    Public Speech As Integer ' l.
    ' Public Start_home_tab As Short ' 2018.09.30. 0 = not set. 1 = yes, start with home tab. 2 = no, start with last used tab.
    Public Start_home_tab As udtStandard_off_on ' 2020.02.15.
    Public Start_maximised As Boolean ' ag.
    Public Taskbar_location As Taskbar_Location
    Public Tab_last As String ' ah.
    Public Theme As udtTheme  ' 2019.03.26. 0 not set. 1 light. 2 dark. 3 system.

    Public Initialised As Boolean

    Sub Initialisation(Optional ByRef tFirst_run_wizard_max As Short = 0)
      ' Internalised 2020.11.02. Revised 2021.04.15.
      If Not Initialised Then
        If Appdata_path = "" Then
          Appdata_path = Environ("APPDATA")
          Appdata_path = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData)
          Profile_path = Environ("UserProfile")
          Resources_path = uLIS.Profile_path & "\Resources"
          Speech = 2

          ' Requires reference System.DirectoryServices.AccountManagement:
          Dim tUser_name As String = System.DirectoryServices.AccountManagement.UserPrincipal.Current.DisplayName ' User full name.
          First_name = Environ("USERNAME")
          First_name = tUser_name
          If InStr(First_name, Chr(32)) > 0 Then
            First_name = Left(First_name, InStr(First_name, Chr(32)) - 1)
          End If
          First_name = Title_case_fn(First_name) ' !
          Main_top = uApp.Default_top
          Main_left = uApp.Default_left

          Initialised = True
        End If

        If tFirst_run_wizard_max > 0 Then First_run_wizard_max = tFirst_run_wizard_max

      End If
    End Sub

  End Structure
  Public uUser As udtUser_Parameters

  Structure udtCon_help
    ' Revised 2021.05.09.
    Public Default_ As String ' 2020.04.03.
    Public Brought As String
    Public F1_default As String
    Public Home_default As String
    Public FRW_default As String
    Public About_default As String
    Public Diagnostic_default As String ' 2020.05.11.
    Public Memory_query_default As String
    Public Password_default As String
    Public Settings_default As String
    Public System_information_default As String ' 2019.02.03.

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.11. Revised 2021.05.09.
      If Not Initialised Then
        uCon_help.FRW_default = "This is the First Run Wizard tab. This guides you through learning to use " & uApp.App_name & "."
        uCon_help.Default_ = "This text box is context-sensitive help. The help shown here dependends on what you point to, or what you select." & vbCrLf
        uCon_help.Default_ = uCon_help.Default_ & "Try pointing to the items in this window, to see."
        uCon_help.F1_default = udtLivingInfoSystems.Name_short & " help is not accessed from the F1 key, but is on screen at all times in this field, Context Sensitive Help." & vbCrLf
        uCon_help.F1_default = uCon_help.F1_default & "If you have a question that has not been answered by that, please email " & udtLivingInfoSystems.Name_short & " Support, at " & uLIS.Support_email & "."

        uCon_help.Home_default = "This the Home tab. It shows a brief introduction to " & uApp.App_name & "."
        uCon_help.About_default = "This tab shows information about " & udtLivingInfoSystems.Name_short & ", and about " & uApp.App_name & "."
        uCon_help.Diagnostic_default = "This tab lets you run diagnostic tests, and shows the results of them."
        uCon_help.Memory_query_default = "This tab lets you remember something."
        uCon_help.Password_default = "This tab lets you enter a password."
        uCon_help.Settings_default = "This tab shows the options (settings, preferences) available to you in " & uApp.App_name & ", and the status of some values."
        uCon_help.System_information_default = "This tab gives detailed system information."

        Initialised = True
      End If
    End Sub

  End Structure
  Public uCon_help As udtCon_help

  ' Encryption:
  Structure udtEncryption
    ' Revised 2020.11.09.
    Public Plaintext As String
    Public Cyphertext As String
    Public Private_key As String
    Public Test As Boolean
    Public Temp As String
    Public i As Integer
    Public j As Integer
    Public User_name As String ' 2019.05.10.
    Public Password As String

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.09.
      If Not Initialised Then
        Private_key = "Copyright © " & uLIS.LLC & ", 1999"

        Initialised = True
      End If
    End Sub

  End Structure
  Public uCrypt As udtEncryption

  ' External:
  Structure udtActive
    ' Revised 2020.11.12.
    ' Continent As String
    Public Country As String ' a
    Public Currency_tla As String ' b. Renamed 2019.11.12.
    Public Currency_symbol As String ' c. Renamed 2018.09.30.
    Public Currency_format As String ' d.
    'Public Comma_format As String
    'Public Comma0_format As String
    'Public Tax_office As String
    'Public GST_loc As String
    'Public GST_name As String
    'Public GST_rate As Single
    'Public GST_changed As Boolean
    'Public PAYE As String

    'Public Corporation_id As Integer
    Public Corporation_name As String ' e.
    'Public Corporation_qty As Integer
    'Public Corporation_id_list() As Integer
    'Public Corporation_name_list() As String
    '' Corp_Id_last As String
    '' Corp_last As String
    'Public City As String

    'Public Branch_id As Integer
    Public Branch_name As String ' f.
    'Public Branch_qty As Integer
    'Public Branch_id_list() As Integer
    'Public Branch_name_list() As String
    '' Branch_id_last As Integer
    '' Branch_name_last As String

    'Public URL As String
    'Public Email As String

    Public Initialised As Boolean

    Sub Initialise()
      ' Internalised 2020.11.12.
      If Not Initialised Then
        Currency_format = "##,###,###,##0.00" ' Default.

        Initialised = True
      End If
    End Sub

  End Structure
  Public uActive As udtActive

  Structure udtInternet
    ' Public Connection_client As Connection_info
    Public Remote_host As IPHostEntry
    Public Server_address As IPAddress
  End Structure

  Structure udtEmail
    ' Separated date unknown. Revised 2021.04.17.
    Public Recipient_address As String
    Public From_address As String ' 2021.02.07 added.
    Public Subject As String
    Public Body As String

    Public Attachment_loc As String ' 2021.02.07 added.
    Public Attachment As System.Net.Mail.Attachment ' 2021.03.02.
    ' Public Attachments As Object    ' Full path to the attachment.

    Public From_password As String  ' 2021.02.07 added.

    Public Smtp_working As Boolean
    Public Smtp_server As String    ' 2021.02.07 added.
    Public Smtp_client As System.Net.Mail.SmtpClient ' 2021.03.02.

    Public Email_message As System.Net.Mail.MailMessage ' 2021.02.07 added.

    Public Initialised As Boolean ' 2021.03.03 added.

    Sub Initialise()
      ' Revised 2021.04.17.
      Recipient_address = ""
      Subject = ""
      Body = ""
      Attachment_loc = ""
      Attachment = Nothing
      Smtp_working = True

      Email_message = New System.Net.Mail.MailMessage

      If Not Initialised Then
        If uCrypt.Private_key = "" Then Call uCrypt.Initialise()
        From_address = "living@outlook.co.nz"

        uCrypt.Plaintext = "13BF27550BBB6AC11B"
        From_password = Crypt_rc4_fn(uCrypt.Plaintext, uCrypt.Private_key, False)

        Smtp_server = "smtp.live.com"

        Smtp_client = New System.Net.Mail.SmtpClient
        With Smtp_client
          .Host = uApp.Email_udt.Smtp_server
          .EnableSsl = True
          .Credentials = New System.Net.NetworkCredential(uApp.Email_udt.From_address, uApp.Email_udt.From_password)
          ' http://en.wikipedia.org/wiki/Simple_Mail_Transfer_Protocol#Ports.
          .Port = 587
        End With
        Initialised = True
      End If
    End Sub

    Sub Prepare_Email_message()
      ' Separated 2021.03.05.
      With uApp.Email_udt.Email_message
        .From = New System.Net.Mail.MailAddress(uApp.Email_udt.From_address, udtLivingInfoSystems.Name_short)
        .To.Add(uApp.Email_udt.Recipient_address) ' Comma delimited.
        .Subject = uApp.Email_udt.Subject

        If uApp.Email_udt.Attachment_loc IsNot Nothing Then
          If uApp.Email_udt.Attachment_loc <> "" Then
            uApp.Email_udt.Attachment = New System.Net.Mail.Attachment(uApp.Email_udt.Attachment_loc)
          End If
        End If

        If uApp.Email_udt.Attachment IsNot Nothing Then .Attachments.Add(uApp.Email_udt.Attachment)
        .Body = uApp.Email_udt.Body
      End With

    End Sub

  End Structure

  Structure udtRegistration
    ' Separated 2018.09.30. Revised 2020.12.27.
    Public Email_address As String ' Implemented 2018.09.09. Revised 2020.11.19. 10 a.
    Public Trial_30_day As Boolean ' Implemented 2018.09.10. 10 b.
    Public Trial_30_day_from_date As Date ' 10 c.
    Public Trial_30_day_20_days_notified As Boolean ' 10 d.
    Public Code As String ' Encrypted. Better name? Implemented 2018.09.24. 10 e.
    Public Code_sent As Boolean ' Implemented 2018.09.25. 10 f.
    Public Message_reported As Boolean ' 2020.12.27.
    Public Registered As Boolean ' Implemented 2018.09.30. 10 g.
  End Structure
  Public uReg As udtRegistration

End Module
