﻿' Module: General.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Revised 2021.03.07.

Module General

  Sub Credit_card_number_check(ByVal tCredit_card_number As String)
    ' Check credit card number is valid.
    ' Constructed from http://www.merriampark.com/anatomycc.htm#Account; and
    ' http://en.wikipedia.org/wiki/Credit_card_number.
    ' Michael Hodgson, 2007.08.17, test track 7184.
    ' Brought into Template 2007.11.18. Revised 2020.11.01.

    'If (Shift down)
    '            Trace()
    '        End If

    Stop
    Dim tContinue As Boolean = False
    Dim tPayment_type As String = Nothing
    Dim tIssuer_identifier_number As String = Nothing
    Dim tAccount_number As String = Nothing
    Dim tCheck_digit As String = Nothing
    Dim mCard(5) As String
    Dim tStatus As Short = 0

    If tCredit_card_number = "" Or tCredit_card_number = "0" Then
      tCredit_card_number = Inputbox_LIS("Enter the credit card number.", "", frmMain.tabControl_main, udtDialogue_position.Top_left)
      tPayment_type = tCredit_card_number
    End If

    If Val(tPayment_type) = 0 Then
      tContinue = False
    Else
      tIssuer_identifier_number = Mid$(tCredit_card_number, 1, 6)
      tAccount_number = Mid$(tCredit_card_number, 7, Len(tCredit_card_number) - 1)
      tCheck_digit = Mid$(tCredit_card_number, Len(tCredit_card_number), 1) ' Check digit

      Select Case tIssuer_identifier_number ' Listed in numerical order, so you can identify a card by it's number.

        Case (Mid$(tIssuer_identifier_number, 1, 2) = "1800") Or (Mid$(tIssuer_identifier_number, 1, 2) = "2131") ' JCB.
          If (Len(tCredit_card_number) = 15) Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 15 digits."
            tContinue = False
          End If
          mCard(0) = "JCB"
          ' SET FORMAT(pmntType;"#### #### #### ###0")  ' ?

        Case (Val(Mid$(tIssuer_identifier_number, 1, 4)) = 2014) Or (Val(Mid$(tIssuer_identifier_number, 1, 4)) = 2149) ' Diners' Club enRoute.
          uError.Message = "The Diners' Club enRoute Credit Card is no longer an active valid card."
          ' If (Length($a)=15)
          ' No validation per Wikipedia. 
          ' Else 
          ' MyAlert ("The Credit Card number entered is incomplete. It should have 15 digits.")
          tContinue = False
          'End if 
          mCard(0) = "Diners' Club enRoute"
          ' SET FORMAT(pmntType;"#### #### #### ##0")

        Case (Val(Mid$(tIssuer_identifier_number, 1, 3)) >= 300) And (Val(Mid$(tIssuer_identifier_number, 1, 3)) <= 305)  ' Diners' Club Carte Blanche. 38?
          If Len(tCredit_card_number) = 14 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 14 digits."
            tContinue = False
          End If
          mCard(0) = "Diners' Club Carte Blanche"
          ' SET FORMAT(pmntType;"#### ###### ###0")

        Case Val(Mid$(tIssuer_identifier_number, 1, 2)) = 35 ' JCB. Per Wikipedia, just 35, however www.darkcoding.net also has 30 and 33.
          If Len(tCredit_card_number) = 16 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
            tContinue = False
          End If
          mCard(0) = "JCB"
          ' SET FORMAT(pmntType;"#### #### #### ###0")

        Case ((Val(Mid$(tIssuer_identifier_number, 1, 2)) = 34) Or (Val(Mid$(tIssuer_identifier_number, 1, 2)) = 37))  ' American Express. 
          If Len(tCredit_card_number) = 15 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 15 digits."
            tContinue = False
          End If
          mCard(0) = "American Express"
          ' SET FORMAT(pmntType;"#### ###### ####0")

        Case (Val(Mid$(tIssuer_identifier_number, 1, 2)) = 36) Or (Val(Mid$(tIssuer_identifier_number, 1, 2)) = 38)  ' Diners' Club International. To do: 38 not included in Wikipedia article. 
          Select Case tCredit_card_number
            Case Len(tCredit_card_number) = 14
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### ###### ###0")
            Case Len(tCredit_card_number) = 16
              Error -1
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0") ' ?
            Case Else
              uError.Message = "The Credit Card number entered is incomplete. It should have 14 or 16 digits."
              tContinue = False
          End Select
          mCard(0) = "Diners' Club International"


        Case (Mid$(tIssuer_identifier_number, 1, 4) = "4903") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "4905") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "4911") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "4936") _
           Or (Mid$(tIssuer_identifier_number, 1, 6) = "564182") _
           Or (Mid$(tIssuer_identifier_number, 1, 6) = "633110") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "6333") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "6759") ' Switch debit card.
          ' Per Wikipedia, in use until mid 2007, after that it will be branded as Maestro. 
          If (Len(tCredit_card_number) = 16) Or (Len(tCredit_card_number) = 18) Or (Len(tCredit_card_number) = 19) Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 or 19 digits."
            tContinue = False
          End If
          mCard(0) = "Maestro"
          ' SET FORMAT(pmntType;"#### #### #### ###0")

        Case (Mid$(tIssuer_identifier_number, 1, 6) = "417500") Or (Mid$(tIssuer_identifier_number, 1, 4) = "4917") Or (Mid$(tIssuer_identifier_number, 1, 4) = "4913")  ' Visa "Electron."
          If Len(tCredit_card_number) = 16 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
            tContinue = False
          End If
          mCard(0) = "Maestro"
          ' SET FORMAT(pmntType;"#### #### #### ###0")  ' ?

        Case Mid$(tIssuer_identifier_number, 1, 1) = "4"  ' Visa
          If (Len(tCredit_card_number) = 13) Or (Len(tCredit_card_number) = 16) Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 13 or 16 digits."
            tContinue = False
          End If
          mCard(0) = "Visa"
          ' SET FORMAT(pmntType;"#### #### #### ###0")

        Case (Mid$(tIssuer_identifier_number, 1, 4) = "5020") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "5038") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "6304") _
           Or (Mid$(tIssuer_identifier_number, 1, 4) = "6759") ' Maestro debit card.
          Select Case tCredit_card_number
            Case Len(tCredit_card_number) = 16
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0")  ' ?
            Case Len(tCredit_card_number) = 18
              Error -1
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0")  '?
            Case Else
              uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
              tContinue = False
          End Select
          mCard(0) = "Maestro"

        Case (Val(Mid$(tIssuer_identifier_number, 1, 2)) >= 51) And (Val(Mid$(tIssuer_identifier_number, 1, 2)) <= 55) ' MasterCard. 55 is also Diners, U.S. and Canada, being due to an alliance between the two. The details are the same, so no problem.
          If Len(tCredit_card_number) = 16 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
            tContinue = False
          End If
          mCard(0) = "MasterCard"
          ' SET FORMAT(pmntType;"#### #### #### ###0")

        Case ((Val(tIssuer_identifier_number) >= 560221) And (Val(tIssuer_identifier_number) <= 560225) Or (Mid$(tIssuer_identifier_number, 1, 4) = "5610")) ' Bank Card. Inactive from 1988.
          uError.Message = "The BankCard Credit Card is no longer an active valid card!"
          ' If (Length($a)=16)
          ' Luhn_check ($a)
          ' Else 
          ' MyAlert ("The Credit Card number entered is incomplete. It should have 16 digits.")
          tContinue = False
          ' End if 
          ' mCard:=0
          mCard(0) = "BankCard"

        Case (Mid$(tIssuer_identifier_number, 1, 1) = "5") ' Dankort (PBS). To do: Possible clash; that's why included last.
          If Len(tCredit_card_number) = 16 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
            tContinue = False
          End If
          mCard(0) = "Dankort"
          ' SET FORMAT(pmntType;"#### #### #### ###0")  ' ?

        Case (Mid$(tIssuer_identifier_number, 1, 4) = "6011") Or (Mid$(tIssuer_identifier_number, 1, 2) = "65") ' Discover Financial Services, Inc.
          If Len(tCredit_card_number) = 16 Then
            Call Luhn_check(tCredit_card_number)
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 16 digits."
            tContinue = False
          End If
          mCard(0) = "Discover"
          ' SET FORMAT(pmntType;"#### #### #### ###0")

        Case (Val(Mid$(tIssuer_identifier_number, 1, 6)) >= 622126) And (Val(Mid$(tIssuer_identifier_number, 1, 6)) <= 622925) ' China Union Pay.
          ' Validation check unknown. 
          Select Case tCredit_card_number
            Case Len(tCredit_card_number) = 16
              ' SET FORMAT(pmntType;"#### #### #### ###0")
            Case Len(tCredit_card_number) = 17
              ' SET FORMAT(pmntType;"#### #### #### ###0")' ?
            Case Len(tCredit_card_number) = 18
              ' SET FORMAT(pmntType;"#### #### #### ###0")' ?
            Case Len(tCredit_card_number) = 19
              ' SET FORMAT(pmntType;"#### #### #### ###0")' ?
            Case Else
              uError.Message = "The Credit Card number entered is incomplete. It should have between 16 and 19 digits."
              tContinue = False
          End Select
          mCard(0) = "China Union Pay"

        Case (Mid$(tIssuer_identifier_number, 1, 4) = "6334") Or (Mid$(tIssuer_identifier_number, 1, 4) = "6767") ' Solo (Paymentech) debit card.
          Select Case tCredit_card_number
            Case Len(tCredit_card_number) = 16
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0") ' ?
            Case Len(tCredit_card_number) = 18
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0") ' ?
            Case Len(tCredit_card_number) = 19
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0") ' ?
            Case Else
              uError.Message = "The Credit Card number entered is incomplete. It should have 16, 18 or 19 digits."
              tContinue = False
          End Select
          mCard(0) = "Solo"

        Case Mid$(tIssuer_identifier_number, 1, 1) = "7" ' Dankort (PBS).
          If Len(tCredit_card_number) = 11 Then
            ' Luhn_check ($a) ' Doesn't appear to conform to the Luhn check. 
          Else
            uError.Message = "The Credit Card number entered is incomplete. It should have 11  digits."
            tContinue = False
          End If
          mCard(0) = "Dankort"
          ' SET FORMAT(pmntType;"#### #### #### ###0")  ' ?

        Case Mid$(tIssuer_identifier_number, 1, 4) = "8699" ' Voyager.
          Select Case tCredit_card_number
            Case Len(tCredit_card_number) = 15
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0")  
              ' ? Web site currently shows a Visa logo and number format, in 16 digits. 
            Case Len(tCredit_card_number) = 16
              Call Luhn_check(tCredit_card_number)
              ' SET FORMAT(pmntType;"#### #### #### ###0")
            Case Else
              uError.Message = "The Credit Card number entered is incomplete. It should have 15 or 16 digits."
              tContinue = False
          End Select
          mCard(0) = "Voyager"

        Case Mid$(tIssuer_identifier_number, 1, 1) = "9" ' Nationally assigned. 
          ' Quoted from www.merriampark.com/anatomycc.htm#Account:
          ' "If the Major Industry Identifier (MII) digit is 9, then the next three digits of the issuer identifier are the 3-digit country codes defined in ISO 3166, 
          ' and the remaining final two digits of the issuer identifier can be defined by the national standards body of the 
          ' specified country in whatever way it wishes."
          mCard(0) = "Other"

        Case Else
          uError.Message = "The Credit Card number entered is invalid. Check the first six digits, which are the Issuer Identification."
          tContinue = False

      End Select
      If uError.Message <> "" Then
        uOp.Result = Msgbox_LIS(uError.Message, MsgBoxStyle.Critical, udtDialogue_types.Okay_only, frmMain.tabControl_main, udtDialogue_position.Top_left)
      End If

    End If

    ' Ender_full:
    If tContinue Then
      ' [Transactions]Drawer = mCard(0)
    Else
      tStatus = -1
      ' [Transactions]Status=-1
    End If
    ' $0=tContinue 

    uOp.Result = Nothing
  End Sub

  Sub Luhn_check(ByVal tCredit_card_number)
    Stop
    ' To do. 
  End Sub

  Sub Nested_if_01()
    ' Revised 2020.02.01.
    uOp.Subr = udtApplication_subroutine_names.Nested_if_01
    Dim tNest As Short ' Say 1 to 10. 
    Dim tVar1(10) As String ' Could be more than 10. 
    Dim tVar2(10) As String

    If _
       (tNest = 1 And tVar1(1) = tVar2(1)) Or _
       (tNest = 1 And tVar1(1) = tVar2(1)) Or _
       (tNest = 2 And tVar1(2) = tVar2(2)) Or _
       (tNest = 3 And tVar1(3) = tVar2(3)) Or _
       (tNest = 4 And tVar1(4) = tVar2(4)) Or _
       (tNest = 5 And tVar1(5) = tVar2(5)) Or _
       (tNest = 6 And tVar1(6) = tVar2(6)) Or _
       (tNest = 7 And tVar1(7) = tVar2(7)) Or _
       (tNest = 8 And tVar1(8) = tVar2(8)) Or _
       (tNest = 9 And tVar1(9) = tVar2(9)) Or _
       (tNest = 10 And tVar1(10) = tVar2(10)) _
    Then

    End If

    uOp.Subr = ""
  End Sub

  Sub Nested_if_02()
    ' Revised 2020.02.01.
    uOp.Subr = udtApplication_subroutine_names.Nested_if_02
    Dim i As Integer
    Dim tNest As Short ' Say 1 to 10. 
    Dim tNext_max As Short = 10
    Dim tVar1(10) As String ' Could be more than 10. 
    Dim tVar2(10) As String

    For i = 1 To tNext_max
      If (tNest = i And tVar1(i) = tVar2(i)) Then
        ' Desired result. 
        Exit For
      End If
    Next

    uOp.Subr = ""
  End Sub

End Module
