﻿' Module: Geo_location.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

' References:
' System.Device.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings.

Imports System.Runtime.InteropServices
Imports System.Device.Location
' Imports Windows.devices.geolocation

Module Geo_location
  Public Const cGeo_location_version As Date = #3/13/2021# ' M/d/y.

  ' Don't change name:
  <DllImport("kernel32.dll")> Private Function GetLocaleInfo(ByVal Locale As UInteger, ByVal LCType As UInteger, <Out()> ByVal lpLCData As System.Text.StringBuilder, ByVal cchData As Integer) As Integer
  End Function

  Private Const Locale_System_Default As UInteger = &H400
  Private Const Locale_SengCountry As UInteger = &H1002

  Private Function Geo_location_get_location_info_fn(ByVal tInfo As UInteger) As String
    ' Sub-routine added 2018.08.19. Revised 2020.02.02.
    uOp.Subr = udtApplication_subroutine_names.Geo_location_get_location_info_fn
    Dim tLpLCData As System.Text.StringBuilder = New System.Text.StringBuilder(256)
    Dim tRet As Integer = GetLocaleInfo(Locale_System_Default, tInfo, tLpLCData, tLpLCData.Capacity)

    If tRet > 0 Then
      Return tLpLCData.ToString().Substring(0, tRet - 1)
    End If
    Return String.Empty

    uOp.Subr = ""
  End Function

  Sub Geo_location_detect()
    ' Sub-routine added 2018.08.19. Revised 2021.03.04.
    uOp.Subr = udtApplication_subroutine_names.Geo_location_detect
    Stop ' Indefinite.
    If uMachine.Development_machine Then
      GoTo Ender_full
    Else
      Stop ' To do, to develop.
    End If

    uActive.Country = Geo_location_get_location_info_fn(Locale_SengCountry) ' Returns Windows Settings value for Country or region: United Kingdom.
    Call Geo_location_get_location_property() ' Returns "Unknown latitude and longitude."
    uOp.Subr = udtApplication_subroutine_names.Geo_location_detect

Ender_full:
    uOp.Subr = ""
  End Sub

  Private Sub Geo_location_get_location_property()
    ' Sub-routine added 2018.08.19. Revised 2020.03.13.
    ' http://social.msdn.microsoft.com/Forums/en-US/bf17faae-1839-4795-bf7a-815af8822a20/help-to-get-my-location-and-longitude-and-latitude-in-vbnet?forum=vbgeneral.
    ' Geolocation reference required:
    ' Imports System.Device.Location

    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = udtApplication_subroutine_names.Geo_location_get_location_property
    Dim tWatcher As New System.Device.Location.GeoCoordinateWatcher()
    tWatcher.TryStart(False, TimeSpan.FromMilliseconds(1000))
    Dim tCoord As GeoCoordinate = tWatcher.Position.Location

    If tCoord.IsUnknown = True Then
      Console.WriteLine("Unknown latitude and longitude.")
      uOp.Result = "Unknown latitude and longitude."
    Else
      Stop ' Indefinite.
      Console.WriteLine("Lat: {0}, Long: {1}", tCoord.Latitude, tCoord.Longitude)
      uOp.Result = String.Format("Lat: {0}, Long: {1}", tCoord.Latitude, tCoord.Longitude)
    End If
    '  tLocals.Msg  = Console.ReadLine
    Console.ReadLine()

    uOp.Subr = ""
  End Sub

End Module
