' Module: Kernel.
' Copyright  LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

' Template Windows application project started 2009.02.01, Sun.

' References:
' Devart.Data.
' Devart.Data.SQLite.
' Microsoft ActiveX Data Objects 6.1 Library. Used by module Database_Adodb.
' Microsoft ADO Ext. 6.0 for DDL Security.
' Microsoft DAO 3.6 Object Library.           Used by module Database_DAO.
' Microsoft Internet Controls.                Used by Imports SHDocVw in module Web.
' Microsoft Shell Controls and Automation.    Used by shell32?
' System.Device                               Used by Geolocation.

' Imported name-spaces:
' System.DirectoryServices.AccountManagement.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings. 

Module Kernel
  Public Const cKernel_version As Date = #5/2/2021# ' M/d/y.

#Region "Shared partial"

  Sub Kernel_local_pre()
    ' Revised 2021.04.28.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Kernel local pre"
    tLocals.Action = udtuStandard_texts.Initiation

    tLocals.Action = "Set uApp"
    With uApp
      .App_name = udtApplication.App_Template
      ' To do; retest with testing name; something goes wrong in log table creation.
      .Internal_name = Replace(uApp.App_name, " ", "_")
      .Short_name = .App_name
      .Name_for_log = .Short_name
      .Name_for_preferences = .App_name
      .English_name = .App_name

      .App_build_number = 20
      If My.Application.Info.Version.Minor > .App_build_number Then .App_build_number = My.Application.Info.Version.Minor
      If My.Application.Info.Version.Build > .App_build_number Then .App_build_number = My.Application.Info.Version.Build
      tLocals.Msg = My.Application.Info.Version.Revision
      tLocals.Msg = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString
      ''  tLocals.Msg  = System.Deployment.Application.ApplicationDeployment.CurrentDeployment.CurrentVersion
      'Dim tVersion As Object = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version
      ''  tLocals.Msg  = Application.ProductVersion
      '' Dim tCd As System.Deployment.Application.ApplicationDeployment = System.Deployment.Application.ApplicationDeployment.CurrentDeployment
      'If System.Deployment.Application.ApplicationDeployment.IsNetworkDeployed Then
      '  tLocals.Msg = System.Deployment.Application.ApplicationDeployment.CurrentDeployment.CurrentVersion.Major
      '  Dim tVersion3 As Version = System.Deployment.Application.ApplicationDeployment.CurrentDeployment.CurrentVersion
      '  Dim tVersion4 As Version = System.Deployment.Application.ApplicationDeployment.CurrentDeployment.UpdatedVersion
      'End If
      'Dim tVersion2 As Version = System.Reflection.Assembly.GetExecutingAssembly.GetName.Version

      .App_revision_date = #05/09/2021# ' M/d/y.
      If Debugger.IsAttached And (uMachine.Development_machine Or uMachine.Testing_machine) Then
        .Debug = True ' True to get more detail.
      End If
      .Progressbar_steps_total = 42 ' Set specifically for each application as needed.

    End With

    tLocals.Action = "Set gStandard_texts"
    ' Call gStandard_texts.Initialise

    tLocals.Action = "Set uApp.Run_count.Iteration_message(x)"

    tLocals.Action = "Set uUser"
    Call uUser.Initialisation(5)

    tLocals.Action = "Set uMemory common"
    If uMachine.Development_machine Then Call uMemory_ProgFiles.Initialise(uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_common)

    tLocals.Action = "Set per application, as desired"

    tLocals.Action = "Set gMemory (application specific)"
    'If uMachine.Name = uMachine.Development_machine_name And (uUser.First_name = "myk_h" Or Environ("USERNAME") = "myk_h") Then
    ' Call gMemory.Initialise("C:\Users\myk_h\OneDrive\Dept 09, RAM\MJBH database.accdb")
    'End If

    tLocals.Action = "Set gUser uncommon (application specific)"

    tLocals.Action = "Set local table names"
    Call gTables.Initialise()

    tLocals.Action = "Set context sensitive help"
    uCon_help.Brought = uApp.App_name & ", the generic programme, is brought to you by " & udtLivingInfoSystems.Name_short & "." & vbCrLf
    uCon_help.Brought = uCon_help.Brought & "www." & udtLivingInfoSystems.Domain & "."

    tLocals.Action = "Set help uncommon"
    ' Call gCon_help.Initialise

    tLocals.Action = "Set enumerations"

Ender_full:
    uOp.Subr = ""
    uOp.Success = True
  End Sub

  Sub Kernel_local_post()
    ' Revised 2021.03.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Kernel local post"
    On Error GoTo Error_captured
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    If uApp.Upgraded Then
      tLocals.Action = "Call Kernel distribute"
      Call Kernel_distribute()
      uOp.Subr = tLocals.Subr
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Cancelled
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
    End If
    tLocals.Success = udtBoolean_extended.True_

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
     uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Ender_partial:
    Call frmMain.Statuslabel1_set()
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  ' Alphabetical until End_game_local:

  Sub App_revision_upgrade_local(ByRef tApp_revision_last As Date, ByVal tDb As udtDatabase_oledb, ByRef tUpgraded As Boolean)
    ' Separated 2019.10.15. Revised 2021.02.28.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "App_revision upgrade local"
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 7
    Call frmMain.Statuslabel1_set()
    tLocals.Status = ""

    tLocals.Action = "Check versions"
    tLocals.Msg = uApp.App_revision_date
    If uMachine.Development_machine Or uMachine.Testing_machine Then
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    Select Case uApp.App_build_number
      Case Is > 1
        ' Stop ' Indefinite.
        ' ...
        ' if do something then tUpgraded = True
    End Select

    If tApp_revision_last < uApp.App_revision_date Then
      tLocals.Action = "Upgrade data"

      If tDb.Database <> uMemory.Database Then
        Call tDb_point_to_uMemory_fn(tDb)
        uOp.Subr = tLocals.Subr
      End If

    End If ' If tApp_revision_last < uApp.App_revision_date Then

    If uMachine.Development_machine And uApp.Upgraded And tDb.Upgrade_table = "" Then
      Call Kernel_distribute() ' Template only.
      uOp.Subr = tLocals.Subr
    End If

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uApp.Progressbar_steps_difference = uApp.Progressbar_steps_total - uApp.Progressbar_step
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Run_count_check_local()
    ' Written 2018.09.30 by MJBH. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Check_run_count_local
    uOp.Subr = tLocals.Subr

    ' Sample:
    ' If uUser.Start_home_tab =  udtBoolean_extended.Null And uApp.Run_count >= 4 Then
    ' uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1
    ' tLocals.Msg = "Looks like you've run " & uApp.App_name & " a few times." & vbCrLf
    ' tLocals.Msg = tLocals.Msg & "Do you want to start at the Home tab (or the last tab you used)?" & vbCr
    ' tLocals.Msg = tLocals.Msg & "You can set this in Edit menu: " & frmMain.tabSettings.Text & "."
    ' Call Msgbox_LIS_top (tLocals.Msg, udtAlert_types.Information, udtDialogue_types.Yes_no, uOp.Subr)
    ' If uDialogue.Active Then GoTo Ender_full Else Call Check_run_count_local_after_dialogue_tab()
    ' End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Run_count_check_local_after_dialogue_tab()
    ' Separated 2019.02.06. Revised 2020.05.01.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    tLocals.Subr = "Check run count local after dialogue tab"
    Stop ' Temporary.

    'If gUser.New_fp_copy_from_previous = udtBoolean_extended.Null And uApp.Run_count >= 4 Then
    '   If uDialogue.Dialogue_result = DialogResult.OK Then
    '      gUser.New_fp_copy_from_previous = udtBoolean_extended.Off
    '   Else
    '      gUser.New_fp_copy_from_previous = udtBoolean_extended.On_
    '   End If
    '   uUser.Preferences_changed = True
    'End If

    uOp.Subr = ""
    If tLocals.Success <> udtBoolean_extended.False_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Country_detect_from_uMemory_ProgFiles(ByRef tCountry As String, ByRef tName_two_letter As String, ByRef tISOCurrencySymbol As String, ByRef tCurrency_format As String, ByRef tCurrencySymbol As String, ByRef tTelephone_code As String, ByRef tNote As String)
    Call Country_detect_from_uMemory_ProgFiles_2(tCountry, tName_two_letter, tISOCurrencySymbol, tCurrency_format, tCurrencySymbol, tTelephone_code, tNote)
  End Sub

  Sub First_run_local(ByRef tChanged As Boolean)
    ' Revised 2021.03.18.
    Dim tLocals As udtLocals = Nothing
    ' Dim tDb As udtDatabase_oledb = Nothing

    tLocals.Subr = "First run local"
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Success = False

    tLocals.Success = udtBoolean_extended.Null
    tLocals.Steps_to_complete = 1
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 2
    Call frmMain.Statuslabel1_set(tLocals.Subr)

    'Call tDb_point_to_uMemory_fn(tDb)
    'If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
    '  Stop ' Permanent.
    '        tLocals.Error_count   =       tLocals.Error_count   + 1
    '  tLocals.Status = udtuStandard_texts.Failed
    '  tLocals.Success = udtBoolean_extended.False_
    '  GoTo Error_process
    'End If
    uOp.Success = False

    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    tLocals.Action = "Check any databases specific to this programme"
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": " & LCase(tLocals.Action) & "..." & vbCrLf)

    tLocals.Action = udtApplication_subroutine_names.Insert_sample_data
    Call Insert_sample_data(tChanged)
    uOp.Subr = tLocals.Subr
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    tLocals.Steps_completed = tLocals.Steps_completed + 1
    Call frmMain.ToolStripProgressBar1_set()

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set()

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Log_write_local(ByRef tLocals As udtLocals, ByVal tFreefile As Integer)
    ' Revised 2021.03.18.
    Select Case tLocals.Subr
      Case Else
        ' ...
    End Select

  End Sub

  Sub Version_check_sub_01_local(ByRef tChanged As Boolean)
    ' Separated 2019.11.30. Revised 2020.04.22.
    uOp.Subr = "Version check sub 01 local"

    If cDatabase_common_MsSql_version > uApp.App_revision_date Then uApp.App_revision_date = cDatabase_common_MsSql_version : tChanged = True
    If cDatabase_common_MySql_version > uApp.App_revision_date Then uApp.App_revision_date = cDatabase_common_MySql_version : tChanged = True
    If cDatabase_common_SqLite_version > uApp.App_revision_date Then uApp.App_revision_date = cDatabase_common_SqLite_version : tChanged = True

    uOp.Subr = ""
  End Sub

  Sub End_game_local()
    ' Revised 2021.02.08.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "End game local"
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1

Ender_full:
    Call frmMain.ToolStripProgressBar1_set()
    If uOp.Called_from = tLocals.Subr Then uOp.Called_from = ""
    uOp.Subr = ""
  End Sub

#End Region ' End region, "Shared partial."

#Region "Shared optional"

  Function Sentence_case_fn(ByRef tString As String) As String
    ' Revised 2020.08.16.
    uOp.Subr = udtgApplication_subroutine_names.Sentence_case_fn

    Sentence_case_fn = UCase(Left(tString, 1)) & LCase(Mid(tString, 2))
    uOp.Subr = ""
  End Function

  Function Date_format_detect_fn(ByVal tInput As String, ByVal tFor_sql As Boolean, Optional ByVal tConstruct As Boolean = False) As String
    ' Demoted 2019.11.09. Revised 2021.03.18.
    Dim tLocals As udtLocals = Nothing
    Dim tPos1 As Short = 0, tPos2 As Short = 0
    Dim tSeparator As String = ""
    Dim tDate_input As Date
    Dim tDate_sample_str As String
    Dim tYear_sample As String, tMonth_sample As String, tDay_sample As String
    Dim tDate_sample As Date
    Dim tYear_output As String, tMonth_output As String, tDay_output As String
    tLocals.Subr = "Date format function"
    On Error GoTo Error_captured
    Date_format_detect_fn = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    If InStr(tInput, ".") <> 0 Then
      tSeparator = "."
    ElseIf InStr(tInput, "/") <> 0 Then
      tSeparator = "/"
    ElseIf InStr(tInput, "-") <> 0 Then
      tSeparator = "-"
    ElseIf InStr(tInput, " ") <> 0 Then
      tSeparator = " "
    End If
    If tSeparator = "" Then
      Stop ' Permanent.
    End If

    tDate_input = DateAndTime.DateValue(tInput)
    For i As Short = 1 To Len(tDate_input)
      tLocals.Msg = Mid(tInput, i, 1)
      If Mid(tInput, i, 1) = tSeparator Then
        If tPos1 = 0 Then
          tPos1 = i
        Else
          tPos2 = i
          Exit For
        End If
      End If
    Next
    If tPos1 = tPos2 Or tPos2 = 0 Then
      Stop ' Permanent.
    End If
    If tFor_sql Then tSeparator = "-"

    tDate_sample_str = "01.02.03"
    tLocals.Msg = DateAndTime.DateValue(tDate_sample_str)
    tDate_sample = DateAndTime.DateValue(tDate_sample_str) ' M.d.y.
    tYear_sample = Year(tDate_sample)
    tMonth_sample = Month(tDate_sample)
    tDay_sample = DateAndTime.Day(tDate_sample)

    Select Case tYear_sample
      Case 2001
        Select Case tMonth_sample
          Case 2
            ' ISO standard, year-month-day. 
            tYear_output = Left(tDate_input, tPos1 - 1)
            tMonth_output = Mid(tDate_input, tPos1 + 1, tPos2 - tPos1 - 1)
            tDay_output = Mid(tDate_input, tPos2 + 1)
            Date_format_detect_fn = tYear_output & tSeparator & tMonth_output & tSeparator & tDay_output
          Case 3
            Stop ' Permanent.
          Case Else
            Stop ' Permanent.
        End Select

      Case 2003
        Select Case tMonth_sample
          Case 1
            Stop
            ' Should be month-day-year. 
            tYear_output = Left(tDate_input, tPos1 - 1)
            tMonth_output = Mid(tDate_input, tPos1 + 1, tPos2 - tPos1 - 1)
            tDay_output = Mid(tDate_input, tPos2 + 1)
            Date_format_detect_fn = tYear_output & tSeparator & tMonth_output & tSeparator & tDay_output

          Case 2
            ' Day-month-year. 
            tYear_output = Mid(tDate_input, tPos2 + 1) ' 2012.03.28. 
            tMonth_output = Mid(tDate_input, tPos1 + 1, tPos2 - tPos1 - 1)
            tDay_output = Left(tDate_input, tPos1 - 1) ' 2012.03.28. 
            Date_format_detect_fn = tDay_output & tSeparator & tMonth_output & tSeparator & tYear_output
            If tFor_sql Then Date_format_detect_fn = tMonth_output & tSeparator & tDay_output & tSeparator & tYear_output

          Case Else
            Stop ' Permanent.
        End Select
      Case Else
        Stop ' Permanent.
    End Select

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

    uOp.Subr = ""
  End Function

  Function Find_in_array_fn(tSource_array() As String, tKeyword As String) As Integer
    ' Findinarray, find in array.
    ' Revised 2019.03.26. Copied from Excel 2021.04.29.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    Dim i As Integer
    Dim tArray() As String
    tArray = Filter(tSource_array, tKeyword, True, vbTextCompare)

    For i = 0 To UBound(tSource_array)
      If tSource_array(i) = tKeyword Then
        If uApp.Debug Then tLocals.Msg = i
        Find_in_array_fn = i
        Exit For
      End If
    Next

  End Function

  Sub Date_detect_system_format(ByRef tNext As Boolean)
    ' Written MJBH 2019.10.19. Demoted 2020.03.31. Revised 2021.03.21.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Detect system date format"

    Call Culture_detect()

    Stop ' To develop.

    If tNext Then
      If tLocals.Error_count > 0 Then tLocals.Error_count = tLocals.Error_count - 1
      Select Case LCase(uMachine.Date_format)
        Case "dd.mm.yyyy" ' Fail.
          uMachine.Date_format = "dd-mm-yyyy" ' Fail.

        Case "dd-mm-yyyy"
          uMachine.Date_format = "dd/mm/yyyy" ' Fail.

        Case "dd/mm/yyyy"
          uMachine.Date_format = "d.m.yyyy" ' Fail.

        Case "d.m.yyyy"
          uMachine.Date_format = "d-m-yyyy" ' Fail.

        Case "d-m-yyyy"
          uMachine.Date_format = "d/m/yyyy" ' Fail.

        Case "d/m/yyyy"
          uMachine.Date_format = "m.d.yyyy" ' Fail.

        Case "m.d.yyyy"
          uMachine.Date_format = "m-d-yyyy" ' Fail.

        Case "m-d-yyyy"
          uMachine.Date_format = "m/d/yyyy" ' Fail.

        Case "m/d/yyyy"
          uMachine.Date_format = "mm.dd.yyyy" ' Fail.

        Case "mm.dd.yyyy"
          uMachine.Date_format = "mm-dd-yyyy" ' Fail.

        Case "mm-dd-yyyy"
          uMachine.Date_format = "mm/dd/yyyy" ' Fail.

        Case "mm/dd/yyyy"
          uMachine.Date_format = "yyyy.mm.dd" ' Fail. 

        Case "yyyy.mm.dd"
          uMachine.Date_format = "yyyy-mm-dd" ' Fail. 

        Case "yyyy-mm-dd"
          uMachine.Date_format = "yyyy/mm/dd" ' Fail. 

        Case "yyyy/mm/dd"
          Stop ' Permanent. End of line.
          uMachine.Date_format = ""

        Case Else
          Stop ' Permanent.
          uMachine.Date_format = ""

      End Select
    Else
      Select Case System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern
        Case "dd.mm.yyyy"
          uMachine.Date_format = "yyyy.mm.dd"
        Case "dd/MM/yyyy"
          uMachine.Date_format = "yyyy.mm.dd"
          uMachine.Date_format = "dd.mm.yyyy" ' Fail.
        Case "M/d/yyyy"
          uMachine.Date_format = "mm/dd/yyyy" ' Works.
        Case Else
          Stop ' Permanent.
          uMachine.Date_format = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern
      End Select
    End If
    tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
  End Sub

#End Region ' End region, "Shared optional."

#Region "Unshared"

  Sub Kernel_distribute()
    ' Separated 2014.04.28. Revised 2021.04.07.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    If uMachine.Name <> uMachine.Development_machine_name Then
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_partial
    End If
    If gOp.Kernel_distributed Then tLocals.Success = udtBoolean_extended.True_ : GoTo Ender_partial
    If Not uApp.Debug Then GoTo Ender_partial
    tLocals.Subr = "Kernel distribute"
    On Error GoTo Error_captured
    uOp.Subr = tLocals.Subr

    tLocals.Action = udtuStandard_texts.Main_start
    ' Difference before 38.
    If uOp.Form_initialising Then
      uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 154
    Else
      uApp.Progressbar_show = True
      uApp.Progressbar_step = 0
      uApp.Progressbar_steps_total = 154
    End If
    Call frmMain.Statuslabel1_set()

    tLocals.Action = "Copy common files to other applications"
    tFile.File_name = "Declarations_common.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "Kernel_common.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "Database_common.vb"
    Call Kernel_distribute_sub_01(tFile)

    'tFile.File_name = "Geo_location.vb" ' Demoted 2019.10.16.
    '    Call Kernel_distribute_sub_01(tFile) 

    tFile.File_name = "Ftp_common.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "Web_common.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "clsEncrypt.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "Crypt_rc4.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "Email_Smtp.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "frmDialogue.Designer.vb"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "frmDialogue.resx"
    Call Kernel_distribute_sub_01(tFile)

    tFile.File_name = "frmDialogue.vb"
    Call Kernel_distribute_sub_01(tFile)

    '      tFile.File_name  = "Devart.Data.dll"
    '    Call Kernel_distribute_sub_01(tFile) 

    '      tFile.File_name  = "Devart.Data.SQLite.dll"
    '    Call Kernel_distribute_sub_01(tFile) 

    tLocals.Action = "Resources"
    For Each tFile.File_full_name In My.Computer.FileSystem.GetFiles(uLIS.Development_path & "\Dept 11A-2, Application development\Template\Template\Resources\")
      If File_properties_determine_fn(tFile) Then Call Kernel_distribute_sub_01(tFile, True)
    Next

    Call Preferences_common_write()
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    gOp.Kernel_distributed = True

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            Call Error_clear(tLocals)
            Resume
          Case 5 ' Access to the path 'C:\...' is denied.
            Call Sound_alert()
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
      tLocals.Action = tLocals.Status
    End If
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 2)
    uOp.Subr = tLocals.Subr
    frmMain.txtKernel_distribute.Text = Date.Today & ", " & DateAndTime.TimeOfDay

Ender_partial:
    If uOp.Form_initialising Then
      ' Dim tuApp_Startup_steps_total_after As Integer = uApp.Progressbar_steps_total
      '  tLocals.Msg  = tuApp_Startup_steps_total_after - tuApp_Startup_steps_total_before
      ' Stop
    Else
      uApp.Progressbar_show = False
      uApp.Progressbar_step = 0
      uApp.Progressbar_steps_total = 0
    End If
    Call frmMain.ToolStripProgressBar1_set()
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Kernel_distribute_sub_01(ByRef tFile As udtFile, Optional ByVal tResources As Boolean = False)
    ' Separated 2014.06.01. Revised 2021.04.06.
    Dim tLocals As udtLocals = Nothing

    Dim tApp_dev As String = ""
    Dim tClient_dev As String = ""
    Dim tTemplate_path As String = ""

    Dim tSeer_loc As String = ""
    Dim tLIS_server_loc As String = ""
    Dim tFinancier_loc As String = ""
    Dim tFinancier_commercial_loc As String = ""
    Dim tImmersion_loc As String = ""
    Dim tContacts_merge As String = ""

    tApp_dev = uLIS.Development_path & "\Dept 11A-2, Application development"
    tClient_dev = uLIS.Development_path & "\Dept 11A-5, Clients' software\Clients"
    tTemplate_path = tApp_dev & "\Template\Template\"

    tSeer_loc = tApp_dev & "\Seer\Windows application\Seer\Seer\"
    tLIS_server_loc = tApp_dev & "\LIS server\LIS server\LIS server\"
    tFinancier_loc = tApp_dev & "\Financier\Financier\Financier\"
    tFinancier_commercial_loc = tApp_dev & "\Financier Commercial Edition\Financier Commercial Edition\Financier Commercial Edition\"
    tImmersion_loc = tClient_dev & "\Immersion Studios\Immersion r.p.g. character sheet, VB\Immersion\"
    tContacts_merge = tApp_dev & "\Contacts merge\Contacts_merge\"

    tFile.Copy_source = tTemplate_path & tFile.File_name
    If tResources Then tFile.Copy_source = tTemplate_path & "Resources\" & tFile.File_name
    ' If InStr(      tFile.File_name , "Devart") > 0 Then tFile.Copy_src = tTemplate_path & "SqLite\" &       tFile.File_name 

    Call Kernel_distribute_sub_02(tFile, tLIS_server_loc, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call Kernel_distribute_sub_02(tFile, tSeer_loc, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call Kernel_distribute_sub_02(tFile, tFinancier_loc, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call Kernel_distribute_sub_02(tFile, tFinancier_commercial_loc, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call Kernel_distribute_sub_02(tFile, tImmersion_loc, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call Kernel_distribute_sub_02(tFile, tContacts_merge, tResources)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Cancelled
      tLocals.Success = udtBoolean_extended.False_
    End If
    uOp.Success = False

    Call frmMain.ToolStripProgressBar1_set()

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Kernel_distribute_sub_02(ByRef tFile As udtFile, ByVal tPath_dest As String, ByVal tResources As Boolean)
    ' Revised 2021.04.07.
    Dim tLocals As udtLocals = Nothing

    If tResources Then tPath_dest = tPath_dest & "Resources\"
    If Dir(tPath_dest, FileAttribute.Directory) = "" Then
      Stop ' Permanent.
      System.IO.Directory.CreateDirectory(tPath_dest)
    End If

    tFile.Copy_destination = tPath_dest & tFile.File_name
    If Dir(tFile.Copy_source, FileAttribute.Normal) = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    Else
      FileCopy(tFile.Copy_source, tFile.Copy_destination)
      tLocals.Success = udtBoolean_extended.True_
    End If
    tFile.Copy_destination = ""

    Call frmMain.ToolStripProgressBar1_set()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Dialogue_behaviour_testing()
    ' Promoted 2019.12.08. Revised 2021.03.21.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Dialogue behaviour testing"
    Stop ' Indefinite.

    tLocals.Action = "2019.12.05, dialogue behaviour testing"
    tLocals.Response = Nothing
    tLocals.Response = Msgbox_LIS("Hello world.", udtAlert_types.Information, udtDialogue_types.Okay_only, frmMain.txtHome_description, udtDialogue_position.Top_left)

    If frmDialogue.Visible Or uDialogue.Active Then

    Else
      Stop ' To do, to develop:
      tLocals.Action = "2019.12.05, dialogue behaviour testing"
      tLocals.Response = Nothing
      tLocals.Response = Inputbox_LIS("Hello world.", "Hello back.", frmMain.txtHome_description, udtDialogue_position.Top_left)

      If frmDialogue.Visible Or uDialogue.Active Then

      Else
        Stop ' To do, to develop:
        tLocals.Action = "2019.12.05, dialogue behaviour testing"
        tLocals.Response = Nothing
        frmDialogue.cboDialogue_response.Items.Add("1")
        frmDialogue.cboDialogue_response.Items.Add("2")
        tLocals.Response = Input_chooser_LIS("Hello world.", frmMain.txtHome_description, udtDialogue_position.Top_left)

        If frmDialogue.Visible Or uDialogue.Active Then

        Else
          tLocals.Response = Nothing
          tLocals.Action = "2019.12.05, dialogue behaviour testing"
          If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
             uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
            tLocals.Success = udtBoolean_extended.False_
          Else
            tLocals.Success = udtBoolean_extended.True_
          End If
          Call Log_write(tLocals, 0)
        End If
      End If
    End If

    Stop
    tLocals.Action = "Test TopMost = True"
    uDialogue.Alert = udtAlert_types.Information
    uDialogue.Dialogue_type = udtDialogue_types.Okay_only
    uDialogue.Parent = frmMain.tabHome
    uDialogue.Prompt = "Hello world"
    uDialogue.Position = udtDialogue_position.Top_left
    Dim tDialogue As New frmDialogue
    tDialogue.Enabled = True
    tDialogue.IsMdiContainer = False
    tDialogue.TopLevel = False
    tDialogue.Visible = True
    tDialogue.Parent.Controls.Add(tDialogue)
    ' tDialogue.TopLevel = True

    uOp.Subr = ""
  End Sub

#End Region ' End region, "Unshared."

End Module
