﻿' Module: Mp3.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Revised 2021.04.07.

Option Explicit On
Imports System.IO
' Imports System.Reflection ' Used by tabSystem_information.

Module Mp3

#Region "MP3_Id3_v1"

  Public Class MP3_Id3_v1
    ' www.codeproject.com/Articles/5590/Reading-and-writing-MP3-ID3v1-tags.

    Public Sub New(Optional ByVal tFilename As String = "")
      MyBase.New()
      uOp.Subr = "MP3_Id3_v1 constructor"
      If tFilename <> "" Then Me.File_name = tFilename
      uOp.Subr = ""
    End Sub

    Public Enum Genres As Byte
      Blues = 0
      ClassicRock = 1
      Country = 2
      Dance = 3
      Disco = 4
      Funk = 5
      Grunge = 6
      HipHop = 7
      Jazz = 8
      Metal = 9
      NewAge = 10
      Oldies = 11
      Other = 12
      Pop = 13
      RnB = 14
      Rap = 15
      Reggae = 16
      Rock = 17
      Techno = 18
      Industrial = 19
      Alternative = 20
      Ska = 21
      DeathMetal = 22
      Pranks = 23
      Soundtrack = 24
      EuroTechno = 25
      Ambient = 26
      TripHop = 27
      Vocal = 28
      JazzFunk = 29
      Fusion = 30
      Trance = 31
      Classical = 32
      Instrumental = 33
      Acid = 34
      House = 35
      Game = 36
      SoundClip = 37
      Gospel = 38
      Noise = 39
      AlternRock = 40
      Bass = 41
      Soul = 42
      Punk = 43
      Space = 44
      Meditative = 45
      InstrumentalPop = 46
      InstrumentalRock = 47
      Ethnic = 48
      Gothic = 49
      Darkwave = 50
      TechnoIndustrial = 51
      Electronic = 52
      PopFolk = 53
      Eurodance = 54
      Dream = 55
      SouthernRock = 56
      Comedy = 57
      Cult = 58
      Gangsta = 59
      Top40 = 60
      ChristianRap = 61
      PopFunk = 62
      Jungle = 63
      NativeAmerican = 64
      Cabaret = 65
      NewWave = 66
      Psychadelic = 67
      Rave = 68
      Showtunes = 69
      Trailer = 70
      LoFi = 71
      Tribal = 72
      AcidPunk = 73
      AcidJazz = 74
      Polka = 75
      Retro = 76
      Musical = 77
      RocknRoll = 78
      HardRock = 79
      None = 255
    End Enum

    Public Enum Frame_types As Byte
      ' Revised 2019.01.05.
      Title = 0
      Artist = 1
      Album = 2
      Year = 3
      Track = 4
      Comment = 5
      Genre = 6
    End Enum

    Private mstrFilename As String
    Public Property File_name() As String
      ' Revised 2019.01.05.
      Get
        Return mstrFilename
      End Get
      Set(ByVal tValue As String)
        If System.IO.File.Exists(tValue) Then
          mstrFilename = tValue
          Call Refresh()
        Else
          Throw New System.IO.FileLoadException("The specified file does not exist.", tValue)
        End If
      End Set
    End Property

    Private mblnTagExists As Boolean
    Public ReadOnly Property Tag_exists() As Boolean
      ' Revised 2021.02.17.
      Get
        '       tLocals.Action  = "Check if tags exist"
        Return mblnTagExists
      End Get
    End Property

    Private mobjFrame(7) As Object
    Public Property Frame(ByVal Frame_type As Frame_types)
      ' Revised 2021.02.17.
      Get
        Return Trim(mobjFrame(Frame_type))
      End Get
      Set(ByVal Value)
        ' tLocals.Action = "Set tag"
        mobjFrame(Frame_type) = Value
      End Set
    End Property

    Public Sub Refresh()
      ' Revised 2021.04.06.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
      Dim tFile As udtFile = Nothing

      Dim strTag As New String(" ", 3)
      Dim strTitle As New String(" ", 30)
      Dim strArtist As New String(" ", 30)
      Dim strAlbum As New String(" ", 30)
      Dim strYear As New String(" ", 4)
      Dim strComment As New String(" ", 28)
      Dim bytDummy As Byte
      Dim bytTrack As Byte
      Dim bytGenre As Byte
      tLocals.Action = "Refresh (gets all tags from the specified file)"
      ' Declarations:
      On Error GoTo Error_captured
      tLocals.Action = "Open the file"
      Dim tFreefile As Integer = FreeFile()
      tFile.Open_file = mstrFilename
      FileOpen(tFreefile, mstrFilename, OpenMode.Binary, OpenAccess.Read, OpenShare.LockWrite)

      tLocals.Action = "Get length of file"
      Dim lngLOF As Long = LOF(tFreefile)
      If lngLOF > 128 Then
        tLocals.Action = "Check for the ID3v1 tag"
        FileGet(tFreefile, strTag, lngLOF - 127, True)
        If strTag.ToUpper = "TAG" Then
          ' Id3v1 tag found:
          mblnTagExists = True

          ' Read all frames from the file
          FileGet(tFreefile, strTitle)
          FileGet(tFreefile, strArtist)
          FileGet(tFreefile, strAlbum)
          FileGet(tFreefile, strYear)
          FileGet(tFreefile, strComment)
          FileGet(tFreefile, bytDummy)
          FileGet(tFreefile, bytTrack)
          FileGet(tFreefile, bytGenre)

          ' Assign the frame content to the properties:
          mobjFrame(0) = strTitle
          mobjFrame(1) = strArtist
          mobjFrame(2) = strAlbum
          mobjFrame(3) = strYear
          mobjFrame(4) = bytTrack
          mobjFrame(5) = strComment
          mobjFrame(6) = bytGenre
        Else
          tLocals.Action = "No Id3v1 tag found"
          mblnTagExists = False
          mobjFrame(0) = ""
          mobjFrame(1) = ""
          mobjFrame(2) = ""
          mobjFrame(3) = ""
          mobjFrame(4) = 0
          mobjFrame(5) = ""
          mobjFrame(6) = 0
        End If
      End If

      tLocals.Action = "Close the file"
      FileClose(tFreefile)
      tFile.Open_file = ""

Error_captured:
      tLocals.Error_number = Err.Number
      tLocals.Error_description = Err.Description
      tLocals.Error_source = Err.Source
      If Err.Number <> 0 Then
        Stop ' Indefinite.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
        tLocals.Success = udtBoolean_extended.False_

        uError.Count = uError.Count + 1
        If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

        If uApp.Terminating = 0 Then
          Select Case tLocals.Error_number
            Case Else
              Stop ' Permanent.
              Call Error_display(tLocals)
          End Select
        End If
      End If

      If uOp.Called_from = uOp.Subr Then uOp.Called_from = ""
      uOp.Subr = ""
    End Sub

    Public Sub Update()
      ' Revised 2021.04.06.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
      Dim tFile As udtFile

      Dim strTag As New String(" ", 3)
      Dim strTitle As New String(" ", 30)
      Dim strArtist As New String(" ", 30)
      Dim strAlbum As New String(" ", 30)
      Dim strYear As New String(" ", 4)
      Dim strComment As New String(" ", 28)
      Dim bytDummy As Byte
      Dim bytTrack As Byte
      Dim bytGenre As Byte
      tLocals.Subr = "MP3_Id3_v1 update"
      tLocals.Action = "Set declarations"
      tLocals.Success = udtBoolean_extended.Null
      Call Error_clear(tLocals)

      Dim tFreefile As Integer = FreeFile()
      Try
        tLocals.Action = "Open the file"
        tFile.Open_file = mstrFilename
        FileOpen(tFreefile, mstrFilename, OpenMode.Binary, OpenAccess.ReadWrite, OpenShare.LockWrite)

        tLocals.Action = "Get length of file"
        Dim lngLOF As Long = LOF(tFreefile)
        If lngLOF > 128 Then
          tLocals.Action = "Check for an existing Id3v1 tag"
          FileSystem.FileGet(tFreefile, strTag, lngLOF - 127)
          If strTag.ToUpper <> "TAG" Then
            tLocals.Action = "No Id3v1 tag found, so just add one"
            FileSystem.Seek(tFreefile, lngLOF)
            strTag = "TAG"
            FileSystem.FilePut(tFreefile, strTag)
          End If

          tLocals.Action = "Fix the length of the frames"
          strTitle = Strings.LSet(gMusic.Track_title, Len(strTitle))
          strArtist = LSet(gMusic.Track_artist, Len(strArtist))
          strAlbum = LSet(gMusic.Track_album, Len(strAlbum))
          strYear = LSet(gMusic.Track_year, Len(strYear))
          bytTrack = Val(gMusic.Track_number)
          strComment = LSet(gMusic.Track_comment, Len(strComment))
          ' bytGenre = LSet(gMusic.Track_genre, Len(bytGenre))
          bytGenre = mobjFrame(6)

          tLocals.Action = "Write the frames to the file"
          FilePut(tFreefile, strTitle)
          FilePut(tFreefile, strArtist)
          FilePut(tFreefile, strAlbum)
          FilePut(tFreefile, strYear)
          FilePut(tFreefile, strComment)
          FilePut(tFreefile, bytDummy)
          FilePut(tFreefile, bytTrack)
          FilePut(tFreefile, bytGenre)

        End If
        tLocals.Success = udtBoolean_extended.True_

      Catch ex As Exception
        Stop ' Permanent.
        tLocals.Error_description = ex.Message
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process
      End Try
      tLocals.Action = "Close the file"
      FileClose(tFreefile)
      tFile.Open_file = ""

Ender_full:
      If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
      Else
        tLocals.Status = udtuStandard_texts.Completion
      End If
      If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
         uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        tLocals.Success = udtBoolean_extended.False_
      Else
        tLocals.Success = udtBoolean_extended.True_
      End If
      Call Log_write(tLocals, 0)
      uOp.Success = True
      uOp.Subr = ""
    End Sub

  End Class

#End Region ' End region, "MP3_Id3_v1."

#Region "Mp3_id3_v2"
  ' http://vbcity.com/forums/t/147802.aspx.

  Public Class mfcReadID3v2Tag

    Private Structure mcsFrame
      Dim Tag As String
      Dim Flag As Integer
      Dim Size As Integer
      Dim Data1 As Byte()
      Dim Data2 As String
    End Structure

    Public Shared Function Read_id3_v2_tag(ByVal sFilename As String) As mcTrack
      ' Revised 2021.04.07.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
      Dim tFile As udtFile = Nothing

      uOp.Subr = udtApplication_subroutine_names.Read_id3_v2_tag

      Try
        Dim tFile_stream As FileStream
        Dim tHeader_byte(15) As Byte
        Dim tmcTrack As New mcTrack

        ' Load object with file path:
        tmcTrack.Filename = sFilename
        tFile_stream = New FileStream(sFilename, FileMode.Open)
        tFile.Open_file = sFilename

        ' Get first 10 bytes, which holds the header tag, size and version:
        tFile_stream.Read(tHeader_byte, 0, 10)

        If (tHeader_byte(0) = &H49) And (tHeader_byte(1) = &H44) And (tHeader_byte(2) = &H33) Then ' It's a proper mp3.
          Dim oFrame As mcsFrame

          ' Start reading tags. I only use 5. There are more!
          oFrame = GetFrame(tFile_stream)
          Do While Not IsNothing(oFrame.Tag)
            Select Case oFrame.Tag
              Case "TIT2" : tmcTrack.Track = oFrame.Data2    ' Song title.
              Case "TPE1" : tmcTrack.Artist = oFrame.Data2   ' Artist.
              Case "TALB" : tmcTrack.Album = oFrame.Data2    ' Album.
              Case "TRCK" : tmcTrack.CdTrack = CInt(oFrame.Data2.Replace("/", "")) ' Cd track number.
              Case "TCON" : tmcTrack.Genre = oFrame.Data2    ' Genre description.
            End Select

            oFrame = New mcsFrame()
            oFrame = GetFrame(tFile_stream)
          Loop
        End If
        Read_id3_v2_tag = tmcTrack

        tFile_stream.Close()
        tFile.Open_file = ""

      Catch ex As Exception
        Stop ' Permanent.
        tLocals.Error_description = ex.Message
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process

        Read_id3_v2_tag = Nothing
      End Try

    End Function

    Public Function Read_id3_v2_tag_v2(ByVal sFile As String) As mcTrack
      ' Version 2. Revised 2021.03.24.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

      Try
        ' MP3 specification:  
        ' --------------------------------------------------  
        ' Field Name        ID3v1   ID3v1.1 ID3v2.2 ID3v2.3     ID3v2.4     Lyrics3v1   Lyrics3v2  
        ' Song Title        title   title   TT2     TIT2        TIT2        --          ETT   
        ' Lead Artist       artist  artist  TP1     TPE1        TPE1        --          EAR  
        ' Album Title       album   album   TAL     TALB        TALB        --          EAL  
        ' Year Released     year    year    TYE     TYER        TDRC        --          --  
        ' Comment           comment comment COM     COMM        COMM        --          INF  
        ' Song Genre        genre   genre   TCO     TCON        TCON        --          --  
        ' Track number      --      track   TRK     TRCK        TRCK        --          --  
        ' Lyrics            --      --      SYL/ULT SYLT/USLT   SYLT/USLT   lyric       LYR  
        ' Auth/Composer     --      --      TCM     TCOM        TCOM        --          AUT  
        ' Popularimeter                             POPM  

        ' POPM frame in the ID3v2 specification - most software all map roughly the same ranges of 0–255 to a 0–5 stars value for display.  
        ' This chart details how windows explorer reads and writes the POPM frame:  
        '    224-255 = 5 stars when READ with windows explorer, writes 255  
        '    160-223 = 4 stars when READ with windows explorer, writes 196  
        '    096-159 = 3 stars when READ with windows explorer, writes 128  
        '    032-095 = 2 stars when READ with windows explorer, writes 64  
        '    001-031 = 1 stars when READ with windows explorer, writes 1  
        ' --------------------------------------------------  
        Dim oFS As FileStream
        Dim baHeader As Byte()
        Dim oTrack As New mcTrack ' !

        ' Load object with file path:
        oTrack.Filename = sFile
        oTrack.Title = sFile.Substring(sFile.LastIndexOf("\"c) + 1) ' Default track to file name, in case nothing is picked up below.  

        oFS = New FileStream(sFile, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)

        ' Get first 10 bytes, which holds the header tag, size and version.  
        ' The ID3v2 tag header, which should be the first information in the file, is 10 bytes as follows:  
        ' ID3v2/file identifier   "ID3"             - 3 bytes  
        ' ID3v2 version           $03 00            - 2 bytes  
        ' ID3v2 flags             %abc00000         - 1 byte  
        ' ID3v2 size              4 * %0xxxxxxx     - 4 bytes  
        ReDim baHeader(2)
        oFS.Read(baHeader, 0, 3) ' Read mp3 header information.
        If (baHeader(0) = &H49) And (baHeader(1) = &H44) And (baHeader(2) = &H33) Then ' It's a proper mp3.
          ReDim baHeader(1)
          oFS.Read(baHeader, 0, 2)

          If baHeader(0) = 2 And baHeader(1) = 3 Then ' ID3v2.3.0.

          End If

          ReDim baHeader(5)
          oFS.Read(baHeader, 0, 5) ' Not currently using flags or size.  

          Dim oFrame As mcsFrame
          ' Start reading tags:
          oFrame = GetFrame_v2(oFS)
          If oFrame.Data2.Length > 255 Then oFrame.Data2 = oFrame.Data2.Substring(0, 255)

          Do While Not IsNothing(oFrame.Tag)
            Select Case oFrame.Tag
              Case "TIT2", "TT2" : oTrack.Title = oFrame.Data2  ' Song title.
              Case "TPE1", "TP1" : oTrack.Artist = oFrame.Data2 ' Artist.
              Case "TALB", "TAL" : oTrack.Album = oFrame.Data2  ' Album.
              Case "TRCK", "TRK"
                Dim sCd As String = oFrame.Data2
                sCd = sCd.Replace("/", "").Replace("?", "")
                If sCd.Length > 2 Then sCd = sCd.Substring(0, sCd.Length - 2)
                ' If IsNumeric(sCD) Then oTrack.TrackNumber = CInt(sCD) ' Cd track number.
                If IsNumeric(sCd) Then oTrack.CdTrack = CInt(sCd) ' Cd track number.
              Case "TCON", "TCO" : oTrack.Genre = oFrame.Data2     ' Genre description.  
              Case "POPM"
            End Select

            oFrame = GetFrame_v2(oFS)
            If IsNothing(oFrame) Then
              ' goLog.Post("Frame Clip", sFile)
              Stop ' To check, develop.
            End If
          Loop
        End If

        oFS.Close()

        Return oTrack

      Catch ex As Exception ' Only log on errors:
        Stop ' Permanent.
        tLocals.Error_description = ex.Message
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process

        ' goLog.StartProc("Metallisoft.MP3Player.Audio.ReadID3v2Tag(" & sFile & ")")
        ' goLog.StopProc("", True, oError, False)
        ' If gbIsIDE Then Stop
        uError.Message = ex.Message
        Return Nothing
      End Try

      uOp.Subr = ""
    End Function

    Private Shared Function GetFrame(ByVal oFile As Stream) As mcsFrame
      ' Revised 2021.03.24.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
      uOp.Subr = "GetFrame"

      Try
        Dim oFrame As New mcsFrame()
        Dim baFrame() As Byte
        Dim oEncoding As New System.Text.ASCIIEncoding()
        ReDim baFrame(5)

        ' Pull frame name:
        oFile.Read(baFrame, 0, 4)
        oFrame.Tag = oEncoding.GetString(baFrame)

        If baFrame(0) <> 0 Then
          oFrame.Tag = oFrame.Tag.Substring(0, 4).Trim

          ' Get 4 bytes for frame size:
          oFile.Read(baFrame, 0, 4)
          oFrame.Size = (65536 * (baFrame(0) * 256 + baFrame(1))) + (baFrame(2) * 256 + baFrame(3))

          ' Skip padding
          oFile.Read(baFrame, 0, 3)

          If oFrame.Size > 0 Then
            ReDim baFrame(oFrame.Size + 1)

            oFile.Read(baFrame, 0, oFrame.Size - 1)
            oFrame.Data1 = baFrame

            If oFrame.Tag.Substring(0, 1) = "T" Then
              oFrame.Data2 = oEncoding.GetString(baFrame).Trim.Replace(Chr(0), "")
            End If
          End If

          GetFrame = oFrame
        Else
          GetFrame = Nothing
        End If

      Catch ex As Exception
        Stop ' Permanent.
        tLocals.Error_description = ex.Message
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process
        GetFrame = Nothing
      End Try

      uOp.Subr = ""
    End Function

    Private Function GetFrame_v2(ByVal oFile As Stream) As mcsFrame
      ' Revised 2021.03.24.
      Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
      uOp.Subr = "GetFrame_v2"

      Try
        Dim oFrame As New mcsFrame
        Dim baFrame As Byte()
        Dim oEncoding As New System.Text.ASCIIEncoding()
        ReDim baFrame(4)

        ' Pull frame name:
        oFile.Read(baFrame, 0, 4)
        oFrame.Tag = oEncoding.GetString(baFrame)

        If baFrame(0) <> 0 Then
          oFrame.Tag = oFrame.Tag.Substring(0, 4).Trim.Replace(Chr(0), "")
          If Not oFrame.Tag.Contains("?") Then
            ' Get 4 bytes for frame size:
            oFile.Read(baFrame, 0, 4)
            oFrame.Size = (65536 * (baFrame(0) * 256 + baFrame(1))) + (baFrame(2) * 256 + baFrame(3))

            ' Skip padding:
            oFile.Read(baFrame, 0, 3)

            If oFrame.Size > 0 Then
              ReDim baFrame(oFrame.Size + 1)
              oFile.Read(baFrame, 0, oFrame.Size - 1)

              If oFrame.Tag.Substring(0, 1) = "T" Then
                oFrame.Data1 = baFrame
                oFrame.Data2 = oEncoding.GetString(baFrame).Trim.Replace(Chr(0), "")
              End If
            End If
          End If
          Return oFrame
        Else
          Return Nothing
        End If

      Catch ex As Exception ' Only log on errors:
        Stop ' Permanent.
        tLocals.Error_description = ex.Message
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process

        ' goLog.StartProc("Metallisoft.MP3Player.Audio.GetFrame()")
        ' goLog.StopProc("", True, oError, False)
        ': If gbIsIDE Then Stop  
        Return Nothing
      End Try

      uOp.Subr = ""
    End Function

  End Class

  Public Class mcTrack

#Region "Public Declarations"
    Private psAlbum As String = ""           ' dbo.Album.Album_Name.
    Private piAlbumId As Integer = 0         ' dbo.Album.Album_Id PK.
    Private psArtist As String = ""          ' dbo.Artist.Artist_name.
    Private piArtistId As Integer = 0        ' dbo.Artist.Artist_Id PK.
    Private pbBookmarked As Boolean = False
    Private piCdTrack As Integer = 0
    Private psFileName As String = ""
    Private psGenre As String = ""
    Private piId As Integer = 0             ' dbo.Tracks.Track_Id PK.
    Private piIndex As Integer = -1         ' Listview index (to enable playing tracks in order regardless of listview's item focus, selections, etc.).
    Private piMaxVolume As Integer = 100    ' To allow per-track volume so a particularly loud track can be kicked down a few decibels.
    Private pbPlayed As Boolean = False
    Private psPlayedOn As String = ""
    Private piPlayCount As Integer = 0
    Private psTrack As String = ""
    Private psTitle As String = ""         ' MJBH, 2019.08.10.

#Region "Properties"

    Public Property Album() As String
      <DebuggerStepThrough()> _
      Get
        Album = psAlbum
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sAlbum As String)
        psAlbum = sAlbum
      End Set
    End Property
    Public Property AlbumId() As Integer
      <DebuggerStepThrough()> _
      Get
        AlbumId = piAlbumId
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iAlbumId As Integer)
        piAlbumId = iAlbumId
      End Set
    End Property
    Public Property Artist() As String
      <DebuggerStepThrough()> _
      Get
        Artist = psArtist
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sArtist As String)
        psArtist = sArtist
      End Set
    End Property
    Public Property ArtistID() As Integer
      <DebuggerStepThrough()> _
      Get
        ArtistID = piArtistId
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iArtistID As Integer)
        piArtistId = iArtistID
      End Set
    End Property
    Public Property Bookmarked() As Boolean
      <DebuggerStepThrough()> _
      Get
        Bookmarked = pbBookmarked
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal bBookmarked As Boolean)
        pbBookmarked = bBookmarked
      End Set
    End Property
    Public Property CdTrack() As Integer
      <DebuggerStepThrough()> _
      Get
        CdTrack = piCdTrack
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iCdTrack As Integer)
        piCdTrack = iCdTrack
      End Set
    End Property
    Public Property Filename() As String
      <DebuggerStepThrough()> _
      Get
        Filename = psFileName
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sFileName As String)
        psFileName = sFileName
      End Set
    End Property
    Public Property Genre() As String
      <DebuggerStepThrough()> _
      Get
        Genre = psGenre
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sGenre As String)
        psGenre = sGenre
      End Set
    End Property
    Public Property Id() As Integer
      <DebuggerStepThrough()> _
      Get
        Return piId
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iId As Integer)
        piId = iId
      End Set
    End Property
    Public Property Index() As Integer
      <DebuggerStepThrough()> _
      Get
        Index = piIndex
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iIndex As Integer)
        piIndex = iIndex
      End Set
    End Property
    Public Property MaxVolume() As Integer
      <DebuggerStepThrough()> _
      Get
        MaxVolume = piMaxVolume
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iMaxVolume As Integer)
        piMaxVolume = iMaxVolume
      End Set
    End Property
    Public Property PlayCount() As Integer
      <DebuggerStepThrough()> _
      Get
        PlayCount = piPlayCount
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal iPlayCount As Integer)
        piPlayCount = iPlayCount
      End Set
    End Property
    Public Property Played() As Boolean
      <DebuggerStepThrough()> _
      Get
        Played = pbPlayed
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal bPlayed As Boolean)
        pbPlayed = bPlayed
      End Set
    End Property
    Public Property PlayedOn() As String
      <DebuggerStepThrough()> _
      Get
        PlayedOn = psPlayedOn
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sPlayedOn As String)
        psPlayedOn = sPlayedOn
      End Set
    End Property
    Public Property Track() As String
      <DebuggerStepThrough()> _
      Get
        Track = psTrack
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sTrack As String)
        psTrack = sTrack
      End Set
    End Property
    Public Property Title() As String
      <DebuggerStepThrough()> _
      Get
        Title = psTitle
      End Get
      <DebuggerStepThrough()> _
      Set(ByVal sTitle As String)
        psTitle = sTitle
      End Set
    End Property

#End Region

    Public Sub New()
      MyBase.New()
      piId = -1
    End Sub

#End Region

  End Class

  Sub Mp3_tag_v2_set()
    ' Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    Stop ' To do, to develop.
    uOp.Subr = "Mp3_tag_v2_set"

    Dim tmfcReadID3v2Tag As New mfcReadID3v2Tag
    Dim tmcTrack As mcTrack
    tmfcReadID3v2Tag = New mfcReadID3v2Tag
    tmcTrack = tmfcReadID3v2Tag.Read_id3_v2_tag_v2(gMusic.File_original_full_name)

    uOp.Subr = ""
  End Sub

#End Region ' End region, "Mp3_id3_v2."

  Sub Mp3_tag_v1_get()
    ' Separated 2019.08.31. Revised 2021.02.14.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim objMP3_v1 As New MP3_Id3_v1(gMusic.File_original_full_name) ' !
    tLocals.Subr = "Mp3_tag v1 get"

    If objMP3_v1.Tag_exists Then ' !
      ' Stop ' To do, to develop:
      frmMain.dgvMusic_details.Item(1, 0).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Artist)
      frmMain.dgvMusic_details.Item(1, 1).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Album)
      frmMain.dgvMusic_details.Item(1, 2).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Year)
      frmMain.dgvMusic_details.Item(1, 3).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Track)
      frmMain.dgvMusic_details.Item(1, 4).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Title)
      frmMain.dgvMusic_details.Item(1, 5).Value = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Comment)

      tLocals.Msg = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Genre)
      gMusic.Track_genre = objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Genre)
      If objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Genre) <> 255 Then
        frmMain.dgvMusic_details.Item(1, 6).Value = [Enum].GetName(GetType(MP3_Id3_v1.Genres), CInt(objMP3_v1.Frame(MP3_Id3_v1.Frame_types.Genre)))
      End If

      frmMain.Refresh()
    End If
    tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    Call Log_write(tLocals, 0)
    uOp.Success = True
    uOp.Subr = ""
  End Sub

  Sub Mp3_tag_v1_set()
    ' Separated 2019.08.03. Revised 2021.02.15.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim objMP3_v1 As New MP3_Id3_v1(gMusic.File_original_full_name) ' !
    tLocals.Subr = "Mp3_tag_v1 set"

    Call objMP3_v1.Update()

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uOp.Success = True
    uOp.Subr = ""
  End Sub

  Sub Mp3_tag_v1_set_02(ByRef tgMusic_item As String, ByVal tMp3_meta_data_item As String, Optional tCheck_size As Boolean = False)
    ' Revised 2019.10.14.
    uOp.Subr = "Mp3_tag_v1_set_02"

    Stop ' To do, to develop.
    If tMp3_meta_data_item = Nothing Then
    Else
      If tCheck_size Then
        If Len(tMp3_meta_data_item) = 1 Then
          If Asc(tMp3_meta_data_item) = 0 Then
            Stop ' Indefinite.
          End If
        End If
      End If

      If tgMusic_item = Nothing Then
        Stop ' Check:
        tgMusic_item = tMp3_meta_data_item
      End If
    End If

    uOp.Subr = ""
  End Sub

End Module
