﻿' Module: Network.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Revised 2018.08.19.

Option Explicit On

Module Network_
  Public Const cNetwork_version As Date = #5/2/2021# ' M/d/y.
  Private Const No_ERROR = 0
  Private Declare Function inet_addr Lib "wsock32.dll" (ByVal s As String) As Long
  Private Declare Function SendARP Lib "iphlpapi.dll" (ByVal DestIp As Long, ByVal ScrIP As Long, ByVal pMacAddr As Long, ByVal PhyAddrLen As Long) As Long
  Private Declare Sub CopyMemory Lib "kernel32.dll" Alias "RtlMoveMemory" (ByVal dst As Object, ByVal src As Object, ByVal bCount As Long)

  Private Function Get_remote_mac_address_fn(ByVal sRemoteIP As String, ByVal sRemoteMacAddress As String) As Boolean
    ' Revised MJBH 2017.11.11.
    Stop ' Temporary, to do, to test.
    Dim dwRemoteIp As Long
    Dim pMac_addr As Long
    Dim bpMac_addr() As Byte
    Dim PhyAddr_len As Long
    Dim tCnt As Long
    Dim tTmp As String = ""

    dwRemoteIp = inet_addr(sRemoteIP)
    If dwRemoteIp <> 0 Then
      PhyAddr_len = 6
      If SendARP(dwRemoteIp, 0&, pMac_addr, PhyAddr_len) = No_ERROR Then
        If pMac_addr <> 0 And PhyAddr_len <> 0 Then
          ReDim bpMac_addr(0 To PhyAddr_len - 1)
          CopyMemory(bpMac_addr(0), pMac_addr, PhyAddr_len)
          For tCnt = 0 To PhyAddr_len - 1
            If bpMac_addr(tCnt) = 0 Then
              tTmp = tTmp & "00-"
            Else
              tTmp = tTmp & Hex$(bpMac_addr(tCnt)) & "-"
            End If
          Next

          If Len(tTmp) > 0 Then
            sRemoteMacAddress = Left$(tTmp, Len(tTmp) - 1)
            Get_remote_mac_address_fn = True
          End If

          Exit Function
        Else
          Get_remote_mac_address_fn = False
        End If
      Else
        Get_remote_mac_address_fn = False
      End If
    Else
      Get_remote_mac_address_fn = False
    End If
  End Function

  Function Get_mac_address_fn(ByVal tIp_address As String) As String
    ' Revised 2018.11.11.
    Dim sRemote_mac_address As String = ""
    Get_mac_address_fn = ""

    If Len(tIp_address) > 0 Then
      If Get_remote_mac_address_fn(tIp_address, sRemote_mac_address) Then
        Get_mac_address_fn = sRemote_mac_address
      Else
        Get_mac_address_fn = "SendARP call failed."
      End If
    End If
  End Function

  Public Sub Port_scan()
    ' Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    ' Console.Clear()
    Console.WriteLine("Please enter internet protocol address: ")
    Dim tIp As String = Console.ReadLine()
    tIp = "192.168.1.2"
    Console.WriteLine("Please enter starting port: ")
    Dim tStart_port As Integer = Console.ReadLine()
    tStart_port = 0
    Console.WriteLine("Please enter end port: ")
    Dim tEnd_port As Integer = Console.ReadLine()
    tEnd_port = 5

    Dim tList_of_ports As New List(Of Integer)
    Try
      For tPort As Integer = tStart_port To tEnd_port
        ' Console.SetCursorPosition(0, 7)
        tLocals.Msg = "Current port: " & tPort & "." & vbCr
        Console.Write(tLocals.Msg)
        Call frmMain.Statuslabel1_set(tLocals.Msg)
        If Scan_port(System.Net.IPAddress.Parse(tIp), tPort) = True Then
          tList_of_ports.Add(tPort)
        End If
      Next

      tLocals.Msg = vbCrLf & "List of open ports on: " & tIp & " (" & System.Net.Dns.GetHostEntry(System.Net.IPAddress.Parse(tIp)).HostName & ")"
      Console.WriteLine(tLocals.Msg)
      Stop ' To do, to re-develop:
      '     uOp.Result  = Msgbox_LIS_top (        tLocals.Msg , vbInformation, udtDialogue_types.Okay_only, uOp.Subr)
      ' uDialogue.Dialogue_result = Msgbox_LIS_top (        tLocals.Msg , vbInformation, udtDialogue_types.Okay_only, uOp.Subr)
      Call Msgbox_LIS_top(tLocals.Msg, vbInformation, udtDialogue_types.Okay_only, uOp.Subr)
      For Each tPort As Integer In tList_of_ports
        Console.Write(tPort & ", ")
      Next
    Catch ex As Exception
      Stop ' Permanent.
      tLocals.Error_description = ex.Message
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process

      Console.WriteLine(vbCrLf & "Invalid internet protocol specified, please try again...")
    End Try

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       tLocals.Success = udtBoolean_extended.False_ Or tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      ' Reduce all error responses down to tLocals.Success = udtBoolean_extended.False_
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    uOp.Result = Nothing
  End Sub

  ReadOnly Property Scan_port(ByVal Ip As System.Net.IPAddress, ByVal Port As Integer) As Boolean
    ' Revised MJBH 2017.08.06.
    Get
      Dim TCP As New System.Net.Sockets.TcpClient
      Try
        TCP.Connect(Ip, Port)
      Catch
      End Try
      If TCP.Connected = True Then
        Scan_port = True
        TCP.Close()
      Else
        Scan_port = False
        TCP.Close()
      End If
    End Get
  End Property

End Module
