﻿' Module: Web.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Revised 2021.05.06.

Option Explicit On
Imports SHDocVw ' Reference: Microsoft Internet Controls.
Imports System
Imports System.IO
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Web
Imports System.Xml
Imports VB = Microsoft.VisualBasic

' SOAP system: 
' Imports System
' Imports System.Web.Services
' Imports System.Web.Services.Protocols
' Imports System.IO
' Declaration:

Public MustInherit Class SoapExtension
  ' Usage: 
  Dim instance As SoapExtension
End Class
' Define a SOAP Extension that traces the SOAP request and SOAP response for the XML Web service method the SOAP extension is applied to.

Public Class TraceExtension
  Inherits SoapExtension
  Private OldStream As Stream
  Private NewStream As Stream
  Private M_filename As String

  ' Save the Stream representing the SOAP request or SOAP response into a local memory buffer.
  ' Public Overrides Function ChainStream(ByVal stream As Stream) As Stream
  Public Function Chain_stream_fn(ByVal tStream As Stream) As Stream
    ' Revised 2018.11.11.
    OldStream = tStream
    NewStream = New MemoryStream()
    Return NewStream
  End Function

  ' When the SOAP extension is accessed for the first time, the XML Web service method it is applied to is accessed to store the file
  ' name passed in, using the corresponding SoapExtensionAttribute.    
  'Public Overloads Overrides Function GetInitialiser(ByVal methodInfo As LogicalMethodInfo, ByVal attribute As SoapExtensionAttribute) As Object
  '   Return CType(attribute, TraceExtensionAttribute).Filename
  'End Function

  ' The SOAP extension was configured to run using a configuration file instead of an attribute applied to a specific XML Web service
  ' method. Return a file name based on the class implementing the Web Service's type.
  'Public Overloads Overrides Function GetInitialiser(ByVal WebServiceType As Type) As Object
  '   ' Return a file name to log the trace information to, based on the type.
  '   Return "C:\" + WebServiceType.FullName + ".log"
  'End Function

  ' Receive the file name stored by GetInitializer and store it in a member variable for this specific instance.
  'Public Overrides Sub Initialize(ByVal initializer As Object)
  '   m_filename = CStr(initializer)
  'End Sub

  ' If the SoapMessageStage is such that the SoapRequest or SoapResponse is still in the SOAP format to be sent or received over the network, save it out to file.
  'Public Overrides Sub ProcessMessage(ByVal tMessage As SoapMessage)
  '   Select Case tMessage.Stage
  '      Case SoapMessageStage.BeforeSerialize
  '      Case SoapMessageStage.AfterSerialize
  '         WriteOutput(Message)
  '      Case SoapMessageStage.BeforeDeserialize
  '         WriteInput(Message)
  '      Case SoapMessageStage.AfterDeserialize
  '      Case Else
  '         Throw New Exception("invalid stage")
  '   End Select
  'End Sub

  ' Write the SOAP message out to a file.
  'Public Sub WriteOutput(ByVal tMessage As SoapMessage)
  '   newStream.Position = 0
  '   Dim fs As New FileStream(m_filename, FileMode.Append, FileAccess.Write)
  '   Dim w As New StreamWriter(fs)

  '   w.WriteLine("-----Response at " + DateTime.Now.ToString())
  '   w.Flush()
  '   Copy(newStream, fs)
  '   w.Close()
  '   newStream.Position = 0
  '   Copy(newStream, oldStream)
  'End Sub

  ' Write the SOAP message out to a file.
  'Public Sub WriteInput(ByVal tMessage As SoapMessage)
  '   Copy(oldStream, newStream)
  '   Dim fs As New FileStream(m_filename, FileMode.Append, FileAccess.Write)
  '   Dim w As New StreamWriter(fs)

  '   w.WriteLine("----- Request at " + DateTime.Now.ToString())
  '   w.Flush()
  '   newStream.Position = 0
  '   Copy(newStream, fs)
  '   w.Close()
  '   newStream.Position = 0
  'End Sub

  Sub Copy(ByVal fromStream As Stream, ByVal toStream As Stream)
    Dim reader As New StreamReader(fromStream)
    Dim writer As New StreamWriter(toStream)
    writer.WriteLine(reader.ReadToEnd())
    writer.Flush()
  End Sub
End Class

' Create a SoapExtensionAttribute for our SOAP Extension that can be applied to an XML Web service method.
<AttributeUsage(AttributeTargets.Method)> Public Class TraceExtensionAttribute
  ' Inherits SoapExtensionAttribute
  Private m_Filename As String = "c:\log.txt"
  Private m_Priority As Integer

  'Public Overrides ReadOnly Property ExtensionType() As Type
  '   Get
  '      Return GetType(TraceExtension)
  '   End Get
  'End Property

  'Public Overrides Property Priority() As Integer
  '   Get
  '      Return m_priority
  '   End Get
  '   Set(ByVal Value As Integer)
  '      m_Priority = Value
  '   End Set
  'End Property

  Public Property Filename() As String
    Get
      Return m_Filename
    End Get
    Set(ByVal Value As String)
      m_Filename = Value
    End Set
  End Property
End Class

Module Web_
  Public Const cWeb_version As Date = #5/2/2021# ' M/d/y.
  Private _Connection_client As Connection_info ' 2021.03.21 To do; replace with mInternet.*.
  Private _Server_address As IPAddress

  Sub Instantiate_IE(ByVal tUrl As String, ByRef tIE As InternetExplorer)
    ' Written by Michael J. B. Hodgson, 2011.11. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Instantiate Internet Explorer"

    tLocals.Action = udtuStandard_texts.Main_start
    frmMain.Cursor = Cursors.WaitCursor

    If Not Url_validate_fn(tUrl, False) Then
      uError.Message = "This web page address (u.r.l.) is invalid."
      Stop ' Permanent.
      Stop ' To do, to re-develop:
      '     uOp.Result  = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
      ' uDialogue.Dialogue_result = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
      Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
      uOp.Subr = ""
      Exit Sub
    End If
    uOp.Subr = "Instantiate Internet Explorer"

    ' objIE = CreateObject("InternetExplorer.Application")
    tIE = New InternetExplorer
    With tIE
      .AddressBar = False
      .StatusBar = False
      .MenuBar = False
      .ToolBar = 0
      .Visible = uLIS.Dev
      .Navigate(tUrl)
    End With

    While tIE.Busy
    End While
    tLocals.Msg = tIE.Document.ReadyState.ToString
    While tIE.Document.ReadyState IsNot "complete"
      tLocals.Msg = tIE.Document.ReadyState
    End While

    frmMain.Cursor = Cursors.Default
    uOp.Result = Nothing
    uOp.Subr = ""
  End Sub

  Public Class Connection_info
    ' Encapuslates the client connection and provides a state object for async read operations.
    Private _Append_method As Action(Of String)
    Public ReadOnly Property AppendMethod As Action(Of String)
      Get
        Return _Append_method
      End Get
    End Property

    Private _Client As TcpClient
    Public ReadOnly Property Client As TcpClient
      Get
        Return _Client
      End Get
    End Property

    Private _Stream As NetworkStream
    Public ReadOnly Property Stream As NetworkStream
      Get
        Return _Stream
      End Get
    End Property

    Private _Last_read_length As Integer
    Public ReadOnly Property LastReadLength As Integer
      Get
        Return _Last_read_length
      End Get
    End Property

    Private _Buffer(63) As Byte

    Public Sub New(ByVal tAddress As IPAddress, ByVal tPort As Integer, ByVal tAppend As Action(Of String))
      ' Revised 2015.12.30.
      _Append_method = tAppend
      _Client = New TcpClient
      _Client.Connect(tAddress, tPort)
      _Stream = _Client.GetStream
    End Sub

    Public Sub Await_data()
      _Stream.BeginRead(_Buffer, 0, _Buffer.Length, AddressOf Do_read_data, Me)
    End Sub

    Public Sub Close()
      If _Client IsNot Nothing Then _Client.Close()
      _Client = Nothing
      _Stream = Nothing
    End Sub

    Private Sub Do_read_data(ByVal tResult As IAsyncResult)
      ' Revised 2015.12.30.
      Dim tInfo As Connection_info = CType(tResult.AsyncState, Connection_info)
      Try
        If tInfo._Stream IsNot Nothing AndAlso tInfo._Stream.CanRead Then
          tInfo._Last_read_length = tInfo._Stream.EndRead(tResult)
          If tInfo._Last_read_length > 0 Then
            Dim tMessage As String = System.Text.Encoding.ASCII.GetString(tInfo._Buffer)
            tInfo._Append_method(tMessage)
          End If
          Call tInfo.Await_data()
        End If
      Catch ex As Exception
        tInfo._Last_read_length = -1
        tInfo._Append_method(ex.Message)
      End Try
    End Sub
  End Class

  Sub Invoke_append_output(ByVal tMessage As String)
    ' The InvokeAppendOutput method could easily be replaced with a lambda method passed to the ConnectionInfo contstructor in the ConnectButton_CheckChanged event handler.
    Dim tDo_append_output As New Action(Of String)(AddressOf Append_output)
    ' Me.Invoke(doAppendOutput, message)
    ' frmMain.Invoke(doAppendOutput, tMessage)
  End Sub

  Private Sub Append_output(ByVal tMessage As String)
    'If RichTextBox1.TextLength > 0 Then
    '   RichTextBox1.AppendText(ControlChars.NewLine)
    'End If
    'RichTextBox1.AppendText(message)
    'RichTextBox1.ScrollToCaret()
  End Sub

  Sub Facebook_page_likes_crawler()
    ' Written by Michael J. B. Hodgson, 2012.10.17. Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tFile As udtFile = Nothing

    Dim objIE As Object
    ' Dim varTables, varTable
    Dim varLists, varList
    Dim varH3s, varH3
    Dim tHyperlinks, tHyperlink
    ' Dim varRows, varRow
    ' Dim varCells, varCell
    ' Dim lngRow As Long
    ' Dim lngColumn As Long
    Dim strBuffer As String

    Dim tUrl As String
    Dim tPackage As String

    Dim tFreeFile As Short = FreeFile()

    tLocals.Subr = "Facebook page likes crawler"

    uError.Description = ""
    On Error GoTo Error_captured
    uOp.Subr = tLocals.Subr
    tLocals.Action = udtuStandard_texts.Initiation
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set()
    uOp.Subr = tLocals.Subr
    tFile.File_full_name = "c:\temp\temp.txt"

Main:
    tUrl = "http://www.facebook.com/AetherCon/likes"
    Call Instantiate_IE(tUrl, objIE)
    uOp.Subr = tLocals.Subr
    tLocals.Action = udtuStandard_texts.Main_start

    tFile.Open_file = tFile.File_full_name
    FileOpen(tFreeFile, tFile.File_full_name, OpenMode.Output)

    tHyperlinks = objIE.document.all.tags("a")
    For Each tHyperlink In tHyperlinks
      If tHyperlink.innertext <> "" Then
        tLocals.Msg = tHyperlink.innertext
        PrintLine(tFreeFile, "A href: " & tLocals.Msg & ".")
      End If
    Next
    PrintLine(tFreeFile)

    varH3s = objIE.Document.All.tags("h3")
    For Each varH3 In varH3s
      ' Use the innerText to see if this is the list item we want:
      tLocals.Msg = varH3.innerText
      PrintLine(tFreeFile, "H3: " & tLocals.Msg & ".")
      If varH3.innerText Like "*Friends Who Like AetherCon! The Steampunk Convention*" Then
        Stop
        varLists = objIE.Document.All.tags("li")
        For Each varList In varLists
          tLocals.Msg = varList.innerText
          If varList.innerText <> "" Then
            Stop
            PrintLine(tFreeFile, "List: " & tLocals.Msg & ".")
            ' Query:
            'tDatabase.Table = uTables.Addo_people
            'tDatabase.Cmd_str = "SELECT Product_service FROM " & uTables.Addo_people & " WHERE Product_service='" & Trim(varList.innerText) & "'"
            'Call Memory_general(tDatabase)
            'uOp.Subr =  tLocals.Subr 

            'tDatabase.Table = "Product_service"
            'If tDatabase.Rows_affected = -1 Or tDatabase.Rows_affected = 0 Then
            '   ' Insert:
            ' tDb.Cmd_str = "INSERT INTO " & tDb.Table_name & vbCrLf
            ' tDb.Cmd_str = tDb.Cmd_str & "(OrganisationID, Product_type_id, Product_service, Updated)" & vbCrLf
            ' tDb.Cmd_str = tDb.Cmd_str & "VALUES (85, 18, '" & Trim(varList.innerText) & "', '" & DateAndTime.Today & "')"
            'Else
            '   ' Update:
            '   stop
            'End If
            'Call Memory_general(tDatabase)
            'uOp.Subr =  tLocals.Subr 

          End If
        Next varList
      End If
    Next varH3
    FileClose(tFreeFile)
    tFile.Open_file = ""
    uOp.Success = True

Ender_full:
    Stop
    objIE.quit()
    objIE = Nothing
    varLists = Nothing
    varList = Nothing
    varH3s = Nothing
    varH3 = Nothing
    strBuffer = Nothing
    tUrl = Nothing
    tPackage = Nothing
    Call File_properties_determine_fn(tFile)

    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then uOp.Success = True
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 91 ' Object variable or With block variable not set.
            Resume Next
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

    uOp.Subr = ""
  End Sub

  Sub Web_api()
    ' Revised 2021.04.04.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tRequest As HttpWebRequest
    Dim tResponse As HttpWebResponse = Nothing
    Dim tReader As StreamReader
    tLocals.Subr = "Web api"
    uOp.Subr = tLocals.Subr
    On Error GoTo Error_captured

    uOp.Subr = tLocals.Subr
    tLocals.Action = udtuStandard_texts.Initiation
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Main:
    tLocals.Action = udtuStandard_texts.Main_start

    ' Try
    ' Create the web request: 
    tRequest = DirectCast(WebRequest.Create("http://developer.yahoo.com/"), HttpWebRequest)
    ' Get response: 
    tResponse = DirectCast(tRequest.GetResponse(), HttpWebResponse)
    ' Get the response stream into a reader: 
    tReader = New StreamReader(tResponse.GetResponseStream())
    ' Console application output: 
    Console.WriteLine(tReader.ReadToEnd())
    tLocals.Msg = tReader.ReadToEnd
    ' Finally
    If Not tResponse Is Nothing Then tResponse.Close()
    ' End Try

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    tLocals.Action = ""

    uOp.Subr = ""
  End Sub

  Public Function Get_page_html_fn(ByVal tUrl As String) As String
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.
    uOp.Subr = "Get page html"
    ' Retrieves the HTML from the specified url. 
    ' The address needs to start with the "http://
    Dim objWC As New System.Net.WebClient()

    If VB.Left(tUrl, 7) <> "http://" Then
      tUrl = "http://" & tUrl
    End If

    Return New System.Text.UTF8Encoding().GetString(objWC.DownloadData(tUrl))
    uOp.Subr = ""
  End Function

  Public Function Get_page_html_fn(ByVal tUrl As String, ByVal tTimeout_seconds As Integer) As String
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.
    uOp.Subr = "Get page html"
    ' Retrieves the HTML from the specified URL, using a default timeout of 10 seconds. 
    Dim objRequest As Net.WebRequest
    Dim objResponse As Net.WebResponse
    Dim objStreamReceive As System.IO.Stream
    Dim objEncoding As System.Text.Encoding
    Dim objStreamRead As System.IO.StreamReader

    Try
      ' Setup our web request:
      objRequest = Net.WebRequest.Create(tUrl)
      objRequest.Timeout = tTimeout_seconds * 1000
      ' Retrieve data from request
      objResponse = objRequest.GetResponse
      objStreamReceive = objResponse.GetResponseStream
      objEncoding = System.Text.Encoding.GetEncoding("utf-8")
      objStreamRead = New System.IO.StreamReader(objStreamReceive, objEncoding)
      ' Set function return value:
      Get_page_html_fn = objStreamRead.ReadToEnd()
      ' Check if available, then close response:
      If Not objResponse Is Nothing Then
        objResponse.Close()
      End If
    Catch
      ' Error occured grabbing data, simply return nothing: 
      Return ""
    End Try

    uOp.Subr = ""
  End Function

  Public Function Parse_links_fn(ByVal HTML As String) As ArrayList
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.
    uOp.Subr = "ParseLinks"
    ' Remember to add the following at top of class:
    ' - Imports System.Text.RegularExpressions
    Dim objRegEx As System.Text.RegularExpressions.Regex
    Dim objMatch As System.Text.RegularExpressions.Match
    Dim arrLinks As New System.Collections.ArrayList()

    ' Create regular expression:
    objRegEx = New System.Text.RegularExpressions.Regex("a.*href\s*=\s*(?:""(?<1>[^""]*)""|(?<1>\S+))", System.Text.RegularExpressions.RegexOptions.IgnoreCase Or System.Text.RegularExpressions.RegexOptions.Compiled)

    ' Match expression to html:
    objMatch = objRegEx.Match(HTML)

    ' Loop through matches and add <1> to array list:
    While objMatch.Success
      Dim strMatch As String
      strMatch = objMatch.Groups(1).ToString
      arrLinks.Add(strMatch)
      objMatch = objMatch.NextMatch()
    End While

    ' Pass back results: 
    Return arrLinks

    uOp.Subr = ""
  End Function

  Public Function Parse_images_fn(ByVal tHtml As String) As ArrayList
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.
    uOp.Subr = "ParseImages"
    ' Remember to add the following at top of class:
    ' - Imports System.Text.RegularExpressions
    Dim objRegEx As System.Text.RegularExpressions.Regex
    Dim objMatch As System.Text.RegularExpressions.Match
    Dim arrLinks As New System.Collections.ArrayList()

    ' Create regular expression: 
    objRegEx = New System.Text.RegularExpressions.Regex("img.*src\s*=\s*(?:""(?<1>[^""]*)""|(?<1>\S+))", System.Text.RegularExpressions.RegexOptions.IgnoreCase Or System.Text.RegularExpressions.RegexOptions.Compiled)

    ' Match expression to html:
    objMatch = objRegEx.Match(tHtml)

    ' Loop through matches and add <1> to array list: 
    While objMatch.Success
      Dim strMatch As String
      strMatch = objMatch.Groups(1).ToString
      arrLinks.Add(strMatch)
      objMatch = objMatch.NextMatch()
    End While

    ' Pass back results: 
    Return arrLinks

    uOp.Subr = ""
  End Function

  ' Using the Data Unison A.P.I.: 
  ' imports System
  ' imports System.Xml
  ' imports System.Net
  ' Namespace DataUnison
  ' Class Program

  Sub Data_unison_a_p_i()
    ' Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Data unison a.p.i."
    tLocals.Msg = ""

    Try
      ' Private didn't work inside this method
      ' Dim restWebUrl As String = "http://open.api.dataunison.com/rest/?CallName=GetPriceResearch" & "&AppID=YourAppID&Version=2&ResearchPeriod=30&QueryKeywords=ipod%20mp3%20player"
      Dim restWebUrl As String = "http://open.api.dataunison.com/rest/?CallName=GetPriceResearch" & "&AppID=f4a88aea-1f74-4c1e-a65d-9bcf8e27be18&Version=2&ResearchPeriod=30&QueryKeywords=ipod%20mp3%20player"
      ' f4a88aea-1f74-4c1e-a65d-9bcf8e27be18
      Dim tTotalSales As XmlNodeList
      Dim tTotalItems As XmlNodeList
      ' AverageSoldPrice, MaxSoldPrice, MinSoldPrice

      Dim tDoc As XmlDocument = New XmlDocument()
      tDoc.PreserveWhitespace = True

      tDoc.Load(restWebUrl)
      ' Console.WriteLine(tDoc.DocumentElement.OuterXml)
      tLocals.Msg = tDoc.DocumentElement.OuterXml
      Dim tTotalSalesString As String
      tTotalSales = tDoc.GetElementsByTagName("TotalSales")
      tTotalItems = tDoc.GetElementsByTagName("TotalItems")

      ' Console.WriteLine("TotalSales" & tTotalSales(0).InnerXml)
      tLocals.Msg = "TotalSales" & tTotalSales(0).InnerXml
      tTotalSalesString = tDoc.GetElementsByTagName("TotalSales").Item(0).ChildNodes(0).Value
      Dim tTotalSales_dbl As Double = Convert.ToDouble(tTotalSales)
      Console.WriteLine("TotalSalesString" & tTotalSales_dbl.ToString("N") & tDoc.GetElementsByTagName("TotalSales").Item(0).ChildNodes(0).NodeType)

    Catch ex As Exception
      Stop ' Permanent.
      tLocals.Error_description = ex.Message
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process

      ' Console.WriteLine("Exception on loading document" + e.Message)
      tLocals.Msg = "Exception on loading document" + ex.Message

    End Try

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       tLocals.Success = udtBoolean_extended.False_ Or tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      ' Reduce all error responses down to tLocals.Success = udtBoolean_extended.False_
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    uOp.Subr = ""
  End Sub

  ' End Class
  ' End Namespace

  Public Function Get_table_text_fn(ByVal sHTML As String) As String
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.

    '   Dim myDoc As mshtml.IHTMLDocument2 = New mshtml.HTMLDocument
    '   Dim mElement As mshtml.IHTMLElement
    '   Dim mElement2 As mshtml.IHTMLElement
    '   Dim mECol As mshtml.IHTMLElementCollection
    '   Dim I As Integer

    Get_table_text_fn = ""

    '   Initialise the document object within the HTMLDocument class:
    '   myDoc.close()
    '   myDoc.open("about:blank")
    '   'write the HTML to the document using the MSHTML "write" method...

    '   Dim clsHTML() As Object = {sHTML}
    '   myDoc.write(clsHTML)
    '   clsHTML = Nothing
    '   mElement = myDoc.body()
    '   mECol = mElement.getElementsByTagName("TD")
    '   For I = 0 To mECol.length - 1
    '      mElement2 = mECol.item(I)
    '      lstResults.Items.Add(mElement2.tagName & " : " & mElement2.innerText)
    '   Next
  End Function

  Sub Html_read()
    uOp.Subr = "Html read"

    Dim tHtml As String = My.Computer.FileSystem.ReadAllText("C:\test.htm")
    Dim tText As String = ""
    Dim c As Char
    Dim tReading As Boolean = True
    For Each c In tHtml
      If c = "<" Then
        tReading = False
      ElseIf c = ">" Then
        tText += " "
        tReading = True
      Else
        If tReading Then tText += c
      End If
    Next

    uOp.Subr = ""
  End Sub

  Function Get_web_page_fn(ByVal tURL As String) As String
    ' Revised 2018.11.11.
    Stop ' Temporary, to do, to test.

    '      Dim XML As MSXML2.XMLHTTP
    Get_web_page_fn = ""

    '      XML = New MSXML2.XMLHTTP
    '      On Error GoTo daEND
    '      getWebPage = ""
    '      XML.Open("GET", strURL, False)
    '      XML.send()
    '      getWebPage = XML.responseText
    'daEND:
  End Function

End Module
