﻿' Module Xml.vb.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson. Revised 2021.04.07.

Option Explicit On

Module Xml_

  Function Xml_read_file(ByVal tFile_name As String, ByVal tTag As String) As String
    ' Written by Michael J. B. Hodgson, 2010.11. Revised 2021.03.07.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tFile As udtFile = Nothing

    Dim tLine As String = ""
    Dim tValue As String = ""

    uOp.Subr = "Xml_read"
    Xml_read_file = ""

    If tFile_name = "" Then Exit Function

    tLocals.Msg = Dir(tFile_name, vbNormal)
    If Dir(tFile_name, vbNormal) <> "" Then
      tFile.Open_file = tFile_name
      FileOpen(tFile.Free_file, tFile_name, OpenMode.Input)
      Do
        Input(tFile.Free_file, tLine)
      Loop Until InStr(tLine, "<" & tTag & ">") <> 0

      If InStr(tLine, "</" & tTag & ">") <> 0 Then
        tValue = tLine
        tValue = Replace(tValue, "<" & tTag & ">", "")
        tValue = Replace(tValue, "</" & tTag & ">", "")

      Else
        Input(tFile.Free_file, tValue)
      End If
      tValue = Trim(tValue)

      FileClose(tFile.Free_file)
    End If

    uOp.Subr = ""
  End Function

  ' -------------------XML Writer Functions ----------------------------------
  ' ----------------------------XML Export Functions----------------------------------------
  Private Function Xml_clean_value(ByRef tTag_value As String) As String
    ' Escapes control characters: 
    tTag_value = Replace(tTag_value, "&", "&amp;")
    tTag_value = Replace(tTag_value, "<", "&lt;")
    tTag_value = Replace(tTag_value, ">", "&gt;")
    tTag_value = Replace(tTag_value, """", "&quot;")
    tTag_value = Replace(tTag_value, "'", "&apos;")

    Return tTag_value
  End Function

  Public Sub Xml_write_header(ByVal tFreefile As Short, ByRef tTxt As String)
    ' Writes a header for the xml file. 
    PrintLine(tFreefile, "<?xml version=""1.0"" encoding=""UTF-8"" ?>")
  End Sub

  Public Sub Xml_write_start_tag(ByRef tFreefile As Short, ByVal tTag_name As String)
    ' Writes out the tag only e.g. <AGENTS_PURCHASE_ORDER>. 
    PrintLine(tFreefile, "<" & tTag_name & ">")
  End Sub

  Public Sub Xml_write_close_tag(ByRef tFreefile As Short, ByVal tTag_name As String)
    ' Writes the closing tag for the specified tagname e.g </AGENTS_PURCHASE_ORDER>. 
    PrintLine(tFreefile, "</" & tTag_name & ">")
  End Sub

  Public Sub Xml_write_tag_with_value(ByRef tFreefile As Short, ByVal tTag_name As String, ByVal tTag_value As String)
    ' Writes out a tag with its value e.g <SOLD_TO>101280</SOLD_TO>. 
    PrintLine(tFreefile, "<" & tTag_name & IIf(tTag_value = "", "/>", ">" & Xml_clean_value(tTag_value) & "</" & tTag_name & ">"))
  End Sub

  Public Sub Xml_write_tag_with_value_and_attribute(ByRef tFreefile As Short, ByVal tTag_name As String, ByVal tTag_value As String, ByVal tAttribute_name As String, ByVal tAttribute_value As String)
    ' Writes out a node with an attribute e.g. <ORG_ID OrgType="EAN_Prefix">789</ORG_ID>. 
    PrintLine(tFreefile, "<" & tTag_name & " " & tAttribute_name & "=""" & tAttribute_value & """" & IIf(tTag_value = "", "/>", ">" & Xml_clean_value(tTag_value) & "</" & tTag_name & ">"))
  End Sub

  ' --------------- XML Reader Functions ---------------------------
  Function Xml_load_xml_document(ByVal tXml_file As String, ByRef tError_text As String) As Boolean
    ' This function loads a given xml file and returns true if the load was successful, or false otherwise. 
    ' It also takes in a string byref which is null if no error or filled with error text if errors are encountered. 
    Dim tFreefile As Short = FreeFile()

    Input(tFreefile, tXml_file)
    tError_text = ""
    ' xmlDocument.Async = False    ' Load synchronously (blocks until loaded), since Maxbasic does not support function callbacks. 

    ' Detect if error loading xml file: 
    'If xmlDocument.ParseError.ErrorCode <> 0 Then
    '   strErrorText = strErrorText & "Line No: " & xmlDocument.parseError.line & Chr(13) & Chr(10)
    '   strErrorText = strErrorText & "Character: " & xmlDocument.ParseError.linepos & Chr(13) & Chr(10)
    '   strErrorText = strErrorText & "File Position:" & xmlDocument.ParseError.filepos & Chr(13) & Chr(10)
    '   strErrorText = strErrorText & "SourceText: " & xmlDocument.ParseError.srcText & Chr(13) & Chr(10)
    '   strErrorText = strErrorText & "ErrorCode: " & xmlDocument.ParseError.errorCode & Chr(13) & Chr(10)
    '   strErrorText = strErrorText & "Description: " & xmlDocument.ParseError.reason & Chr(13) & Chr(10)
    '   Return False
    'Else
    '   Return True
    'End If
  End Function

  Function Xml_get_root() As Object
    ' This function returns the root element of an xml document. 
    ' Return xmlDocument.DocumentElement
    Return ""
  End Function

  Function Xml_get_child_list(ByRef tNode As Object) As Object
    ' Given a node with children, this functions returns it's child nodes as a list. 
    Return tNode.ChildNodes
  End Function

  Function Xml_get_item_in_list(ByRef tNode_list As Object, ByVal tItem_index As Short) As Object
    ' This function returns a specific item from within a list, 0 based. 

    If tNode_list = "" Then
      Return ""
    Else
      If tNode_list.Length > 0 And tNode_list.ItemIndex <= (tNode_list.Length - 1) Then
        Return tNode_list(tItem_index)
      Else
        Return ""
      End If
    End If
  End Function

  Function Xml_has_children(ByRef tNode As Object) As Boolean
    ' This function takes in a node, then returns true if the node has any children, false otherwise. 
    Return tNode.hasChildNodes
  End Function

  Function Xml_get_list_by_tag_name(ByRef tNode As Object, ByVal tTag_name As String) As Object
    ' This function retrieves all the child nodes matching the passed tagname and returns in a nodelist. 
    Return tNode.GetElementsByTagName(tTag_name)
  End Function

  Function Xml_get_first_by_tag_name(ByRef tNode As Object, ByVal tTag_name As String) As Object
    ' This function retrieves all the child nodes matching the passed tagname and returns the first node. 
    Dim tNode_list As Object

    tNode_list = Xml_get_list_by_tag_name(tNode, tTag_name)
    If tNode_list.Length > 0 Then
      Return tNode_list(0)
    Else
      Return ""
    End If
  End Function

  Function Xml_get_first_in_list(ByRef tNode_list As Object) As Object
    ' Returns the first item in a node list. 
    If tNode_list.Length > 0 Then
      Return tNode_list(0)
    Else
      Return ""
    End If
  End Function

  Function Xml_get_last_in_list(ByRef tNode_list As Object) As Object
    ' Returns the last item in a node list. 
    If tNode_list.Length > 0 Then
      Return tNode_list(tNode_list.Length - 1)
    Else
      Return ""
    End If
  End Function

  Function Xml_get_next_in_list(ByRef tNode_list As Object) As Object
    Dim tNode_to_return As Object

    tNode_to_return = tNode_list.NextNode
    ' If Not IsNull(NodeToReturn) And ISObject(NodeToReturn) And Not IsEmpty(NodeToReturn) Then
    If tNode_to_return = "" Then
      Return ""
    Else
      Return tNode_to_return
    End If
  End Function

  Function Xml_get_atrribute_value(ByRef tNode As Object, ByVal tAttribute_name As String) As String
    If tNode = "" Then
      Return ""
    Else
      Return tNode.GetAttribute(tAttribute_name)
    End If
  End Function

  ' End of xml functions. 

End Module
