﻿' Module: clsEncrypt.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

Option Explicit On
Option Strict Off

Module clsEncrypt
  Public Const cclsEncrypt_version As Date = #4/8/2021# ' M/d/y.

  Public Class clsEncrypt
    ' Copyright 2005 www.vbusers.com. 
    ' Encryption class for encrypting and decrypting text using a password and an XOR operator.
    ' Written by Andrew Baker. 
    ' For the terms and conditions see www.vbusers.com/termsandconditions.asp.

    '---------------Private class variables-----------------
    Private mPrivate_key As String
    Private mPlaintext As String ' MJBH, 2014.03.11.

    Property Private_key As String
      ' Purpose: Gets or sets the password used by the encryption routines. 
      ' Inputs : N A. 
      ' Outputs: N A. 
      ' Date   : 2005.03.11.
      Get
        Return mPrivate_key
      End Get
      Set(ByVal value As String)
        mPrivate_key = value
      End Set
    End Property

    Property Plaintext As String
      ' Revised MJBH, 2014.03.11.
      Get
        Return mPlaintext
      End Get
      Set(value As String)
        mPlaintext = value
      End Set
    End Property

    Function Encrypt_or_decrypt_fn(ByVal sData As String, Optional ByVal tUse_unicode As Boolean = False) As String
      ' Purpose: Encrypts or decrypts the supplied string.
      ' Inputs : sData:		 The data to process.
      '          bUseUnicode: If true, converts unicode text else converts ASCII text.
      ' Outputs: Returns an encrypted string if the input was unencrypted, or a decrypted string if the input was encrypted.
      ' Date   : 2005.03.25.
      ' Revised: MJBH, 2021.03.04.
      Stop
      Dim abData() As Byte
      Dim abPassword() As Byte

      If Len(sData) > 0 And Len(mPrivate_key) > 0 Then
        If tUse_unicode Then
          ' abData = sData
          ' abPassword = zsPassword
          abData = String_to_byte_array_fn(sData)
          abPassword = String_to_byte_array_fn(mPrivate_key)
        Else
          ' All VB text is stored in Unicode, so must convert back to ASCII first
          ' abData = StrConv(sData, vbFromUnicode)
          ' abPassword = StrConv(zsPassword, vbFromUnicode)
          abData = String_to_byte_array_fn(sData) ' !
          abPassword = String_to_byte_array_fn(mPrivate_key) ' !
        End If
        Encrypt_or_decrypt_fn = Unicode_bytes_to_string_fn(Encrypt_array_fn(abData, abPassword, tUse_unicode)) ' !
        ' Encrypt_or_decrypt_fn = Encrypt_array_fn(abData, abPassword, tUse_unicode) ' !
      Else
        ' Either no data or password has been set: 
        Encrypt_or_decrypt_fn = sData
      End If
      uOp.Result = Encrypt_or_decrypt_fn
    End Function

    Private Function Encrypt_array_fn(ByVal abData() As Byte, ByVal abPassword() As Byte, ByVal tUse_unicode As Boolean) As Byte()
      ' Purpose   : Encrypts or decrypts the specified byte array using the password string with an XOR operator.
      ' Inputs    : abData              A byte array containing the data to encrypt.
      '             abPassword          The password to encrypt the data with.
      ' Outputs   : Returns the encrypted or decrypted byte array.
      ' Date      : 2001.03.25
      ' Revised   : MJBH, 2018.11.10.
      ' Notes     :
      Stop
      Dim tThis_line As Long
      Dim abResults() As Byte
      Dim tPassword_index As Long
      On Error GoTo Error_failed

      ' Size the array to store the resulting data:
      ' ReDim abResults(LBound(abData) To UBound(abData))
      ReDim abResults(UBound(abData))
      ReDim Encrypt_array_fn(UBound(abData))
      ' Loop over each byte in the data array, apply the encryption algorithm
      For tThis_line = 0 To UBound(abData)
        ' Encrypt the byte:
        abResults(tThis_line) = Encrypt_byte_fn(abData(tThis_line), abPassword, tPassword_index)
      Next

      ' Return the results:
      If tUse_unicode Then
        ' Array will already be in unicode format:
        Encrypt_array_fn = abResults
      Else
        ' Have to convert back to Unicode before storing as a string.
        ' Encrypt_array_fn = StrConv(abResults, vbUnicode)
        Encrypt_array_fn = abResults
        ' Encrypt_array_fn = Unicode_bytes_to_string_fn(abResults) ' 2014.03.09.
      End If

      Exit Function

Error_failed:
      Debug.Print("Error in zEncryptArray: " & Err.Description)
      Debug.Assert(False)
    End Function

    Private Function Encrypt_byte_fn(ByVal bytValue As Byte, ByVal abPassword() As Byte, ByRef tPassword_index As Long) As Byte
      ' Purpose  : This function encrypts one byte, then modifies the password.
      ' Inputs   : bytValue            A byte array containing the data to encrypt.
      '            abPassword          The password to encrypt the data with.
      ' Outputs  : Returns the input byte after the encryption algorithm has been applied to it.
      ' Author   : Andrew Baker. 
      ' Date     : 2000.09.04.
      ' Notes    : Modifies the password bytes after each iteration to make decryption harder.
      ' Revisions: MJBH, 2018.11.10.
      Stop
      On Error GoTo Error_failed

      If tPassword_index = UBound(abPassword) - 1 Then
        ' Text exceeded password, reset password array:
        tPassword_index = 0
      End If

      ' Exclusive or the byte with the current password byte
      Encrypt_byte_fn = bytValue Xor abPassword(tPassword_index)
      ' Exclusive or the byte with the first character of the password
      ' multiplied by the current index into the password. And the result with
      ' 256 to avoid possible overflow errors:
      Encrypt_byte_fn = (Encrypt_byte_fn Xor CInt(abPassword(tPassword_index)) * tPassword_index) And &HFF

      ' Modify the password.
      ' Set the current byte in the password to the current byte plus the next byte.
      abPassword(tPassword_index) = (CInt(abPassword(tPassword_index)) + abPassword(tPassword_index + 1)) And &HFF

      ' Increment the password index:
      tPassword_index = tPassword_index + 1

      Exit Function

Error_failed:
      Debug.Print("Error in zEncryptByte: " & Err.Description)
      Debug.Assert(False)
    End Function

    Public Function String_to_byte_array_fn(tStr As String) As Byte()
      ' Revised 2018.11.10.
      Dim tEncoding As New System.Text.UTF8Encoding
      Return tEncoding.GetBytes(tStr)
    End Function

    Public Function Unicode_bytes_to_string_fn(ByVal tBytes() As Byte) As String
      ' Revised 2018.11.10.
      Return System.Text.Encoding.Unicode.GetString(tBytes)
      ' Test, 2014.03.06: 
      ' Return System.Text.Encoding.UTF8.GetString(tBytes)
    End Function

    Private newPropertyValue As String

    Public Property NewProperty() As String
      Get
        If True Then
          newPropertyValue = 1
        Else
          newPropertyValue = 2
        End If
        Return newPropertyValue
      End Get
      Set(ByVal value As String)
        newPropertyValue = value
      End Set
    End Property

  End Class

#Region "Encryption"

  Function Encrypt_LIS_fn() As String
    ' Written 2013.12.04. Revised 2021.04.08.
    Encrypt_LIS_fn = Encrypt_LIS_two_fn(4)
    ' uOp.Result = Encrypt_LIS_fn
  End Function

  Function Encrypt_LIS_two_fn(Optional ByVal tPassword_number As Integer = 4) As String
    ' Written 2013.12.04. Revised 2021.01.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Encrypt_LIS_two_fn
    uCrypt.Cyphertext = ""
    uCrypt.Plaintext = ""
    uCrypt.Private_key = ""

    Select Case tPassword_number
      Case 1
      Case 2
      Case 3
        uCrypt.Plaintext = "<w«3¹VŒeûö"
        uCrypt.Plaintext = "pw1ao2sr3sd4"
      Case 4
        ' tPlaintext = "n8P[=']e"
        uCrypt.Plaintext = "Jv#NIR]Ti0"
    End Select

    Encrypt_LIS_two_fn = Encrypt_it_fn(uCrypt.Plaintext, True, tPassword_number)

    If Encrypt_LIS_two_fn = uCrypt.Plaintext Then
      Stop
    End If

  End Function

  Function Encrypt_it_fn(ByVal tPlaintext As String, ByVal tEncrypt As Boolean, Optional ByVal tPassword_number As Integer = 0) As String
    ' Written by Michael John Boyack Hodgson, 2013.12.04. Revised 2021.03.13.
    Dim tLocals As udtLocals = Nothing
    ' Dim tIs_unicode As Boolean
    tLocals.Subr = udtApplication_subroutine_names.Encrypt_it_fn
    Encrypt_it_fn = ""

    If tPassword_number = 0 Then tPassword_number = 4
    uCrypt.Private_key = "Copyright © " & uLIS.LLC & ", 1999"
    If tPassword_number = 3 Then uCrypt.Private_key = uLIS.LLC
    uCrypt.Plaintext = tPlaintext
    uCrypt.Test = False

    Select Case tPassword_number
      Case 1
        uCrypt.Plaintext = ""
      Case 2
        uCrypt.Plaintext = "pw1ao2sr3sd4"
      Case 3
        Stop
        ' oEncrypt = New clsEncrypt
        'tIs_unicode = False
        'oEncrypt.Private_key = uCrypt.Private_key
        'oEncrypt.Plaintext = uCrypt.Plaintext

        'fnEncrypt_it = oEncrypt.fnEncrypt_or_decrypt(oEncrypt.Plaintext, tIs_unicode)
        '  tLocals.Msg  = Encrypt_it_fn

        'If uCrypt.Test Then
        '   ' Test decrypt test:
        '   oEncrypt.Plaintext =  tLocals.Msg 
        ' Encrypt_it_fn = oEncrypt.Encrypt_or_decrypt_fn(oEncrypt.Plaintext, tIs_unicode)
        '  tLocals.Msg  = Encrypt_it_fn
        'End If

      Case 4
        If tEncrypt Then
          Encrypt_it_fn = Crypt_rc4_fn(uCrypt.Plaintext, uCrypt.Private_key, True)
          tLocals.Msg = Encrypt_it_fn
        Else
          Encrypt_it_fn = Crypt_rc4_fn(uCrypt.Plaintext, uCrypt.Private_key, False)
        End If

        If uCrypt.Test Then
          tLocals.Action = "Test decrypt test"
          uCrypt.Plaintext = tLocals.Msg
          Encrypt_it_fn = Crypt_rc4_fn(uCrypt.Plaintext, uCrypt.Private_key, False)
          tLocals.Msg = Encrypt_it_fn
        End If

    End Select

    If uCrypt.Test Then
      If tLocals.Msg = tPlaintext Then
        Stop
      Else
        Stop
      End If
    End If

    uOp.Subr = ""
  End Function

#End Region

End Module

