﻿' Module: Database_common.
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner.
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.
' Oledb version.

' Internal reference:

' External reference:
' www.connectionstrings.com.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings. 

Module Database_common
  Public Const cDatabase_common_version As Date = #5/7/2021# ' M/d/y.

#Region "Create, instantiate, set-up"

  Sub Memory_create(ByRef tChanged As Boolean)
    ' Revised 2021.03.07.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Memory_create
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    Call frmMain.Statuslabel1_set(tLocals.Subr & ".")
    tLocals.Success = udtBoolean_extended.Null
    uMemory.Success = False
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check memory.")

Main:
    tLocals.Action = udtApplication_subroutine_names.uMemory_copy_from_progfiles
    Call uMemory_copy_from_progfiles(tChanged)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

    tLocals.Action = udtApplication_subroutine_names.Memory_instantiate
    Call Memory_instantiate()
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      If Not uApp.Initialising Then Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If
    uOp.Called_from = "" ' Allows writing to Memory database.
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Status = udtuStandard_texts.Completion
    Call Log_write(tLocals, 0)
    Call frmMain.ToolStripProgressBar1_set()

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub uMemory_copy_from_progfiles(ByRef tChanged As Boolean)
    ' Revised 2021.04.08.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    tLocals.Subr = udtApplication_subroutine_names.uMemory_copy_from_progfiles
    uOp.Called_from = tLocals.Subr
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 3

    tLocals.Action = udtuStandard_texts.Main_start
    If uMachine.Development_machine Or uMachine.Testing_machine Then
      ' Requires permissions users full control.

      tFile.File_name = udtLivingInfoSystems.LIS_memory_populated_password
      tFile.File_full_name = uLIS.Resources_path & "\" & tFile.File_name
      tLocals.Action = "Check exists 1, " & tFile.File_full_name
      If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
        tFile.Copy_source = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
        If uMachine.Testing_machine Then tFile.Copy_source = uUser.My_dox & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
        If Dir(tFile.Copy_source, FileAttribute.Normal) = "" Then
          Stop ' Permanent, but continue.
        Else
          tFile.Copy_destination = uLIS.Resources_path & "\" & tFile.File_name
          tLocals.Action = "Copy LIS_memory_populated_password from " & tFile.Copy_source & " to " & tFile.Copy_destination
          If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
            Try
              FileCopy(tFile.Copy_source, tFile.Copy_destination)
              tChanged = True
              tLocals.Success = udtBoolean_extended.True_
            Catch ex As Exception
              Stop ' Permanent.
              tLocals.Error_description = ex.Message
              'tLocals.Error_count = tLocals.Error_count + 1
              'tLocals.Success = udtBoolean_extended.False_
              'GoTo Error_process
              uError.Message = tFile.File_name & " could not be copied from" & vbCrLf
              uError.Message = uError.Message & uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources to" & vbCrLf
              uError.Message = uError.Message & uLIS.Resources_path & "." & vbCrLf
              ' uError.Message = uError.Message & "This is a non-critical error."
              uDialogue.Dialogue_result = Msgbox_LIS(uError.Message, udtAlert_types.Information, udtDialogue_types.Okay_only, frmMain.tabControl_main, udtDialogue_position.Top_left)
              If Dir(uMemory.Database) = "" Then uApp.Terminating = uApp.Terminating + 1
            End Try
          End If
        End If
      Else
        tLocals.Continue_ = True
        tLocals.Success = udtBoolean_extended.True_
      End If ' If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
      Call Log_write(tLocals, 0)

      Call frmMain.ToolStripProgressBar1_set()
      tFile.Copy_source = ""
      tFile.Copy_destination = ""
      tFile.File_name = ""
      tFile.File_full_name = ""
      If uApp.Terminating > 0 Then GoTo Error_process
      tLocals.Success = udtBoolean_extended.Null

      tFile.File_name = udtLivingInfoSystems.LIS_memory_populated_no_password
      tFile.File_full_name = uLIS.Resources_path & "\" & tFile.File_name
      tLocals.Action = "Check exists 2, " & tFile.File_full_name
      If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
        tFile.Copy_source = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
        If uMachine.Testing_machine Then tFile.Copy_source = uUser.My_dox & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
        If Dir(tFile.Copy_source, FileAttribute.Normal) = "" Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        Else
          tFile.Copy_destination = uLIS.Resources_path & "\" & tFile.File_name
          tLocals.Action = "Copy LIS_memory_populated_no_password from " & tFile.Copy_source & " to " & tFile.Copy_destination
          If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
            Try
              FileCopy(tFile.Copy_source, tFile.Copy_destination)
              tChanged = True
              tLocals.Success = udtBoolean_extended.True_
            Catch ex As Exception
              Stop ' Permanent.
              tLocals.Error_description = ex.Message
              'tLocals.Error_count = tLocals.Error_count + 1
              'tLocals.Success = udtBoolean_extended.False_
              'GoTo Error_process
              uError.Message = tFile.File_name & " could not be copied from" & vbCrLf
              uError.Message = uError.Message & uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources to" & vbCrLf
              uError.Message = uError.Message & uLIS.Resources_path & "." & vbCrLf
              uError.Message = uError.Message & "This is a non-critical error."
              uDialogue.Dialogue_result = Msgbox_LIS(uError.Message, udtAlert_types.Information, udtDialogue_types.Okay_only, frmMain.tabControl_main, udtDialogue_position.Top_left)
            End Try
          End If
        End If
      Else
        tLocals.Continue_ = True
        tLocals.Success = udtBoolean_extended.True_
      End If
      Call Log_write(tLocals, 0)

      Call frmMain.ToolStripProgressBar1_set()
      tFile.Copy_source = ""
      tFile.Copy_destination = ""
      tFile.File_name = ""
      tFile.File_full_name = ""
    End If ' If uMachine.Development_machine Or uMachine.Testing_machine Then
    tLocals.Success = udtBoolean_extended.Null

    tFile.File_full_name = uLIS.Profile_path & "\" & udtLivingInfoSystems.LIS_memory
    tLocals.Action = "Check exists 3, " & tFile.File_full_name
    If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
      tFile.Copy_source = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_populated_password
      If uMachine.Development_machine Then
        tLocals.Action = "No password"
        tFile.Copy_source = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & udtLivingInfoSystems.LIS_memory_populated_no_password
      End If
      If uMachine.Testing_machine Then
        tLocals.Action = "No password"
        If uUser.My_dox = "" Then
          ' Not defined in Environment_determine yet. Come back here after it is.
          ' tLocals.Success = udtBoolean_extended.True_
          GoTo Ender_partial
        Else
          tFile.Copy_source = uUser.My_dox & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & udtLivingInfoSystems.LIS_memory_populated_no_password
        End If
      End If
      If Dir(tFile.Copy_source, FileAttribute.Normal) = "" Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      Else
        tFile.Copy_destination = uLIS.Profile_path & "\" & udtLivingInfoSystems.LIS_memory
        tLocals.Action = "Copy LIS_memory_populated from " & tFile.Copy_source & " to " & tFile.Copy_destination
        Try
          FileCopy(tFile.Copy_source, tFile.Copy_destination)
          tChanged = True
          tLocals.Success = udtBoolean_extended.True_
        Catch ex As Exception
          uError.Description = ex.Message
          Stop ' Permanent.
          tLocals.Error_description = ex.Message
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End Try
      End If
    Else
      tLocals.Continue_ = True
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

    Call frmMain.ToolStripProgressBar1_set()
    tFile.Copy_source = ""
    tFile.Copy_destination = ""
    tFile.File_name = ""
    tFile.File_full_name = ""

Error_process:
    tLocals.Error_number = Err.Number
    If Err.Description <> "" Then tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    tFile.File_full_name = ""
    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Database_table_create(ByRef tDb As udtDatabase_oledb, ByVal tTable_def As udtMemory_Fields)
    ' Revised 2021.03.29.
    Dim tLocals As udtLocals = Nothing
    Dim i As Integer
    tLocals.Subr = "Create table " & tDb.Table_name
    tDb.Table_name = tTable_def.Table
    uOp.Subr = tLocals.Subr

    tLocals.Action = udtuStandard_texts.Main_start
    If uOp.Called_from <> "First run" Then uOp.Called_from = tLocals.Subr

    ' Watch for Allow Null, etc.
    tLocals.Action = "Create table"
    tDb.Cmd_str = "CREATE TABLE " & tDb.Table_name & " (" & vbCr
    For i = 1 To tTable_def.Length
      tDb.Cmd_str = tDb.Cmd_str & "[" & tTable_def.Name_(i) & "] " & tTable_def.Type_(i)
      If tTable_def.Name_(i) = "Id" Then
        tDb.Cmd_str = tDb.Cmd_str & " NOT NULL PRIMARY KEY"
      End If
      If i = tTable_def.Length Then
        tDb.Cmd_str = tDb.Cmd_str & ")"
      Else
        tDb.Cmd_str = tDb.Cmd_str & ", " & vbCr
      End If
    Next
    Call Database_general(tDb)
    If Not uOp.Success Then
      If InStr(uOp.Result, "already exists.") = 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process
      End If
    End If
    uOp.Success = False

Ender_full:
    tDb.Upgrade_counter = tDb.Upgrade_counter + 1
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " completion." & vbCrLf)
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

    uOp.Called_from = ""
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Database_instantiate(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.06.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    tLocals.Subr = udtApplication_subroutine_names.Database_instantiate
    tLocals.Status = udtuStandard_texts.Initiation
    If uOp.Called_from = udtApplication_subroutine_names.Log_write_memory Then
    Else
      tFile.File_full_name = tDb.Database
      Call Log_write(tLocals, 0)
      tFile.File_full_name = tDb.Database
    End If
    tLocals.Status = ""
    uOp.Success = False

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    With tDb
      If .Database = "" Then
        Call Sound_alert()
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      ElseIf .Database = "Off" Then
        Stop ' Old technology; need to update Kernel with:
        ' uMemory_ProgFiles.Database = ""
        ' or
        ' uMemory_ProgFiles.Database = uLIS.Prog_file_path & "\Resources\" & udtLivingInfoSystems.LIS_memory_common
      Else
        tLocals.Action = "Instantiate database " & .Database
        If Dir(.Database, FileAttribute.Normal) = "" Then
          Stop ' Permanent.
          Call Memory_create(Nothing)
          If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
          End If
          uOp.Success = False
        End If
        .Indent = Space(8)
        tLocals.Action = "Set tDb.Conn_str"
        If LCase(Microsoft.VisualBasic.Right(.Database, 4)) = ".mdb" Then
          .Database_version = 11
          .Conn_str = "Provider=Microsoft.Jet.OLEDB.4.0; Data Source=" & .Database
        ElseIf LCase(Microsoft.VisualBasic.Right(.Database, 6)) = ".accdb" Then
          .Conn_str = "Provider=Microsoft.ACE.OLEDB.12.0; Data Source=" & .Database ' Should work with mdb as well.
          .Database_version = 12
        Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        tLocals.Action = "Set tDb.Use_password"
        If InStr(.Database, "password") > 0 Then
          If InStr(.Database, "no_password") = 0 Then
            .Use_password = True
          End If
        End If
        If (Not uMachine.Development_machine And Not uMachine.Testing_machine) Or .Use_password Then
          .Conn_str = .Conn_str & ";Jet OLEDB:Database Password=" & Encrypt_LIS_fn()
        End If
        tLocals.Action = "Set tDb.Conn"
        .Conn = New OleDb.OleDbConnection(.Conn_str)
        tLocals.Action = "Set tDb.Cmd"
        .Cmd = New OleDb.OleDbCommand
        tLocals.Action = "Set tDb.Cmd.Connection"
        .Cmd.Connection = .Conn
        tLocals.Action = "Set tDb.Cmd.CommandType"
        .Cmd.CommandType = CommandType.Text
        .Success = True
        tLocals.Success = udtBoolean_extended.True_
      End If ' If .Database = "" Then
    End With

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If uOp.Called_from <> udtApplication_subroutine_names.Log_write_memory Then
      Call Log_write(tLocals, 0)
      uOp.Success = False
    End If

Ender_partial:
    tDb.Success = False
    tFile.File_full_name = ""
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Memory_instantiate()
    ' Separated 2014.01.24. Revised 2021.03.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Memory_instantiate
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    uOp.Called_from = tLocals.Subr
    tLocals.Status = ""

    tLocals.Action = udtuStandard_texts.Main_start
    With uMemory
      .Database = uLIS.Profile_path & "\" & udtLivingInfoSystems.LIS_memory
      ' If that fails, use:
      ' .Database = ulis.Profile_path & "\" & uApp.Internal_name & "\Memory.mdb"
      If Dir(.Database, FileAttribute.Normal) = "" Then
        If uApp.First_run_in_progress Then
          tLocals.Success = udtBoolean_extended.False_
          GoTo Ender_full
        Else
          tLocals.Action = "Call First_run"
          Call First_run() ' Check everything.
          If Dir(.Database, FileAttribute.Normal) = "" Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
            .Database = uLIS.Profile_path & "\" & uApp.Internal_name & "\" & udtLivingInfoSystems.LIS_memory
          End If
        End If
      Else
        tLocals.Action = "Call Database_instantiate(uMemory)"
        Call Database_instantiate(uMemory)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        .Initialised = True
      End If
    End With

    If uMemory_ProgFiles.Database <> "" Then
      tLocals.Action = "Call Database_instantiate(uMemory_ProgFiles)"
      Call Database_instantiate(uMemory_ProgFiles)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
    End If
    uOp.Success = False

    tLocals.Action = "Call Database_instantiate_local"
    Call Database_instantiate_local()
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

    Call frmMain.Statuslabel1_set()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Function tDb_point_to_uMemory_fn(ByRef tDb As udtDatabase_oledb) As Boolean
    ' Written MJBH. Revised 2021.02.08.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.tDb_point_to_uMemory

    If uMemory.Database Is Nothing Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

    With tDb
      .Database = uMemory.Database
      Call Database_instantiate(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      .Use_password = uMemory.Use_password

      tDb_point_to_uMemory_fn = True
    End With

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Sub Memory_define_Log()
    ' Revised 2021.04.04.
    With uDef_Log
      .Table = uTables.Log
      uOp.Subr = "Memory define " & .Table
      .Test_field = "User"
      .Length = 20
      ReDim .Name_(.Length)
      ReDim .Type_(.Length)
      ReDim .Value(.Length)
      ReDim .Condition_operator(.Length)
      ReDim .Condition_value(.Length)

      .Name_(1) = "Id"
      .Type_(1) = "Autoincrement"
      .Name_(2) = "Date_"
      .Type_(2) = "Date"
      .Name_(3) = "Time_"
      .Type_(3) = "Time"
      .Name_(4) = "Computer_name"
      .Type_(4) = "Text"
      .Name_(5) = "Application"
      .Type_(5) = "Text"
      .Name_(6) = "Corporation"
      .Type_(6) = "Text"
      .Name_(7) = "Branch"
      .Type_(7) = "Text"
      .Name_(8) = "User"
      .Type_(8) = "Text"
      .Name_(9) = "Event_"
      .Type_(9) = "Text"
      .Name_(10) = "Subr"
      .Type_(10) = "Text"
      .Name_(11) = "Action"
      .Type_(11) = "Text"
      .Name_(12) = "OpenF"
      .Type_(12) = "Text"
      .Name_(13) = "Document"
      .Type_(13) = "Text"
      .Name_(14) = "Error_number"
      .Type_(14) = "Integer"
      .Name_(15) = "Error_description"
      .Type_(15) = "Text"
      .Name_(16) = "Error_extra"
      .Type_(16) = "Text"
      .Name_(17) = "Error_message"
      .Type_(17) = "Text"
      .Name_(18) = "Error_source"
      .Type_(18) = "Text"
      .Name_(19) = "Status"
      .Type_(19) = "Text"
      .Name_(20) = "Success"
      .Type_(20) = "Bit"
    End With

    uOp.Subr = ""
  End Sub

#End Region ' Create, instantiate, set-up.

#Region "Database diagnostic"

  Sub Database_diagnostic_00(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.04.

    ' A Level 1 diagnostic: A thorough troubleshooting test series, which requires taking the associated system off-line.
    ' Requires human input and can take hours. A level 1 diagnostic requires a team of crew members to physically
    ' verify operation of system mechanisms and to system readings, rather than depending on the automated programmes,
    ' thereby guarding against possible malfunctions in self-testing hardware and software.
    ' A level 2 diagnostic: A thorough troubleshooting test series, which does not require taking the associated system off-line.
    ' Like a level 1, involves extensive automated routines, but requires crew verification of fewer operational elements.
    ' This yields a somewhat less reliable system analysis, but is a procedure that can be conducted in less than half the
    ' time of the more complex tests.
    ' A level 3 diagnostic: Involves crew verification of only key mechanics and systems readings. Level 3 diagnostics are
    ' intended to be performed in ten minutes or less.
    ' A Level 4 diagnostic: Involves more sophisticated batteries of automated diagnostics. For most systems, Level 4
    ' diagnostics can be performed in under 30 seconds.
    ' A Level 5 diagnostic: Aa fully automated programme, fully run by a computer and only takes seconds;

    ' http://scifi.stackexchange.com/questions/31189/meaning-of-levels-of-diagnostic-tests-in-star-trek:
    ' 1    Take system out of service, full shutdown, manual inspection, observed re-initialisation, manual testing.
    ' 2    Take system out of service, limited manual inspection, crew monitored automated tests.
    ' 3    Take system out of service, minimal manual inspection, automated tests.
    ' 4    Take system out of service, automated tests.
    ' 5    Automated tests.

    ' Warning: speculation (although it is supported by good engineering princples):
    ' The higher (or lower depending on how you do it) number means that you get more detailed information.
    ' Some people ask why you would not put it on the highest level at all times, the reason you do not and you
    ' try to pick your level carefully is that you risk too much information being spewed out and causing you to
    ' miss something important or to have to spend a lot of time sifting through information to find what you need.
    ' Also, more output means more time spent producing output - not much of a consideration normally but on a
    ' real time machine (like a nuclear power plant monitor) a couple of microseconds writing output and not
    ' monitoring the systems can be serious. There is also the fact that more diagnostics means bigger log files
    ' which can cause an issue on some systems where users do not delete them and you end up with, literally,
    ' years of log files eating up disc space.

    Dim tLocals As udtLocals = Nothing
    Dim tCalled_from As String = uOp.Called_from
    Dim tTable_name As String
    Dim tCmd_str As String
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_00
    tLocals.Status = udtuStandard_texts.Initiation
    uOp.Diagnostics.In_progress = True
    tLocals.Msg = "Performing database level five diagnostic..."
    Call frmMain.Statuslabel1_set(tLocals.Msg)
    tTable_name = ""
    tCmd_str = ""
    tLocals.Status = ""

    If tDb.Cmd_str = "" Then
      Call Error_write_mega(tLocals, "", tDb.Table_name)
    Else
      Call Error_write_mega(tLocals, "Command string:" & vbCr & tDb.Cmd_str, tDb.Table_name)
    End If
    uOp.Called_from = tCalled_from

    Call Error_clear(tLocals)

    tTable_name = tDb.Table_name
    tCmd_str = tDb.Cmd_str
    If tDb.Upgrade_table = "" Then tDb.Upgrade_table = tDb.Table_name
    Call Database_diagnostic_level_05(tDb)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Continue_ = False
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uOp.Success = False
      ' GoTo Error_process
    End If
    uOp.Success = False

    If tTable_name <> "" And tCmd_str <> "" Then
      tDb.Table_name = tTable_name
      tDb.Cmd_str = tCmd_str
      Call Database_general(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Continue_ = False
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        ' GoTo Error_process
      End If
      uOp.Success = False

      tDb.Upgrade_table = ""
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Database_diagnostic_level_05(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.01.
    ' A Level 5 diagnostic is a fully automated programme, fully run by a computer and only takes seconds.
    ' Top level. Knows each kernel table, and calls the first upgrade step if necessary.
    If tDb.Upgrade_table = "" Then Exit Sub
    Dim tLocals As udtLocals = Nothing
    Dim tTable As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_level_05
    tLocals.Status = udtuStandard_texts.Initiation
    On Error GoTo Error_captured
    Call frmMain.Statuslabel1_set("Running system wide memory diagnostic tests.", Color.Red, FontStyle.Bold)
    uApp.Progressbar_step = uApp.Progressbar_step + 1
    uOp.Diagnostics.In_progress = True
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    If Not (System.IO.File.Exists(tDb.Database)) Then
      Call Memory_create(Nothing)
    End If
    tLocals.Action = "Checking database " & tDb.Database
    Call frmMain.txtDiagnostic_result_append(tLocals.Action & "." & vbCrLf)

    If tDb.Database <> uMemory_ProgFiles.Database Then
      tLocals.Continue_ = True

      If tDb.Upgrade_table = uTables.Log Or tDb.Upgrade_table = "All" Then
        tDb.Table_name = uTables.Log
        tLocals.Action = "Checking table " & tDb.Table_name
        Call frmMain.txtDiagnostic_result_append(tLocals.Action & "...")
        If uDef_Log.Length = 0 Then Call Memory_define_Log()
        Call Database_diagnostic_level_05_part_01(tDb, uDef_Log)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process ' Continue.
        End If
        uOp.Success = False
      End If

      If tDb.Upgrade_table = udtTables.Preference Or tDb.Upgrade_table = "All" Then
        tDb.Table_name = udtTables.Preference
        tLocals.Action = "Checking table " & tDb.Table_name
        Call frmMain.txtDiagnostic_result_append(tLocals.Action & "...")
        If uDef_Preference.Length = 0 Then Call Memory_define_Preference()
        Call Database_diagnostic_level_05_part_01(tDb, uDef_Preference)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process ' Continue.
        Else
          If uOp.Preferences_loaded Then
            uOp.Success = False
            uUser.Preferences_changed = True
            Call Preferences_write()
          End If
        End If
        uOp.Success = False
      End If

    End If

    If tDb.Upgrade_table <> "" Then
      Call Database_diagnostic_local(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process ' Continue.
      End If
      uOp.Success = False
    End If
    If tDb.Upgrade_table <> "" And tDb.Upgrade_table <> "All" Then Stop ' No upgrade was done.

    If tDb.Upgrade_table = "" Then
      tDb.Upgrade_in_progress = False
      If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    ElseIf tDb.Upgrade_table = "All" Then
      ' So far so good, carry on.
      tLocals.Success = udtBoolean_extended.True_
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process ' Continue.
    End If

    If uMachine.Development_machine Then
      If Not uOp.Diagnostics.All_tables_in_progress Then
        Call Diagnostic_all_tables()
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process ' Continue.
        End If
        uOp.Success = False
      End If
    End If

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            Call Error_clear(tLocals)
            Resume
          Case Else
            Call Database_error_handler(tLocals, tDb)
        End Select
      End If
    End If

Ender_full:
    If Not uOp.Diagnostics.All_tables_in_progress Then uOp.Diagnostics.In_progress = False
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed." & vbCrLf & vbCrLf)
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf & vbCrLf)
    End If
    Call Log_write(tLocals, 0)

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Database_diagnostic_level_05_part_01(ByRef tDb As udtDatabase_oledb, ByRef tTable_def As udtMemory_Fields)
    ' Revised 2021.04.07.
    ' Knows and co-ordinates each step of the diagnostic process.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing
    Dim tField As String = ""
    On Error GoTo Error_captured
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_level_05_part_01
    tLocals.Action = udtuStandard_texts.Initiation
    tLocals.Status = udtuStandard_texts.Initiation
    tLocals.Continue_ = True ' Until set false.
    tLocals.Error_count = 0
    tLocals.Success = udtBoolean_extended.False_
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    tDb.Upgrade_in_progress = True
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1

    tDb.Table_name = tTable_def.Table
    tLocals.Action = "Check if the table exists"
    Call Database_diagnostic_level_05_part_02(tDb)

    If tDb.Success Or uOp.Success Then
      tDb.Success = False
      uOp.Success = False
      tLocals.Action = "If the table exists, check each field"
      Call Database_diagnostic_level_05_part_04(tDb, tTable_def, tTable_def.Test_field)

      If tDb.Upgrade_counter < 2 And tField <> "" And tDb.Upgrade_table <> "All" Then
        tField = Replace(tField, "The INSERT INTO statement contains the following unknown field name: '", "")
        tField = Replace(tField, "'.  Make sure you have typed the name correctly, and try the operation again.", "")
      Else
        tField = ""
      End If
    End If

Upgrade_section:
    If tLocals.Action = udtuStandard_texts.Upgrade Then
      Call frmMain.txtDiagnostic_result_append("Adding fields to " & tDb.Table_name & "... ")

      If tField = "" And tDb.Error_ <> "" Then tField = tDb.Error_
      If tField = "" Then
        Call Database_upgrade_insert(tDb, tTable_def)
        tSearch.Keyword = Left$(tSearch.Keyword, Len(tSearch.Keyword) - 2)
        tLocals.Action = "Fields added to " & tTable_def.Table & " table; " & Trim(tSearch.Keyword)
      Else
        tField = Replace(tField, "The INSERT INTO statement contains the following unknown field name: '", "")
        tField = Replace(tField, "'.  Make sure you have typed the name correctly, and try the operation again.", "")
        Call Database_upgrade_insert(tDb, tTable_def, tField)
        If uOp.Success Then
          tLocals.Action = "Fields added to " & tTable_def.Table & " table; " & tField
          tField = ""
        Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
      End If ' If tField = "" Then

      tLocals.Action = "Test it again"
      tDb.Table_name = tTable_def.Table
      Call Database_diagnostic_level_05_part_04(tDb, tTable_def, tTable_def.Test_field)
      Call frmMain.ToolStripProgressBar1_set()

    ElseIf tLocals.Action = "Create db" Then
      Call Memory_create(Nothing)
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If

    ElseIf tLocals.Action = "Create" Then
      Call Database_table_create(tDb, tTable_def)
      If tDb.Success Then
        If tTable_def.Table = udtTables.Preference Then
          Call Preference_insert_first(tDb)
        End If

        tLocals.Success = udtBoolean_extended.True_
        tLocals.Action = "Table " & tTable_def.Table & " added to " & tDb.Database
        Call Log_write(tLocals, 0)

      Else
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If

    End If ' If tLocals.Action = udtuStandard_texts.Upgrade Then

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        uOp.Called_from = tLocals.Subr
        Call Database_error_handler(tLocals, tDb)
        Call Error_clear(tLocals)
        If tLocals.Action = "" Then
          Resume Error_captured
        Else
          If tLocals.Error_count > 0 Then tLocals.Error_count = tLocals.Error_count - 1
          tLocals.Success = udtBoolean_extended.Null
          Resume Upgrade_section
        End If
      End If
    End If

Ender_full:
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)

Ender_partial:
    If tDb.Upgrade_table <> "All" Then tDb.Upgrade_table = ""
    tDb.Success = False
    Call frmMain.Statuslabel1_set("Running system wide memory diagnostic tests.", Color.Red, FontStyle.Bold)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub Database_diagnostic_level_05_part_02(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.03.29.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_level_05_part_02
    tLocals.Action = "Check if the table exists"

    With tDb
      .Rows_affected = 0
      tLocals.Action = "Check table " & .Table_name
      .Cmd_str = "SELECT TOP 1 * FROM " & .Table_name
      If .Table_name = udtTables.Preference Then
        .Cmd_str = .Cmd_str & " WHERE " & uDef_Preference.Test_field & "='" & uApp.Name_for_preferences & "'"
      End If
      tDb.Data_adapter = New OleDb.OleDbDataAdapter(tDb.Cmd_str, tDb.Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, tDb.Table_name)
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count

      .Cmd_str = ""
      .Success = True
      Call frmMain.txtDiagnostic_result_append(" Table " & .Table_name & " exists. ")

      tLocals.Action = udtuStandard_texts.Dataset_clear
      .Dataset.Clear()
      .Dataset.Dispose()
      .Dataset = Nothing
      .Data_adapter.Dispose()
      .Data_adapter = Nothing
    End With

    uOp.Success = True
  End Sub

  Private Sub Database_diagnostic_level_05_part_03(ByRef tDb As udtDatabase_oledb, ByVal tField As String)
    ' Revised 2021.04.01.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_level_05_part_03

    tLocals.Action = "Check if a record exists"
    With tDb
      .Rows_affected = 0
      tLocals.Action = "Check " & .Table_name
      .Cmd_str = "SELECT TOP 1 " & tField & " FROM " & .Table_name & " WHERE App_name='" & uApp.Internal_name & "'"

      .Data_adapter = New OleDb.OleDbDataAdapter(.Cmd_str, .Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, .Table_name)
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count

      tLocals.Action = udtuStandard_texts.Dataset_clear
      .Dataset.Clear()
      .Dataset.Dispose()
      .Dataset = Nothing
      .Data_adapter.Dispose()
      .Data_adapter = Nothing
      .Cmd_str = ""

    End With

    uOp.Subr = ""
  End Sub

  Private Sub Database_diagnostic_level_05_part_04(ByRef tDb As udtDatabase_oledb, ByRef tTable_var As udtMemory_Fields, ByVal tTest_field As String)
    ' Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing
    Dim i As Integer
    Dim tTest_type As String = ""
    Dim tInsert As String
    Dim tDelete As String
    tLocals.Subr = udtApplication_subroutine_names.Database_diagnostic_level_05_part_04
    tLocals.Status = ""

    tLocals.Action = "Test write, to check that all fields are there"
    Call frmMain.txtDiagnostic_result_append("Check all fields... ")

    tLocals.Action = udtuStandard_texts.Sql_build_insert
    tInsert = "INSERT INTO [" & tDb.Table_name & "] (" & vbCr
    For i = 1 To tTable_var.Length
      If uApp.Debug Then tLocals.Msg = tTable_var.Name_(i)
      If tTable_var.Type_(i) = "Autoincrement" Then
        ' Skip.
      Else
        tInsert = tInsert & tDb.Indent & "[" & tTable_var.Name_(i) & "]"
      End If
      If i <> tTable_var.Length And tTable_var.Type_(i) <> "Autoincrement" Then
        tInsert = tInsert & ", " & vbCr
      End If
      If tTable_var.Name_(i) = tTest_field Then
        tTest_type = tTable_var.Type_(i)
        If tTest_type <> "Text" Then
          Stop ' Always ensure the test field is a text field.
          tLocals.Continue_ = False
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          uApp.Terminating = uApp.Terminating + 1 ' Critical error.
          GoTo Error_process
        End If
      End If
    Next
    tInsert = tInsert & vbCr & ") VALUES (" & vbCr
    For i = 1 To tTable_var.Length
      If uApp.Debug Then tLocals.Msg = tTable_var.Type_(i)
      Select Case tTable_var.Type_(i)
        Case "Autoincrement"
          ' Do nothing. 
        Case "Bit"
          tInsert = tInsert & tDb.Indent & "True"
        Case "Decimal", "Integer", "Single", "Short"
          tInsert = tInsert & tDb.Indent & "1"
        Case "Text", "VARCHAR(MAX)"
          tInsert = tInsert & tDb.Indent & "'Test'"
        Case "Date"
          ' tInsert = tInsert & tDb.Indent & "'" & Date.Today & "'"
          tInsert = tInsert & tDb.Indent & "'" & udtuStandard_texts.Date_null & "'"
        Case "Time"
          tInsert = tInsert & tDb.Indent & "'" & TimeString & "'"
        Case Else
          tLocals.Msg = tTable_var.Type_(i)
          ' Check that the gDef_*.length is the correct value.
          Stop
      End Select
      If i = tTable_var.Length Then
        tInsert = tInsert & ")"
      Else
        If tTable_var.Type_(i) <> "Autoincrement" Then
          tInsert = tInsert & ", " & vbCr
        End If
      End If
    Next

    tLocals.Action = udtuStandard_texts.Sql_build_delete
    tDelete = "DELETE FROM " & tDb.Table_name & " WHERE " & tTest_field & "="
    If uApp.Debug Then tLocals.Msg = tTest_field.GetTypeCode()
    Select Case tTest_type
      Case "Autoincrement"
        Stop ' Invalid. Do nothing.
        'Case "Bit"
        '   stop
        '   tDelete = tDelete & tDb.Indent & "True"
        'Case "Decimal", "Integer", "Single", "Short"
        '   stop
        '   tDelete = tDelete & "'1'"
      Case "Text", "VARCHAR(MAX)"
        tDelete = tDelete & "'Test'"
        'Case "Date"
        '  Stop ' Always ensure the test field is a text field.
        '  Stop ' Use Sql_builder_format_date_fn.
        '  tDelete = tDelete & "Format(#" & Replace(udtuStandard_texts.Date_null, ".", "-") & "#, " & Chr(34) & "yyyy.mm.dd" & Chr(34) & ")"
        'Case "Time"
        '   tDelete = tDelete & "'" & TimeString & "'"
      Case Else
        Stop ' Always ensure the test field is a text field.
    End Select

    Call Sql_write(tInsert & vbCr & tDelete)

    With tDb
      .Rows_affected = -1
      .Success = False
      If .Conn.State <> ConnectionState.Open Then .Conn.Open()
      tLocals.Action = "Check " & .Table_name
      .Cmd_str = tInsert
      .Cmd.CommandText = .Cmd_str
      .Cmd.ExecuteNonQuery()
      .Success = True ' INSERT INTO table with all fields works.
      Call frmMain.txtDiagnostic_result_append("All fields exist." & vbCrLf)
      .Cmd_str = tDelete
      .Cmd.CommandText = .Cmd_str
      .Cmd.ExecuteNonQuery()
      .Success = True ' DELETE FROM table with all fields works.
      .Conn.Close()
      .Cmd_str = ""
      .Cmd.CommandText = .Cmd_str
    End With

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check cancelled." & vbCrLf)
    End If

Ender_full:
    uOp.Subr = ""
  End Sub

  Private Sub Database_upgrade_insert(ByRef tDb As udtDatabase_oledb, ByVal tTable_var As udtMemory_Fields, Optional ByVal tField As String = "")
    ' Revised 2021.04.07.
    ' Insert any omitted fields.
    ' Try to insert each field; for any that do, the Insert will fail gracefully; for any that don't exist,
    ' the Insert will create it.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing

    Dim i As Integer
    Dim tTable As String
    tLocals.Subr = udtApplication_subroutine_names.Database_upgrade_insert
    tLocals.Status = ""
    tTable = tDb.Table_name
    tSearch.Keyword = ""

    If tField = "" Then
      For i = 1 To tTable_var.Length
        tDb.Table_name = tTable
        tLocals.Action = Replace(Replace("Insert field " & tTable_var.Name_(i), "[", ""), "]", "")

        tLocals.Action = udtuStandard_texts.Sql_build_alter
        tDb.Cmd_str = "ALTER TABLE " & tDb.Table_name & " ADD [" & tTable_var.Name_(i) & "] " & tTable_var.Type_(i)
        If tTable_var.Name_(i) = "Id" Or tTable_var.Name_(i) = "[Id]" Then
          tDb.Cmd_str = tDb.Cmd_str & " NOT NULL PRIMARY KEY"
        End If
        Call Database_general(tDb)
        If tDb.Success Then
          tSearch.Keyword = tSearch.Keyword & tTable_var.Name_(i) & ", "
          Call frmMain.txtDiagnostic_result_append("Added field '" & tTable_var.Name_(i) & "'. ")
        Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
      Next
    Else
      Dim tType As String = ""
      tType = tTable_var.Type_(Array.IndexOf(tTable_var.Name_, tField))
      Call Database_upgrade_insert_one(tDb, tField, tType)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If
    uOp.Success = tDb.Success

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub Database_upgrade_insert_one(ByRef tDb As udtDatabase_oledb, ByVal tField As String, ByVal tType As String)
    ' Separated 2016.09.03. Revised 2021.01.30.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_upgrade_insert_one
    tLocals.Action = "Insert field " & tField & " into " & tDb.Table_name

    Call Sql_builder_alter_one(tDb, tField, tType)
    If uOp.Success Then
      tDb.Upgrade_counter = tDb.Upgrade_counter + 1
      Call frmMain.txtDiagnostic_result_append("Added field '" & tField & "'. ")
      uOp.Success = True
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uOp.Success = False
    End If

    uOp.Subr = ""
  End Sub

  Sub Diagnostic_all_tables(Optional ByVal tDatabase As String = "")
    ' Written 2020.05.18. Revised 2021.04.08.
    If uOp.Diagnostics.All_tables_done Then Exit Sub
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim tUpgraded As Boolean = False
    Dim tDb As udtDatabase_oledb = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Diagnostic_all_tables
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 135
    Call frmMain.Statuslabel1_set("Running system wide memory diagnostic tests.", Color.Red, FontStyle.Bold)
    tLocals.Msg = ""
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    uOp.Diagnostics.All_tables_in_progress = True

    If tDatabase = "" Then tDatabase = uMemory.Database
    If tDatabase = "" Or Not System.IO.File.Exists(tDatabase) Then GoTo Ender_full
    tDb.Database = tDatabase
    If tDb.Conn_str = "" Then
      Call Database_instantiate(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
    End If

    tDb.Upgrade_table = "All"
    Call Database_diagnostic_level_05(tDb)
    ' No error diversion; must continue processing.

    tLocals.Status = udtuStandard_texts.Initiation
    If tDb.Upgrade_counter > 0 Then tUpgraded = True
    tFile.File_full_name = tDatabase
    Call Log_write(tLocals, 0)
    tLocals.Status = ""

    If tDatabase = uMemory.Database Then
      tDatabase = uMemory_ProgFiles.Database
      If tDatabase <> "" Then Call Diagnostic_all_tables(tDatabase)
      uOp.Diagnostics.All_tables_in_progress = True
      uOp.Diagnostics.All_tables_done = False
      uOp.Success = False

      If tDatabase = uMemory_ProgFiles.Database Then
        tDatabase = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_populated_password
        If tDatabase <> "" Then Call Diagnostic_all_tables(tDatabase)
        uOp.Diagnostics.All_tables_in_progress = True
        uOp.Diagnostics.All_tables_done = False
        uOp.Success = False

        If tDatabase = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_populated_password Then
          tDatabase = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_populated_no_password
          If tDatabase <> "" Then Call Diagnostic_all_tables(tDatabase)
          uOp.Diagnostics.All_tables_in_progress = True
          uOp.Diagnostics.All_tables_done = False
          uOp.Success = False

          If tDatabase = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_populated_no_password Then
            If uMachine.Development_machine Or uMachine.Testing_machine Then
              tDatabase = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & udtLivingInfoSystems.LIS_memory_populated_no_password
              If tDatabase <> "" Then Call Diagnostic_all_tables(tDatabase)
              uOp.Diagnostics.All_tables_in_progress = True
              uOp.Diagnostics.All_tables_done = False
              uOp.Success = False

              If tDatabase = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & udtLivingInfoSystems.LIS_memory_populated_no_password Then
                tDatabase = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & udtLivingInfoSystems.LIS_memory_populated_password
                If tDatabase <> "" Then Call Diagnostic_all_tables(tDatabase)
                uOp.Success = False
              End If
            End If ' If uMachine.Development_machine Or uMachine.Testing_machine Then

          End If

        Else
          uOp.Diagnostics.All_tables_in_progress = False
          uOp.Diagnostics.All_tables_done = True
        End If
      Else
        uOp.Diagnostics.All_tables_in_progress = False
        uOp.Diagnostics.All_tables_done = True
      End If
    Else
      uOp.Diagnostics.All_tables_in_progress = False
      uOp.Diagnostics.All_tables_done = True
    End If

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    tFile.File_full_name = tDatabase
    Call Log_write(tLocals, 0)
    tDb.Upgrade_table = ""
    tFile.File_full_name = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

#Region "Data fixes"
  ' Written MJBH 2019.07.18. Revised 2020.04.06.

  Sub Data_fix_currency_sign_to_currency_symbol(ByRef tUpgraded As Boolean)
    ' Written MJBH 2019.07.18. Revised 2021.02.21.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Data_fix_currency_sign_to_currency_symbol
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = "Data fix"
    tLocals.Status = ""

    Call Data_fix_field_rename(udtTables.Preference, "Currency_sign", "Currency_symbol", "Text", tUpgraded)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
      ' GoTo Error_handler
    Else
      tLocals.Success = udtBoolean_extended.True_
      tUpgraded = True
    End If

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Data_fix_currency_to_currency_(ByRef tUpgraded As Boolean)
    ' Written MJBH 2019.10.15. Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Data_fix_currency_to_currency_
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = uOp.Subr
    Call Error_clear(tLocals)
    uOp.Subr = tLocals.Subr
    tLocals.Status = ""

    tLocals.Action = "Call " & udtApplication_subroutine_names.Data_fix_field_rename
    Call Data_fix_field_rename(udtTables.Preference, "Currency", "Currency_", "Text", tUpgraded)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_handler 
    Else
      tLocals.Success = udtBoolean_extended.True_
      tUpgraded = True
    End If

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Data_fix_status(ByRef tUpgraded As Boolean)
    ' Written 2021.01.25. Revised 2021.03.21.
    ' Where uMemory.~log.Action = a status, move it to uMemory.~log.Status.
    Dim tLocals As udtLocals = Nothing
    Dim tDb As udtDatabase_oledb = Nothing
    tLocals.Subr = "Data fix Log.Status"
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = uOp.Subr
    tLocals.Status = ""

    Call tDb_point_to_uMemory_fn(tDb)
    uOp.Success = False

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    With tDb
      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Initiation & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Initiation & "'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Initiation & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Initiation & ".'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Failed & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Failed & "'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Failed & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Failed & ".'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Cancelled & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Cancelled & "'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Cancelled & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Cancelled & ".'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Completion & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Completion & "'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

      .Table_name = uTables.Log
      .Cmd_str = udtuStandard_texts.Sql_update & " " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "SET   Status='" & udtuStandard_texts.Completion & ".'" & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Action='" & udtuStandard_texts.Completion & ".'"
      Call Database_general(tDb)
      If Not uOp.Success Then
        tLocals.Status = udtuStandard_texts.Failed
      End If
      uOp.Result = ""

    End With

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Data_fix_date_to_date_(ByRef tUpgraded As Boolean)
    ' Written MJBH 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Data fix date to date_"
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = uOp.Subr
    Call Error_clear(tLocals)
    uOp.Subr = tLocals.Subr
    tLocals.Status = ""

    tLocals.Action = "Call " & udtApplication_subroutine_names.Data_fix_field_rename
    Call Data_fix_field_rename(uTables.Log, "Date", "Date_", "Date", tUpgraded)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_handler 
    Else
      tLocals.Success = udtBoolean_extended.True_
      tUpgraded = True
    End If

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Data_fix_time_to_time_(ByRef tUpgraded As Boolean)
    ' Written MJBH 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Data fix time to time_"
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = uOp.Subr
    Call Error_clear(tLocals)
    uOp.Subr = tLocals.Subr
    tLocals.Status = ""

    tLocals.Action = "Call " & udtApplication_subroutine_names.Data_fix_field_rename
    Call Data_fix_field_rename(uTables.Log, "Time", "Time_", "Time", tUpgraded)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_handler 
    Else
      tLocals.Success = udtBoolean_extended.True_
      tUpgraded = True
    End If

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Data_fix_field_rename(ByVal tTable As String, ByVal tField_name_old As String, ByVal tField_name_new As String, ByVal tType As String, ByRef tUpgraded As Boolean)
    ' Separated MJBH 2019.10.20. Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Data_fix_field_rename
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Called_from = tLocals.Subr
    Call Error_clear(tLocals)
    uOp.Subr = tLocals.Subr
    tLocals.Status = ""

    tLocals.Action = "Test if field " & tField_name_old & " exists"
    uMemory.Table_name = tTable
    uMemory.Cmd_str = "SELECT Id, " & tField_name_old & " FROM " & uMemory.Table_name
    Call Database_general(uMemory)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    If uDialogue.Active Then
      tLocals.Success = udtBoolean_extended.False_
      uApp.Terminating = uApp.Terminating + 1
      GoTo Ender_full
    End If
    uOp.Success = False

    If uMemory.Rows_affected > 0 Then
      tLocals.Action = "Create the new field, " & tField_name_new
      tLocals.Success = udtBoolean_extended.False_
      uMemory.Table_name = tTable
      uMemory.Cmd_str = "ALTER TABLE " & uMemory.Table_name & vbCrLf
      uMemory.Cmd_str = uMemory.Cmd_str & "ADD " & tField_name_new & " " & tType
      Call Database_general(uMemory)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      If uDialogue.Active Then
        tLocals.Success = udtBoolean_extended.False_
        uApp.Terminating = uApp.Terminating + 1
        GoTo Ender_full
      End If
      uOp.Success = False

      tLocals.Action = "Copy all data from " & tField_name_old & " to " & tField_name_new
      tUpgraded = True
      tLocals.Success = udtBoolean_extended.Null
      uMemory.Table_name = udtTables.Preference
      uMemory.Cmd_str = "UPDATE " & uMemory.Table_name & vbCrLf
      uMemory.Cmd_str = uMemory.Cmd_str & "SET " & tField_name_new & " = " & tField_name_old
      Call Database_general(uMemory)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      If uDialogue.Active Then
        tLocals.Success = udtBoolean_extended.False_
        uApp.Terminating = uApp.Terminating + 1
        GoTo Ender_full
      End If
    End If
    uOp.Success = False

    tLocals.Action = "Delete the old field, " & tField_name_old
    uMemory.Table_name = udtTables.Preference
    uMemory.Cmd_str = "ALTER TABLE " & uMemory.Table_name & vbCrLf
    uMemory.Cmd_str = uMemory.Cmd_str & "DROP COLUMN [" & tField_name_old & "]"
    Call Database_general(uMemory)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    If uDialogue.Active Then
      tLocals.Success = udtBoolean_extended.False_
      uApp.Terminating = uApp.Terminating + 1
    End If
    uOp.Success = False

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       tLocals.Success = udtBoolean_extended.False_ Or tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      ' Reduce all error responses down to tLocals.Success = udtBoolean_extended.False_
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If
    Call Log_write(tLocals, 0)

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

#End Region ' End region, "Data_fixes"

#End Region ' End region, "Database diagnostic."

#Region "Preferences"

  Sub Preference_insert_first(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.04.17.
    Dim tLocals As udtLocals = Nothing
    Dim tCmd_str As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Preference_insert_first
    uApp.First_run_in_progress = True

    tLocals.Action = udtuStandard_texts.Main_start
    If uDef_Preference.Length = 0 Then Call Memory_define_Preference()
    uDef_Preference.Value(2) = uApp.Name_for_preferences
    uDef_Preference.Value(4) = uApp.App_revision_date
    Call Sql_builder_insert(tDb, uDef_Preference)
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      ' If Not tLocals.Continue_ Then GoTo Error_process
    End If
    If uOp.Success Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)

    uOp.Subr = ""
    If tLocals.Success <> udtBoolean_extended.False_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Preferences_load(Optional ByVal tMultiple As Boolean = False)
    ' Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tSearch As udtSearch = Nothing

    Dim tBool As Boolean
    Dim tNum As Single
    Dim tDate As Date
    Dim tUpgraded As Boolean
    Dim tApp_revision_last As Date
    Dim tString As String
    Dim tCalled_from As String
    tLocals.Subr = udtApplication_subroutine_names.Preferences_load
    tLocals.Status = udtuStandard_texts.Initiation
    On Error GoTo Error_captured
    uOp.Subr = tLocals.Subr
    tCalled_from = uOp.Called_from
    tBool = False
    tNum = 0
    tDate = udtuStandard_texts.Date_null
    tString = ""
    tUpgraded = False
    uUser.Preferences_changed = False
    tLocals.Success = udtBoolean_extended.Null
    Call frmMain.Statuslabel1_set(tLocals.Subr & ".")

    If Not tDb_point_to_uMemory_fn(tDb) Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If ' If tDb_point_to_uMemory_fn(tDb) Then

    If Not System.IO.File.Exists(tDb.Database) Then
      Stop ' Permanent.
      If uOp.Called_from = tLocals.Subr Then
        Stop ' Permanent.
      Else
        uOp.Called_from = tLocals.Subr
        Stop ' To do, to test to check uApp.Progressbar_steps_total.
        tLocals.Action = "Call " & udtApplication_subroutine_names.Memory_instantiate
        Call Memory_instantiate()
        uOp.Subr = tLocals.Subr
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
      End If
    End If ' If System.IO.File.Exists(tDb.Database) Then

    tLocals.Action = "Prepare database"
    With tDb
      .Rows_affected = -1
      .Success = False
      .Table_name = udtTables.Preference
      .Cmd_str = "SELECT * FROM " & .Table_name & " WHERE App_name='" & uApp.Name_for_preferences & "'"
      Call Sql_write(.Cmd_str)
      tDb.Data_adapter = New OleDb.OleDbDataAdapter(tDb.Cmd_str, tDb.Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, .Table_name)
      ' This "Preference" is _within_ the dataset. DataAdapters open the connection implicitly;
      ' Fill invokes the Select command; then the connection is closed.
      .Success = True
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count
    End With
    Select Case tDb.Rows_affected
      Case 0
        tLocals.Action = "Call " & udtApplication_subroutine_names.Preference_insert_first
        Call Preference_insert_first(tDb)
        uOp.Subr = tLocals.Subr
        tUpgraded = True
        uUser.Preferences_changed = True

        uOp.Called_from = tLocals.Subr
        tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_load
        Call Preferences_load()
        uOp.Subr = tLocals.Subr
        uUser.Preferences_changed = True

      Case 1
        tLocals.Action = udtuStandard_texts.Dataset_read

        tLocals.Action = "uApp"
        ' Check build number, to see if new build, because if it is, there might need to be new stuff done.
        tSearch.Keyword = "App_build"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          tUpgraded = True
          uUser.Preferences_changed = True
        Else
          tNum = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          If tNum = 0 Then
            Stop ' Permanent.
            uUser.Preferences_changed = True
          Else
            Select Case tNum
              Case Is = uApp.App_build_number
                ' Nothing to do.
              Case Is > uApp.App_build_number
                ' Ignore.
              Case Else
                ' Stop ' Indefinite.
                uUser.Preferences_changed = True
                tUpgraded = True
            End Select
          End If
        End If
        tNum = 0

        tSearch.Keyword = "App_revision"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          tUpgraded = True
          uUser.Preferences_changed = True
          tApp_revision_last = uApp.App_revision_date
        Else
          tApp_revision_last = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          If tApp_revision_last < uApp.App_revision_date Then
            uUser.Preferences_changed = True
            tUpgraded = True
          ElseIf tApp_revision_last = uApp.App_revision_date Then
            ' Do nothing.
          Else
            Stop
            uDialogue.Prompt = "This is not the latest version of " & uApp.App_name & ". Please use the latest version. Contact " & udtLivingInfoSystems.Name_long & " if you do not have it." & vbCrLf
            uDialogue.Prompt = uDialogue.Prompt & uLIS.Support_email & vbCrLf
            uDialogue.Prompt = uDialogue.Prompt & "www." & udtLivingInfoSystems.Domain & vbCrLf
            Call Msgbox_LIS_top(uDialogue.Prompt, udtAlert_types.Critical, udtDialogue_types.Okay_only, tLocals.Subr)
            tLocals.Status = udtuStandard_texts.Cancelled
            GoTo Ender_full
          End If

        End If

        tSearch.Keyword = "Environment_dump_emailed"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          If uApp.Environment_dump_emailed = False Then uApp.Environment_dump_emailed = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tSearch.Keyword = "First_run_completed_date"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          uApp.First_run_completed_date = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tSearch.Keyword = "Memory_check_completed_date"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uApp.Memory_check_completed_date = udtuStandard_texts.Date_null
        Else
          uApp.Memory_check_completed_date = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tLocals.Action = "uActive"
        If uActive.Country = Nothing Then
          tSearch.Keyword = "Country"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Preferences_changed = True
          Else
            If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uActive.Country = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        If uActive.Currency_tla = Nothing Then
          tSearch.Keyword = "Currency_"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Preferences_changed = True
          Else
            If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uActive.Currency_tla = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tSearch.Keyword = "Currency_format"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uActive.Currency_format = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        If uActive.Currency_symbol = Nothing Then
          tSearch.Keyword = "Currency_symbol"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Preferences_changed = True
          Else
            If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uActive.Currency_symbol = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tLocals.Action = "uUser"

        If uUser.First_run_wizard = False Then
          tSearch.Keyword = "First_run_wizard"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.First_run_wizard = True
            uUser.Preferences_changed = True
          Else
            uUser.First_run_wizard = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tSearch.Keyword = "First_run_wizard_step"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.First_run_wizard = True
          uUser.First_run_wizard_step = 1
        Else
          uUser.First_run_wizard_step = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          If uUser.First_run_wizard_step <= 0 Then
            uUser.First_run_wizard_step = 1
            uUser.Preferences_changed = True

          ElseIf uUser.First_run_wizard_step < uUser.First_run_wizard_max Then
            If Not uUser.First_run_wizard Then uUser.Preferences_changed = True
            uUser.First_run_wizard = True

          ElseIf uUser.First_run_wizard_step > uUser.First_run_wizard_max Then ' Allows for something going wrong.
            uUser.First_run_wizard = False
            uUser.First_run_wizard_step = uUser.First_run_wizard_max
            uUser.Preferences_changed = True
          End If
        End If

        tSearch.Keyword = "First_run_wizard_completed"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          uUser.First_run_wizard_completed = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tSearch.Keyword = "Main_top"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
          uUser.Main_top = uApp.Default_top
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> 0 Then uUser.Main_top = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If
        If uUser.Main_top < uApp.Default_top Then uUser.Main_top = uApp.Default_top

        tSearch.Keyword = "Main_left"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
          uUser.Main_left = uApp.Default_left
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> 0 Then uUser.Main_left = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If
        If uUser.Main_left < uApp.Default_left Then uUser.Main_left = uApp.Default_left

        tSearch.Keyword = "Main_height"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
          uUser.Main_height = frmMain.Height
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> 0 Then uUser.Main_height = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If
        If uUser.Main_height < frmMain.MinimumSize.Height Then uUser.Main_height = frmMain.MinimumSize.Height

        tSearch.Keyword = "Main_width"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Main_width = frmMain.Width
          uUser.Preferences_changed = True
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> 0 Then uUser.Main_width = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If
        If uUser.Main_width < frmMain.MinimumSize.Width Then uUser.Main_width = frmMain.MinimumSize.Width

        If uUser.Min_sys_tray = False Then
          tSearch.Keyword = "Min_sys_tray"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Preferences_changed = True
          Else
            uUser.Min_sys_tray = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tSearch.Keyword = "Speech"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
          uUser.Speech = 2
          Call Preferences_write_one("Speech", uUser.Speech) ' To set default; as it could be changed during normal operation.
        Else
          If uApp.First_run_in_progress Then ' ?
            uUser.Speech = 2
          Else
            If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> 0 Then uUser.Speech = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tSearch.Keyword = "Start_home_tab"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          uUser.Start_home_tab = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tSearch.Keyword = "Start_maximised"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Start_maximised = True
          uUser.Preferences_changed = True
        Else
          uUser.Start_maximised = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        If uUser.Tab_last = Nothing Then
          tSearch.Keyword = "Tab_last"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Tab_last = uApp.Default_tab
            uUser.Preferences_changed = True
          Else
            If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) = "" Then
              uUser.Tab_last = uApp.Default_tab
              uUser.Preferences_changed = True
            Else
              uUser.Tab_last = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
            End If
          End If
        End If

        If uUser.Theme = 0 Then
          tSearch.Keyword = "Theme"
          If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
            uUser.Preferences_changed = True
            uUser.Theme = udtTheme.Null
          Else
            uUser.Theme = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
          End If
        End If

        tSearch.Keyword = "User_first_name"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
        Else
          If tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword) <> "" Then uUser.First_name = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
        End If

        tLocals.Action = "uReg"
        tSearch.Keyword = "Registered"
        If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
          uUser.Preferences_changed = True
        Else
          uReg.Registered = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)

          If Not uReg.Registered Then

            If uReg.Trial_30_day = False Then
              tSearch.Keyword = "Reg_30_day_trial"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                uUser.Preferences_changed = True
              Else
                uReg.Trial_30_day = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

            If uReg.Trial_30_day_from_date = udtuStandard_texts.Time_null Then
              tSearch.Keyword = "Reg_30_day_trial_from_date"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                ' uUser.Preferences_changed = True
              Else
                uReg.Trial_30_day_from_date = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

            If uReg.Trial_30_day_20_days_notified = False Then
              tSearch.Keyword = "Reg_30_day_trial_20_days_notified"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                uUser.Preferences_changed = True
              Else
                uReg.Trial_30_day_20_days_notified = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

            If uReg.Code = Nothing Then
              tSearch.Keyword = "Reg_code"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                ' uUser.Preferences_changed = True
              Else
                uReg.Code = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

            If uReg.Code_sent = False Then
              tSearch.Keyword = "Reg_code_sent"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                uUser.Preferences_changed = True
              Else
                uReg.Code_sent = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

            If uReg.Email_address = Nothing Then
              tSearch.Keyword = "Reg_email"
              If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)) Then
                ' uUser.Preferences_changed = True
              Else
                uReg.Email_address = tDb.Dataset.Tables(tDb.Table_name).Rows(0).Item(tSearch.Keyword)
              End If
            End If

          End If

        End If
        tLocals.Action = "Common end"

        tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_load_local
        Call Preferences_load_local(tDb)
        uOp.Subr = tLocals.Subr
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

        tLocals.Action = "Call Preferences load environment_dump completed date"
        Call Preferences_load_environment_dump_completed_date(tDb)
        uOp.Subr = tLocals.Subr
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

        uOp.Preferences_loaded = True
        tLocals.Success = udtBoolean_extended.True_

      Case Else
        Stop ' Permanent.
        tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_duplicate_fix
        Call Preferences_duplicate_fix(tDb)
        uOp.Subr = tLocals.Subr
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

        tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_load
        Call Preferences_load()
        uOp.Subr = tLocals.Subr
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        GoTo Ender_full

    End Select ' Select Case tDb.Pref_rows_affected

    tLocals.Action = "Close"
    If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
    tDb.Cmd_str = ""
    tDb.Table_name = ""

Upgrade_section:
    If uUser.Preferences_changed Or tDb.Upgrade_table <> "" Or tUpgraded Then
      tLocals.Action = "Upgrade section"
      tLocals.Success = udtBoolean_extended.Null
      uOp.Success = False
      uApp.Progressbar_steps_difference = uApp.Progressbar_steps_total - uApp.Progressbar_step
      uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 17
      Call frmMain.Statuslabel2_set(udtuStandard_texts.Upgrading)
      If uApp.Debug Then tLocals.Msg = tDb.Upgrade_table
      ' Update stuff.
      ' tDb.Upgrade_table = "A table"
      ' uUser.Preferences_changed = True
      ' Update stuff.
      ' Create tables.
      ' Create table columns.

      If tDb.Upgrade_table <> "" Or tUpgraded Then
        tLocals.Action = "Call " & udtApplication_subroutine_names.App_revision_upgrade
        If uApp.App_name <> udtApplication.App_Financier_Personal_Edition Then
          uApp.Progressbar_steps_difference = uApp.Progressbar_steps_total - uApp.Progressbar_step ' Before. 32.
        End If
        uOp.Called_from = tLocals.Subr
        tLocals.Action = "Call" & udtApplication_subroutine_names.App_revision_upgrade
        Call App_revision_upgrade(tApp_revision_last, tDb, tUpgraded)
        uOp.Subr = tLocals.Subr
        If uDialogue.Active Then GoTo Ender_full
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

        tLocals.Action = "Upgrade section"
        If uApp.App_name <> udtApplication.App_Financier_Personal_Edition Then
          uApp.Progressbar_steps_difference = uApp.Progressbar_steps_total - uApp.Progressbar_step
        End If
      End If

      If tDb.Upgrade_table <> "" Then
        uOp.Called_from = uOp.Subr
        tLocals.Action = "Call " & udtApplication_subroutine_names.Database_diagnostic_level_05
        Call Database_diagnostic_level_05(tDb)
        uOp.Subr = tLocals.Subr
        If Not tDb.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

      End If
      uOp.Subr = tLocals.Subr
      tLocals.Action = "Upgrade section"

      If uUser.Preferences_changed And uOp.Preferences_loaded And tDb.Upgrade_table = "" Then
        tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_write
        Call Preferences_write()
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
      Else
        uOp.Called_from = tCalled_from
        uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 2
        If Not uOp.Preferences_loaded Then
          tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_load
          Call Preferences_load()
          If uOp.Success Then
            tLocals.Error_count = 0
          Else
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
          End If
        End If
        uOp.Success = False
        If uDialogue.Active Then
          GoTo Ender_full
        End If
      End If
    End If ' If uUser.Preferences_changed Or tDb.Upgrade_table <> "" Or tUpgraded Then

Upgrade_section_end:

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 13 ' Conversion from type 'DBNull' to type 'String' is not valid. Field is empty. 
            tNum = 0
            tString = ""
            tDate = Nothing
            tBool = False
            uUser.Preferences_changed = True
            Resume Next
          Case 91 ' Object reference not set to an instance of an object.
            Call Error_clear(tLocals)
            tDb.Upgrade_table = tDb.Table_name
            Resume Upgrade_section
          Case Else
            Stop ' Indefinite.
            Call Database_error_handler(tLocals, tDb)
            uOp.Subr = tLocals.Subr
            If tLocals.Action = "" Then
              Resume Error_captured
            Else
              Resume Upgrade_section
            End If
        End Select
      End If
    End If

Ender_full:
    tBool = False
    tNum = 0
    tDate = Nothing
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
      tLocals.Success = udtBoolean_extended.True_
    End If
    If tDb.Attempt = 0 And uError.Message = "" And uError.Description = "" And uError.Number = 0 And tLocals.Error_count = 0 Then tLocals.Success = udtBoolean_extended.True_
    uOp.Preferences_loaded = True
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Ender_partial:
    If tUpgraded Then uApp.Upgraded = True
    uOp.Called_from = ""
    Call frmMain.Statuslabel1_set()
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Preferences_load_environment_dump_completed_date(ByRef tDb As udtDatabase_oledb)
    ' Separated 2021.04.07.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing
    tLocals.Subr = "Preferences load environment_dump completed date"

    With tDb
      .Rows_affected = -1
      .Success = False
      .Table_name = udtTables.Preference
      .Cmd_str = "SELECT TOP 1 Environment_dump_completed_date FROM " & .Table_name & vbCrLf
      .Cmd_str = .Cmd_str & "WHERE Environment_dump_completed_date IS NOT NULL" & vbCrLf
      .Cmd_str = .Cmd_str & "ORDER BY Environment_dump_completed_date DESC" & vbCrLf
      Call Sql_write(.Cmd_str)
      tDb.Data_adapter = New OleDb.OleDbDataAdapter(.Cmd_str, .Conn_str)
      .Dataset = New DataSet
      tLocals.Action = udtuStandard_texts.Dataset_fill
      .Data_adapter.Fill(.Dataset, .Table_name)
      .Success = True
      tLocals.Action = udtuStandard_texts.Dataset_read
      .Rows_affected = .Dataset.Tables(.Table_name).Rows.Count

      If .Success And .Rows_affected > 0 Then
        tSearch.Keyword = "Environment_dump_completed_date"
        If IsDBNull(.Dataset.Tables(.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          uApp.Environment_dump_completed_date = .Dataset.Tables(.Table_name).Rows(0).Item(tSearch.Keyword)
        End If
      End If
      uOp.Success = True

    End With

  End Sub

  Sub Preferences_write()
    ' Revised 2021.04.06.
    Dim tLocals As udtLocals = Nothing
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tFile As udtFile

    tLocals.Subr = udtApplication_subroutine_names.Preferences_write
    If uOp.Preferences_loaded = False Or uOp.Form_loaded = False Or uUser.Preferences_changed = False Then GoTo Ender_full
    If uApp.Terminating >= 3 Then GoTo Ender_full

    If uError.Last_error_sub = tLocals.Subr Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    End If
    On Error GoTo Error_captured
    If uOp.Form_initialising Then uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1

    tLocals.Status = udtuStandard_texts.Initiation
    If tDb.Database = "" Then
      If Not tDb_point_to_uMemory_fn(tDb) Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
    End If
    uOp.Success = False
    tDb.Success = False

    If System.IO.File.Exists(tDb.Database) Then
      tLocals.Action = "Data conditioning"
      Dim tMaximised As Boolean
      If (Not uUser.First_run_wizard Or uUser.First_run_wizard_step >= uUser.First_run_wizard_max) And uUser.Tab_last = "tabFirst_run_wizard" Then
        uUser.Tab_last = frmMain.tabHome.Name
      End If

      tLocals.Success = udtBoolean_extended.Null
      If uActive.Country = "" Or uActive.Country = "Null" Then tLocals.Success = udtBoolean_extended.False_
      If uActive.Currency_tla = "" Then tLocals.Success = udtBoolean_extended.False_
      If uActive.Currency_symbol = "" Then tLocals.Success = udtBoolean_extended.False_
      If uActive.Currency_format = "" Then tLocals.Success = udtBoolean_extended.False_
      If tLocals.Success = udtBoolean_extended.False_ Then
        tLocals.Action = "Call " & udtApplication_subroutine_names.Detect_country
        If uActive.Country = "" Then
          Call Country_detect(System.Globalization.RegionInfo.CurrentRegion.DisplayName, True)
        Else
          Call Country_detect(uActive.Country, True)
        End If
      End If
      tLocals.Success = udtBoolean_extended.Null

      tLocals.Action = udtuStandard_texts.Sql_build_update
      tDb.Table_name = udtTables.Preference
      tDb.Cmd_str = "UPDATE " & tDb.Table_name & " SET" & vbCr

      tLocals.Action = "uApp"
      tDb.Cmd_str = tDb.Cmd_str & "App_revision                     ='" & uApp.App_revision_date & "', " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "App_build                        =" & uApp.App_build_number & ", " & vbCr
      ' tDb.Cmd_str = tDb.Cmd_str & "Email_smtp_working               =" & uApp.Email_smtp_working & ", " & vbCr ' Set only from a change in user preferences.
      If uApp.Environment_dump_emailed Then tDb.Cmd_str = tDb.Cmd_str & "Environment_dump_emailed         =" & uApp.Environment_dump_emailed & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Environment_dump_completed_date  ='" & uApp.Environment_dump_completed_date & "', " & vbCr ' 2021.02.16.
      tDb.Cmd_str = tDb.Cmd_str & "First_run_completed_date         ='" & uApp.First_run_completed_date & "', " & vbCr ' 2021.02.16.
      tDb.Cmd_str = tDb.Cmd_str & "Memory_check_completed_date      ='" & uApp.Memory_check_completed_date & "', " & vbCr ' 2021.02.16.

      tLocals.Action = "uActive"
      If uActive.Country <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Country                          ='" & uActive.Country & "', " & vbCr
      If uActive.Currency_tla <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Currency_                        ='" & uActive.Currency_tla & "', " & vbCr
      If uActive.Currency_format <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Currency_format                  ='" & uActive.Currency_format & "', " & vbCr
      If uActive.Currency_symbol <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Currency_symbol                  ='" & uActive.Currency_symbol & "', " & vbCr
      If uActive.Corporation_name <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Corporation                      ='" & uActive.Corporation_name & "', " & vbCr
      If uActive.Branch_name <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Branch                           ='" & uActive.Branch_name & "', " & vbCr

      tLocals.Action = "uUser"
      tDb.Cmd_str = tDb.Cmd_str & "First_run_wizard                 =" & uUser.First_run_wizard & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "First_run_wizard_step            =" & uUser.First_run_wizard_step & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "First_run_wizard_completed       =" & uUser.First_run_wizard_completed & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Main_top                         =" & uUser.Main_top & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Main_left                        =" & uUser.Main_left & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Main_height                      =" & uUser.Main_height & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Main_width                       =" & uUser.Main_width & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Min_sys_tray                     =" & uUser.Min_sys_tray & ", " & vbCr
      ' tDb.Cmd_str = tDb.Cmd_str & "Speech                           =" & uUser.Speech & ", " & vbCr ' Set only from a change in user preferences. Allows for temporaryily turning off speach due to a computer error.
      If frmMain.WindowState = FormWindowState.Maximized Then tMaximised = True
      tDb.Cmd_str = tDb.Cmd_str & "Start_maximised                  =" & tMaximised & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Start_home_tab                   =" & uUser.Start_home_tab & ", " & vbCr
      If uUser.Tab_last = frmMain.tabAbout.Name Or uUser.Tab_last = frmMain.tabSettings.Name Then uUser.Tab_last = frmMain.tabHome.Name
      tDb.Cmd_str = tDb.Cmd_str & "Tab_last                         ='" & uUser.Tab_last & "', " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "Theme                            =" & uUser.Theme & ", " & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "User_first_name                  ='" & uUser.First_name & "', " & vbCr

      tLocals.Action = "uReg"
      If Not uReg.Registered Then
        tDb.Cmd_str = tDb.Cmd_str & "Reg_30_day_trial                 =" & uReg.Trial_30_day & ", " & vbCr
        tDb.Cmd_str = tDb.Cmd_str & "Reg_30_day_trial_20_days_notified=" & uReg.Trial_30_day_20_days_notified & ", " & vbCr
        tDb.Cmd_str = tDb.Cmd_str & "Reg_code                         ='" & uReg.Code & "', " & vbCr
        tDb.Cmd_str = tDb.Cmd_str & "Reg_code_sent                    =" & uReg.Code_sent & ", " & vbCr
        If uReg.Email_address <> "" Then tDb.Cmd_str = tDb.Cmd_str & "Reg_email                        ='" & uReg.Email_address & "', " & vbCr
      End If
      tDb.Cmd_str = tDb.Cmd_str & "Registered                       =" & uReg.Registered & ", " & vbCr

      Call Preferences_write_local(tDb.Cmd_str)
      tDb.Cmd_str = tDb.Cmd_str & "Updated                          ='" & Date.Today & "'" & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "WHERE App_name                   ='" & uApp.Name_for_preferences & "'"

      ' Call Sql_conditioning(tDb.Cmd_str, udtuStandard_texts.Sql_update)
      Call Sql_write(tDb.Cmd_str)

      tFile.Open_file = tDb.Database
      With tDb
        If .Conn.State <> ConnectionState.Open Then .Conn.Open()
        .Cmd.CommandText = .Cmd_str
        tLocals.Action = udtuStandard_texts.Sql_write
        .Rows_affected = .Cmd.ExecuteNonQuery()
        .Conn.Close()
        If .Upgrade_table = "" Then .Success = True
      End With
      tFile.Open_file = ""

      If tDb.Rows_affected = 0 Then
        Call Preference_insert_first(tDb)
        Call Preferences_write()
      End If

      If tDb.Success And tDb.Upgrade_table = "" Then
        ' uApp.Upgraded = False
        uUser.Preferences_changed = False
      Else
        Stop ' Permanent.
        Call Database_diagnostic_level_05(tDb)
        Call Preferences_write()
        tDb.Success = True
      End If
    End If

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 1 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Stop ' Temporary, check  uStandard_texts .Sql_update.
        Call Database_error_handler(tLocals, tDb)
        If uOp.Success Then
          Call Preferences_write()
        Else
          Resume Ender_full
        End If
      End If
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Action = "" Then tLocals.Action = tLocals.Status
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 1 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_process
    End If

    Call frmMain.ToolStripProgressBar1_set()
    tFile.Open_file = ""
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Preferences_write_one(ByVal tField As String, ByVal tValue As Object)
    ' Written by Michael John Boyack Hodgson. Revised 2021.05.06.
    ' This has to use tDb, not uMemory, because it's called by Environment_dump_email.
    Dim tLocals As udtLocals = Nothing
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tFile As udtFile

    If Not uUser.Preferences_changed Then Exit Sub
    tLocals.Subr = udtApplication_subroutine_names.Preferences_write_one
    On Error GoTo Error_captured
    tLocals.Action = udtuStandard_texts.Main_start

    With tDb
      .Database = uMemory.Database
      .Use_password = uMemory.Use_password
      Call tDb_point_to_uMemory_fn(tDb)
      tLocals.Action = udtuStandard_texts.Sql_build_update
      .Table_name = udtTables.Preference
      tDb.Cmd_str = "UPDATE " & .Table_name & " SET " & vbCr
      If Val(tValue) = 0 Then
        tDb.Cmd_str = tDb.Cmd_str & tField & "='" & tValue & "', " & vbCr
      Else
        If IsNumeric(tValue) Or tValue.GetType Is GetType(Boolean) Then
          tDb.Cmd_str = tDb.Cmd_str & tField & "=" & tValue & ", " & vbCr
        Else
          tDb.Cmd_str = tDb.Cmd_str & tField & "='" & tValue & "', " & vbCr
        End If
      End If
      tDb.Cmd_str = tDb.Cmd_str & "Updated       ='" & Date.Today & "'" & vbCr
      tDb.Cmd_str = tDb.Cmd_str & "WHERE App_name='" & uApp.Name_for_preferences & "'"

      Call Sql_write(tDb.Cmd_str)

      tFile.Open_file = .Database
      If .Conn.State <> ConnectionState.Open Then .Conn.Open()
      tLocals.Action = udtuStandard_texts.Sql_write
      .Cmd.CommandText = tDb.Cmd_str
      .Rows_affected = .Cmd.ExecuteNonQuery()
      .Conn.Close()
      If .Upgrade_table = "" And .Rows_affected = 1 Then
        .Success = True
      Else
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If

      tFile.Open_file = ""
      If .Success And .Upgrade_table = "" Then
        If uOp.Called_from <> udtApplication_subroutine_names.Preferences_load Then uUser.Preferences_changed = False
      Else
        Call Database_diagnostic_level_05(tDb)
        Call Preferences_write()
      End If
    End With

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' To do, use Database_error_handler.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case -2147217904
            ' No value given for one or more required parameters.
            Select Case tLocals.Action
              Case udtuStandard_texts.Sql_write
                tDb.Upgrade_table = tDb.Table_name
                Resume Next
              Case Else
                uError.Extra = tDb.Cmd_str
                Stop ' Permanent.
                Call Error_display(tLocals)
            End Select
          Case 3704
            ' Operation is not allowed when the object is closed.
          Case Else
            Call Database_error_handler(tLocals, uMemory)
            If tLocals.Action = udtuStandard_texts.Retry Then
              Resume
            Else
              Stop ' Permanent.
              Call Error_display(tLocals)
            End If
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub Preferences_duplicate_fix(ByRef tDb As udtDatabase_oledb)
    ' Written by Michael John Boyack Hodgson, 2014.08.01. Revised 2021.04.17.
    Stop ' Permanent.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing

    Dim i As Integer
    Dim tId As Integer
    tLocals.Subr = udtApplication_subroutine_names.Preferences_duplicate_fix
    uOp.Subr = tLocals.Subr
    ' On Error GoTo Error_handler
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set()
    uOp.Subr = tLocals.Subr

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    tLocals.Action = "Check id"
    For i = 0 To tDb.Rows_affected - 1
      tLocals.Action = udtuStandard_texts.Dataset_read
      tSearch.Keyword = "Id"
      If IsDBNull(tDb.Dataset.Tables(tDb.Table_name).Rows(i).Item(tSearch.Keyword)) Then
      Else
        If tDb.Dataset.Tables(tDb.Table_name).Rows(i).Item(tSearch.Keyword) > tId Then
          tId = tDb.Dataset.Tables(tDb.Table_name).Rows(i).Item(tSearch.Keyword)
        End If
      End If
    Next

    tDb.Cmd_str = "DELETE FROM " & tDb.Table_name & " WHERE" & vbCr
    tDb.Cmd_str = tDb.Cmd_str & "App_name='" & uApp.Name_for_preferences & "' AND" & vbCrLf
    tDb.Cmd_str = tDb.Cmd_str & "Id      =" & tId
    Call Database_general(tDb)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      ' If Not tLocals.Continue_ Then GoTo Error_process
    End If
    uOp.Success = False

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

    uOp.Subr = ""
    Call frmMain.Statuslabel1_set()
  End Sub

#End Region ' End region, "Preferences."

#Region "Database operations general"

#Region "Sql builders"
  ' Region started 2020.03.21. Revised 2020.12.27.

  Sub Sql_builder_alter_one(ByRef tDb As udtDatabase_oledb, ByVal tField As String, ByRef tData_type As String)
    ' Written 2020.04.08. Revised 2021.03.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtuStandard_texts.Sql_build_alter
    tLocals.Status = udtuStandard_texts.Initiation

    If tDb.Conn_str = "" Then Call Database_instantiate(tDb)
    tLocals.Action = udtuStandard_texts.Main_start
    tLocals.Status = ""

    tDb.Cmd_str = udtuStandard_texts.Sql_alter & " " & tDb.Table_name & vbCrLf
    tDb.Cmd_str = tDb.Cmd_str & "ADD " & tField & " " & tData_type

    Call Database_general(tDb)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Sql_builder_alter_many(ByRef tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields)
    ' Written 2020.08.16.
    Stop ' To do, to develop.
  End Sub

  Sub Sql_builder_delete_one(tDb As udtDatabase_oledb, ByVal tField As String, ByRef tOperator As String, ByRef tCondition As String)
    ' Written 2020.03.21. Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Sql builder delete"
    tLocals.Action = udtuStandard_texts.Sql_build_delete

    If tDb.Table_name = "" Then Stop
    If tField = "" Then Stop
    If tOperator = "" Then Stop
    If tCondition = "" Then Stop

    tDb.Cmd_str = "DELETE FROM " & tDb.Table_name & vbCrLf
    If IsNumeric(tCondition) Or tCondition.GetType = GetType(Boolean) Then
      tDb.Cmd_str = tDb.Cmd_str & " WHERE " & tField & tOperator & tCondition
    Else
      tDb.Cmd_str = tDb.Cmd_str & " WHERE " & tField & tOperator & "'" & tCondition & "'" & vbCrLf
    End If
    Call Database_general(tDb)
    uOp.Subr = "Sql builder delete"

    uOp.Subr = ""
  End Sub

  Sub Sql_builder_delete_many(tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields)
    ' Written 2020.08.16.
    Stop ' To do, to develop.
  End Sub

  Sub Sql_builder_insert(ByRef tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields)
    ' Separated 2020.04.27. Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing
    Dim i As Integer
    uOp.Subr = udtuStandard_texts.Sql_build_insert
    tLocals.Action = uOp.Subr

    If tDb.Table_name = "" Then Stop : GoTo Ender_full

    With tDb
      .Cmd_str = udtuStandard_texts.Sql_insert & " " & .Table_name & " ("

      For i = 2 To tTable_definition.Length
        If tTable_definition.Value(i) <> Nothing Then
          If i > 2 And i <= tTable_definition.Length Then tDb.Cmd_str = tDb.Cmd_str & ", "
          tDb.Cmd_str = tDb.Cmd_str & tTable_definition.Name_(i)
        End If
      Next
      .Cmd_str = .Cmd_str & ")" & vbCrLf
      .Cmd_str = .Cmd_str & "VALUES (" & vbCrLf

      For i = 2 To tTable_definition.Length
        If tTable_definition.Value(i) <> Nothing Then
          If i > 2 And i <= tTable_definition.Length Then tDb.Cmd_str = tDb.Cmd_str & ", " & vbCrLf
          If IsNumeric(tTable_definition.Value(i)) Or tTable_definition.Value(i).GetType = GetType(Boolean) Then
            tDb.Cmd_str = tDb.Cmd_str & tTable_definition.Value(i)
          Else
            tDb.Cmd_str = tDb.Cmd_str & "'" & tTable_definition.Value(i) & "'"
          End If
        End If
      Next
      .Cmd_str = .Cmd_str & ")" & vbCrLf
    End With
    Call Database_general(tDb)

Ender_full:
    uOp.Subr = ""
  End Sub

  Sub Sql_builder_select_one(ByRef tDb As udtDatabase_oledb, ByVal tCondition_field As String, ByRef tCondition_operator As String, ByRef tCondition_value As String)
    ' Written 2020.03.21. Revised 2021.05.06.
    ' For one field.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = "Sql builder select"
    tLocals.Action = udtuStandard_texts.Sql_build_select

    If tDb.Table_name = "" Then Stop : GoTo Ender_full
    If tCondition_field = "" Then Stop : GoTo Ender_full
    If tCondition_operator = "" Then Stop : GoTo Ender_full
    If tCondition_value = "" Then Stop : GoTo Ender_full

    tDb.Cmd_str = "SELECT " & tCondition_field & " FROM " & tDb.Table_name & vbCrLf
    If IsNumeric(tCondition_value) Or tCondition_value.GetType Is GetType(Boolean) Then
      tDb.Cmd_str = tDb.Cmd_str & " WHERE " & tCondition_field & tCondition_operator & tCondition_value
    Else
      tDb.Cmd_str = tDb.Cmd_str & " WHERE " & tCondition_field & tCondition_operator & "'" & tCondition_value & "'" & vbCrLf
    End If

    Call Database_general(tDb)
    uOp.Subr = "Sql builder select"

Ender_full:
    uOp.Subr = ""
  End Sub

  Sub Sql_builder_select_many(ByRef tDb As udtDatabase_oledb, ByRef tTable_definition As udtMemory_Fields)
    ' Written 2020.08.09. Revised 2021.05.06.
    ' For all fields.
    Dim tLocals As udtLocals = Nothing
    Dim tSomething_set As Boolean = False
    tLocals.Subr = "Sql builder select"
    tLocals.Action = udtuStandard_texts.Sql_build_select
    With tDb
      If .Table_name = "" Then Stop : GoTo Ender_full
      tLocals.Action = udtuStandard_texts.Sql_build_select
      .Cmd_str = udtuStandard_texts.Sql_select & " * FROM " & .Table_name
      With tTable_definition
        For i As Short = 2 To tTable_definition.Length
          If .Condition_value(i) <> Nothing Then
            Stop ' To do, if values given, WHERE values:
            tDb.Cmd_str = tDb.Cmd_str & vbCrLf
            tDb.Cmd_str = tDb.Cmd_str & "WHERE" & vbCrLf
            For j As Short = 2 To tTable_definition.Length
              If .Condition_operator(j) = "" Or .Condition_value(j) = "" Then
              Else
                If IsNumeric(.Condition_value(j)) Or .Condition_value(j).GetType Is GetType(Boolean) Then
                  tDb.Cmd_str = tDb.Cmd_str & .Name_(j) & .Condition_operator(j) & .Condition_value(j) & vbCrLf
                Else
                  tDb.Cmd_str = tDb.Cmd_str & .Name_(j) & .Condition_operator(j) & "'" & .Condition_value(j) & "'" & vbCrLf
                End If
                tSomething_set = True
              End If
            Next
            Exit For
          End If
        Next
      End With
    End With

    If tSomething_set Then
      Call Database_general(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        '        tLocals.Error_count    =        tLocals.Error_count    + 1
        tLocals.Status = udtuStandard_texts.Failed
        uOp.Success = False
        ' GoTo Error_process
      End If
      uOp.Success = False
    Else
      tLocals.Status = udtuStandard_texts.Failed
    End If

Ender_full:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Sql_builder_update(ByRef tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields, Optional ByVal tCondition_field As String = "", Optional ByVal tCondition_operator As String = "", Optional ByVal tCondition_value As Object = "")
    ' Separated 2020.04.27. Revised 2021.05.06.
    ' For all fields.
    Dim tLocals As udtLocals = Nothing
    Dim tSomething_set As Boolean = False
    tLocals.Subr = "Sql builder update"
    tLocals.Status = udtuStandard_texts.Initiation

    If tDb.Table_name = "" Or tTable_definition.Length = 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    Else
      tLocals.Action = udtuStandard_texts.Sql_build_update
      tDb.Cmd_str = "UPDATE " & tDb.Table_name & vbCrLf
      tDb.Cmd_str = tDb.Cmd_str & "SET "
      With tTable_definition
        For i As Integer = 2 To .Length
          If .Value(i) <> Nothing Then
            If (i > 2 And i < .Length) Or tSomething_set Then tDb.Cmd_str = tDb.Cmd_str & ", " & vbCrLf
            tDb.Cmd_str = tDb.Cmd_str & .Name_(i) & "="
            If IsNumeric(.Value(i)) Or .Value(i).GetType Is GetType(Boolean) Then
              tDb.Cmd_str = tDb.Cmd_str & .Value(i)
              tSomething_set = True
            Else
              tDb.Cmd_str = tDb.Cmd_str & "'" & .Value(i) & "'"
              tSomething_set = True
            End If
          End If
        Next

        If tSomething_set Then
          If tCondition_field = "" Or tCondition_operator = "" Or tCondition_value = "" Then
          Else
            tDb.Cmd_str = tDb.Cmd_str & vbCrLf
            tDb.Cmd_str = tDb.Cmd_str & "WHERE " & tCondition_field & tCondition_operator
            If IsNumeric(tCondition_value) Then
              tDb.Cmd_str = tDb.Cmd_str & tCondition_value
            Else
              tDb.Cmd_str = tDb.Cmd_str & "'" & tCondition_value & "'"
            End If
          End If
          For i As Integer = 2 To tTable_definition.Length
            If .Condition_operator(i) <> "" Then
              tDb.Cmd_str = tDb.Cmd_str & vbCrLf
              tDb.Cmd_str = tDb.Cmd_str & "WHERE "
              For j As Integer = 2 To .Length
                If .Condition_operator(j) <> "" Then
                  If .Condition_value(j) = "" Then
                    tDb.Cmd_str = tDb.Cmd_str & .Name_(j) & .Condition_operator(j) & "'" & .Value(j) & "'" & vbCrLf
                  Else
                    tDb.Cmd_str = tDb.Cmd_str & .Name_(j) & .Condition_operator(j) & "." & .Condition_value(j) & "." & vbCrLf
                  End If
                End If
              Next
              Exit For
            End If
          Next

          Call Database_general(tDb)
          uOp.Subr = tLocals.Subr
          If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
          End If
          tLocals.Success = udtBoolean_extended.True_

        Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          GoTo Error_process
        End If ' If tSomething_set Then
      End With
    End If ' If tDb.Table_name = "" Or tTable_definition.Length = 0 Then

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Sql_builder_upsert_one(ByRef tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields)
    ' Written 2020.04.25. Revised 2021.04.09.
    ' For one field.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Sql builder upsert"
    tLocals.Action = udtuStandard_texts.Sql_build_upsert

    Call Sql_builder_update(tDb, tTable_definition)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process
    End If
    uOp.Success = False
    uOp.Result = uOp.Result & " Updated."

    If tDb.Rows_affected = 0 Then
      Call Sql_builder_insert(tDb, tTable_definition)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process
      End If
      uOp.Success = False
      uOp.Result = uOp.Result & " Inserted."
    End If

Error_process:
    If (tLocals.Continue_ = False And (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Sql_builder_upsert_many(ByRef tDb As udtDatabase_oledb, ByVal tTable_definition As udtMemory_Fields)
    ' Separated 2020.04.27. Revised 2021.01.24.
    ' For all fields.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Sql builder upsert"
    tLocals.Action = udtuStandard_texts.Sql_build_upsert

    Call Sql_builder_select_many(tDb, tTable_definition)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    End If
    uOp.Success = False

    If tDb.Rows_affected = 0 Then
      Call Sql_builder_insert(tDb, tTable_definition)
    Else
      Call Sql_builder_update(tDb, tTable_definition)
    End If
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      uOp.Success = False
    End If

Ender_full:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Function Sql_builder_format_date_fn(ByVal tDate As Date) As String
    ' Written 2020.04.08. Revised 2021.03.01.
    '' Means failed.
    ' Means also works.
    Dim tLocals As udtLocals = Nothing
    Dim tDate_str As String = ""

    tLocals.Subr = "Sql builder format date"
    Sql_builder_format_date_fn = ""

    If uApp.Debug Then tLocals.Msg = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern ' dd/MM/yyyy.
    If uApp.Debug Then tLocals.Msg = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern ' yyyy.MM.dd.

    tDate_str = DateAndTime.DateValue(tDate)
    tDate_str = Replace(tDate_str, ".", "-")
    tDate_str = Replace(tDate_str, "/", "-")
    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy-mm-dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy-mm-dd")'.
    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy.mm.dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy.mm.dd")'.
    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy/mm/dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy/mm/dd")'.
    ' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "mm/dd/yyyy" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "mm/dd/yyyy")'.

    Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & "mm/dd/yyyy" & Chr(34) & ")"
    ' Works! yyyy.mm.dd, d.m.yyyy, m/d/yyyy.

    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & "yyyy-mm-dd" & Chr(34) & ")"
    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & Replace(System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern, ".", "-") & Chr(34) & ")" ' Works.
    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")" ' Works.

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy.MM.dd")'.
    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "dd/MM/yyyy")'.

    ' 2020.05.12:
    ' From lis-vm-test-01 (lis-test-vm-01), set to d.m.yyyy:
    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy-mm-dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy-mm-dd")'.
    ' d.m.yyyy yyyy-mm-dd

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy.mm.dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy.mm.dd")'.
    ' d.m.yyyy yyyy.mm.dd

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "yyyy/mm/dd" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy/mm/dd")'.
    ' d.m.yyyy yyyy/mm/dd

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & "mm/dd/yyyy" & Chr(34) & ")"
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "mm/dd/yyyy")'.
    ' d.m.yyyy mm/dd/yyyy

    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & "yyyy-mm-dd" & Chr(34) & ")"
    ' d-m-yyyy yyyy-mm-dd

    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & Replace(System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern, ".", "-") & Chr(34) & ")" ' Works.
    ' d-m-yyyy d-m-yyyy

    ' Sql_builder_format_date_fn = "Format(#" & tDate_str & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")" ' Works.
    ' d-m-yyyy d.m.yyyy

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")"
    ' d.m.yyyy d.m.yyyy
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "yyyy.MM.dd")'.

    '' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern & Chr(34) & ")"
    ' d.m.yyyy d.m.yyyy
    ' Syntax error in date in query expression '[Date]=Format(#2000.01.09#, "dd/MM/yyyy")'.

    If uMachine.Date_format <> "" Then
      ' Sql_builder_format_date_fn = "Format(#" & tDate & "#, " & Chr(34) & uMachine.Date_format & Chr(34) & ")"
    End If

    uOp.Subr = ""
  End Function

#End Region ' End region, "Sql builders."

  Sub Log_write_memory(ByVal tLocals_external As udtLocals)
    ' Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing
    Dim tError As udtError
    Dim tFile As udtFile

    Dim tuFile_File_full_name As String
    Dim tDb As udtDatabase_oledb = Nothing

    If uTables.Log = "" Or uMemory.Database = "" Then
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_full
    End If

    tError.Description = uError.Description
    tError.Extra = uError.Extra
    tError.Message = uError.Message
    tError.Number = uError.Number
    tError.Source = uError.Source

    tuFile_File_full_name = tFile.File_full_name

    tLocals.Subr = udtApplication_subroutine_names.Log_write_memory
    If uError.Last_error_sub = tLocals.Subr Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    End If

    tLocals.Action = "Condition the text to be inserted"
    tLocals_external.Subr = Replace(tLocals_external.Subr, "'", "_")
    If tError.Description = "" Then
      tError.Description = ""
    Else
      If InStr(tError.Description, "INSERT INTO") > 0 Then
        tError.Description = ""
      End If
      tError.Description = Left(tError.Description, 255)
      tError.Description = Replace(tError.Description, Chr(34), "'")
      tError.Description = Replace(tError.Description, "''", "`")
    End If
    If tError.Extra = tDb.Cmd_str Then
      tError.Extra = ""
    End If
    If tError.Message = tError.Description Then
      tError.Message = ""
    Else
      tError.Message = Replace(tError.Message, Chr(34), "'")
    End If

    uOp.Called_from = tLocals.Subr
    If Not tDb_point_to_uMemory_fn(tDb) Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uOp.Success = False
      GoTo Error_process
    End If ' If tDb_point_to_uMemory_fn(tDb) Then
    uOp.Success = False

    tLocals.Action = udtuStandard_texts.Sql_build_insert
    tDb.Table_name = uTables.Log
    tDb.Cmd_str = "INSERT INTO " & tDb.Table_name & " (" & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Date_, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Time_, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[User], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Computer_name, " & vbCr

    If uApp.App_name = udtApplication.App_Seer Or uApp.App_name = udtApplication.App_Financier_Personal_Edition Or uApp.App_name = udtApplication.App_Financier_Commercial_Edition Then
      tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Application, " & vbCr
    End If

    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Event_, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Subr, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[Status], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[Action], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "OpenF, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Error_Number, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Error_Description, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[Error_Extra], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[Error_Message], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Error_Source, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Corporation, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "[Branch], " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Success, " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "Document" & vbCr

    tDb.Cmd_str = tDb.Cmd_str & ") VALUES (" & vbCr

    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & Date.Today & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & TimeString & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uUser.First_name & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uMachine.Name & "', " & vbCr

    If uApp.App_name = udtApplication.App_Seer Or uApp.App_name = udtApplication.App_Financier_Personal_Edition Or uApp.App_name = udtApplication.App_Financier_Commercial_Edition Then
      tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uApp.App_name & "', " & vbCr
    End If

    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uOp.Event_ & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tLocals_external.Subr & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tLocals_external.Status & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tLocals_external.Action & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tuFile_File_full_name & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tError.Number & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tError.Description & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tError.Extra & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tError.Message & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tError.Source & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uActive.Corporation_name & "', " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & uActive.Branch_name & "', " & vbCr
    If tLocals_external.Success = udtBoolean_extended.True_ Then uOp.Success = True
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "" & uOp.Success & ", " & vbCr
    tDb.Cmd_str = tDb.Cmd_str & tDb.Indent & "'" & tFile.File_full_name & "'" & vbCr
    tDb.Cmd_str = tDb.Cmd_str & ")"
    Call Database_general(tDb)
    If uOp.Event_ = "" Or Not uOp.Success Or Not tDb.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

Error_process:
    If uOp.Success = False Or tLocals.Error_count <> 0 Or uOp.Event_ = "" Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

Ender_full:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Database_general(ByRef tDb As udtDatabase_oledb)
    ' Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tTable_name As String = ""
    Dim tCmd_str As String
    Dim tStatement As String = ""
    On Error GoTo Error_captured
    tLocals.Subr = udtApplication_subroutine_names.Database_general
    uOp.Subr = tLocals.Subr
    uError.Message = ""

    tLocals.Action = "Check database and sql basics"
    tDb.Error_ = ""
    tDb.Rows_affected = -1
    tDb.Success = False
    If tDb.Table_name = "" Then
      Call Sound_alert()
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      uError.Message = "Table not specified in tDb.Table_name."
      GoTo Error_process
    End If
    If tDb.Cmd_str = "" Then
      Stop ' Permenent.
      uError.Message = "No s.q.l. command."
      tLocals.Status = udtuStandard_texts.Failed
      GoTo Error_process
    Else
      tDb.Cmd_str = Trim(tDb.Cmd_str)
      tCmd_str = tDb.Cmd_str
    End If
    If tDb.Conn_str = "" Or tDb.Conn Is Nothing Or tDb.Cmd Is Nothing Or tDb.Database Is Nothing Then ' 2018.09.30.
      Stop ' Permenent.
      tLocals.Subr = uOp.Subr
      tTable_name = tDb.Table_name
      tCmd_str = tDb.Cmd_str

      Call Database_instantiate(tDb)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        uError.Message = "Database not instantiated."
        GoTo Error_process
      End If

      tDb.Table_name = tTable_name
      tDb.Cmd_str = tCmd_str
    End If
    If tDb.Database <> "" And Not (System.IO.File.Exists(tDb.Database)) Then
      Call Sound_alert()
      Stop ' Permanent.
      tCmd_str = tDb.Cmd_str
      Call Memory_create(Nothing)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Continue_ = False
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
      tDb.Cmd_str = tCmd_str
    End If
    uError.Message = ""

    tLocals.Action = "Check sql syntax"
    tStatement = UCase(Left(tDb.Cmd_str, InStr(tDb.Cmd_str, " ") - 1))
    Call Sql_conditioning(tDb.Cmd_str, tStatement)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    tLocals.Action = "Write sql to text file"
    Call Sql_write(tDb.Cmd_str)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    tLocals.Action = "Detect and check table"
    If Not Database_general_detect_table_fn(tDb, tTable_name, tStatement) Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    With tDb
      Select Case tStatement
        Case udtuStandard_texts.Sql_create, udtuStandard_texts.Sql_insert, udtuStandard_texts.Sql_update, udtuStandard_texts.Sql_delete, udtuStandard_texts.Sql_alter
          tLocals.Action = tStatement & " " & tDb.Table_name
          If .Conn.State <> ConnectionState.Open Then .Conn.Open()
          .Cmd.CommandText = tDb.Cmd_str
          tLocals.Action = udtuStandard_texts.Sql_write
          .Rows_affected = .Cmd.ExecuteNonQuery()
          .Conn.Close()
          tLocals.Msg = "records returned: " & .Rows_affected & "."

        Case udtuStandard_texts.Sql_select
          .Dataset = New DataSet
          .Data_adapter = New OleDb.OleDbDataAdapter(.Cmd_str, .Conn_str)
          tLocals.Action = udtuStandard_texts.Dataset_fill
          .Rows_affected = .Data_adapter.Fill(.Dataset, .Table_name) ' A momentary connection. 
          tLocals.Msg = "records returned: " & .Rows_affected & "."
          If .Rows_affected = 0 Then
            ' Check server case sensitivity. 
            ' Stop
          End If

        Case Else
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
      End Select

      .Success = True
      .Cmd_str = ""
      .Cmd.CommandText = ""
    End With
    If tDb.Success Then tLocals.Success = udtBoolean_extended.True_

Upgrade_section:
    If tLocals.Action = udtuStandard_texts.Sql_write Or tLocals.Action = udtuStandard_texts.Dataset_fill Then

    ElseIf tLocals.Action = udtuStandard_texts.Dataset_fill Then
      ' Normal operation, continue.

    ElseIf InStr(tLocals.Action, tStatement) > 0 Then
      ' Normal operation, continue.

    ElseIf tLocals.Action = udtuStandard_texts.Upgrade Then
      If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
      Call Database_diagnostic_00(tDb)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
      tLocals.Action = "Upgrade section completed"

    ElseIf tLocals.Action = udtuStandard_texts.Retry Then
      Stop ' Permanent.
      Call Database_general(tDb)
      uOp.Subr = tLocals.Subr
      If uDialogue.Active Then GoTo Error_process

    ElseIf tLocals.Action = udtuStandard_texts.Continue_ Then
      tDb.Success = True
      tLocals.Success = udtBoolean_extended.True_

    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

Ender_part_1:
    Select Case uError.Message
      Case ""
        If tLocals.Success = udtBoolean_extended.True_ Then
          tDb.Attempt = 0
          tDb.Error_ = ""
          If tLocals.Msg = "" Then
            tLocals.Response = "Success."
          Else
            tLocals.Response = "Success; " & tLocals.Msg
          End If
        Else
          If tLocals.Msg = "" Then
            tLocals.Response = "Failure."
          Else
            tLocals.Response = "Failure;" & tLocals.Msg
          End If
          Stop ' Permanent.
        End If
        uOp.Result = tLocals.Response

      Case "Database not instantiated."
        Stop ' Permanent. Use udtuStandard_texts.x.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process

      Case "No s.q.l. command."
        Call Sound_alert()
        Stop ' Indefinite. Use udtuStandard_texts.x.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        uError.Message = "The s.q.l. command was empty."
        Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)

      Case "Table not specified in tDb.Table_name."
        Stop ' Permanent. Use udtuStandard_texts.x.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process

      Case Else
        Stop ' Indefinite.
        If uApp.Terminating = 0 Then Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
        If InStr(uError.Description, "The process cannot access the file") > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
          frmMain.Close()
        End If
    End Select

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uError.Number <> 0 Or uApp.Terminating > 1 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If tLocals.Action = "Fill" Or Left$(tLocals.Action, 6) = udtuStandard_texts.Sql_insert Or Left$(tLocals.Action, 6) = udtuStandard_texts.Sql_update Then tLocals.Action = ""
      If uApp.Terminating = 0 Then
        tDb.Error_ = uError.Description
        Call Database_error_handler(tLocals, tDb)
        tLocals.Error_count = tLocals.Error_count - 1
        tLocals.Success = udtBoolean_extended.Null
        Call Error_clear(tLocals)
        'If (tLocals.Action <> "" Or tDb.Upgrade_table <> "") And Not tDb.Upgrade_in_progress Then
        '  Resume Upgrade_section
        'Else
        '  Resume Error_captured
        'End If
        If tLocals.Action = "" Then
          Resume Error_captured
        Else
          Resume Upgrade_section
        End If
      End If
    End If
    uError.Message = ""

Ender_full:
    If (tStatement = udtuStandard_texts.Sql_select Or tStatement = udtuStandard_texts.Sql_update) And tDb.Success Then
    Else
      Call Database_clear_partial(tDb, False)
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Function Database_general_micro_execute(ByRef tDb As udtDatabase_oledb) As Object
    ' Separated 2021.04.17. Revised 2021.04.22.
    With tDb
      If .Cmd.CommandText <> "" Then
        If .Cmd_str <> .Cmd.CommandText Then
          Database_general_micro_execute = False
        End If
      End If
      Call Sql_write(.Cmd_str)
      .Cmd.CommandText = .Cmd_str
      If .Conn.State <> ConnectionState.Open Then .Conn.Open()
      Database_general_micro_execute = .Cmd.ExecuteScalar
      .Conn.Close()
      .Cmd_str = Nothing
      .Cmd.CommandText = Nothing
    End With
  End Function

  Sub Database_general_micro_fill(ByRef tDb As udtDatabase_oledb)
    ' Separated 2021.04.17. Revised 2021.04.22.
    With tDb
      If .Cmd.CommandText <> "" Then
        If .Cmd_str <> .Cmd.CommandText Then
          .Rows_affected = -2
        End If
      End If
      Call Sql_write(.Cmd_str)
      .Cmd.CommandText = .Cmd_str
      .Data_adapter = New OleDb.OleDbDataAdapter(.Cmd_str, .Conn)
      ' tLocals.Action = udtuStandard_texts.Dataset_clear
      .Dataset.Clear()
      ' tLocals.Action = udtuStandard_texts.Dataset_fill
      ' tFile.Open_file = .Db.Table_name
      .Rows_affected = .Data_adapter.Fill(.Dataset, .Table_name)
      ' tFile.Open_file = ""
      .Cmd_str = Nothing
      .Cmd.CommandText = Nothing
    End With
  End Sub

  Private Function Database_general_detect_table_fn(ByVal tDb As udtDatabase_oledb, ByRef tTable_name As String, ByRef tStatement As String) As Boolean
    ' Separated 2018.08.21. Revised 2021.02.16.
    Dim tLocals As udtLocals = Nothing
    Dim tTable_position As Short = 0       ' The position of the table name in the command string.
    Dim tParenthesis_position As Short = 0 ' The position of the first open parenthesis in the command string.
    uOp.Subr = udtApplication_subroutine_names.Database_general_detect_table

    tLocals.Action = "Determine table name"
    tTable_position = 0
    ' If UCase(tStatement) = udtuStandard_texts.Sql_insert Then
    If InStr(LCase(tDb.Cmd_str), "insert") > 0 Then
      tStatement = udtuStandard_texts.Sql_insert
      tTable_position = InStr(tDb.Cmd_str, "INTO", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "into ", "")
      tTable_name = Replace(tTable_name, "INTO ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_select Then
      tTable_position = InStr(tDb.Cmd_str, "FROM", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "from ", "")
      tTable_name = Replace(tTable_name, "FROM ", "")
      ' ElseIf UCase(tStatement) = udtuStandard_texts.Sql_create Then
    ElseIf InStr(LCase(tDb.Cmd_str), "create") > 0 Then
      tStatement = udtuStandard_texts.Sql_create
      tTable_position = InStr(tDb.Cmd_str, "CREATE TABLE", CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "create table ", "")
      tTable_name = Replace(tTable_name, "CREATE TABLE ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_update Then
      tTable_position = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_update, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "update ", "")
      tTable_name = Replace(tTable_name, "UPDATE ", "")
    ElseIf UCase(tStatement) = udtuStandard_texts.Sql_delete Then
      tTable_position = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_delete, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "delete from ", "")
      tTable_name = Replace(tTable_name, "DELETE FROM ", "")
      ' ElseIf UCase(tStatement) = udtuStandard_texts.Sql_alter Then
    ElseIf InStr(LCase(tDb.Cmd_str), "alter") > 0 Then
      tStatement = udtuStandard_texts.Sql_alter
      tTable_position = InStr(tDb.Cmd_str, udtuStandard_texts.Sql_alter, CompareMethod.Text)
      tTable_name = Mid(tDb.Cmd_str, tTable_position)
      tTable_name = Replace(tTable_name, "alter table ", "")
      tTable_name = Replace(tTable_name, "ALTER TABLE ", "")
    ElseIf UCase(tStatement) = "PROCEDURE" Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    End If
    If tTable_name = "" Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
      GoTo Ender_full
    End If

    tTable_position = InStr(tTable_name, " ")
    tParenthesis_position = InStr(tTable_name, "(")
    If UCase(tStatement) = udtuStandard_texts.Sql_insert And tTable_position > tParenthesis_position Then
      If tTable_position < tParenthesis_position Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_full
        uError.Message = "Alert! The table is not specified in the command string."
        Exit Function
      End If
    End If
    If tTable_position > 0 Then
      tTable_name = Mid(tTable_name, 1, tTable_position - 1)
    End If
    tTable_position = InStr(tTable_name, vbCr, CompareMethod.Text)
    If tTable_position > 0 Then
      tTable_name = Mid(tTable_name, 1, tTable_position - 1)
    End If
    tTable_name = Replace(tTable_name, "[", "")
    tTable_name = Replace(tTable_name, "]", "")
    tTable_name = Replace(tTable_name, "(", "")

    If UCase(tDb.Table_name) <> UCase(tTable_name) Then
      Stop ' Permenent.
      tLocals.Status = udtuStandard_texts.Failed
      tDb.Table_name = tTable_name
    Else
      Database_general_detect_table_fn = True
    End If

Ender_full:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Public Sub Sql_conditioning(ByRef tCmd_str As String, tStatement As String)
    ' Revised 2021.05.07.
    ' Sql conditioning:
    ' Modify the sql command string depending on the sql server being used.
    ' References:
    ' Having www.w3schools.com/sql/sql_having.asp.
    ' Top    www.w3schools.com/sql/sql_top.asp.
    Dim tLocals As udtLocals = Nothing

    If tCmd_str = "" Then
      Call Sound_alert()
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
    Else

      If InStr(LCase(tCmd_str), "format") = 0 And InStr(tCmd_str, Chr(34)) > 0 Then
        Stop ' Fix the code.
        ' Sql should use only single speech marks, not double speech marks.
      End If

      If InStr(tCmd_str, udtuStandard_texts.Sql_alter) > 0 Then

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_create) > 0 Then

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_delete) > 0 Then

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_insert) > 0 Then
        If InStr(tCmd_str, Chr(37)) <> 0 Then ' %.
          ' Stop ' Indefinite. ' Works in LIS-Vm-Test-07, 2021.05.07.
          tCmd_str = Replace(tCmd_str, "%", "' & Chr(37) & '")
          ' tCmd_str = Replace(tCmd_str, "%", " percent") ' Until solved.
          ' tCmd_str = Replace(tCmd_str, " & ''", "")
        End If

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_modify) > 0 Then
        tCmd_str = Replace(tCmd_str, "MODIFY COLUMN", "ALTER TABLE")

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_select) > 0 Then

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_show) > 0 Then
        uError.Message = "Command 'SHOW' is not understood in an Access database."

      ElseIf InStr(tCmd_str, udtuStandard_texts.Sql_update) > 0 Then
        If InStr(UCase(tCmd_str), " LIKE ") > 0 Then ' 2017.08.06.
          ' Use % not *.
          tCmd_str = Replace(tCmd_str, Chr(42), Chr(37)) ' Replace * with %.
        End If

      Else
        uError.Description = "Sql command has not been properly formed."
        tLocals.Success = udtBoolean_extended.False_

      End If
    End If
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Database_error_handler(ByRef tLocals_external As udtLocals, ByRef tDb As udtDatabase_oledb)
    ' Separated by Michael John Boyack Hodgson, 2017.12.03. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Database error handler"
    tLocals.Action = ""
    tDb.Error_ = tLocals_external.Error_description
    tLocals_external.Action = ""
    uError.Last_error_sub = tLocals_external.Subr

    If Not uOp.Diagnostics.In_progress And InStr(uOp.Called_from, "Data fix") = 0 Then Stop ' Do it here, then resume.

    Select Case tLocals_external.Error_number
      Case 0
        Stop ' Permanent.
        tLocals_external.Action = udtuStandard_texts.Retry

      Case 5 ' Description varies: 
        If InStr(tLocals_external.Error_description, "already exists") > 0 Then
          ' Skip and continue.
          tLocals.Msg = tLocals_external.Error_description
          Call Error_clear(tLocals)
          tLocals_external.Action = udtuStandard_texts.Continue_

        ElseIf tLocals_external.Error_description = "Cannot find table or constraint." Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals_external.Action = udtuStandard_texts.Upgrade

        ElseIf InStr(tLocals_external.Error_description, "Could not find file") > 0 And InStr(tLocals_external.Error_description, UCase(Left(udtTables.Information_Schema_Columns, InStr(udtTables.Information_Schema_Columns, ".") - 1))) Then
          Stop ' Do nothing, exit gracefully.

        ElseIf InStr(tLocals_external.Error_description, "Could not find output table") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals_external.Action = udtuStandard_texts.Upgrade

        ElseIf tLocals_external.Error_description = "Could not update; currently locked." Then
          Stop ' Indefinite.
          tLocals_external.Action = udtuStandard_texts.Retry

        ElseIf InStr(tLocals_external.Error_description, "Data type mismatch in criteria expression.") > 0 Then
          Stop ' Permanent.
          uApp.Terminating = uApp.Terminating + 1

        ElseIf InStr(tLocals_external.Error_description, "does not belong to table") > 0 Then
          Stop ' Indefinite, to test:
          tDb.Upgrade_table = tDb.Table_name
          Call Database_diagnostic_level_05(tDb)
          uOp.Subr = tLocals.Subr
          If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
            ' Else
            ' tUpgraded = True
          End If
          tLocals.Success = udtBoolean_extended.True_
          tLocals_external.Action = udtuStandard_texts.Retry

          'Stop ' Otherwise:
          'tDb.Upgrade_table = tDb.Table_name
          ' tLocals.Action  = udtuStandard_texts.Upgrade

        ElseIf InStr(tLocals_external.Error_description, "ERROR [08001] Client unable to establish connection") > 0 Then
          Stop ' Permanent.
          Call Error_display(tLocals)

        ElseIf InStr(tLocals_external.Error_description, "is already opened exclusively by another user") > 0 Then
          If tDb.Attempt <= 2 Then
            tDb.Attempt = tDb.Attempt + 1
            tLocals_external.Action = udtuStandard_texts.Retry
          Else
            ' uError.Message = "Please close the memory database. Clicking Cancel will close " & uApp.App_name & "."
            ' Call Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_cancel, uOp.Subr)
            uError.Message = "Please close the memory database."
            Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
          End If

        ElseIf InStr(tLocals_external.Error_description, "is exclusively locked by user ") > 0 Then
          uError.Message = "Exclusive access denied."
          If uApp.Debug Then
            tLocals_external.Action = udtuStandard_texts.Retry
          Else
            tLocals_external.Action = ""
            Stop ' Permanent.
            uApp.Terminating = uApp.Terminating + 1
            Call Error_display(tLocals)
          End If

        ElseIf InStr(tLocals_external.Error_description, "No value given for one or more required parameters.") > 0 Then
          ' Try a memory upgrade; but that may not be the solution. 
          If uOp.Called_from = udtApplication_subroutine_names.Data_fix_field_rename Then
            tDb.Upgrade_table = ""
            tLocals.Success = udtBoolean_extended.True_
          Else
            tDb.Upgrade_table = tDb.Table_name
            tLocals_external.Action = udtuStandard_texts.Upgrade

            Stop ' Indefinite, to test: added 2021.02.25:
            Call Database_diagnostic_level_05(tDb)
            If Not uOp.Success Then tLocals.Success = udtBoolean_extended.False_
          End If

        ElseIf tLocals_external.Error_description = "Operation must use an updateable query." Then
          Call frmMain.txtDiagnostic_result_append("Access denied. Skipping this database." & vbCrLf & vbCrLf)
          tLocals_external.Action = udtuStandard_texts.Retry
          tLocals.Status = udtuStandard_texts.Failed
          If tDb.Database = uMemory.Database Then
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Status = udtuStandard_texts.Failed
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
          Else
            Call Error_clear(tLocals)
          End If

        ElseIf InStr(tLocals_external.Error_description, "Syntax error (missing operator) in query expression ") > 0 Then
          ' Look for a missing comma.
          ' Could also be a date format problem.
          Stop ' Permanent.
          Call Error_display(tLocals)

        ElseIf InStr(tLocals_external.Error_description, "Syntax error in date in query expression") > 0 Then
          Stop ' Permanent.

        ElseIf InStr(tLocals_external.Error_description, "Syntax error in field definition.") > 0 Then
          Stop ' Permanent.

        ElseIf InStr(tLocals_external.Error_description, "Syntax error in number in query expression") > 0 Then
          Stop ' Permanent.

        ElseIf InStr(tLocals_external.Error_description, "The database engine could not lock table") > 0 Then
          ' The database engine could not lock table 'Preference' because it is already in use by another person or process.
          tLocals_external.Action = udtuStandard_texts.Retry

        ElseIf InStr(tLocals_external.Error_description, "The INSERT INTO statement contains the following unknown field name") > 0 Then
          If tDb.Upgrade_table <> "All" Then tDb.Upgrade_table = tDb.Table_name
          tDb.Error_ = tLocals_external.Error_description
          tLocals_external.Action = udtuStandard_texts.Upgrade

          If uOp.Called_from <> udtApplication_subroutine_names.Database_diagnostic_level_05_part_01 Then
            If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
            Call Database_diagnostic_00(tDb)
            If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
              Stop ' Permanent.
              tLocals.Continue_ = False
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Status = udtuStandard_texts.Failed
              tLocals.Success = udtBoolean_extended.False_
              GoTo Error_process
            End If
            uOp.Success = False
            tLocals_external.Action = udtuStandard_texts.Retry
          End If

        ElseIf tLocals_external.Error_description = "The 'Microsoft.ACE.OLEDB.12.0' provider is not registered on the local machine." Then
          tDb.Error_ = "Access to the database " & gMemory.Database & " failed." & vbCrLf
          tDb.Error_ = tDb.Error_ & tLocals_external.Error_description

        ElseIf tLocals_external.Error_description = "The 'Microsoft.Jet.OLEDB.4.0' provider is not registered on the local machine." Then
          tDb.Error_ = "Access to the database " & gMemory.Database & " failed." & vbCrLf
          tDb.Error_ = tDb.Error_ & tLocals_external.Error_description

        ElseIf InStr(tLocals_external.Error_description, "The Microsoft Jet database engine cannot find the input table or query") > 0 Then
          tDb.Upgrade_table = tDb.Table_name
          tLocals_external.Action = "Create"

        ElseIf InStr(tLocals_external.Error_description, "The table '") > 0 And _
               InStr(tLocals_external.Error_description, "' is already opened exclusively by another user, or it is already open through the user interface and cannot be manipulated programmatically.") > 0 Then
          Stop ' Permanent.
          uApp.Terminating = uApp.Terminating + 1
          Call Error_display(tLocals)

        ElseIf InStr(tLocals_external.Error_description, "There is no field named") > 0 Then
          ' There is no field named 'Currency' in table 'Preference'.
          If uOp.Called_from = udtApplication_subroutine_names.Data_fix_field_rename Then
            tDb.Upgrade_table = ""
            tLocals.Success = udtBoolean_extended.True_
          Else
            tDb.Upgrade_table = tDb.Table_name
            tLocals_external.Action = udtuStandard_texts.Upgrade
          End If

        ElseIf InStr(tLocals_external.Error_description, "Unrecognized database format") > 0 Then
          Stop ' Permanent.
          ' Kill it.
          ' If not resolved, use uUser.Ap_data_path & "\" & ulis.Name & "\" & uApp.App_name & "\" & uApp.internal & " Memory.mdb"

        Else
          Stop ' Permanent.
          If tLocals_external.Error_description = "" Then
            ' Another process can nullify it.
            tLocals_external.Action = udtuStandard_texts.Retry
          Else
            Call Error_display(tLocals)
          End If

        End If

      Case 9 ' Subscript out of range
        Stop ' Indefinite.
        Select Case tLocals_external.Error_description
          '    Case "Check memory db"
          '      ' Field doesn't exist; create.
          '      tBool = False
          '      tNum = 0
          '      tDate = Nothing
          '      tString = ""
          '      uUser.Preferences_changed = True
          '      Resume Next
          '    Case "Read" ' Field does not exist. 
          '       tLocals.Msg  = ""
          '      tBool = False
          '      tNum = 0
          '      tDate = Nothing
          '      tString = ""
          '      uUser.Preferences_changed = True
          '      Resume Next
          '    Case Else
          '      Stop ' Permanent.
          '                  Call Error_display(tLocals) 
          '      uOp.Subr =  tLocals.Subr 
        End Select

        Select Case tLocals.Action
          Case "Check memory db"
            ' Field doesn't exist; create.
            Resume Next
          Case "Read" ' Field does not exist. 
            tLocals.Msg = ""
            Resume Next
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select

      Case Else
        Stop ' Permanent.
        Call Error_display(tLocals)

    End Select ' tLocals_external.Error_number

    ' Call Database_clear_partial(tDb, False) ' 2021.03.21.

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

  End Sub

  Function Database_query_one_fn(ByVal tCmd_str As String, ByRef tDatabase As String) As String
    ' Written by Michael John Boyack Hodgson, 2012.05.04. Revised 2021.03.17.
    ' Generalised version.
    Dim tLocals As udtLocals = Nothing
    Dim tDb As udtDatabase_oledb = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_query_one_fn
    tLocals.Success = udtBoolean_extended.Null
    On Error GoTo Error_captured
    Database_query_one_fn = ""

    Call Sql_write(tCmd_str)

    With tDb
      .Error_ = ""
      .Cmd_str = tCmd_str
      .Database = tDatabase
      Call Database_instantiate(tDb)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_partial
      End If
      uOp.Success = False

      If .Conn.State <> ConnectionState.Open Then .Conn.Open()
      .Cmd.CommandText = tCmd_str
      Database_query_one_fn = .Cmd.ExecuteScalar
      .Conn.Close()
      .Cmd.CommandText = ""
      .Error_ = ""
    End With
    tLocals.Success = udtBoolean_extended.True_

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 13 ' Conversion from type 'DBNull' to type 'String' is not valid.
            Resume Next
          Case Else
            Call Database_error_handler(tLocals, tDb)
        End Select
      End If
    End If

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Function Memory_query_one_fn(ByVal tCmd_str As String) As String
    ' Revised 2021.03.31.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Memory_query_one_fn
    On Error GoTo Error_captured
    tLocals.Msg = ""
    tLocals.Success = udtBoolean_extended.Null
    Memory_query_one_fn = ""
    Call Sql_write(tCmd_str)

    With uMemory
      .Error_ = ""

      If .Conn.State <> ConnectionState.Open Then .Conn.Open()
      .Cmd.CommandText = tCmd_str
      Memory_query_one_fn = .Cmd.ExecuteScalar
      .Conn.Close()
      .Cmd.CommandText = ""
      .Error_ = ""
    End With
    tLocals.Success = udtBoolean_extended.True_

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 13 ' Conversion from type 'DBNull' to type 'String' is not valid.
            Resume Next
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Function

  Function Memory_query_one_fn(ByVal tTable_name As String, ByVal tSelect_field As String, ByVal tWhere_field As String, ByVal tWhere_value As String) As String
    ' Revised 2021.05.06.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    uOp.Subr = udtApplication_subroutine_names.Memory_query_one_fn
    uOp.Success = False
    Memory_query_one_fn = ""
    uMemory.Error_ = ""

    tLocals.Action = udtuStandard_texts.Sql_build_select
    If tWhere_field = "Date" Then
      tWhere_value = Replace(tWhere_value, ".", "-")
      uMemory.Cmd_str = "SELECT " & tSelect_field & " FROM " & tTable_name & " WHERE [" & tWhere_field & "]=Format (#" & tWhere_value & "#, " & Chr(34) & "yyyy.mm.dd" & Chr(34) & ")"
      ' uMemory.Cmd_str = "SELECT " & tSelect_field & " FROM " & tTable_name & " WHERE [" & tWhere_field & "]=Format (#" & tWhere_value & "#, " & Chr(34) & uMachine.Date_format & Chr(34) & ")"
    Else
      If Val(tWhere_value) = 0 Then
        uMemory.Cmd_str = "SELECT " & tSelect_field & " FROM " & tTable_name & " WHERE [" & tWhere_field & "]=" & "'" & tWhere_value & "'"
      Else
        uMemory.Cmd_str = "SELECT " & tSelect_field & " FROM " & tTable_name & " WHERE [" & tWhere_field & "]=" & tWhere_value
      End If
    End If
    Call Sql_write(uMemory.Cmd_str)
    If uMemory.Dataset Is Nothing Then uMemory.Dataset = New DataSet

    tLocals.Action = udtuStandard_texts.Dataset_fill
    uMemory.Rows_affected = uMemory.Data_adapter.Fill(uMemory.Dataset, tTable_name)
    uMemory.Error_ = ""

    tLocals.Action = udtuStandard_texts.Dataset_read
    Select Case uMemory.Rows_affected
      Case 0
      Case 1
        If Not IsDBNull(uMemory.Dataset.Tables(tTable_name).Rows(0).Item(tSelect_field)) Then
          Memory_query_one_fn = uMemory.Dataset.Tables(tTable_name).Rows(0).Item(tSelect_field).ToString
        End If
      Case Else
        Stop
    End Select

    uOp.Success = True
    uOp.Subr = ""
  End Function

  Function Last_use_fn() As Boolean
    ' Revised 2021.04.07.
    Dim tLocals As udtLocals = Nothing
    Dim tSearch As udtSearch = Nothing

    Dim tToday As Date
    Dim tToday_str As String
    Dim tLast_date As Date
    tLocals.Subr = udtApplication_subroutine_names.Last_use_fn
    uOp.Subr = tLocals.Subr

    tToday = DateAndTime.Today ' M/d/y.
    tToday_str = CStr(tToday) ' Y/m/d.
    tToday_str = Replace(tToday_str, ".", "/")
    tLocals.Action = udtuStandard_texts.Sql_build_select
    uMemory.Table_name = uTables.Log
    uMemory.Cmd_str = "SELECT TOP 1 Date_ FROM " & uMemory.Table_name & " WHERE " & vbCr
    uMemory.Cmd_str = uMemory.Cmd_str & "Date_=Format(#" & tToday_str & "#, " & Chr(34) & "yyyy/mm/dd" & Chr(34) & ") AND" & vbCr
    uMemory.Cmd_str = uMemory.Cmd_str & "Subr='End game'" & vbCr
    uMemory.Cmd_str = uMemory.Cmd_str & "ORDER BY Id DESC"
    Call Database_general(uMemory)
    uOp.Subr = tLocals.Subr
    If uMemory.Rows_affected <> 1 Then
      uMemory.Table_name = uTables.Log
      uMemory.Cmd_str = "SELECT TOP 1 Date_ FROM " & uMemory.Table_name & " WHERE" & vbCr
      uMemory.Cmd_str = uMemory.Cmd_str & "Date_<>Format(#" & tToday_str & "#, " & Chr(34) & "yyyy/mm/dd" & Chr(34) & ")" & vbCr
      If uApp.App_name = udtApplication.App_Seer Then
        uMemory.Cmd_str = uMemory.Cmd_str & "AND Application=" & "'" & uApp.App_name & "'" & vbCr
      End If
      uMemory.Cmd_str = uMemory.Cmd_str & "ORDER BY Id DESC"
      Call Database_general(uMemory)
      uOp.Subr = tLocals.Subr
      tSearch.Keyword = "Date_"
      If uMemory.Rows_affected > 0 Then
        tLocals.Action = udtuStandard_texts.Dataset_read
        If IsDBNull(uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)) Then
        Else
          tLast_date = CDate(uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item(tSearch.Keyword)) ' Mm/dd/yyyy.
          If DateAdd(DateInterval.Day, 7, tLast_date) < DateAndTime.Today Then
            Last_use_fn = True
          End If
        End If
      End If
    End If

    uOp.Subr = ""
  End Function

  Sub Database_clear_partial(ByRef tDb As udtDatabase_oledb, ByVal tBefore As Boolean)
    ' Re-implemented MJBH 2017.08.13. Revised 2021.02.17.
    ' tBefore: If before a database action; to reset values.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Database_clear_partial

    With tDb
      If .Conn IsNot Nothing Then
        If .Conn.State = ConnectionState.Open Then .Conn.Close()
      End If
      If tBefore Then .Cmd_str = Nothing
      If tBefore Then If .Cmd IsNot Nothing Then .Cmd.CommandText = Nothing
      ' If .Cmd IsNot Nothing Then .Cmd = Nothing
      .Data_adapter = Nothing
      If .Dataset IsNot Nothing Then
        If tBefore Then
          tLocals.Action = udtuStandard_texts.Dataset_clear
          .Dataset.Clear()
          tLocals.Action = ""
        End If
      End If
      If tBefore Then .Error_ = Nothing
      .Rows_affected = 0
      If tBefore Then .Success = False
      ' If tBefore Then .Table_name = Nothing
    End With

    uOp.Subr = ""
  End Sub

#End Region ' End region, "Database operations general."

End Module
