﻿' Module: Kernel_common
' Copyright © LivingInfo Systems, 1999. All rights reserved.
' Except as provided by the Copyright Act 1994, no part of this publication may be reproduced or stored in a retrieval system in any form or by any means without the prior written permission of the copyright owner. 
' www.livinginfosystems.com.
' Written by Michael John Boyack Hodgson.

' References:
' Microsoft.VisualBasic.PowerPacks.Vs.

' Imported name-spaces:
' Microsoft.VisualBasic.
' System.
' System.Data.
' System.Drawing.
' System.Management.
' System.Speech.
' System.Windows.Forms.
' System.Xml.

Option Explicit On
Option Strict Off ' Allows translation of numeric variables into strings.
Imports System.Globalization
' Imports System.DirectoryServices.AccountManagement.UserPrincipal
Imports System.IO.Compression
Imports System.Runtime.InteropServices

Module Kernel_common
  Public Const cKernel_common_version As Date = #5/7/2021# ' M/d/y.
  <DllImport("Shell32", CharSet:=CharSet.Auto, SetLastError:=True)> _
  Public Function ShellExecute(ByVal hwnd As IntPtr, ByVal lpOperation As String, ByVal lpFile As String, ByVal pParameters As String, ByVal lpDirectory As String, ByVal nShowCmd As Integer) As IntPtr
  End Function
  ' Public Declare Function SetLocaleInfo& Lib "kernel32" Alias "SetLocaleInfoA" (ByVal Locale As Long, ByVal LCType As Long, ByVal lpLCData As String)
  ' Public Declare Function PostMessage Lib "user32" Alias "PostMessageA" (ByVal hwnd As Long, ByVal wMsg As Long, ByVal wParam As Long, ByVal lParam As Object) As Long
  Dim mModule As New Collection
  Public mObjSynth As New System.Speech.Synthesis.SpeechSynthesizer

#Region "Kernel"

  Sub Environment_determine()
    ' Revised 2021.04.28.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Environment_determine
    uOp.Called_from = tLocals.Subr
    tLocals.Action = udtuStandard_texts.Initiation
    tLocals.Status = udtuStandard_texts.Initiation
    uApp.Initialising = True
    On Error GoTo Error_captured

    tLocals.Action = "Set variables"
    If uApp.Startup_sequence < 7 Then
      tLocals.Action = "Set uLIS"
      Call uLIS.Initialise()

      uApp.Startup_sequence = 7
    End If

    If uApp.Startup_sequence < 8 Then
      tLocals.Action = "Set uMachine"
      Call uMachine.Initialise()

      uApp.Startup_sequence = 8
    End If

    If uApp.Startup_sequence < 9 Then
      tLocals.Action = "Kernel local pre"
      Call Kernel_local_pre()
      uOp.Success = False

      uApp.Startup_sequence = 9
    End If

    If uApp.Startup_sequence < 10 Then
      tLocals.Action = "Set uApp"
      Call uApp.Initialise()

      uApp.Startup_sequence = 10
    End If

    If uApp.Startup_sequence < 11 Then
      tLocals.Action = "Set table names"
      Call uTables.Initialise()

      uApp.Startup_sequence = 11
    End If

    If uApp.Startup_sequence < 12 Then
      tLocals.Action = "Set uCrypt"
      Call uCrypt.Initialise()

      uApp.Startup_sequence = 12
    End If

    If uApp.Startup_sequence < 13 Then
      tLocals.Action = "Set uUser"
      Call uUser.Initialisation()

      uApp.Startup_sequence = 13
    End If

    If uApp.Startup_sequence < 14 Then
      tLocals.Action = "Set uActive"
      Call uActive.Initialise()
      ' Call Geo_location_detect()

      uApp.Startup_sequence = 14
    End If

    If uApp.Startup_sequence < 15 Then
      tLocals.Action = "Set uCon_help"
      Call uCon_help.Initialise()

      uApp.Startup_sequence = 15
    End If

    tLocals.Action = "Detect environment"
    If uApp.Startup_sequence < 16 Then
      tLocals.Action = "Set operating system version dependants"
      Call Operating_system_detect()
      uOp.Subr = tLocals.Subr
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      If uMachine.Testing_machine Then uLIS.Development_path = uUser.My_dox
      uApp.Startup_sequence = 16
    End If

    If uApp.Startup_sequence < 17 Then
      tLocals.Action = "Start-up sequence begins"
      tLocals.Success = udtBoolean_extended.True_
      Call Log_write(tLocals, 0)
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      tLocals.Success = udtBoolean_extended.Null
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 17
    End If

    If uApp.Startup_sequence < 18 Then
      tLocals.Action = "Memory instantiate"
      Call Memory_instantiate()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 18
    End If
    tLocals.Status = ""

    If uApp.Startup_sequence < 19 Then
      uOp.Called_from = tLocals.Subr
      tLocals.Action = "First Memory log entry"
      tLocals.Success = udtBoolean_extended.True_
      Call Log_write(tLocals, 0)
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      tLocals.Success = udtBoolean_extended.Null
      uOp.Success = False

      uOp.Subr = tLocals.Subr
      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 19
    End If

    If uApp.Startup_sequence < 20 Then
      tLocals.Action = "Installed check"
      Call Installed_check()
      uOp.Subr = tLocals.Subr
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 20
    End If

    If uApp.Startup_sequence < 21 Then
      tLocals.Action = udtApplication_subroutine_names.Version_check
      Call Version_check()
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 21
    End If

    tLocals.Action = "Detect user settings and act on them"
    If uApp.Startup_sequence < 22 Then
      tLocals.Action = udtApplication_subroutine_names.Preferences_load
      Call Preferences_load()
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      If uApp.First_run_in_progress Then
        ' Set first run defaults that may be set differently in Preferences:
        uUser.Start_maximised = True
        uUser.Preferences_changed = True
      End If

      If uApp.First_run_completed_date = udtuStandard_texts.Date_null Then
        Call First_run()
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
      End If

      uOp.Called_from = tLocals.Subr
      If uUser.First_name = "" Then uUser.First_name = Environ("UserName")
      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 22
    End If

    If uApp.Startup_sequence < 23 Then
      tLocals.Action = "Registration"
      Call Registration()
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 23
    End If

    If uApp.Startup_sequence < 24 Then
      tLocals.Action = "Kernel local post"
      Call Kernel_local_post()
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      uOp.Called_from = tLocals.Subr
      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 24
    End If

    If uApp.Startup_sequence < 25 Then
      tLocals.Action = udtApplication_subroutine_names.Check_run_count
      Call Run_count_check()
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 25
    End If

    If uApp.Startup_sequence < 26 Then
      tLocals.Action = uApp.App_name & udtuStandard_texts.Session_begin ' This session begins.
      tLocals.Success = udtBoolean_extended.True_
      Call Log_write(tLocals, 0)
      tLocals.Success = udtBoolean_extended.Null
      uOp.Success = False

      Call frmMain.Statuslabel1_set()
      uApp.Startup_sequence = 26
    End If

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Or uApp.Terminating <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      uError.Last_error_sub = tLocals.Subr
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            Call Error_clear(tLocals)
            Resume
            Resume Next
          Case 5
            If InStr(Err.Description, "An error occurred creating the form. See Exception.InnerException for details. The error is: Could not find any resources appropriate for the specified culture or the neutral culture.") > 0 Then
              Stop ' Permanent.
              Call Sound_alert()
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Success = udtBoolean_extended.False_
              frmMain.Close()
            Else
              Stop ' Permanent.
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Success = udtBoolean_extended.False_
              Call Error_display(tLocals)
            End If
          Case 75 ' A required privilege is not held by the client.
            Resume Next
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    tLocals.Action = "Ender full"
    If uOp.Called_from = tLocals.Subr Then uOp.Called_from = ""
    ' If Not uDialogue.Active Then uApp.Initialising = False
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    Call frmMain.Statuslabel1_set()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub Operating_system_detect()
    ' Separated 2014.01.26. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Operating_system_determine
    uOp.Subr = tLocals.Subr

    tLocals.Action = "Set operating system version dependants"
    Dim tOperating_system_version_major As Short = System.Environment.OSVersion.Version.Major
    Dim tOperating_system_version_major_revision As Short = System.Environment.OSVersion.Version.MajorRevision
    Dim tOperating_system_version_minor As Short = System.Environment.OSVersion.Version.Minor
    Dim tOperating_system_version_minor_revision As Short = System.Environment.OSVersion.Version.MinorRevision
    Dim tOperating_system_version_build As Short = System.Environment.OSVersion.Version.Build
    Dim tOperating_system_version_revision As Short = System.Environment.OSVersion.Version.Revision
    Dim tOperating_system_version_string As String = System.Environment.OSVersion.VersionString
    Dim tMy_computer_info_osversion As String = My.Computer.Info.OSVersion
    '  tLocals.Msg  = Get_mac_address("10.1.1.2")
    Dim tOperating_system_64_bit As Boolean = Environment.Is64BitOperatingSystem
    Dim tProcess_64_bit As Boolean = Environment.Is64BitProcess

    uMachine.Os_major = System.Environment.OSVersion.Version.Major
    uMachine.Os_minor = System.Environment.OSVersion.Version.Minor
    uMachine.Os_name = System.Environment.OSVersion.ToString
    ' http://en.wikipedia.org/wiki/List_of_Microsoft_Windows_versions
    ' 3.0       : Win NT 3.51.

    ' 4.0       : Win 95.
    ' 4.0.0.0   : Win NT 4.
    ' 4.1       : Win 98.
    ' 4.1       : Win 98SE.
    ' 4.3       : Win NT server standard.
    ' 4.3       : Win NT server enterprise.
    ' 4.9       : Win ME.

    ' 5.0       : Win2000.
    ' 5.1       : XP.
    ' 5.2       : XP Pro x64.
    ' 5.2       : Windows home server.
    ' 5.2       : WinSer2003 enterprise, data centre, web edition, standard.
    ' 5.2       : WinSer2003 R2.

    ' 6.0.0     : Vista.
    ' 6.0.1     : 
    ' 6.0.3     : WinSer2008 standard, enterprise, data centre.
    ' 6.1.3     : WinSer2008 R2.
    ' 6.1.1     : Windows 7.
    ' 6.2.      : Windows 8.
    ' 6.3.      : Windows 8.1.
    ' 10.0.     : Windows 10.
    Select Case uMachine.Os_major
      Case Is < 5
        ' 3.0       : Win NT 3.51.

        ' 4.0       : Win 95.
        ' 4.0.0.0   : Win NT 4.
        ' 4.1       : Win 98.
        ' 4.1       : Win 98SE.
        ' 4.3       : Win NT server standard.
        ' 4.3       : Win NT server enterprise.
        ' 4.9       : Win ME.
        uError.Message = udtLivingInfoSystems.Name_short & " " & uApp.App_name & " needs testing on this operating system."
        Call Msgbox_LIS_top(uError.Message, vbExclamation, udtDialogue_types.Okay_only, tLocals.Subr)
        uMachine.Os_name = "<=Win98"
        uUser.Appdata_path = Environ("windir") & "\Profiles\" & Environment.UserName & "\Application Data"

      Case Is <= 6 ' Known systems, between W2k and Windows 7.
        ' 5.0       : Win2000.
        ' 5.1       : XP.
        ' 5.2       : XP Pro x64.
        ' 5.2       : Windows home server.
        ' 5.2       : WinSer2003 enterprise, data centre, web edition, standard.
        ' 5.2       : WinSer2003 R2.
        Stop ' Indefinite.
        uMachine.Environment_known = True ' Disable for testing.
        uUser.My_dox = Environ("Userprofile") & "\Documents"
        Select Case uMachine.Os_minor
          Case 0
            uApp.Default_top = 25
          Case Is >= 1
            uApp.Default_top = 40
          Case Else
            Stop ' Permanent.
        End Select

        ' If compiled, detects Win10 as W6:
        Dim tMy_dox As String = Replace(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments), "\Documents", "")
        If Dir(tMy_dox, FileAttribute.Directory) <> "" Then uUser.My_dox = tMy_dox

      Case Is <= 10 ' Windows 10.
        ' 6.0.0     : Vista.
        ' 6.0.1     : 
        ' 6.0.3     : WinSer2008 standard, enterprise, data centre.
        ' 6.1.3     : WinSer2008 R2.
        ' 6.1.1     : Windows 7.
        ' 6.2.      : Windows 8.
        ' 6.3.      : Windows 8.1.
        ' 10.0.     : Windows 10.
        uMachine.Environment_known = True ' Disable for testing.
        uApp.Default_left = -8
        uUser.My_dox = Environ("Userprofile") & "\OneDrive" ' uUser.My_dox = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments)

      Case Else
        Stop
        uMachine.Environment_known = False
        uMachine.Os_name = "Unknown"

        uOp.Called_from = uOp.Subr
        Call Environment_dump(True)
        uOp.Subr = tLocals.Subr
        uError.Message = "Operating system detection failue."
        Call Appdata_email()
        uOp.Subr = tLocals.Subr

    End Select

    ' Call Culture_detect()
    uOp.Subr = tLocals.Subr

    Call frmMain.Statuslabel1_set()
    uOp.Success = True
  End Sub

  Sub Culture_detect(Optional tSet As Boolean = False)
    ' Written MJBH 2019.10.19. Revised 2021.04.17.
    ' Was Kernel.Date_detect_system_format.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Culture detect"

    tLocals.Msg = System.Globalization.CultureInfo.CurrentCulture.Name
    Dim tDate_separator As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.DateSeparator
    Dim tFull_date_time_pattern As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.FullDateTimePattern
    Dim tAbbreviated_era_name As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.GetAbbreviatedEraName(0)
    '  tLocals.Msg  = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.GetAllDateTimePatterns
    ' https://docs.microsoft.com/en-us/dotnet/api/system.globalization.datetimeformatinfo.getalldatetimepatterns?redirectedfrom=MSDN&view=netframework-4.8#overloads
    Dim tEra_name As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.GetEraName(0)
    Dim tFirst_day_of_week As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.FirstDayOfWeek.ToString
    Dim tLong_date_pattern As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.LongDatePattern ' dd MMMM yyyy.
    Dim tNative_calendar_name As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.NativeCalendarName
    Dim tShort_date_pattern As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern
    Dim tTime_separator As String = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.TimeSeparator

    tLocals.Msg = System.Globalization.CultureInfo.CurrentUICulture.Name
    Dim tUI_date_separator As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.DateSeparator
    Dim tUI_full_date_time_pattern As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.FullDateTimePattern
    Dim tUI_abbreviated_era_name As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.GetAbbreviatedEraName(0)
    '  tLocals.Msg  = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.GetAllDateTimePatterns
    ' https://docs.microsoft.com/en-us/dotnet/api/system.globalization.datetimeformatinfo.getalldatetimepatterns?redirectedfrom=MSDN&view=netframework-4.8#overloads
    Dim tUI_era_name As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.GetEraName(0)
    Dim tUI_long_date_pattern As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.LongDatePattern ' dd MMMM yyyy.
    Dim tUI_native_calendar_name As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.NativeCalendarName
    Dim tUI_short_date_pattern As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern
    Dim tUI_time_separator As String = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.TimeSeparator

    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.CurrencyEnglishName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.CurrencyNativeName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.CurrencySymbol
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.DisplayName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.EnglishName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.GeoId
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.IsMetric
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.ISOCurrencySymbol
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.Name
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.NativeName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.ThreeLetterISORegionName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.ThreeLetterWindowsRegionName
    tLocals.Msg = System.Globalization.RegionInfo.CurrentRegion.TwoLetterISORegionName

    uMachine.Date_format = tShort_date_pattern

    If tSet Then
      tSet = True

      tLocals.Action = "Set"
      Dim tDtfi As DateTimeFormatInfo = CultureInfo.CreateSpecificCulture("en-US").DateTimeFormat
      tDtfi = CultureInfo.CreateSpecificCulture("en-US").DateTimeFormat
      tDtfi.ShortDatePattern = "dd.mm.yyyy"
      tDtfi.LongDatePattern = "dddd, dd mmmm, yyyy"

      tDtfi.ShortDatePattern = "dd-mm-yyyy"
      tDtfi.LongDatePattern = "dddd, dd mmmm, yyyy"

      tDtfi.ShortDatePattern = "dd/mm/yyyy"
      tDtfi.LongDatePattern = "dddd, dd mmmm, yyyy"

      tDtfi.ShortDatePattern = "mm/dd/yyyy"
      tDtfi.LongDatePattern = "dddd, MMMM d, yyyy"

      tDtfi.ShortDatePattern = "yyyy.mm.dd"
      tDtfi.LongDatePattern = "yyyy MMMM dd, dddd"
    End If

Ender_full:
    tLocals.Success = udtBoolean_extended.True_
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
  End Sub

  Sub Country_detect(ByVal tCountry As String, Optional ByVal tSet_active As Boolean = False)
    ' Revised 2021.04.17.
    ' tSet_active: If false, detect only temporary data. If true, set gActive with those data.
    Dim tLocals As udtLocals = Nothing
    Dim tCurrencyEnglishName As String
    Dim tCurrencyNativeName As String
    Dim tCurrency_format As String = Nothing ' ##,###,###,##0.00, etc.
    Dim tISO_currency_tla As String ' Three letter.
    Dim tCurrency_symbol As String ' One character.
    Dim tDisplay_name As String
    Dim tEnglishName As String
    Dim tName_two_letter As String
    Dim tNativeName As String
    Dim tTelephone_code As String = Nothing
    Dim tNote As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Detect_country
    uOp.Subr = tLocals.Subr

    ' Reference: 
    ' http://msdn.microsoft.com/en-us/library/system.globalization.regioninfo.aspx
    tLocals.Action = "Detect system values"
    tCurrencyEnglishName = System.Globalization.RegionInfo.CurrentRegion.CurrencyEnglishName
    tCurrencyNativeName = System.Globalization.RegionInfo.CurrentRegion.CurrencyNativeName
    tISO_currency_tla = System.Globalization.RegionInfo.CurrentRegion.ISOCurrencySymbol ' Three letter.
    tCurrency_symbol = System.Globalization.RegionInfo.CurrentRegion.CurrencySymbol ' One character.
    tDisplay_name = System.Globalization.RegionInfo.CurrentRegion.DisplayName
    tEnglishName = System.Globalization.RegionInfo.CurrentRegion.EnglishName  ' Country name, given in English.
    tName_two_letter = System.Globalization.RegionInfo.CurrentRegion.Name ' Two letter.
    tNativeName = System.Globalization.RegionInfo.CurrentRegion.NativeName

    If tCountry = "" Or tISO_currency_tla = "" Or tCurrency_format = "" Or tCurrency_symbol = "" Then
      Select Case tName_two_letter
        Case "AU"
          If tCountry = "" Then tCountry = "Australia"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "Au dollar"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "AUd"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"
        Case "GB"
          If tCountry = "" Then tCountry = "United Kingdom of Great Britain and Northern Ireland"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "G.B. pound"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "GBp"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"
        Case "IE"
          If tCountry = "" Then tCountry = "Ireland"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "G.B. pound"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "GBp"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"
        Case "JP"
          If tCountry = "" Then tCountry = "Japan"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "Japanese yen"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "Jpy"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0"
        Case "NZ"
          If tCountry = "" Then tCountry = "New Zealand"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "N.Z. dollar"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "NZd"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"
        Case "US"
          If tCountry = "" Then tCountry = "United States"
          If tEnglishName = "" Then tEnglishName = tCountry
          tCurrencyEnglishName = "U.S. dollar"
          tCurrencyNativeName = tCurrencyEnglishName
          tISO_currency_tla = "USd"
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"
        Case Else
          Stop ' Permanent.
          ' tLocals.Error_count = tLocals.Error_count + 1
          ' tLocals.Status = udtuStandard_texts.Failed
          ' tLocals.Success = udtBoolean_extended.False_
          ' GoTo Error_process
          tLocals.Action = "Set default values"
          If tCountry = "" Then tCountry = System.Globalization.RegionInfo.CurrentRegion.DisplayName
          If tEnglishName = "" Then tEnglishName = tCountry
          If tCurrency_format = "" Then tCurrency_format = "##,###,###,##0.00"

      End Select
    End If

    If tCountry = "" Or tISO_currency_tla = "" Or tCurrency_symbol = "" Then
      Stop ' Permanent.

      If uMemory_ProgFiles.Database = "" Or uMemory_ProgFiles.Database = "Off" Then
      Else
        Call Country_detect_from_uMemory_ProgFiles(tCountry, tName_two_letter, tISO_currency_tla, tCurrency_format, tCurrency_symbol, tTelephone_code, tNote)
        uOp.Subr = tLocals.Subr
      End If

      If tCountry = "" Or tISO_currency_tla = "" Or tCurrency_symbol = "" Then
        Call Country_settings_scrape()
        uOp.Subr = tLocals.Subr
      End If

      If tCountry = "" Or tISO_currency_tla = "" Or tCurrency_symbol = "" Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If

    End If

    If tCountry = "" Or tEnglishName = "" Or tISO_currency_tla = "" Or tCurrency_symbol = "" Or tCurrency_format = "" Or tCurrency_symbol = "" Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    If tSet_active Then
      If uActive.Country <> tEnglishName Then
        uActive.Country = tEnglishName
        uUser.Preferences_changed = True
      End If
      If uActive.Currency_tla <> tISO_currency_tla Then
        uActive.Currency_tla = tISO_currency_tla
        uUser.Preferences_changed = True
      End If
      If uActive.Currency_format <> tCurrency_format Then
        uActive.Currency_format = tCurrency_format
        uUser.Preferences_changed = True
      End If
      If uActive.Currency_symbol <> tCurrency_symbol Then
        uActive.Currency_symbol = tCurrency_symbol
        uUser.Preferences_changed = True
      End If
    End If

Error_process:
    If uOp.Event_ = "" Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub First_run()
    ' Revised 2021.04.17.
    ' Called by Installed_check.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim tCheck_it As Boolean = False
    Dim tCalled_from As String

    If uApp.First_run_completed_date <> udtuStandard_texts.Date_null Then
      uOp.Success = True
      GoTo Ender_partial
    End If
    tLocals.Subr = udtApplication_subroutine_names.First_run
    tLocals.Status = udtuStandard_texts.Initiation
    tLocals.Steps_to_complete = 4
    tCalled_from = uOp.Called_from
    uOp.Subr = tLocals.Subr
    uOp.Called_from = tLocals.Subr

    If Dir(uApp.Appdata_path, FileAttribute.Directory) <> "" Then
      Call Log_write(tLocals, 0)
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

    On Error GoTo Error_captured
    Call frmMain.Statuslabel1_set(tLocals.Subr & " check.")
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check initiation." & vbCrLf)
    uApp.First_run_in_progress = True
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 5
    If Not IsNothing(uApp.Resources) Then uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + uApp.Resources.Length + 1
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    uApp.Email_udt.Smtp_working = True

    tLocals.Action = "Call " & udtApplication_subroutine_names.First_run_make_folders
    uOp.Called_from = tLocals.Subr
    Call First_run_make_folders(tLocals.Changed)
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    If uDialogue.Active Then GoTo Ender_full
    tLocals.Steps_completed = tLocals.Steps_completed + 1

    If Not File_properties_determine_fn(tFile) Then Stop
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set(tLocals.Subr & " check.")

    tLocals.Action = "Check Memory"
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check memory exists.")
    tCheck_it = False
    If uMemory.Database = "" Then
      tLocals.Action = "Call Memory instantiate to get the database name"
      Call Memory_instantiate()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If
    If Not System.IO.File.Exists(uMemory.Database) Then tCheck_it = True
    If uMemory_ProgFiles.Database <> "" Then
      If Not System.IO.File.Exists(uMemory_ProgFiles.Database) Then tCheck_it = True
    End If
    If tCheck_it Then
      Call frmMain.txtDiagnostic_result_append(vbCrLf)
      uOp.Called_from = uOp.Subr
      Call Memory_create(tLocals.Changed)
      uOp.Subr = tLocals.Subr
      If uMachine.Testing_machine And uUser.My_dox = "" And uApp.Startup_sequence < 10 Then
        tLocals.Action = "Memory_create tried to copy a uMemory database from OneDrive, but failed because I don't know where OneDrive is yet."
        tLocals.Continue_ = True
      Else
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        tLocals.Steps_completed = tLocals.Steps_completed + 1
        Call frmMain.Statuslabel1_set()
        Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)
      End If
    Else
      tLocals.Steps_completed = tLocals.Steps_completed + 1
    End If ' If tCheck_it Then
    uOp.Success = False
    Call frmMain.Statuslabel1_set()

    tLocals.Action = "Check resources"
    uOp.Called_from = tLocals.Subr
    Call First_run_check_resources(tLocals.Changed)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    If uDialogue.Active Then GoTo Ender_full
    tLocals.Steps_completed = tLocals.Steps_completed + 1
    uOp.Success = False

    If uMachine.Development_machine Then
      tLocals.Steps_to_complete = tLocals.Steps_to_complete + 1
      tFile.File_name = udtLivingInfoSystems.LIS_memory_common
      tLocals.Action = "Check " & tFile.File_name
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check memory: " & tFile.File_name & ".")
      tFile.File_full_name = uLIS.Resources_path & "\" & tFile.File_name
      If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
        Stop ' Indefinite. To do, to test uApp.Progressbar_steps_total.
        'tFile.Copy_src = "ftp://ftp.livinginfosystems.com/download/resources/" &       tFile.File_name 
        'tFile.Copy_dest = ulis.Prog_file_path & "\" &       tFile.File_name 
        'Call Ftp_download(tFile.Copy_src, tFile.Copy_dest)
        tLocals.Changed = True
      End If
      tLocals.Steps_completed = tLocals.Steps_completed + 1
      Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)
    End If

    tLocals.Action = "Call First run local"
    Call First_run_local(tLocals.Changed)
    uOp.Subr = tLocals.Subr
    If Not uOp.Success Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    tLocals.Steps_completed = tLocals.Steps_completed + 1
    uOp.Success = False

    If tLocals.Changed Then
      tLocals.Action = "Re-check all."
      uOp.Success = False
      Call frmMain.txtDiagnostic_result_append(tLocals.Action & vbCrLf)
      Call First_run()
    End If

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    uCrypt.User_name = ""
    uCrypt.Password = ""
    If uApp.Initialising And uUser.Speech = 0 Then uUser.Speech = 1

    tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
      If tLocals.Steps_completed = tLocals.Steps_to_complete Then
        uApp.First_run_completed_date = Date.Today
        uUser.Preferences_changed = True
      End If
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.ToolStripProgressBar1_set()
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf & vbCrLf)

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub First_run_make_folders(ByRef tChanged As Boolean)
    ' Revised 2021.04.08.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim tApplication_data_created As Boolean
    Dim cProfile_path As String = "Check Profile_path"
    Dim cAppdata_path As String = "Check Appdata_path"
    Dim cProg_file_path As String = "Check Prog_file_path"
    Dim cResources As String = "Check Resources"
    Dim cShort_name As String = "Check Short_name"
    tLocals.Subr = udtApplication_subroutine_names.First_run_make_folders

    If uOp.Called_from <> udtApplication_subroutine_names.First_run Then
      If uApp.First_run_completed_date = udtuStandard_texts.Date_null Then
        Call First_run()
        If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          ' GoTo Error_process
        Else
          tLocals.Success = udtBoolean_extended.True_
          GoTo Ender_partial
        End If
        uOp.Success = False
      End If
    End If

    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set(tLocals.Subr & ": check folder structure...")
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check folder structure..." & vbCrLf)
    On Error GoTo Error_captured
    uOp.Called_from = tLocals.Subr
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1

Profile_path:
    tLocals.Action = cProfile_path
    tFile.File_path = uLIS.Profile_path
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check Profile path, " & tFile.File_path & ".")
    If Dir(tFile.File_path, vbDirectory) = "" Then
      tLocals.Action = "Create Profile_path"
      MkDir(tFile.File_path)
      If uError.Number = 0 Then
        tChanged = True
        tApplication_data_created = True
        tLocals.Success = udtBoolean_extended.True_
      End If
      uError.Extra = "Path created: " & tFile.File_path & "."
      tLocals.Success = udtBoolean_extended.False_
    End If
    Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)

Appdata_path:
    tLocals.Action = cAppdata_path
    tFile.File_path = uApp.Appdata_path
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check Appdata path, " & tFile.File_path & ".")
    If Dir(tFile.File_path, vbDirectory) = "" Then
      tLocals.Action = "Create Appdata_path"
      MkDir(tFile.File_path)
      tChanged = True
      If uOp.Subr = "" Then GoTo Ender_full
      tLocals.Success = udtBoolean_extended.True_
      uError.Extra = "Path created: " & tFile.File_path & "."
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
      uOp.Success = False
    End If
    Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)

Prog_file_path:
    tLocals.Action = cProg_file_path
    tFile.File_path = uLIS.Prog_file_path
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check Prog_file path, " & tFile.File_path & ".")
    If Dir(tFile.File_path, vbDirectory) = "" Then
      tLocals.Action = "Create Prog_file_path"
      MkDir(uLIS.Prog_file_path)
      tChanged = True
      If uOp.Subr = "" Then GoTo Ender_full
      If uError.Number = 0 Then
        tLocals.Success = udtBoolean_extended.True_
        uError.Extra = "Path created: " & tFile.File_path & "."
        Call Log_write(tLocals, 0)
        uOp.Subr = tLocals.Subr
      Else
        If tLocals.Action = "Fail" Then
          tLocals.Msg = uApp.App_name & " cannot be installed for all users due to a permissions error."
          tLocals.Msg = tLocals.Msg & "If you want " & uApp.App_name & " to be installed for all users, please contact your systems administrator, to elevate your permissions."
          tLocals.Msg = tLocals.Msg & "The " & uApp.App_name & " Installer will continue, and install " & uApp.App_name & " for you only."
          uDialogue.Dialogue_result = Msgbox_LIS(tLocals.Msg, udtAlert_types.Information, udtDialogue_types.Okay_only, frmMain.tabHome, udtDialogue_position.Top_left)
          If uDialogue.Dialogue_result = DialogResult.OK Then
            uLIS.Prog_file_path = uLIS.Profile_path
            tLocals.Action = ""
            Call First_run_make_folders(tChanged)
            uOp.Subr = tLocals.Subr
            If Not uOp.Success Then
              Stop ' Permanent.
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Success = udtBoolean_extended.False_
              GoTo Error_process
            End If
            uOp.Success = False
          Else
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Success = udtBoolean_extended.False_
            GoTo Error_process
          End If
        End If
      End If
    End If
    Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)

Resources:
    tLocals.Action = cResources
    tFile.File_path = uLIS.Resources_path
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check Resources, " & tFile.File_path & ".")
    If Dir(tFile.File_path, vbDirectory) = "" Then
      tLocals.Action = "Create Resources"
      MkDir(tFile.File_path)
      tChanged = True
      If uOp.Subr = "" Then GoTo Ender_full
      tLocals.Success = udtBoolean_extended.True_
      uError.Extra = "Path created: " & tFile.File_path & "."
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
      uOp.Success = False
    End If
    Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)

Short_name:
    tLocals.Action = cShort_name
    tFile.File_path = uLIS.Prog_file_path & "\" & uApp.Short_name
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check Short_name, " & tFile.File_path & ".")
    If Dir(tFile.File_path, vbDirectory) = "" Then
      tLocals.Action = "Create Short_name"
      MkDir(tFile.File_path)
      tChanged = True
      If uOp.Subr = "" Then GoTo Ender_full
      uOp.Success = True
      uError.Extra = "Path created: " & tFile.File_path & "."
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
      uOp.Success = False
    End If
    Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)

Error_process:
    If tLocals.Error_count <> 0 Or uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 5, 75
            '  5: Access to the path 'C:\Program Files (x86)\LivingInfo Systems' is denied.
            ' 75: Path, file access error.
            Select Case tLocals.Action
              Case cProfile_path
                Err.Clear()
                Call Error_clear(tLocals)
                tLocals.Success = udtBoolean_extended.False_
                Resume Appdata_path
              Case cAppdata_path
                Call Error_clear(tLocals)
                tLocals.Success = udtBoolean_extended.False_
                Resume Prog_file_path
              Case cProg_file_path, cResources, cShort_name
                uError.Number = 0
                Call frmMain.Statuslabel1_set(uError.Description, Color.Red)
                uOp.Subr = tLocals.Subr
                uLIS.Prog_file_path = uLIS.Profile_path
                Call Error_clear(tLocals)
                Call First_run_make_folders(tChanged)
                tLocals.Error_count = 0
              Case Else
                Stop ' Permanent.
                Call Error_display(tLocals)
            End Select
            Resume Next
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Error_count = 0 And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " completion." & vbCrLf)

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub First_run_check_resources(ByRef tChanged As Boolean)
    ' Separated 2020.05.20. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim i As String = 0
    tLocals.Subr = udtApplication_subroutine_names.First_run_check_resources
    tLocals.Continue_ = True

    If uOp.Called_from <> udtApplication_subroutine_names.First_run Then
      If uApp.First_run_completed_date = udtuStandard_texts.Date_null Then
        Stop ' Permanent.
        Call First_run()
        If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process
        Else
          tLocals.Success = udtBoolean_extended.True_
          GoTo Ender_partial
        End If
        uOp.Success = False
      End If
    End If

    uOp.Subr = tLocals.Subr

    tLocals.Action = "Check resources"
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & "." & vbCrLf)
    For Each tFile.File_name In uApp.Resources
      tFile.Copy_source = ""
      tFile.Copy_destination = ""
      tFile.File_full_name = ""

      If tFile.File_name <> "" Then
        tLocals.Action = "Check " & tFile.File_name
        Call frmMain.txtDiagnostic_result_append(tLocals.Subr & ": check resource " & i & " of " & uApp.Resources.GetLength(0) & ": " & tFile.File_name & ".")
        tFile.File_full_name = uLIS.Resources_path & "\" & tFile.File_name
        If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
          Dim tUse_ftp As Boolean = False
          Call frmMain.txtDiagnostic_result_append(" not found. Attempting to replace it." & vbCrLf)
          tLocals.Success = udtBoolean_extended.Null
          If uMachine.Development_machine Or uMachine.Testing_machine Then
            tFile.Copy_source = uLIS.Development_path & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
            If uMachine.Testing_machine Then
              tFile.Copy_source = uUser.My_dox & "\Dept 11A-2, Application development\0 Programme files shared\Resources\" & tFile.File_name
            End If
            If tFile.Copy_source = "" Or Dir(tFile.Copy_source, FileAttribute.Normal) = "" Then
              tUse_ftp = True
            Else
              tFile.Copy_destination = uLIS.Resources_path & "\" & tFile.File_name
              If Dir(tFile.Copy_destination, FileAttribute.Normal) = "" Then
                Try
                  FileCopy(tFile.Copy_source, tFile.Copy_destination)
                  tChanged = True
                  tLocals.Success = udtBoolean_extended.True_
                Catch ex As Exception
                  If uLIS.Resources_path = uUser.Resources_path Then
                    tLocals.Error_description = ex.Message
                    tLocals.Error_count = tLocals.Error_count + 1
                    tLocals.Success = udtBoolean_extended.False_
                    If Not tLocals.Continue_ Then GoTo Error_process
                  Else
                    If InStr(ex.Message, "Access to the path") > 0 And InStr(ex.Message, "is denied") > 0 Then
                      uLIS.Resources_path = uUser.Resources_path
                      If Dir(uLIS.Resources_path, FileAttribute.Directory) = "" Then MkDir(uLIS.Resources_path)
                      Call First_run_check_resources(tChanged)
                      Exit For
                    End If
                  End If
                End Try
              End If
            End If
          Else
            tUse_ftp = True
          End If
          ' If tUse_ftp Then
          '  Stop ' To do, review.
          '  tFile.Copy_source = tFile.File_name
          '  tFile.Copy_destination = tFile.File_full_name
          '  Call Ftp_download_prep(tFile)
          '  If Not uOp.Success Then

          '    uLIS.Resources_path = uUser.Resources_path
          '    tFile.Copy_destination = uLIS.Resources_path & "\" & tFile.File_name

          '    Call Ftp_download_prep(tFile)
          '    If uOp.Event_ = "" Or Not uOp.Success Then
          'Stop ' Permanent.
          'tLocals.Error_count = tLocals.Error_count + 1
          'tLocals.Success = udtBoolean_extended.False_
          'uError.Message = uApp.App_name & " is not properly installed; a file is missing:" & vbCrLf
          'uError.Message = uError.Message & tFile.File_name & vbCrLf
          'uError.Message = uError.Message & "Please re-install " & uApp.App_name & "."
          'Call Msgbox_LIS(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, frmMain.tabHome, udtDialogue_position.Top_left)
          'uOp.Subr = tLocals.Subr
          'uApp.Terminating = uApp.Terminating + 1
          'If Not tLocals.Continue_ Then GoTo Error_process
          '    End If

          '  End If

          '  uOp.Success = False
          'End If ' If tUse_ftp Then

          ' If tLocals.Success = udtBoolean_extended.False_ Then
          tLocals.Msg = "A required file is missing, " & Replace(tLocals.Action, "Check ", "") & "." & vbCrLf
          tLocals.Msg = tLocals.Msg & "Please re-install " & uApp.App_name & "."
          Call Msgbox_LIS_top(tLocals.Msg, udtAlert_types.Critical, udtDialogue_types.Okay_only, tLocals.Subr)
          Exit For
          ' End If

        Else
          Call frmMain.txtDiagnostic_result_append(" Passed." & vbCrLf)
        End If ' If Dir(tFile.File_full_name, FileAttribute.Normal) = "" Then
      End If
      Call frmMain.ToolStripProgressBar1_set()
      i = i + 1
    Next
    tFile.Copy_source = ""
    tFile.Copy_destination = ""
    tFile.File_name = ""
    tFile.File_full_name = ""
    Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " completion." & vbCrLf)

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

Ender_full:
    tLocals.Action = "Ender full"
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If

    Call Log_write(tLocals, 0)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
    End If
    uOp.Success = False

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Installed_check()
    ' Separated 2014.06.25. Revised 2021.04.16.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile = Nothing

    Dim tInstalled As Boolean
    Dim tFiles_checked As Short ' To do, replace with tLocals.Steps_completed.
    Dim tFiles_to_check As Short

    tLocals.Subr = udtApplication_subroutine_names.Installed_check
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    tInstalled = True ' Until found false.
    tFile.File_full_name = ""
    tFiles_checked = 0
    tFiles_to_check = 0
    tLocals.Status = ""

    tLocals.Action = udtuStandard_texts.Main_start
    tFiles_to_check = tFiles_to_check + 1
    tLocals.Action = "Check " & tFiles_checked + 1 & ", " & uApp.Log_loc
    If Dir(uApp.Log_loc) = "" Then tInstalled = False : Stop ' Permanent.
    tFiles_checked = tFiles_checked + 1

    If tInstalled Then
      tFiles_to_check = tFiles_to_check + 1
      tLocals.Action = "Check " & tFiles_checked + 1 & ", " & uMemory.Database
      If Dir(uMemory.Database) = "" Then tInstalled = False : Stop ' Permanent.
      tFiles_checked = tFiles_checked + 1

      If tInstalled Then
        tLocals.Action = "Check resources"
        tFiles_to_check = tFiles_to_check + UBound(uApp.Resources)
        For Each tFile.File_name In uApp.Resources
          If tFile.File_name <> "" Then
            tFile.File_full_name = uLIS.Resources_path & "\" & tFile.File_name
            tLocals.Action = "Check " & tFiles_checked + 1 & ", " & tFile.File_name
            If Dir(tFile.File_full_name) = "" Then
              Stop ' Permanent.

              tFile.File_full_name = uUser.Resources_path & "\" & tFile.File_name
              If Dir(tFile.File_full_name) = "" Then
                Stop ' Permanent.
                tInstalled = False
                Exit For
              End If

              ' tInstalled = False
              ' Exit For
            End If
            tFiles_checked = tFiles_checked + 1
          End If
          tFile.File_full_name = ""
          tFile.File_name = ""
        Next

        If tInstalled Then
          If uMachine.Development_machine Then
            tFiles_to_check = tFiles_to_check + 1
            tFile.File_full_name = uLIS.Resources_path & "\" & udtLivingInfoSystems.LIS_memory_common
            tLocals.Action = "Check " & tFiles_checked + 1 & ", " & tFile.File_full_name
            If Dir(tFile.File_full_name) = "" Then tInstalled = False : Stop ' Permanent.
            tFiles_checked = tFiles_checked + 1
          End If
        End If

        tFiles_to_check = tFiles_to_check + 1
        tLocals.Action = "Check " & tFiles_checked + 1 & ", " & uMachine.Environment_dump
        If Dir(uMachine.Environment_dump) = "" Then Call Environment_dump(False)
        uOp.Subr = tLocals.Subr
        ' No error check, because emailing the file can fail.
        tFiles_checked = tFiles_checked + 1

      End If
    End If

    If tInstalled Or uOp.Called_from = uOp.Subr Then
      '
    Else
      ' Added 2016.11.17. Revised 2021.02.21.
      tLocals.Action = "Call " & udtApplication_subroutine_names.First_run
      tFile.File_name = ""
      tFile.File_full_name = ""
      Call First_run()
      uOp.Subr = tLocals.Subr
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
      If uDialogue.Active Then GoTo Ender_full

      tLocals.Action = "Call " & udtApplication_subroutine_names.Installed_check
      uOp.Called_from = tLocals.Subr
      Call Installed_check()
      uOp.Subr = tLocals.Subr
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

    tFile.File_full_name = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) & "\*" & udtLivingInfoSystems.Name_short & " appdata.zip"
    tLocals.Action = "Attempt to delete " & tFile.File_full_name
    Dim tFile_name As String = Dir(tFile.File_full_name, FileAttribute.Normal)
    If tFile_name <> "" Then
      tFile.File_full_name = Replace(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) & "\*" & udtLivingInfoSystems.Name_short & " appdata.zip", "*", "")
      Call File_properties_determine_fn(tFile)
      tFile.File_full_name = tFile.File_path & "\" & tFile_name
      Try
        Kill(tFile.File_full_name)
      Catch ex As Exception
        tLocals.Error_description = ex.Message
      End Try
    End If

    tLocals.Action = "Check the number of files checked: " & tFiles_checked
    If tInstalled And tFiles_checked = tFiles_to_check Then
      tLocals.Success = udtBoolean_extended.True_
    Else
      Stop ' Indefinite.
      Call Installed_check()
      uOp.Subr = tLocals.Subr
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      tLocals.Success = udtBoolean_extended.True_
    End If

    Call frmMain.Statuslabel1_set()

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

    If uOp.Called_from = uOp.Subr Then uOp.Called_from = ""
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Run_count_check()
    ' Written 2018.09.30 by MJBH. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Check_run_count
    uOp.Subr = tLocals.Subr
    If uUser.First_run_wizard Then GoTo Ender_full

    If uApp.Run_count.Count = 0 Then
      With uMemory
        .Table_name = uTables.Log
        .Cmd_str = "SELECT COUNT(Action) as Run_count" & vbCr
        .Cmd_str = .Cmd_str & "FROM " & .Table_name & vbCr
        ' .Cmd_str = .Cmd_str & "WHERE Action='" & uApp.App_name & udtuStandard_texts.Session_begin & "'"
        .Cmd_str = .Cmd_str & "WHERE Action LIKE '%" & udtuStandard_texts.Session_begin & "%'"
      End With
      Call Database_general(uMemory)
      uOp.Subr = tLocals.Subr
      uOp.Success = False
      tLocals.Action = udtuStandard_texts.Dataset_read
      If uMemory.Rows_affected = 0 Then
        Stop ' Should occur only on first run.
      Else
        uApp.Run_count.Count = uMemory.Dataset.Tables(uMemory.Table_name).Rows(0).Item("Run_count")
        If uApp.Run_count.Count = 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process
        End If
      End If
    End If

    If uUser.Start_home_tab <= udtStandard_off_on.Null And uApp.Run_count.Count >= uApp.Run_count.Iteration(1) Then
      uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1
      uDialogue.Called_from = uOp.Subr
      Call Msgbox_LIS_top(uApp.Run_count.Iteration_message(1), udtAlert_types.Information, udtDialogue_types.Yes_no, tLocals.Subr)
      uOp.Subr = tLocals.Subr
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process
      End If
      uOp.Success = False
    End If

    If uUser.Speech < 1 And uApp.Run_count.Count >= uApp.Run_count.Iteration(2) Then
      uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 1
      uDialogue.Called_from = uOp.Subr
      Call Msgbox_LIS_top(uApp.Run_count.Iteration_message(2), udtAlert_types.Information, udtDialogue_types.Yes_no, tLocals.Subr)
      uOp.Subr = tLocals.Subr
      If uDialogue.Active Then GoTo Ender_full
      If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process
      End If
      uOp.Success = False
    End If

    Call Run_count_check_local()
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process
    End If
    uOp.Success = False

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)

    uOp.Result = ""
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Run_count_check_after_dialogue_tab()
    ' Revised 2021.05.01.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Check_run_count_after_dialogue_tab

    Select Case uDialogue.Prompt
      Case uApp.Run_count.Iteration_message(1)
        If uDialogue.Dialogue_result = DialogResult.OK Then
          uUser.Start_home_tab = udtStandard_off_on.On_
        Else
          uUser.Start_home_tab = udtStandard_off_on.Off
        End If
        uUser.Preferences_changed = True

      Case uApp.Run_count.Iteration_message(2)
        If uDialogue.Dialogue_result = DialogResult.OK Then
          Call frmMain.SettingsToolStripMenuItem_Click(Nothing, Nothing)
        End If
        uUser.Preferences_changed = True

      Case Else
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        ' If Not tLocals.Continue_ Then GoTo Error_process

    End Select

    uOp.Subr = ""
    If tLocals.Success <> udtBoolean_extended.False_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Speak_it(ByVal tCalled_from As String, ByVal tText As String, Optional ByVal tSpeech_level As Integer = 0)
    ' Revised 2021.03.14.
    ' Speech is asynchronous, so don't set any global variables here, until after speech (mObjSynth.Speak( tText )) has finished, and code execution has returned here.
    ' Set tSpeech_level <= uUser.Speech to have a spoken response.
    ' 0 = No speech (silent).
    ' 1 = Some speech (depends on the calling sub.).
    ' 2 = All available speech (verbose).
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Speak_it
    If tText = "" Then
      GoTo Ender_partial
    End If
    If uUser.Speech = 0 Or tSpeech_level = 0 Or tSpeech_level < uUser.Speech Then
      GoTo Ender_partial
    End If
    On Error GoTo Error_captured
    tLocals.Action = tLocals.Subr

Main:
    tLocals.Action = "Set statusbar"
    If Right$(tText, 1) <> "." Then tText = tText & "."
    frmMain.ToolStripStatusLabel1.Text = "Speaking, '" & tText & "'"
    tLocals.Action = ""

    tLocals.Action = "Speak"
    If uApp.Debug Or uOp.Subr = udtApplication_subroutine_names.End_game Then
      mObjSynth.Speak(tText)
    Else
      mObjSynth.SpeakAsync(tText)
    End If

    If tCalled_from <> "" Then
      tCalled_from = Replace(tCalled_from, "'", "")
      uOp.Subr = tCalled_from ' Temporarily.
    End If
    uError.Extra = tText
    If uUser.Speech <> 0 Then tLocals.Success = udtBoolean_extended.True_
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uError.Extra = ""

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            Resume Error_captured
          Case 5
            If uError.Description = "Audio device error encountered. - Error Code: 0x2" Then
              uUser.Speech = 0
            Else
              Stop ' Permanent.
              Call Error_display(tLocals)
            End If
          Case 53 ' Retrieving the COM class factory for component with CLSID {00024500-0000-0000-C000-000000000046} failed due to the following error: 8007007e The specified module could not be found. (Exception from HRESULT: 0x8007007E).
            ' Skip and try again later.
          Case 462 ' The RPC server is unavailable. (Exception from HRESULT: 0x800706BA).
            uUser.Speech = 0
            Resume Ender_partial
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_partial:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Speak_greeting()
    ' Written by Michael J. B. Hodgson, 2003.05.15. Revised 2021.05.05.
    Dim tLocals As udtLocals = Nothing
    Dim tHour_val As Short = 0
    Dim tHour_str As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Speak_greeting
    uOp.Subr = tLocals.Subr
    uOp.Called_from = uOp.Subr

    tHour_str = Left$(TimeOfDay, InStr(TimeOfDay, ":") - 1)
    tHour_str = Replace(tHour_str, "#", "")
    tHour_val = Val(tHour_str)
    If InStr(UCase(TimeOfDay), "PM") > 0 Or InStr(UCase(TimeOfDay), "P.M.") > 0 Then
      If tHour_val < 12 Then tHour_val = tHour_val + 12
    End If
    Select Case tHour_val
      Case Is < 12
        tLocals.Msg = "Good morning, "
      Case Is < 18
        tLocals.Msg = "Good afternoon, "
        If InStr(Now.ToString, "am") > 0 Or InStr(Now.ToString, "AM") > 0 Or _
           InStr(Now.ToString, "a.m.") > 0 Or InStr(Now.ToString, "A.M.") > 0 Then tLocals.Msg = "Good morning, "
      Case Else
        tLocals.Msg = "Good evening, "
    End Select

    tLocals.Msg = tLocals.Msg & uUser.First_name & ". "
    tLocals.Msg = tLocals.Msg & "Welcome "
    If Last_use_fn() Then tLocals.Msg = tLocals.Msg & "back " ' !
    uOp.Subr = tLocals.Subr
    tLocals.Msg = tLocals.Msg & "to " & uApp.English_name & "."
    Call Speak_it(tLocals.Subr, tLocals.Msg, 2)

    uOp.Subr = ""
  End Sub

  Function English_convert_fn(ByRef tString As String, Optional ByVal tSilent As Boolean = False, Optional ByVal tSource_is_scn As Boolean = False) As String
    ' Written by Michael J. B. Hodgson. Imported from Excel 2017.09.07. Revised 2021.03.17.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.English_convert_fn
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 1)
    tLocals.Status = ""

    tLocals.Action = udtuStandard_texts.Main_start
    If Not tSilent Then Call frmMain.Statuslabel1_set()

    tString = Replace(tString, "  ", " ") ' Two space to one.

    tString = Replace(tString, "&", "and")

    tString = Replace(tString, "acqua", "aqua")
    tString = Replace(tString, "anti ", "anti-")

    If tSource_is_scn Then tString = Replace(tString, "bklt", "booklet")

    If tSource_is_scn Then tString = Replace(tString, "c/s", "C.S.")
    If tSource_is_scn Then tString = Replace(tString, "cass ", "cassette ")
    If tSource_is_scn Then tString = Replace(tString, "cass)", "cassette)")
    tString = Replace(tString, "center", "centre")
    tString = Replace(tString, "check #", "cheque #")
    tString = Replace(tString, "color", "colour")
    tString = Replace(tString, "counseling", "counselling")

    tString = Replace(tString, "formulas", "formulae")

    If tSource_is_scn Then tString = Replace(tString, "HNDBK", "handbook")
    If tSource_is_scn Then tString = Replace(tString, "Hndbk", "handbook")
    tString = Replace(tString, "honor", "honour")

    If tSource_is_scn Then tString = Replace(tString, "i/c", "I.C.")

    tString = Replace(tString, " ive ", " I've ")

    tString = Replace(tString, "ize", "ise")

    tString = Replace(tString, "jewelry", "jewellery")

    If tSource_is_scn Then tString = Replace(tString, "LATAM", "LatAm")
    tString = Replace(tString, "ltd ", "Ltd. ")

    tString = Replace(tString, "mag ", "magazine ")
    tString = Replace(tString, "mens", "men's")
    tString = Replace(tString, "mold", "mould")

    tString = Replace(tString, "non ", "non-")

    tString = Replace(tString, "pal", "PAL")
    tString = Replace(tString, "pkg", "package")
    If tSource_is_scn Then tString = Replace(tString, "pts/sp", "PTS-SP")
    tString = Replace(tString, "pty ", "Pty. ")
    tString = Replace(tString, "p/l", "Pty. Ltd. ")

    If tSource_is_scn Then tString = Replace(tString, "r/d", "rundown")
    If tSource_is_scn Then tString = Replace(tString, "rons", "Ron's")

    If tSource_is_scn Then tString = Replace(tString, "Scientoloy", "Scientology")
    If tSource_is_scn Then tString = Replace(tString, "SCN", "Scientology")
    If tSource_is_scn Then tString = Replace(tString, "Scn", "Scientology")

    tString = Replace(tString, "theater", "theatre")
    If tSource_is_scn Then tString = Replace(tString, "TROUBLESHOOTER", "trouble-shooter")
    If tSource_is_scn Then tString = Replace(tString, "trs ", "TR's ")
    If tSource_is_scn Then tString = Replace(tString, " trs", " TR's")

    tString = Replace(tString, "vol ", "volume ")

    If tSource_is_scn Then tString = Replace(tString, "wth", "TWTH")

    If tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.True_
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)

    English_convert_fn = tString
    If Not tSilent Then Call frmMain.Statuslabel1_set("", True)
    uOp.Subr = ""
  End Function

  Function Title_case_fn(ByRef tString As String) As String
    ' Written 2018.04.08. Revised 2018.11.11.
    Title_case_fn = StrConv(tString, VbStrConv.ProperCase)
  End Function

  Sub Clipboard_copy_to(ByVal tText As String)
    If tText <> "" Then
      Try
        My.Computer.Clipboard.SetText(tText)
      Catch ex As Exception
        Stop ' Indefinite.
        uError.Description = ex.Message
      End Try
    End If
  End Sub

  Sub Sound_alert(Optional ByVal tWait_to_complete As Boolean = False)
    ' Written by Michael John Boyack Hodgson, 2013.08.18. Revised 2021.04.08.
    ' Flag:
    ' 0 (Default) Play the sound synchronously. Code execution pauses until sound is complete.
    ' 1           Play the sound asynchronously. Code execution does not wait for sound to complete.
    ' 2           If the specified sound is not found, do not play the default sound (no sound is played).
    ' 4           lpszSoundName is a memory file of the sound. Not used in VBA, VB6.
    ' 8           Continue playing sound in a loop until the next call to sndPlaySound.
    ' &H10        Do not stop playing the current sound before playing the specified sound.
    Dim tFull_name As String = uLIS.Resources_path & "\Alert.wav"
    If Dir(tFull_name, FileAttribute.Normal) = "" Then
      Stop ' Permanent.
      tFull_name = uUser.Resources_path & "\Alert.wav"
      ' uError.Number = uError.Number + 1
      uError.Source = System.Reflection.MethodBase.GetCurrentMethod.Name
      uError.Message = "Sound alert file not found."
      Call frmMain.Statuslabel2_set(uError.Message, Color.Red)
      If uApp.Startup_sequence > 20 Then
        Call First_run()
      End If
    End If

    If Dir(tFull_name, FileAttribute.Normal) <> "" Then
      Try
        If tWait_to_complete Or uDialogue.Called_from = udtApplication_subroutine_names.Version_check Then
          My.Computer.Audio.Play(tFull_name, AudioPlayMode.WaitToComplete)
        Else
          My.Computer.Audio.Play(tFull_name, AudioPlayMode.Background)
        End If
      Catch ex As Exception
        uError.Description = ex.Message
      End Try
    End If

  End Sub

  Sub Ready(ByVal tPlay_sound As Boolean, Optional ByVal tText As String = "")
    ' Written by Michael John Boyack Hodgson, 2013.12.09. Revised 2021.04.13.
    Dim tLocals As udtLocals = Nothing
    On Error GoTo Error_captured

    If tText = "" Then tText = uApp.App_name & ": ready."
    frmMain.ToolStripStatusLabel1.Text = tText
    If tPlay_sound Then
      Dim tFull_name As String = uLIS.Resources_path & "\Ready.wav"
      If Dir(tFull_name, FileAttribute.Normal) = "" Then
        Stop ' Permanent.
        tFull_name = uUser.Resources_path & "\Ready.wav"
        If Dir(tFull_name, FileAttribute.Normal) = "" Then
          Stop ' Permanent.
        End If
      End If
      If Dir(tFull_name, FileAttribute.Normal) <> "" Then My.Computer.Audio.Play(tFull_name, AudioPlayMode.Background)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If
    uOp.Subr = ""
  End Sub

  Public Sub Get_taskbar_location()
    ' Revised 2020.05.15.
    Dim tBounds As Rectangle = Screen.PrimaryScreen.Bounds
    Dim tWorking As Rectangle = Screen.PrimaryScreen.WorkingArea
    If tWorking.Height < tBounds.Height And tWorking.Y > 0 Then
      uUser.Taskbar_location = Taskbar_Location.Top
    ElseIf tWorking.Height < tBounds.Height And tWorking.Y = 0 Then
      uUser.Taskbar_location = Taskbar_Location.Bottom
    ElseIf tWorking.Width < tBounds.Width And tWorking.X > 0 Then
      uUser.Taskbar_location = Taskbar_Location.Left
    ElseIf tWorking.Width < tBounds.Width And tWorking.X = 0 Then
      uUser.Taskbar_location = Taskbar_Location.Right
    Else
      uUser.Taskbar_location = Nothing
    End If
    ' Return uUser.Taskbar_location
    uOp.Success = True
  End Sub

  Function Date_format_fn(ByVal tDate As Date, Optional tLong As Boolean = False) As String
    ' Written 2020.11.25. Revised 2021.02.01.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Format date"
    Date_format_fn = ""
    If uApp.Debug Then tLocals.Msg = System.Globalization.CultureInfo.CurrentUICulture.DateTimeFormat.ShortDatePattern ' dd/MM/yyyy.
    If uApp.Debug Then tLocals.Msg = System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern ' yyyy.MM.dd.

    Date_format_fn = Format(tDate, udtApplication.Default_date_format)
    If tLong Then Date_format_fn = Format(tDate, udtApplication.Default_date_long_format)

  End Function

#End Region ' End region, Kernel.

#Region "Dialogue"

  Public Sub Msgbox_LIS_top(ByVal tPrompt As String, ByVal tAlert As udtAlert_types, ByVal tType As udtDialogue_types, ByVal tCalled_from As String)
    ' Separated 2016.05.17. Revised 2021.05.02.
    ' Model:
    ' uOp.Result = MsgBox(tAlert, MsgBoxStyle.Information, tParent)
    Dim tLocals As udtLocals = Nothing
    If uDialogue.Active Or tPrompt = "" Or tAlert = Nothing Or tType = Nothing Or tCalled_from = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = tAlert
      .Called_from = tCalled_from
      .Dialogue_type = tType
      .Prompt = tPrompt
      Call frmMain.Top_dialogue_preparation()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      If Not .Active Then
        If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
        Else
          Stop ' Permanent.
        End If
        Call Dialogue_clear(False)
      End If
    End With

Error_process:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Public Function Msgbox_LIS(ByVal tPrompt As String, ByVal tAlert As udtAlert_types, ByVal tType As udtDialogue_types, ByVal tParent As Object, ByVal tPosition As udtDialogue_position) As DialogResult
    ' Separated 2019.11.29. Revised 2021.05.07.
    ' Model:
    ' uOp.Result = MsgBox(tAlert, MsgBoxStyle.Information, tParent)
    Dim tLocals As udtLocals = Nothing
    If uDialogue.Active Or tPrompt = "" Or tAlert = Nothing Or tType = Nothing Or tParent Is Nothing Or tPosition = Nothing Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    Msgbox_LIS = DialogResult.None
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = tAlert
      .Dialogue_type = tType
      .Prompt = tPrompt
      .Parent = tParent
      .Position = tPosition

      tLocals.Action = "Set top level"
      tLocals.Msg = frmDialogue.TopLevel
      frmDialogue.TopLevel = True ' Required.
      tLocals.Action = "Show dialogue"
      frmDialogue.ShowDialog() ' Modal.
      ' Can't check for success because other form elements can get focus in-between, and complete, so set uOp.Success to false.
      ' If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      If uOp.Event_ = "" Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      If frmDialogue.Visible Then
        .Active = True
      Else
        If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
        Else
          Msgbox_LIS = .Dialogue_result
        End If
        Call Dialogue_clear(False)
      End If

    End With

Error_process:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Sub Inputbox_LIS_top(ByVal tPrompt As String, ByVal tResponse_suggestion As String, ByVal tCalled_from As String)
    ' Separated 2016.05.17. Revised 2021.05.02.
    ' Model:
    ' uOp.Result = InputBox(tPrompt, Suggestion, tParent)
    Dim tLocals As udtLocals = Nothing

    If uDialogue.Active Or tPrompt = "" Or tCalled_from = Nothing Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = udtAlert_types.Information
      .Called_from = tCalled_from
      .Dialogue_type = udtDialogue_types.Question
      .Prompt = tPrompt
      .Response_str = tResponse_suggestion

      Call frmMain.Top_dialogue_preparation()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      If Not .Active Then
        If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
        Else
          Stop ' Permanent.
        End If
        Call Dialogue_clear(False)
      End If

    End With

Error_process:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Function Inputbox_LIS(ByVal tPrompt As String, ByVal tResponse_suggestion As String, ByVal tParent As Object, tPosition As udtDialogue_position) As String
    ' Separated 2019.11.30. Revised 2021.03.28.
    ' Model:
    ' uOp.Result = InputBox(tPrompt, Suggestion, tParent)
    Dim tLocals As udtLocals = Nothing
    Inputbox_LIS = ""
    If uDialogue.Active Or tPrompt = "" Or tParent Is Nothing Or tPosition = Nothing Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = udtAlert_types.Information
      .Dialogue_type = udtDialogue_types.Question
      .Prompt = tPrompt
      .Response_str = tResponse_suggestion
      .Parent = tParent
      .Position = tPosition

      tLocals.Action = "Set top level"
      frmDialogue.TopLevel = True ' Required.
      tLocals.Action = "Show dialogue"
      frmDialogue.ShowDialog() ' Modal.
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False

      If frmDialogue.Visible Then
        .Active = True
      Else
        If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
        Else
          Inputbox_LIS = .Response_str
        End If
        Call Dialogue_clear(False)
      End If

    End With

Error_process:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Sub Input_chooser_LIS_top(ByVal tPrompt As String, ByVal tCalled_from As String)
    ' Separated 2016.05.17. Revised 2021.05.02.
    ' User chooses from the combobox.
    ' Model:
    ' uOp.Result = InputBox(tMessage, tParent)
    ' Requires cboDialogue_response to have been pre-populated.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.

    Stop ' To do, to develop:
    If uDialogue.Active Or tPrompt = "" Or tCalled_from = "" Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      uOp.Success = False
      Exit Sub
    End If
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = udtAlert_types.Information
      .Called_from = tCalled_from
      .Prompt = tPrompt
      .Dialogue_type = udtDialogue_types.Chooser
      If frmDialogue.cboDialogue_response.Items.Count = 0 Then
        Stop ' Indefinite.
        tLocals.Msg = "Please define a set of items to choose from."
        .Alert = udtAlert_types.Critical
        .Dialogue_type = udtDialogue_types.Okay_only
      End If
      Call frmMain.Top_dialogue_preparation()

      If Not .Active Then
        If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
          ' Input_chooser_top_LIS = .Response_str
        End If
        Call Dialogue_clear(False)
      End If
    End With

  End Sub

  Function Input_chooser_LIS(ByVal tPrompt As String, ByVal tParent As Object, tPosition As udtDialogue_position) As String
    ' Separated 2016.05.17. Revised 2021.03.28.
    ' User chooses from the combobox.
    ' Model:
    ' uOp.Result = InputBox(tMessage, tParent)
    ' Requires cboDialogue_response to have been pre-populated.
    Dim tLocals As udtLocals = Nothing
    Input_chooser_LIS = ""
    If uDialogue.Active Or tPrompt = "" Or tParent Is Nothing Or tPosition = Nothing Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    Call Dialogue_clear(True)

    With uDialogue
      .Alert = udtAlert_types.Information
      .Prompt = tPrompt
      .Dialogue_type = udtDialogue_types.Chooser
      If uApp.Debug Then tLocals.Msg = frmDialogue.cboDialogue_response.Items.Count
      If frmDialogue.cboDialogue_response.Items.Count = 0 Then
        Stop ' Indefinite.
        .Prompt = "Please define a set of items to choose from."
        .Alert = udtAlert_types.Critical
        .Dialogue_type = udtDialogue_types.Okay_only
      End If
      .Parent = tParent
      .Position = tPosition

      tLocals.Action = "Set top level"
      frmDialogue.TopLevel = True
      tLocals.Action = "Show dialogue"
      frmDialogue.ShowDialog() ' Modal.
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_full
      End If

      If .Dialogue_result = DialogResult.None Or .Dialogue_result = DialogResult.Cancel Then
      Else
        Input_chooser_LIS = .Response_str
        Call Dialogue_clear(False)
      End If
    End With

Error_process:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_

Ender_full:

  End Function

  Sub Dialogue_clear(ByVal tBefore As Boolean)
    ' Written 2016.06.06. Revised 2020.10.25.
    uUser.Cancel = False
    With uDialogue
      If tBefore Then
        .Dialogue_result = 0
        .Response_num = 0
        .Response_str = Nothing
      Else
        .Active = False
        .Alert = udtAlert_types.Null
        .Called_from = Nothing
        ' .Combobox_last = udtDialogue_combobox_items.Null
        .Dialogue_type = Nothing
        .Parent = Nothing
        .Position = Nothing
        .Prompt = Nothing

        frmDialogue.cboDialogue_response.Items.Clear()
      End If

      ' Reset here, in case custom values were set:
      ' .txtPrompt_MouseHover = "This is asking you a question."
      ' .txtResponse_MouseHover = "This is a suggested response."
      ' .cboResponse_MouseHover = "Select from one of the available options."
      ' .Okay_button_MouseHover = "Click the Okay button when you are happy with your response."
      ' .Cancel_button_MouseHover = "Click the Cancel button to cancel this choice."
    End With
  End Sub

#End Region ' End region, Dialogue.

#Region "External"
  ' External to this software.

  Function XML_read_fn(ByVal tFile_name As String, ByVal tTag As String) As String
    ' Revised 2021.03.17.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile = Nothing
    XML_read_fn = ""
    If tFile_name = "" Then Exit Function
    tLocals.Subr = udtApplication_subroutine_names.XML_read_fn

    With tFile
      .File_full_name = tFile_name
      If Dir(.File_full_name, vbNormal) <> "" Then
        .Free_file = FreeFile()
        .Open_file = .File_full_name
        FileOpen(.Free_file, .File_full_name, OpenMode.Input, OpenAccess.Read)
        Do
          Input(.Free_file, .Text_read_line)
        Loop Until InStr(.Text_read_line, "<" & tTag & ">") <> 0
        If InStr(.Text_read_line, "</" & tTag & ">") = 0 Then
          Input(.Free_file, tLocals.Msg)
        Else
          tLocals.Msg = .Text_read_line
          tLocals.Msg = Replace(tLocals.Msg, "<" & tTag & ">", "")
          tLocals.Msg = Replace(tLocals.Msg, "</" & tTag & ">", "")
          tLocals.Msg = Trim(tLocals.Msg)
        End If
        FileClose(.Free_file)
        .Open_file = ""
        XML_read_fn = tLocals.Msg
      End If
    End With

    uOp.Success = True
    uOp.Subr = ""
  End Function

  Sub Environment_dump(ByRef tRegenerate As Boolean)
    ' Revised 2021.04.04.
    Dim tLocals As udtLocals = Nothing
    Dim tIndex As Short
    Dim tLine As String
    Dim tEnv_string As String
    Dim tFile As udtFile = Nothing
    If uApp.Environment_dump_completed_date = Date.Today Then
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_partial
    End If
    tLocals.Subr = udtApplication_subroutine_names.Environment_dump
    ' On Error GoTo Error_handler

    If uMachine.Name = uMachine.Development_machine_name Then
      If uOp.Called_from = "" Then
        Call frmMain.Statuslabel1_set(tLocals.Subr & ".")
      Else
        Call frmMain.Statuslabel1_set(uOp.Called_from & ": " & tLocals.Subr & ".")
      End If
    Else
      Call frmMain.Statuslabel1_set("Generating diagnostic information...")
    End If
    tIndex = 1

    With tFile
      If Dir(uMachine.Environment_dump, vbNormal) = "" Or tRegenerate Then
        tRegenerate = True
        .Open_file = uMachine.Environment_dump
        Try
          FileOpen(1, uMachine.Environment_dump, OpenMode.Output)
          PrintLine(1, "Environment Dump.")
          PrintLine(1, "Date: " & Date.Today & "; time: " & TimeOfDay & ".")
          PrintLine(1, "Machine name: " & uMachine.Name & ".")
          PrintLine(1, "Operating system: " & uMachine.Os_name & ".")
          PrintLine(1, "Version: " &
            System.Environment.OSVersion.Version.Major & "." &
            System.Environment.OSVersion.Version.Minor & "." &
            System.Environment.OSVersion.Version.Build & ".")
          PrintLine(1, "Environment dump generated by: " & uApp.App_name & ".")
          PrintLine(1, "Screen resolution:" & Str(Get_system_metrics_fn(0)) & " by" & Str(Get_system_metrics_fn(1)) & ".")
          PrintLine(1)
          Do
            tEnv_string = Environ(tIndex) ' Get environment variable.
            If tEnv_string <> "" Then
              If tIndex < 10 Then
                tLine = Str(tIndex) & ", " & tEnv_string
              Else
                tLine = Trim(Str(tIndex)) & ", " & tEnv_string
              End If
              PrintLine(1, tLine)
            End If
            tIndex = tIndex + 1
          Loop Until tEnv_string = ""
          PrintLine(1)
          PrintLine(1, "End of report.")
          PrintLine(1)
          FileClose(1)
          .Open_file = ""
          uApp.Environment_dump_completed_date = Date.Today
          uUser.Preferences_changed = True
        Catch ex As Exception
          Stop ' Permanent.
          tLocals.Error_description = ex.Message
          Call Error_handler_for_text_file(tLocals)
          If tLocals.Action = "Resume" Then
            tLocals.Action = ""
            Call Environment_dump(False)
          End If
        End Try
      Else
        If Dir(uMachine.Environment_dump, vbNormal) <> "" Then
          uApp.First_run_in_progress = False
        End If
      End If
    End With

Email_it:
    tLocals.Action = udtApplication_subroutine_names.Error_email
    tLocals.Msg = uApp.Environment_dump_completed_date
    tLocals.Msg = uApp.Environment_dump_emailed
    If tRegenerate Or (uOp.Called_from <> udtApplication_subroutine_names.First_run And Not uApp.Environment_dump_emailed) Then
      tLocals.Msg = tLocals.Subr & " from " & uApp.App_name & " in " & uMachine.Name & "."
      Call Error_write_mega(tLocals)
      Call Appdata_email()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

    If uUser.Preferences_changed Then
      Call Preferences_write()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed." & vbCrLf)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 53, 55, 75, 76
            Call Error_handler_for_text_file(tLocals)
            ' If tLocals.Action = "Resume" Then tLocals.Action = "" : Resume
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
      uUser.Preferences_changed = True
    End If
    Call Log_write(tLocals, 0)

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Private Sub Environment_dump_email()
    ' Separated 2018.04.25. Revised 2021.03.03.
    ' Replaced by Appdata_email.
    If uMachine.Development_machine And uApp.App_name <> udtApplication.App_LivingInfo_Systems_Server Then Exit Sub
    ' See www.codeguru.com/csharp/.net/net_general/the-process-class-and-.net.html?utm_medium=email&utm_campaign=B2B_NL_MTU_20200518_AR2&mkt_tok=eyJpIjoiTnpaaE9HUmpOamN6TTJOaiIsInQiOiJySXlqOWtzS3hXd0FHZjlRbGd5TzI2THdDUHJER0NmbTZFV2M5aUxKXC94ZzBDNCtLSjVWS2ZqRVFFUlJ2ZHVJUUZta1piQW51bVczbWVFb1dSc2JiRkZwOHVhTTY5Q0h6MmFhc2dMWUROb3haQlwvdkYrOVRQVU1IWjJoVjJTa1BxIn0%3D.
    Dim tLocals As udtLocals = Nothing : Stop ' Retest using tLocals.
    Dim tOp As udtOperation
    tLocals.Subr = udtApplication_subroutine_names.Environment_dump_email
    tOp.Subr = tLocals.Subr
    tOp.Called_from = uOp.Subr

    With uApp.Email_udt
      .Recipient_address = uLIS.Support_email
      .Subject = uApp.App_name & " in " & uMachine.Name & "."
      .Body = ""
      .Attachment_loc = uMachine.Environment_dump
    End With
    ' If Email_smtp_send_fn(uApp.Email_udt, False) Then
    If Email_smtp_send_fn() Then
      tOp.Subr = tLocals.Subr
      uApp.Environment_dump_emailed = True
      uUser.Preferences_changed = True
      Call Preferences_write_one("Environment_dump_emailed", uApp.Environment_dump_emailed)
      tOp.Subr = tLocals.Subr
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_handler
    End If

    '     Call Log_write( tLocals ,  0) 
  End Sub

  Sub Log_write(ByVal tLocals_external As udtLocals, ByVal tSpeech_level As Integer, Optional ByVal tError_dump As String = "")
    ' Revised 2021.05.05.
    Dim tLocals As udtLocals = Nothing ' What I'm doing, as usual.
    Dim tFile As udtFile = Nothing
    Dim tCalled_from As String = ""
    Dim tError As udtError

    tLocals.Subr = udtApplication_subroutine_names.Log_write
    tLocals.Status = udtuStandard_texts.Initiation
    On Error GoTo Error_captured

    tLocals.Action = "Check for serious error state"
    If uOp.Event_ = "" Or tLocals.Error_count > 0 Or uApp.Terminating > 1 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If uError.Number <> 0 Then GoTo Error_captured
      If uApp.Debug Then GoTo Ender_full
    End If
    If uError.Last_error_sub = tLocals.Subr Then
      If uApp.Terminating = 0 Then uApp.Terminating = 1
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full ' To prevent infinite looping.
    End If
    If uApp.Terminating >= 3 Then GoTo Ender_full

    tError.Description = uError.Description
    tError.Extra = uError.Extra
    tError.Message = uError.Message
    tError.Number = uError.Number
    tError.Source = uError.Source

    tLocals.Action = "Update uApp.Log_loc, in case the year turns over during a session"
    Call uApp.Initialise()
    If uApp.Log_loc = "" Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If

    tLocals.Status = ""

    tLocals.Action = "Check folder exists"
    If Dir(uApp.Appdata_path, FileAttribute.Directory) = "" Then
      If Not uApp.First_run_in_progress Then
        tLocals.Action = "Create folders"
        Call First_run_make_folders(tLocals.Continue_)
        If uOp.Event_ = "" Or tLocals.Error_count > 0 Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 1 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Ender_full
        End If
      End If
    End If

    tLocals.Action = "Check uOp.Subr"
    If tLocals_external.Subr = "" Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      tLocals.Msg = "Log_write was called, however uOp.Subr is empty."
      If uOp.Called_from <> udtApplication_subroutine_names.Error_write_mega Then
        Call Error_display(tLocals)
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_full
      End If
    End If

    tLocals.Action = "Condition data"
    If tLocals_external.Status <> "" Then
      If tLocals_external.Status = udtuStandard_texts.Initiation Then tLocals_external.Success = udtBoolean_extended.True_
      If tLocals_external.Success = udtBoolean_extended.True_ And (tLocals_external.Status = udtuStandard_texts.Cancelled Or tLocals_external.Status = udtuStandard_texts.Failed) Then
        Stop ' Permanent.
        If False Then GoTo Ender_full
      End If
    End If

    tLocals.Action = "Check tLocals_external.Action"
    If tLocals_external.Action = "" Then
      If tLocals_external.Status = udtuStandard_texts.Failed Then
        Stop ' Indefinite.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If uApp.Debug Then GoTo Ender_full
      End If
    Else
      If tLocals_external.Action = udtuStandard_texts.Initiation Or _
         tLocals_external.Action = udtuStandard_texts.Cancelled Or _
         tLocals_external.Action = udtuStandard_texts.Failed Or _
         tLocals_external.Action = udtuStandard_texts.Completion Then
        If tLocals_external.Action = udtuStandard_texts.Initiation And tLocals_external.Status = "" Then tLocals_external.Status = tLocals_external.Action
        If tLocals_external.Action = udtuStandard_texts.Cancelled And tLocals_external.Status = "" Then tLocals_external.Status = tLocals_external.Action
        If tLocals_external.Action = udtuStandard_texts.Completion And tLocals_external.Status = "" Then tLocals_external.Status = tLocals_external.Action
        If (tLocals_external.Action = udtuStandard_texts.Initiation And _
           tLocals_external.Status = udtuStandard_texts.Initiation) Or _
           (tLocals_external.Action = udtuStandard_texts.Cancelled And _
           tLocals_external.Status = udtuStandard_texts.Cancelled) Or _
           (tLocals_external.Action = udtuStandard_texts.Completion And _
           tLocals_external.Status = udtuStandard_texts.Completion) Then
          ' Expected, carry on.
        Else
          Stop ' Fix calling sub.
          tLocals_external.Status = tLocals_external.Action
          tLocals_external.Action = ""
          If uApp.Debug Then GoTo Ender_full
        End If
      End If
    End If
    If tLocals_external.Status = udtuStandard_texts.Initiation Then tLocals_external.Success = udtBoolean_extended.True_
    If tLocals_external.Action <> "" And Right$(tLocals_external.Action, 1) <> "." Then tLocals_external.Action = tLocals_external.Action & "."

    tLocals.Action = "Check tLocals_external.Success"
    If tLocals_external.Success <> udtBoolean_extended.True_ Then
      If tLocals_external.Subr = udtApplication_subroutine_names.Memory_create And tLocals_external.Action = udtApplication_subroutine_names.uMemory_copy_from_progfiles & "." Then
        ' Not a critical error, continue.
      Else
        Stop ' Permanent.
        If uApp.Debug Then GoTo Ender_partial ' Optional.
      End If
    End If

    tLocals.Action = "Check tLocals_external.Status"
    If tLocals_external.Status = "" Then
      If tLocals_external.Action = udtuStandard_texts.Initiation Or _
         tLocals_external.Action = udtuStandard_texts.Cancelled Or _
         tLocals_external.Action = udtuStandard_texts.Failed Or _
         tLocals_external.Action = udtuStandard_texts.Completion Then
        Stop ' Permanent, fix calling sub.
        tLocals_external.Status = tLocals_external.Action
        If uApp.Debug Then GoTo Ender_partial
      End If
    Else
      If Right$(tLocals_external.Status, 1) <> "." Then tLocals_external.Status = tLocals_external.Status & "."
    End If

    tLocals.Action = "Condition data, part 2"
    If tError.Message = tError.Description Then tError.Message = Nothing
    If tError.Description <> "" Then
      tError.Description = Replace(tError.Description, Chr(34), "'")
      If Right$(tError.Description, 1) <> "." Then tError.Description = tError.Description & "."
    End If
    If tError.Source <> "" Then
      tError.Source = Replace(tError.Source, Chr(34), "'")
      If Right$(tError.Source, 1) <> "." Then tError.Source = tError.Source & "."
    End If
    If tError.Message <> "" Then
      tError.Message = Replace(tError.Message, Chr(34), "'")
      If Right$(tError.Message, 1) <> "." Then tError.Message = tError.Message & "."
    End If
    If tError.Extra <> "" Then
      tError.Extra = Replace(tError.Extra, Chr(34), "'")
      If Right$(tError.Extra, 1) <> "." Then tError.Extra = tError.Extra & "."
    End If

    tLocals.Action = "Check for first run, or new year"
    If Dir(uApp.Log_loc) = "" Then
      If uOp.Called_from <> udtApplication_subroutine_names.Log_write Then
        uOp.Called_from = tLocals.Subr
        tLocals.Action = uApp.App_name & " text Memory creation"
        tLocals.Status = ""
        tLocals.Success = udtBoolean_extended.True_
        uError = Nothing
        Call Log_write(tLocals, 0)
      End If
    End If

    With tFile
      tLocals.Action = "Write log"
      .Free_file = FreeFile()
      .Open_file = "Log"
      FileOpen(.Free_file, uApp.Log_loc, OpenMode.Append)
      Print(.Free_file, Date.Today & " " & TimeString & ". ")
      Print(.Free_file, "Computer: " & uMachine.Name & ". ")
      Print(.Free_file, "Application: " & uApp.App_name & ". ")
      If uActive.Corporation_name <> "" Then
        Print(.Free_file, "Corporation: " & uActive.Corporation_name)
        If Right$(uActive.Corporation_name, 1) = "." Then
          Print(.Free_file, " ")
        Else
          Print(.Free_file, ". ")
        End If
      End If
      If uActive.Branch_name <> "" Then Print(.Free_file, "Branch: " & uActive.Branch_name & ". ")
      Print(.Free_file, "User: " & Environment.UserName & ". ")
      Print(.Free_file, "Event: " & uOp.Event_ & ". ")
      Print(.Free_file, "Sub-routine: " & tLocals_external.Subr & ". ")
      If tLocals_external.Status <> "" Then Print(.Free_file, "Status: " & tLocals_external.Status & " ")
      If tLocals_external.Action <> "" Then Print(.Free_file, "Action: " & tLocals_external.Action & " ")
      ' If tFile.Open_file <> "" Then Print(.Free_file, "Openf: " & tFile.Open_file & ". ")
      ' If tFile.File_full_name <> "" Then Print(.Free_file, "Active document: " & tFile.File_full_name & ". ")

      If tError.Message <> "" Then Print(.Free_file, "Error message: " & tError.Message & " ")
      If tError.Number <> 0 Then Print(.Free_file, "Error number: " & tError.Number & ". ")
      If tError.Description <> "" Then Print(.Free_file, "Error description: " & tError.Description & " ")
      If tError.Source <> "" Then Print(.Free_file, "Error source: " & tError.Source & " ")
      If tError.Extra <> "" Then Print(.Free_file, "Extra: " & tError.Extra & " ")

      If tLocals_external.Success = udtBoolean_extended.True_ Then uOp.Success = True
      Print(.Free_file, "Success: " & uOp.Success & ".")
      PrintLine(.Free_file)
      If tError_dump <> "" Then PrintLine(.Free_file, tError_dump)
      ' Call Log_write_local(tLocals, .Free_file)

      FileClose(.Free_file)
      .Open_file = ""
    End With

Continue_01:
    tLocals.Action = "Check call Log_write_memory"
    If uApp.Startup_sequence >= 18 Then
      tLocals.Action = "Call Log_write_memory"
      Call Log_write_memory(tLocals_external)
      If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 1 Then
        Stop ' Indefinite.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If

    tLocals.Action = "Check Speech"
    If tSpeech_level > 0 Then
      tLocals.Action = "Speak"
      If tLocals_external.Status = "" Then
        Call Speak_it(tLocals.Subr, tLocals_external.Subr & " " & tLocals_external.Action, tSpeech_level)
      Else
        Call Speak_it(tLocals.Subr, tLocals_external.Subr & " " & tLocals_external.Status, tSpeech_level)
      End If
    End If
    uOp.Called_from = tCalled_from

Error_process:
    If tLocals.Error_count <> 0 Then
      tLocals.Action = "Error process"
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Action = "Error_captured"
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            ' Can occur if another process was using a file.
            If uError.Description = "" Then Resume
          Case 53, 55, 75, 76
            Call Error_handler_for_text_file(tLocals)
            Call Error_clear(tLocals)
            If tLocals.Action = udtuStandard_texts.Retry Then tLocals.Action = "" : Resume
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

Ender_partial:
    If tCalled_from <> "" Then uOp.Called_from = tCalled_from
    Call uOp.Initialise()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Sql_write(ByVal tText As String)
    ' Separated 2017.12.31. Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile = Nothing
    tLocals.Subr = uOp.Subr ' Not the other way around.
    If uApp.Sql_last_loc Is Nothing Or uApp.Sql_last_loc = "" Then Exit Sub
    On Error GoTo Error_captured

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    With tFile
      .Free_file = FreeFile()
      .Open_file = uApp.Sql_last_loc
      FileOpen(.Free_file, uApp.Sql_last_loc, OpenMode.Output)
      Print(.Free_file, tText)
      FileClose(.Free_file)
      .Open_file = ""
    End With

Error_captured:
    If Err.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      Call Error_handler_for_text_file(tLocals)
      If tLocals.Action = "Resume" Then tLocals.Action = "" : Resume
    End If
  End Sub

  Sub Config_xml_create()
    ' Written by Michael John Boyack Hodgson, 2013.11.17. Imported and converted from Excel 2019.01.11. Revised 2021.04.04.
    ' Development_machine_name only.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile
    If uMachine.Name <> UCase(uMachine.Development_machine_name) Then Exit Sub

    With tFile
      .Free_file = FreeFile()
      tLocals.Subr = udtApplication_subroutine_names.Config_xml_create
      tLocals.Action = udtuStandard_texts.Main_start

      tLocals.Action = "Create it"
      .Open_file = uLIS.Config_loc
      FileOpen(.Free_file, uLIS.Config_loc, OpenMode.Output)
      Print(.Free_file, "<?xml version=" & Chr(34) & "1.0" & Chr(34) & " encoding=" & Chr(34) & "UTF-8" & Chr(34) & "?>")
      Print(.Free_file, "<LIS_config>")
      Print(.Free_file, " <Dev>" & uMachine.Development_machine_name & "</Dev>")
      Print(.Free_file, " <Version>" & uApp.App_revision_date & "</Version>")
      Print(.Free_file, " <Build>" & uApp.App_build_number & "</Build>")
      Print(.Free_file, "</LIS_config>")
      FileClose(.Free_file)
      .Open_file = ""
    End With
    uOp.Success = True

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
  End Sub

  Function File_properties_determine_fn(ByRef tFile As udtFile) As Boolean
    ' Revised 2021.04.14.
    Dim tLocals As udtLocals = Nothing
    Dim i As Integer = 0
    Dim tPos As Integer = 0
    Dim tFileInfo As System.IO.FileInfo = Nothing
    tLocals.Subr = udtApplication_subroutine_names.File_properties_determine_fn

    File_properties_determine_fn = False
    If tFile.File_full_name <> "" Then tFileInfo = New System.IO.FileInfo(tFile.File_full_name)

    With tFile
      .Copy_source = Nothing
      .Copy_destination = Nothing
      .File_attributes = Nothing
      .File_creation_time = udtuStandard_texts.Time_null
      .File_exists = False
      .File_extension = Nothing
      ' .File_full_name = Nothing
      .File_is_read_only = False
      .File_last_access_time = udtuStandard_texts.Time_null
      .File_last_write_time = udtuStandard_texts.Time_null
      .File_length = 0
      .File_name = Nothing
      .File_path = Nothing
      .File_short_name = Nothing
      .Open_file = Nothing
      .Path_delimiter = Nothing
      .Text_read_line = Nothing

      If tFile.File_full_name <> "" Then
        If InStr(tFile.File_full_name, ":\") = 0 Then
          .File_name = tFile.File_full_name
        End If

        tPos = 0
        If InStr(1, .File_full_name, "/") <> 0 Then
          ' Internet path:
          .Path_delimiter = "/"
        End If
        If InStr(1, .File_full_name, "\") <> 0 Then
          ' Local path:
          .Path_delimiter = "\"
        End If
        If .Path_delimiter = "" And .File_path <> "" Then
          If InStr(1, .File_path, "/") <> 0 Then
            ' Internet path:
            .Path_delimiter = "/"
          End If
          If InStr(1, .File_path, "\") <> 0 Then
            ' Local path:
            .Path_delimiter = "\"
          End If
          If .Path_delimiter = "" Then
            Stop
          Else
            ' Assume the same path as last: 
            .File_full_name = .File_path & "\" & .File_name
          End If
        End If

        If .Path_delimiter = "\" Then
          If tFileInfo IsNot Nothing Then
            .File_attributes = tFileInfo.Attributes
            .File_creation_time = tFileInfo.CreationTime
            .File_path = tFileInfo.DirectoryName
            .File_exists = tFileInfo.Exists
            .File_extension = tFileInfo.Extension
            .File_is_read_only = tFileInfo.IsReadOnly
            .File_last_access_time = tFileInfo.LastAccessTime
            .File_last_write_time = tFileInfo.LastWriteTime
            If .File_exists Then .File_length = tFileInfo.Length
            .File_name = tFileInfo.Name
          End If

          i = Len(.File_full_name)
          tPos = InStrRev(.File_full_name, .Path_delimiter)
          If .File_path = "" Then .File_path = Mid$(.File_full_name, 1, tPos - 1)
          If .File_name = "" Then .File_name = Mid$(.File_full_name, tPos + 1)

          i = Len(.File_name)
          tPos = 0
          tPos = InStrRev(.File_name, ".")
          If tPos <> 0 Then
            If .File_short_name = "" Then .File_short_name = Left$(.File_name, tPos - 1)
            If .File_extension = "" Then .File_extension = Mid$(.File_name, tPos + 1)
          End If
        Else
          Stop ' Indefinite.
          Exit Function
        End If ' If .Path_delimiter = "\" Then
      End If ' If tFull_name <> "" Then
    End With

    File_properties_determine_fn = True ' Success.
  End Function

  Sub Me_update()
    ' Written by Michael J. B. Hodgson, 2006.08.21. Revised 2021.03.28.
    ' 2021.04.08 not currently implemented.
    Dim tLocals As udtLocals = Nothing
    Dim tRevision As String = ""
    Dim tBuild As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Me_update
    On Error GoTo Error_captured
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    Call frmMain.Statuslabel1_set()
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    tLocals.Msg = "Connecting to " & udtLivingInfoSystems.Name_short & " server..."
    Call frmMain.Statuslabel1_set(tLocals.Msg)

Web:
    tLocals.Action = "Test web access"
    Call Me_web_update()
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process
    End If
    uOp.Success = False

Result:
    Select Case uError.Message
      Case ""
        tLocals.Success = udtBoolean_extended.True_
      Case Else
        Call frmMain.Statuslabel1_set(uError.Message)
    End Select

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       tLocals.Success = udtBoolean_extended.False_ Or tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed ' Reduce all error responses down to tLocals.Status = udtuStandard_texts.Failed.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 5
            ' A network-related or instance-specific error occurred while establishing a connection to SQL server. The server was not found or was not accessible.
            ' Verify that the instance name is correct and that SQL server is configured to allow remote connections.
            ' (provider: SQL Network Interfaces, error: 26 - error locating server instance specified)".
            Stop ' Indefinite.
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    Call Log_write(tLocals, 0)

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Function Registry_get_value_fn(ByVal tKey As String) As String
    ' A basic walk through of the function for those interested:
    ' Firstly, the function opens the registry key @ HKEY_CURRENT_USER\Sofware\Clients\StartmenuInternet
    ' The default value at this registry key is the executable name of your default browser (IEXPLORE9.EXE, FIREFOX.EXE, Chrome.exe... etc), these names aren't overly helpful as it's not the actual browser's "name". However, they do help out, so we will store this in the "baseName" variable for later use.
    ' The next step is to build our next registry location. 
    ' Due to the difference in registry keys on a 32 bit versus 36 bit system the ternary "IIf" operator was used here to apply the "Wow6432Node" if necessary (64 bit systems). 
    ' The StartMenuInternet key contains multiple sub-keys, one for each browser you have installed, coincidentally the sub-key names correspond to the value stored in baseName, so we use that variable to open the default browser's folder.
    ' This is the moneymaker, the default value at this key is the actual name of the browser, so we just assign retVal the default key value and call it quits.
    ' Return retVal <- self explanatory, return the value we found.
    ' (Not yet tested on a 32 bit system so if someone finds an error please point it out).

    Stop ' Temporary, until tested and cleaned up.
    ' Separated MJBH 2018.12.29.
    Dim tRet_val As String = String.Empty
    ' Using tBase_key As Microsoft.Win32.RegistryKey = My.Computer.Registry.CurrentUser.OpenSubKey("Software\Clients\StartmenuInternet")
    Using tBase_key As Microsoft.Win32.RegistryKey = My.Computer.Registry.CurrentUser.OpenSubKey(tKey)
      Dim tBase_name As String = tBase_key.GetValue("").ToString
      ' Dim tSub_key As String = "SOFTWARE\" & IIf(Environment.GetEnvironmentVariable("PROCESSOR_ARCHITECTURE") = "AMD64", "Wow6432Node\", "") & "Clients\StartMenuInternet\" & tBase_name
      Dim tSub_key As String = ""
      'If Environment.GetEnvironmentVariable("PROCESSOR_ARCHITECTURE") = "AMD64" Then
      '   tSub_key = "Wow6432Node\"
      'End If
      'tSub_key = "SOFTWARE\" & tSub_key & "Clients\StartMenuInternet\" & tBase_name
      Using tValue As Microsoft.Win32.RegistryKey = My.Computer.Registry.LocalMachine.OpenSubKey(tSub_key)
        tRet_val = tValue.GetValue("").ToString
      End Using
    End Using
    Return tRet_val
  End Function

  Function Compression_fn(ByVal tPath As String, Optional ByVal tCompress As Boolean = False) As String
    ' Written 2020.11.05. Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    Dim tDestination As String = ""
    Compression_fn = ""

    If tPath <> "" Then
      tDestination = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData)
      tDestination = tDestination & "\" & Date_format_fn(Date.Today)
      tDestination = tDestination & " " & udtLivingInfoSystems.Name_short & " appdata.zip"
      If Dir(tDestination) <> "" Then
        Try
          tLocals.Action = "Kill " & tDestination
          Kill(tDestination)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If
      If Dir(tDestination) = "" Then
        Try
          tLocals.Action = "Zip " & tDestination
          ZipFile.CreateFromDirectory(tPath, tDestination)
          Compression_fn = tDestination
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If
    End If ' If tPath <> "" Then

  End Function

#End Region ' End region, External.

#Region "Registration"
  ' Region revised 2021.04.22.

  Sub Registration()
    ' Separated 2014.05.04. Revised 2021.05.02.
    ' Checks the registration status of this application.
    Dim tLocals As udtLocals = Nothing
    If uMachine.Development_machine Then
      If Not uReg.Registered Then uUser.Preferences_changed = True
      uReg.Registered = True
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_partial
    End If
    tLocals.Subr = udtApplication_subroutine_names.Registration
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)

    tLocals.Action = udtuStandard_texts.Main_start
    If uReg.Registered Then
      Call frmMain.Enter_registration_code_tool_strip_menu_item_visibility(False)
      tLocals.Success = udtBoolean_extended.True_
    Else
      If uReg.Trial_30_day Or uReg.Trial_30_day_from_date <> udtuStandard_texts.Time_null Then
        If uApp.Debug Then tLocals.Msg = DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today)
        If DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) >= 20 And DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) < 30 Then
          If uReg.Trial_30_day_20_days_notified Then
          Else
            If Not uReg.Message_reported Then
              tLocals.Msg = uApp.App_name & " is not registered."
              Call Speak_it(uOp.Subr, tLocals.Msg, 2)
              uOp.Subr = tLocals.Subr
              tLocals.Msg = tLocals.Msg & vbCrLf
              tLocals.Msg = tLocals.Msg & "Your 30 day trial is nearing expiration. Please enter a registration code in order to continue to use " & uApp.Short_name & "." & vbCrLf
              tLocals.Msg = tLocals.Msg & "This can be entered here, or in the Help menu."
              If uDialogue.Prompt <> tLocals.Msg Then
                Call Inputbox_LIS_top(tLocals.Msg, "", uOp.Subr)
                uOp.Subr = tLocals.Subr
              End If
              tLocals.Success = udtBoolean_extended.True_
              If uDialogue.Active Then GoTo Ender_full
              If uDialogue.Response_str = "" Then
                uReg.Trial_30_day_20_days_notified = True
                uUser.Preferences_changed = True
                Call Preferences_write_one("Reg_30_day_trial_from_date", Date.Today)
              Else
                Call Register(uDialogue.Response_str)
              End If
              uOp.Subr = tLocals.Subr
            End If
          End If

        ElseIf DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) >= 30 Then
          If Not uReg.Message_reported Then
            tLocals.Msg = "Your 30 day trial has expired."
            Call Speak_it(uOp.Subr, tLocals.Msg, 2)
            uOp.Subr = tLocals.Subr
            tLocals.Msg = tLocals.Msg & vbCrLf
            tLocals.Msg = tLocals.Msg & "Please enter a registration code in order to continue to use " & uApp.Short_name & "." & vbCrLf
            tLocals.Msg = tLocals.Msg & vbCr & "To request a registration code, enter 'request'."
            Call Inputbox_LIS_top(tLocals.Msg, "", uOp.Subr)
            uOp.Subr = tLocals.Subr
            tLocals.Success = udtBoolean_extended.True_
            If uDialogue.Active Then GoTo Ender_full
            If InStr(LCase(uOp.Result), "request") > 0 Then
              Call Registration_request()
            Else
              Call Register(uOp.Result)
            End If
          End If
        End If
        tLocals.Success = udtBoolean_extended.True_

      Else ' If uReg.Trial_30_day Then
        If uReg.Code_sent Then
          tLocals.Msg = "Please enter a registration code to continue to use " & uApp.Short_name & "."
          Call Speak_it(tLocals.Subr, tLocals.Msg)
          uOp.Subr = tLocals.Subr
          Call Inputbox_LIS_top(tLocals.Msg, "", uOp.Subr)
          uOp.Subr = tLocals.Subr
          If uDialogue.Active Then GoTo Ender_full
          If uOp.Result = "" Then
            frmMain.Close()
          Else
            Call Register(uOp.Result)
          End If
        Else
          If uApp.Debug Then tLocals.Msg = DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today)
          If uReg.Code = "" Or DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) >= 30 Then
            Call Registration_request()
            uOp.Subr = tLocals.Subr
            If uOp.Success Then tLocals.Success = udtBoolean_extended.True_
          Else
            tLocals.Success = udtBoolean_extended.True_
          End If ' If uUser.Reg_code = "" Then
        End If ' If Not uUser.Reg_code_sent Then
      End If ' If uUser.Reg_30_day_trial Then
    End If ' If Not uReg.Registered Then

Ender_full:
    tLocals.Action = "Ender full"

    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    Call Log_write(tLocals, 0)
    If InStr(frmMain.ToolStripStatusLabel1.Text, "registered") = 0 Then Call frmMain.Statuslabel1_set()

Ender_partial:
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Registration_request()
    ' Separated 2015.01.31. Revised 2021.04.06.
    Dim tLocals As udtLocals = Nothing
    Dim tSubmission As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Registration_request
    uOp.Subr = tLocals.Subr
    uOp.Called_from = tLocals.Subr

    uReg.Code = Registration_code_fn() ' !
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    If uDialogue.Active Then
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_full
    End If

    uApp.Email_udt.Smtp_working = True
    tLocals.Action = "Call Registration request email"
    Call Registration_request_email()
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Or uReg.Code = "" Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check cancelled." & vbCrLf)
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr

Ender_partial:
    uOp.Subr = ""
    Call frmMain.Statuslabel1_set("Registration request sent.")
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Registration_request_email()
    ' Separated 2020.10.23. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim tSubmission As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Registration_request_email
    If uOp.Event_ = "" Then uOp.Event_ = tLocals.Subr
    tLocals.Action = "Sending registration request"
    Call frmMain.Statuslabel1_set(tLocals.Action)

    Call Registration_request_detail(tSubmission)
    If tSubmission = "" Then Stop
    uOp.Success = False
    tSubmission = tSubmission & "Submision code: " & uReg.Code & "."

    If uReg.Code_sent Then
      tLocals.Success = udtBoolean_extended.True_
    Else
      ' Checking stops it being sent twice.
      With uApp.Email_udt
        Call .Initialise()
        .Recipient_address = uLIS.Support_email
        .Subject = "Registration submission from " & uReg.Email_address & "."
        .Body = tSubmission
      End With
      Call frmMain.Statuslabel1_set("Registering with " & udtLivingInfoSystems.Name_short)
      If Email_smtp_send_fn() Then
        If uOp.Event_ = "" Then uOp.Event_ = tLocals.Subr
        uReg.Code_sent = True
        uReg.Trial_30_day_from_date = Date.Today
        uReg.Trial_30_day = True
        uUser.Preferences_changed = True
        Call Preferences_write_one("Reg_30_day_trial_from_date", Date.Today)
        tLocals.Success = udtBoolean_extended.True_
      Else
        tLocals.Msg = "The " & udtLivingInfoSystems.Name_short & " server could not be contacted. Please email the following file to " & vbCr
        tLocals.Msg = tLocals.Msg & uLIS.Support_email & ", in order to register " & uApp.App_name & "." & vbCrLf
        tLocals.Msg = tLocals.Msg & "This address has been copied to the clipboard for you to paste into your email client."
        Call Msgbox_LIS_top(tLocals.Msg, vbInformation, udtDialogue_types.Okay_only, tLocals.Subr)

        Call Clipboard_copy_to(uLIS.Support_email)

        tFile.File_full_name = uUser.My_dox & "\" & uApp.App_name & " registration submission.txt"
        Call File_properties_determine_fn(tFile)
        tFile.Open_file = tFile.File_full_name
        FileOpen(1, tFile.File_full_name, OpenMode.Output)
        Print(1, tSubmission)
        FileClose(1)
        tFile.Open_file = ""
        Try
          tLocals.Response = ShellExecute(IntPtr.Zero, "open", tFile.File_name, vbMaximizedFocus, tFile.File_path, 1)
          tLocals.Success = udtBoolean_extended.True_
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If
    End If ' If Not uReg.Code_sent Then

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Sub

  Sub Registration_request_detail(ByRef tSubmission As String)
    ' Separared 2015.02.15, so that Settings registration can use it. Revised 2021.04.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Registration_request_detail

    If uMachine.Motherboard_id = "" Then Call uMachine.Motherboard_id_determine()

    tSubmission = "Machine name: " & uMachine.Name & "." & vbCr
    tSubmission = tSubmission & "Motherboard id: " & uMachine.Motherboard_id & "." & vbCr
    tSubmission = tSubmission & "User: " & Environment.UserName & "." & vbCr
    If Not Registration_validate_email_fn(uReg.Email_address) Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
      GoTo Error_process
    End If
    uOp.Success = False

    tSubmission = tSubmission & "Email: " & uReg.Email_address & "." & vbCr
    tSubmission = tSubmission & "Software: " & uApp.App_name & "." & vbCr

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Failed Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check cancelled." & vbCrLf)
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
  End Sub

  Function Registration_code_submission_fn(ByVal tMachine As String, ByVal tMotherboard_id As String, ByVal tUser_name As String, ByVal tEmail As String, ByVal tApp_name As String) As String
    ' Separated 2018.02.04. Revised 2021.04.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Registration_code_submission_fn
    Registration_code_submission_fn = ""

Main:
    If Not Registration_validate_email_fn(tEmail) Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process
    End If

    tLocals.Response = tMachine & "|" & tMotherboard_id & "|" & tUser_name & "|" & tEmail & "|" & tApp_name & "|" & udtLivingInfoSystems.Name_short & " registration code."
    Registration_code_submission_fn = Crypt_rc4_fn(tLocals.Response, uCrypt.Private_key, True) ' !

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If
    Call Log_write(tLocals, 0)
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Function

  Function Registration_code_fn() As String
    ' Separated 2015.02.08. Revised 2021.04.14.
    Dim tLocals As udtLocals = Nothing
    Dim tSubmission_str As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Registration_code_fn
    Registration_code_fn = ""
    uOp.Subr = tLocals.Subr

    If Not Registration_validate_email_fn(uReg.Email_address) Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_
      GoTo Ender_full
    End If
    uOp.Subr = tLocals.Subr
    If uDialogue.Active Then
      tLocals.Success = udtBoolean_extended.True_
      GoTo Ender_full
    End If

    If uMachine.Motherboard_id = "" Then Call uMachine.Motherboard_id_determine()

    tSubmission_str = uMachine.Name
    tSubmission_str = tSubmission_str & "!" & uMachine.Motherboard_id
    uOp.Subr = tLocals.Subr
    tSubmission_str = tSubmission_str & "!" & uApp.App_name
    tSubmission_str = tSubmission_str & "!" & Environment.UserName
    tSubmission_str = tSubmission_str & "!" & uReg.Email_address
    Registration_code_fn = Encrypt_it_fn(tSubmission_str, True, 4) ' !
    uOp.Subr = tLocals.Subr
    If uReg.Code <> Registration_code_fn Then
      uUser.Preferences_changed = True
      uReg.Code = Registration_code_fn
      Call Preferences_write()
      uOp.Subr = tLocals.Subr
    End If
    tLocals.Success = udtBoolean_extended.True_

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
  End Function

  Function Registration_validate_email_fn(ByVal tEmail As String) As Boolean
    ' Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Registration_validate_email_fn
    uOp.Subr = tLocals.Subr
    Registration_validate_email_fn = False

    If tEmail = "" Then
      uApp.First_run_in_progress = True ' ?
      tLocals.Msg = uApp.App_name & " is not registered. Please enter your email address to register " & uApp.Short_name & "."
      Call Speak_it(tLocals.Subr, tLocals.Msg, 2)
      uOp.Subr = tLocals.Subr
      uOp.Called_from = tLocals.Subr

      tLocals.Msg = uApp.App_name & " is not registered." & vbCrLf & "Please enter your email address to register " & uApp.Short_name & "."
      uDialogue.Called_from = uOp.Subr
      Call Inputbox_LIS_top(tLocals.Msg, "", tLocals.Subr)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_full
      End If
      uOp.Success = False
      If uDialogue.Active Then
        tLocals.Success = udtBoolean_extended.True_
        GoTo Ender_full
      End If

    Else
      If InStr(1, tEmail, "@") > 0 And InStr(1, tEmail, ".") > 0 Then
        tLocals.Success = udtBoolean_extended.True_
        Registration_validate_email_fn = True
        If uOp.Called_from = udtApplication_subroutine_names.tabClient_registration_preparation Or uOp.Called_from = udtApplication_subroutine_names.txtReg_rcv_code_LostFocus Then
        Else
          If tEmail <> uReg.Email_address Then
            uReg.Email_address = tEmail
            uUser.Preferences_changed = True
            uOp.Subr = tLocals.Subr
          End If
        End If
      Else
        tLocals.Msg = "The submitted email ('" & uReg.Email_address & "') was determined to be invalid."
        Call Msgbox_LIS_top(tLocals.Msg, udtAlert_types.Critical, udtDialogue_types.Okay_only, tLocals.Subr)
        uOp.Subr = tLocals.Subr
        uReg.Email_address = ""
        If uDialogue.Active Then GoTo Ender_full
        uUser.Preferences_changed = True
        uError.Message = tLocals.Msg
      End If
    End If

    If uUser.Preferences_changed Then
      Call Preferences_write()
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        GoTo Ender_full
      End If
      uOp.Success = False
    End If

Ender_full:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
      Registration_validate_email_fn = True
    End If
    Call Log_write(tLocals, 0)
    uOp.Subr = ""

    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Function

  Sub Register(ByVal tSubmission As String)
    ' Separated 2014.06.07. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tRegistration_response As String
    tLocals.Subr = udtApplication_subroutine_names.Register
    tLocals.Error_count = 0

    tLocals.Action = "Test request"
    If LCase(tSubmission) = "request" Then
      uReg.Code_sent = False
      Call Registration_request()
      If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process
      End If
      uOp.Success = False

    Else
      tLocals.Action = "Get expected registration response"
      If uMachine.Motherboard_id = "" Then
        Call uMachine.Motherboard_id_determine()
        If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process
        End If
        uOp.Success = False
      End If

      If Right(tSubmission, 1) = "." Then tSubmission = Left(tSubmission, Len(tSubmission) - 1)
      tRegistration_response = Registration_code_submission_fn(uMachine.Name, uMachine.Motherboard_id, Environment.UserName, uReg.Email_address, uApp.App_name)
      If Not uOp.Success Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        If Not tLocals.Continue_ Then GoTo Error_process
      End If
      uOp.Success = False

      tLocals.Action = "Test registration response"
      If tSubmission = tRegistration_response Or tSubmission = Crypt_rc4_fn("12BC360A5CE9389F4621CE18", uCrypt.Private_key, False) Then
        uReg.Registered = True
        uUser.Preferences_changed = True
        Call Preferences_write_one("Registered", True)
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process
        End If
        uOp.Success = False

        Call frmMain.Enter_registration_code_tool_strip_menu_item_visibility(False)

        tLocals.Msg = uApp.App_name & " is now registered."
        Call Speak_it(uOp.Subr, tLocals.Msg, 2)
        Call Log_write(tLocals, 0)
        If Not uOp.Success Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          If Not tLocals.Continue_ Then GoTo Error_process
        End If
        uOp.Success = False

        Call frmMain.Statuslabel1_set(tLocals.Msg)
      Else
        tLocals.Action = "Submitted registration code does not match the expected code"
        Call Sound_alert()
        If DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) <= 30 Then
          Call frmMain.Statuslabel1_set("Your registration submission failed. This application will stop working at the end of the 30 trial.", Color.Red)
        Else
          If DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) <= 40 Then
            tLocals.Msg = "Registration submission failed. Your 30 day trial ended " & DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) & " days ago." & vbCrLf
            tLocals.Msg = tLocals.Msg & "You may request a new registration code by clicking the Request Registration button in " & frmMain.tabSettings.Name & "."
            Call Msgbox_LIS_top(tLocals.Msg, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
            If Not uOp.Success Then
              Stop ' Permanent.
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Status = udtuStandard_texts.Failed
              tLocals.Success = udtBoolean_extended.False_
              If Not tLocals.Continue_ Then GoTo Error_process
            End If
            uOp.Success = False
          Else
            tLocals.Msg = "Registration submission failed. Your 30 day trial ended " & DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) & " days ago." & vbCrLf
            tLocals.Msg = tLocals.Msg & uApp.App_name & " will now quit."
            Call Msgbox_LIS_top(tLocals.Msg, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
            If Not uOp.Success Then
              Stop ' Permanent.
              tLocals.Error_count = tLocals.Error_count + 1
              tLocals.Status = udtuStandard_texts.Failed
              tLocals.Success = udtBoolean_extended.False_
              If Not tLocals.Continue_ Then GoTo Error_process
            End If
            uOp.Success = False

            uApp.Terminating = uApp.Terminating + 1
          End If ' If DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) <= 60 Then
        End If ' If DateDiff(DateInterval.Day, uReg.Trial_30_day_from_date, Date.Today) <= 30 Then
        uOp.Subr = tLocals.Subr

      End If ' If tSubmission = tRegistration_response Or tSubmission = Crypt_rc4_fn("12BC360A5CE9389F4621CE18", uCrypt.Private_key, False) Then
    End If

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If tLocals.Success <> udtBoolean_extended.False_ Then tLocals.Status = udtuStandard_texts.Completion
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)
    uOp.Subr = ""
  End Sub

#End Region ' End region, Registration.

#Region "Error handling"

  Sub Error_clear(ByRef tLocals_external As udtLocals, Optional ByVal tInitialise As Boolean = False)
    ' Written by MJBH, 2018.04.18. Copied from Excel 2019.12.02. Revised 2021.04.24.
    With tLocals_external
      If .Error_count > 0 Then .Error_count = .Error_count - 1
      .Error_description = ""
      .Error_number = 0
      .Error_source = ""
      If .Success = udtBoolean_extended.False_ Then .Success = udtBoolean_extended.Null
    End With

    With uError
      .Description = Nothing
      If .Count > 0 Then .Count = .Count - 1
      If tInitialise Then .Count = 0 ' Use at the beginning of a sub-routine to reset.
      .Extra = Nothing
      .Message = Nothing
      .Number = 0
      .Last_error_sub = ""
      .Source = Nothing
    End With

    ' Err.Clear()
  End Sub

  Sub Error_display(ByRef tLocals_external As udtLocals)
    ' Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tCalled_from As String = uOp.Called_from
    Dim tError As udtError = Nothing

    On Error GoTo Error_captured
    tLocals.Subr = "Error display"
    tLocals.Response = Nothing
    tLocals.Status = udtuStandard_texts.Initiation
    If uOp.Called_from = tLocals.Subr Then
      uApp.Terminating = uApp.Terminating + 1
      GoTo Ender_full ' Prevents looping.
    End If
    If uError.Reported Then GoTo Ender_full
    If uOp.Subr = "" Then uOp.Subr = tLocals.Subr
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start

    tError.Number = uError.Number
    tError.Description = uError.Description
    tError.Count = uError.Count
    tError.Extra = uError.Extra
    tError.Last_error_sub = uError.Last_error_sub
    If tError.Last_error_sub = "" Then
      Stop ' Indefinite; fix the calling sub-routine.
      uError.Last_error_sub = tLocals_external.Subr
      GoTo Ender_partial
    End If
    tError.Message = uError.Message
    tError.Source = uError.Source
    tError.Speech_error_reported = uError.Speech_error_reported
    If tError.Number = 0 Then tError.Number = tLocals_external.Error_number
    If tError.Description = "" Then tError.Description = tLocals_external.Error_description
    If tLocals_external.Error_count > tError.Count Then tError.Count = tLocals_external.Error_count
    If tError.Source = "" Then tError.Source = tLocals_external.Error_source

    If tError.Count >= 20 Then tError.Description = "Twenty or more uncaptured errors have been encountered."
    If tError.Description = "" And tError.Message = "" Then uError.Description = "An error message was called, but the error was not specified."
    If tError.Description = "" And tError.Message <> "" Then
      tError.Description = tError.Message
      tError.Message = ""
    End If
    If tError.Description = "Common Language Runtime detected an invalid program." Then uApp.Terminating = uApp.Terminating + 1

    Call Error_write_mega(tLocals_external)
    Call Sound_alert()
    If uOp.Called_from <> udtApplication_subroutine_names.Error_email And Not uMachine.Development_machine Then
      Call Appdata_email()
      uOp.Subr = tLocals.Subr
    End If
    If uDialogue.Active Then GoTo Ender_full

    mObjSynth.Speak(tError.Description)
    Call Msgbox_LIS_top(tError.Description, udtAlert_types.Information, udtDialogue_types.Okay_only, tLocals.Subr)
    uOp.Subr = tLocals.Subr
    '    If uDialogue.Active Then GoTo Ender_partial
    '    If uOp.Called_from = "" Then uOp.Called_from =  tLocals.Subr 

Error_captured:
    If Err.Number <> 0 Then
      Stop ' Permanent. Now what?
      '      uError.Message = "An error occurred in " & uOp.Subr & "." & vbCr
      '      uError.Message = uError.Message & "Computer name = " & uMachine.Name & "." & vbCr
      '      uError.Message = uError.Message & "User name = " & Environment.UserName & "." & vbCr
      '      uError.Message = uError.Message & "User corporation = " & uActive.Corporation_name & "." & vbCr
      '      uError.Message = uError.Message & "User branch = " & uActive.Branch_name & "." & vbCr
      '      uError.Message = uError.Message & "Action$ = " & uOp.Action & "." & vbCr
      '      uError.Message = uError.Message & "OpenF$ = " & tFile.Open_file & "." & vbCr
      '      uError.Message = uError.Message & "Error source = " & Err.Source & "." & vbCr
      '      uError.Message = uError.Message & "Error description = " & Err.Description & "." & vbCr
      '      uError.Message = uError.Message & "Error number = " & Str$(Err.Number) & "."
      '      If InStr(uError.Description, "Common Language Runtime detected an invalid program.") > 0 Then
      '        uError.Message = uError.Message & "This error requires restarting this computer to resolve." & vbCr
      '      End If
      '      '     uOp.Result  = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
      '      ' uDialogue.Dialogue_result = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
      '      Call Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
      '      uError.Message = ""
    End If

    '    If uApp.Terminating Or InStr(uError.Description, "Common Language Runtime detected an invalid program.") > 0 Then
    '      frmMain.Close()
    '      Call End_game()
    '    End If
    '    If        tLocals.Error_count    <> 0 And uError.Message <> "" Then
    '      '     uOp.Result  = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
    '      ' uDialogue.Dialogue_result = Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
    '      Call Msgbox_LIS_top (uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only,  tLocals.Subr )
    '    End If
    '    uOp.Called_from = tCalled_from
    '        uOp.Result  = Nothing

Ender_full:
    uOp.Called_from = tLocals.Subr
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
    If uApp.Terminating = 0 Then
      Call Log_write(tLocals, 0)
    End If
    Call frmMain.Statuslabel1_set("", True)

Ender_partial:
    uOp.Subr = tLocals.Subr
    Stop ' Permanent.
  End Sub

  Sub Appdata_email()
    ' Written by Michael J. B. Hodgson, 2017.02.16. Revised 2021.04.22.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Error_email
    uOp.Called_from = tLocals.Subr

    With uApp.Email_udt
      Call .Initialise()
      .Recipient_address = uLIS.Support_email
      .Subject = "Application data zip file submission."
      If uError.Message <> "" Then .Subject = "Error detected in " & uApp.App_name & " on " & uMachine.Name & "."
      .Body = "Company name: " & uActive.Corporation_name & vbCr
      .Body = .Body & "Machine name: '" & Environ("COMPUTERNAME") & "'." & vbCr
      .Body = .Body & "Operating system: '" & System.Environment.OSVersion.VersionString & "'." & vbCr
      .Body = .Body & "Application: '" & uApp.App_name & "'." & vbCr
      .Body = .Body & "uLIS.Prog_file_path: '" & uLIS.Prog_file_path & "'." & vbCr
      .Body = .Body & "User name: '" & Environ("UserName") & "'." & vbCr
      .Body = .Body & "uUser.Profile_path: '" & uUser.Profile_path & "'." & vbCr
      .Body = .Body & "uUser.Ap_data_path: '" & uUser.Appdata_path & "'." & vbCr
      If uError.Message <> "" Then .Body = .Body & "Error: " & uError.Message & vbCr
      .Attachment_loc = Compression_fn(uLIS.Profile_path)
    End With
    Call frmMain.Statuslabel1_set("Status report")
    If Email_smtp_send_fn() Then
      uError.Reported = True
      If Not uApp.Environment_dump_emailed Then
        uApp.Environment_dump_emailed = True
        uUser.Preferences_changed = True
      End If
      Try
        Kill(uApp.Email_udt.Attachment_loc)
      Catch ex As Exception
        tLocals.Error_description = ex.Message
      End Try
    Else
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      ' GoTo Error_handler
    End If

Ender_full:
    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uOp.Subr = ""
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Error_write_micron(ByVal tText As String)
    ' Revised 2021.04.15.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile = Nothing

    tLocals.Subr = uOp.Subr ' Not the other way around.
    tLocals.Action = ""
    On Error GoTo Error_captured

    If uApp.Error_log_loc Is Nothing Or uApp.Error_log_loc = "" Then
      Call uApp.Initialise()
    End If
    If uApp.Error_log_loc Is Nothing Or uApp.Error_log_loc = "" Then Exit Sub

    If tText <> "" Then
      With tFile
        .Free_file = FreeFile()
        .Open_file = uApp.Error_log_loc
        FileOpen(.Free_file, uApp.Error_log_loc, OpenMode.Output)
        Print(.Free_file, tText)
        FileClose(.Free_file)
        .Open_file = ""
      End With
    End If
    tLocals.Success = True

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 53, 55, 75, 76
            Call Error_handler_for_text_file(tLocals)
            If tLocals.Action = "Resume" Then tLocals.Action = "" : Resume
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    Err.Clear()
  End Sub

  Sub Error_write_mega(ByVal tLocals_external As udtLocals, Optional ByVal tMessage As String = "", Optional ByVal tTable As String = "")
    ' Written by Michael John Boyack Hodgson, 2014.03.01. Revised 2021.04.28.
    Dim tLocals As udtLocals = Nothing
    Dim tError_desc As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Error_write_mega

    If uMachine.Motherboard_id = "" Then Call uMachine.Motherboard_id_determine()
    If uMachine.Date_format = "" Then Call Culture_detect()
    ' If uMachine.Date_format = "" Then Call Date_detect_system_format(False)
    If uActive.Country = "" Then Call Country_detect(System.Globalization.RegionInfo.CurrentRegion.DisplayName, True)

    If uError.Description = "" And Err.Description <> "" Then tLocals.Error_description = Err.Description
    If uError.Number = 0 And Err.Number <> 0 Then tLocals.Error_number = Err.Number
    If uError.Source = "" And Err.Source <> "" Then tLocals.Error_source = Err.Source

    tError_desc = tError_desc & "Error status detailed report for " & uApp.App_name & "." & vbCr
    tError_desc = tError_desc & "Date: " & Format(DateAndTime.Today, "yyyy.MM.dd ddd") & ", "
    tError_desc = tError_desc & "time: " & DateAndTime.TimeString & "."
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "Computer name: " & uMachine.Name & "." & vbCr
    tError_desc = tError_desc & "User name: " & Environment.UserName & "." & vbCr
    tError_desc = tError_desc & "Sub-routine: " & tLocals_external.Subr & "." & vbCr
    tError_desc = tError_desc & "User corporation: " & uActive.Corporation_name & "." & vbCr
    tError_desc = tError_desc & "User branch: " & uActive.Branch_name & "." & vbCr
    tError_desc = tError_desc & "Action$: " & tLocals_external.Action & "." & vbCr
    ' tError_desc = tError_desc & "OpenF$: " & tFile.Open_file & "." & vbCr

    tError_desc = tError_desc & "Error source: " & uError.Source & "." & vbCr
    tError_desc = tError_desc & "Error description: " & uError.Description
    If Right$(uError.Description, 1) <> "." Then tError_desc = tError_desc & "."
    tError_desc = tError_desc & vbCr
    tError_desc = tError_desc & "Error number:" & Str$(uError.Number) & "." & vbCr
    tError_desc = tError_desc & "1a: " & vbCr ' uError.Continue_ & vbCr
    tError_desc = tError_desc & "1b: " & uError.Count & vbCr
    tError_desc = tError_desc & "1c: " & uError.Extra & vbCr
    If uError.Message <> "" Then
      If Left(uError.Message, 17) = "An error occured" Then
        ' Looping.
        Stop
        uApp.Terminating = uApp.Terminating + 1
      Else
        tError_desc = tError_desc & vbCr & "Error message: " & uError.Message & "." & vbCr
      End If
    End If
    tError_desc = tError_desc & "1d: " & uError.Message & vbCr
    tError_desc = tError_desc & "1e: " & vbCr ' uError.Suppress_error & vbCr
    tError_desc = tError_desc & "1f: " & uError.Speech_error_reported & vbCr
    tError_desc = tError_desc & "1g: " & vbCr ' uError.Error_log_loc & vbCr
    tError_desc = tError_desc & "1h: " & udtError.Infinite_loop_limit & vbCr
    tError_desc = tError_desc & "1i: " & uError.Last_error_sub & vbCr
    tError_desc = tError_desc & "1j: " & uError.Reported & vbCr
    If tMessage <> "" Then
      If Right(tMessage, Len(tMessage)) <> "." Then tMessage = tMessage & "."
      tError_desc = tError_desc & "Extra message: " & tMessage & "." & vbCr
    End If
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "2a: " & uLIS.Config_loc & vbCr
    tError_desc = tError_desc & "2b: " & udtLivingInfoSystems.Name_short & vbCr
    tError_desc = tError_desc & "2c: " & uLIS.Dev & vbCr
    tError_desc = tError_desc & "2d: " & uLIS.Prog_file_path & vbCr
    tError_desc = tError_desc & "2e: " & uLIS.Profile_path & vbCr
    tError_desc = tError_desc & "2f: " & vbCr ' uLIS.Server_network_path & vbCr
    tError_desc = tError_desc & "2g: " & vbCr ' uLIS.Server_ip & vbCr
    tError_desc = tError_desc & "2h: " & uLIS.Resources_path & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "3a: " & uMachine.Name & vbCr
    tError_desc = tError_desc & "3b: " & uMachine.Environment_known & vbCr
    tError_desc = tError_desc & "3c: " & uMachine.Os_major & vbCr
    tError_desc = tError_desc & "3d: " & uMachine.Os_minor & vbCr
    tError_desc = tError_desc & "3e: " & uMachine.Os_name & vbCr
    tError_desc = tError_desc & "3f: " & uMachine.Development_machine_name & vbCr
    tError_desc = tError_desc & "3g: " & uMachine.Environment_dump & vbCr
    tError_desc = tError_desc & "3h: " & uMachine.Development_machine & vbCr
    tError_desc = tError_desc & "3i: " & uMachine.Motherboard_id & vbCr
    tError_desc = tError_desc & "3j: " & uMachine.Testing_machine & vbCr
    tError_desc = tError_desc & "3k: " & uMachine.Date_format & vbCr
    ' See Template.Kernel.Date_detect_system_format for more information:
    tError_desc = tError_desc & "3.1a, Date separator: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.DateSeparator & vbCr
    tError_desc = tError_desc & "3.1b, Full date time pattern: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.FullDateTimePattern & vbCr
    tError_desc = tError_desc & "3.1c, Abbreviated era name: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.GetAbbreviatedEraName(0) & vbCr
    tError_desc = tError_desc & "3.1d, Era name: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.GetEraName(0) & vbCr
    tError_desc = tError_desc & "3.1e, Long date pattern: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.LongDatePattern & vbCr ' dd MMMM yyyy.
    tError_desc = tError_desc & "3.1f, Native calendar name: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.NativeCalendarName & vbCr
    tError_desc = tError_desc & "3.1g, Time separator: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.TimeSeparator & vbCr
    tError_desc = tError_desc & "3.1h, Short date pattern: " & System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.ShortDatePattern & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "4a: " & uApp.App_name & "." & vbCr
    tError_desc = tError_desc & "4b: " & uApp.Internal_name & "." & vbCr
    tError_desc = tError_desc & "4c: " & uApp.Short_name & "." & vbCr
    tError_desc = tError_desc & "4d: " & uApp.App_build_number & "." & vbCr
    tError_desc = tError_desc & "4e: " & uApp.App_revision_date & "." & vbCr
    tError_desc = tError_desc & "4f: " & uApp.Appdata_path & "." & vbCr
    tError_desc = tError_desc & "4g: " & vbCr ' uApp.Default_height & vbCr
    tError_desc = tError_desc & "4h: " & uApp.Default_left & vbCr
    tError_desc = tError_desc & "4i: " & uApp.Default_top & vbCr
    tError_desc = tError_desc & "4j: " & vbCr ' uApp.Default_width & vbCr
    tError_desc = tError_desc & "4k: " & uApp.First_run_in_progress & vbCr
    tError_desc = tError_desc & "4l: " & vbCr ' uApp.Export & vbCr
    tError_desc = tError_desc & "4m: " & uApp.Environment_dump_emailed & vbCr
    tError_desc = tError_desc & "4n: " & uApp.Error_log_loc & vbCr
    tError_desc = tError_desc & "4o: " & vbCr ' uApp.Installed & vbCr
    tError_desc = tError_desc & "4p: " & uApp.Log_loc & vbCr
    tError_desc = tError_desc & "4q: " & vbCr ' uApp.Modules_exported & vbCr
    tError_desc = tError_desc & "4r: " & uApp.Prog_file_path & vbCr
    tError_desc = tError_desc & "4s: " & uApp.Me_ & vbCr
    tError_desc = tError_desc & "4t: " & uApp.Initialising & vbCr
    tError_desc = tError_desc & "4u: " & uApp.End_game & vbCr
    tError_desc = tError_desc & "4v: " & uApp.Terminating & vbCr
    tError_desc = tError_desc & "4w: " & uApp.Debug & vbCr
    tError_desc = tError_desc & "4x: " & vbCr ' uApp.Email_smtp_working & vbCr
    tError_desc = tError_desc & "4y: " & vbCr ' uApp.EditPad_loc & vbCr
    tError_desc = tError_desc & "4z: " & uApp.Run_count.Count & vbCr
    tError_desc = tError_desc & "4aa: " & uApp.Environment_dump_completed_date & vbCr
    tError_desc = tError_desc & "4ab: " & uApp.First_run_completed_date & vbCr
    tError_desc = tError_desc & "4ac: " & uApp.First_run_in_progress & vbCr
    tError_desc = tError_desc & "4ad: " & uApp.Memory_check_completed_date & vbCr
    tError_desc = tError_desc & "4ae: " & uApp.Startup_sequence & vbCr
    tError_desc = tError_desc & "4af: " & uApp.Shutdown & vbCr
    tError_desc = tError_desc & "4ag: " & uApp.Sql_last_loc & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "5a: " & vbCr ' uMemory.Use & vbCr
    tError_desc = tError_desc & "5b: " & vbCr ' uMemory.Db_type & vbCr
    tError_desc = tError_desc & "5c: " & vbCr ' uMemory.Server & vbCr
    tError_desc = tError_desc & "5d: " & vbCr ' uMemory.Port & vbCr
    tError_desc = tError_desc & "5e: " & vbCr ' uMemory.Username & vbCr
    tError_desc = tError_desc & "5f: " & vbCr ' uMemory.Password & vbCr
    tError_desc = tError_desc & "5g: " & vbCr ' uMemory.Exe_loc & vbCr
    tError_desc = tError_desc & "5h: " & uMemory.Database & vbCr
    tError_desc = tError_desc & "5i: " & uMemory.Table_name & vbCr
    If InStr(uMemory.Conn_str, "Password") = 0 Then
      tError_desc = tError_desc & "5j:" & vbCr
    Else
      tError_desc = tError_desc & "5j: " & Mid(uMemory.Conn_str, InStr(uMemory.Conn_str, "Password") + 9) & vbCr
    End If
    tError_desc = tError_desc & "5k: " & uMemory.Cmd_str & vbCr
    tError_desc = tError_desc & "5l: " & vbCr ' uMemory.Conn & vbCr
    tError_desc = tError_desc & "5m: " & vbCr ' uMemory.Cmd & vbCr
    tError_desc = tError_desc & "5n: " & vbCr ' uMemory.Record_set & vbCr
    tError_desc = tError_desc & "5o: " & uMemory.Rows_affected & vbCr
    tError_desc = tError_desc & "5p: " & uMemory.Upgrade_table & vbCr
    tError_desc = tError_desc & "5q: " & uMemory.Changed & vbCr
    tError_desc = tError_desc & "5r: " & uMemory.Indent & vbCr
    tError_desc = tError_desc & "5s: " & uMemory.Success & vbCr
    tError_desc = tError_desc & "5t: " & vbCr ' uMemory.Upgrade_field & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "6a: " & vbCr ' uUser.Country & vbCr
    tError_desc = tError_desc & "6b: " & vbCr ' uUser.Currency & vbCr
    tError_desc = tError_desc & "6c: " & vbCr ' uUser.Currency_format & vbCr
    tError_desc = tError_desc & "6d: " & vbCr ' uUser.Currency_sign & vbCr
    tError_desc = tError_desc & "6e: " & vbCr ' uUser.Environment_dump_emailed & vbCr
    tError_desc = tError_desc & "6f: " & vbCr ' uUser.Reg_30_day_trial & vbCr
    tError_desc = tError_desc & "6g: " & vbCr ' uUser.Reg_30_day_trial_from_date & vbCr
    tError_desc = tError_desc & "6h: " & vbCr ' uUser.Trial_30_day_20_days_notified & vbCr
    tError_desc = tError_desc & "6i: " & vbCr ' uUser.Reg_code & vbCr
    tError_desc = tError_desc & "6j: " & vbCr ' uUser.Code_sent & vbCr
    tError_desc = tError_desc & "6k: " & vbCr ' uUser.Registered & vbCr
    tError_desc = tError_desc & "6l: " & uUser.Speech & vbCr
    tError_desc = tError_desc & "6m: " & vbCr ' uUser.User_first_name & vbCr
    tError_desc = tError_desc & "6n: " & uUser.Appdata_path & vbCr
    tError_desc = tError_desc & "6o: " & uUser.Cancel & vbCr
    tError_desc = tError_desc & "6p: " & uUser.My_dox & vbCr
    tError_desc = tError_desc & "6q: " & vbCr ' uUser.Option_num & vbCr
    tError_desc = tError_desc & "6r: " & vbCr ' uUser.Option_str & vbCr
    tError_desc = tError_desc & "6s: " & uUser.Preferences_changed & vbCr
    tError_desc = tError_desc & "6t: " & uUser.Profile_path & vbCr
    tError_desc = tError_desc & "6u: " & vbCr ' uUser.User_name & vbCr
    tError_desc = tError_desc & "6v: " & uUser.First_name & vbCr
    tError_desc = tError_desc & "6w: " & uUser.First_run_wizard & vbCr
    tError_desc = tError_desc & "6x: " & uUser.First_run_wizard_max & vbCr
    tError_desc = tError_desc & "6y: " & uUser.First_run_wizard_step & vbCr
    tError_desc = tError_desc & "6z: " & uUser.First_run_wizard_loading & vbCr
    tError_desc = tError_desc & "6aa: " & uUser.Main_top & vbCr
    tError_desc = tError_desc & "6ab: " & uUser.Main_left & vbCr
    tError_desc = tError_desc & "6ac: " & uUser.Main_width & vbCr
    tError_desc = tError_desc & "6ad: " & uUser.Main_height & vbCr
    tError_desc = tError_desc & "6ae: " & uUser.Min_sys_tray & vbCr
    tError_desc = tError_desc & "6af: " & vbCr ' uUser.Reg_email & vbCr
    tError_desc = tError_desc & "6ag: " & uUser.Start_maximised & vbCr
    tError_desc = tError_desc & "6ah: " & uUser.Tab_last & vbCr
    tError_desc = tError_desc & "6ai: " & uUser.First_run_wizard_completed & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "7a: " & uOp.Subr & vbCr
    tError_desc = tError_desc & "7b: " & vbCr ' uOp.Action & vbCr
    tError_desc = tError_desc & "7c: " & uOp.Called_from & vbCr
    tError_desc = tError_desc & "7d: " & vbCr ' uOp.Msg & vbCr
    tError_desc = tError_desc & "7e: " & vbCr ' uOp.Open_file & vbCr
    tError_desc = tError_desc & "7f: " & vbCr ' uOp.Debug & vbCr
    tError_desc = tError_desc & "7g: " & vbCr ' uOp.Initialising & vbCr
    tError_desc = tError_desc & "7h: " & vbCr ' uOp.Keyboard_ascii & vbCr
    tError_desc = tError_desc & "7i: " & uOp.Result & vbCr
    tError_desc = tError_desc & "7j: " & Get_system_metrics_fn(0) & vbCr
    tError_desc = tError_desc & "7k: " & Get_system_metrics_fn(1) & vbCr
    tError_desc = tError_desc & "7l: " & uOp.Success & vbCr
    tError_desc = tError_desc & "7m: " & Year(DateAndTime.Now) & vbCr
    tError_desc = tError_desc & "7n: " & vbCr ' uOp.Terminating & vbCr
    tError_desc = tError_desc & "7o: " & vbCr ' uOp.Text_read_line & vbCr
    tError_desc = tError_desc & "7p: " & uOp.Event_ & vbCr
    tError_desc = tError_desc & "7q: " & vbCr ' uOp.Status & vbCr
    tError_desc = tError_desc & "7r: " & uOp.Administer & vbCr
    tError_desc = tError_desc & "7s: " & vbCr ' uOp.Diagnostics & vbCr
    tError_desc = tError_desc & "7t: " & uOp.Diagnostics.All_tables_done & vbCr
    tError_desc = tError_desc & "7u: " & uOp.Diagnostics.All_tables_in_progress & vbCr
    tError_desc = tError_desc & "7v: " & uOp.Diagnostics.In_progress & vbCr
    tError_desc = tError_desc & "7w: " & uOp.Form_initialising & vbCr
    tError_desc = tError_desc & "7x: " & uOp.Form_updating & vbCr
    tError_desc = tError_desc & "7y: " & uOp.Form_refreshing & vbCr
    tError_desc = tError_desc & "7z: " & uOp.Form_loaded & vbCr
    tError_desc = tError_desc & "7aa: " & vbCr ' uOp.Num_err & vbCr
    tError_desc = tError_desc & "7ab: " & uOp.Preferences_loaded & vbCr
    tLocals.Action = "tLocals_external:"
    tError_desc = tError_desc & "7ac: " & tLocals_external.Subr & vbCr
    tError_desc = tError_desc & "7ad: " & tLocals_external.Action & vbCr
    tError_desc = tError_desc & "7ae: " & tLocals_external.Changed & vbCr
    tError_desc = tError_desc & "7af: " & tLocals_external.Status & vbCr
    tError_desc = tError_desc & "7ag: " & tLocals_external.Success & vbCr
    tError_desc = tError_desc & "7ah: " & tLocals_external.Continue_ & vbCr
    tError_desc = tError_desc & "7ai: " & tLocals_external.Msg & vbCr
    tError_desc = tError_desc & "7aj: " & tLocals_external.Response & vbCr
    tError_desc = tError_desc & "7ak: " & tLocals_external.Steps_completed & vbCr
    tError_desc = tError_desc & "7al: " & tLocals_external.Steps_to_complete & vbCr
    tError_desc = tError_desc & "7am: " & tLocals_external.Error_number & vbCr
    tError_desc = tError_desc & "7an: " & tLocals_external.Error_source & vbCr
    tError_desc = tError_desc & "7ao: " & tLocals_external.Error_description & vbCr
    tError_desc = tError_desc & "7ap: " & tLocals_external.Error_count & vbCr
    tError_desc = tError_desc & vbCr

    'tError_desc = tError_desc & "8a: " & vbCr ' tFile.Batch & vbCr
    'tError_desc = tError_desc & "8b: " & tFile.File_extension & vbCr
    'tError_desc = tError_desc & "8c: " & tFile.File_full_name & vbCr
    'tError_desc = tError_desc & "8d: " & tFile.File_path & vbCr
    'tError_desc = tError_desc & "8e: " & tFile.File_name & vbCr
    'tError_desc = tError_desc & "8f: " & tFile.File_short_name & vbCr
    'tError_desc = tError_desc & "8g: " & vbCr ' tFile.Fso & vbCr
    'tError_desc = tError_desc & "8h: " & vbCr ' tFile.Fso_file & vbCr
    'tError_desc = tError_desc & "8i: " & vbCr ' tFile.EditPad_loc & vbCr
    'tError_desc = tError_desc & "8j: " & vbCr ' tFile.Environment_dump & vbCr
    'tError_desc = tError_desc & "8k: " & vbCr ' tFile.Installer_path & vbCr
    'tError_desc = tError_desc & "8l: " & tFile.Copy_source & vbCr
    'tError_desc = tError_desc & "8m: " & tFile.Copy_destination & vbCr
    'tError_desc = tError_desc & "8n: " & tFile.Open_file & vbCr
    'tError_desc = tError_desc & "8o: " & tFile.Text_read_line & vbCr
    'tError_desc = tError_desc & vbCr

    ' tError_desc = tError_desc & "9a: " & uSearch.Keyword & vbCr
    ' tError_desc = tError_desc & "9b: " & uSearch.Sub_keyword & vbCr
    'tError_desc = tError_desc & "9c: " & vbCr ' uSearch.Match_exact & vbCr
    'tError_desc = tError_desc & "9d: " & vbCr ' uSearch.Test_value & vbCr
    'tError_desc = tError_desc & "9e: " & vbCr ' uSearch.Find_next & vbCr
    'tError_desc = tError_desc & "9f: " & vbCr ' uSearch.By_row & vbCr
    'tError_desc = tError_desc & "9g: " & vbCr ' uSearch.Cell_addr & vbCr
    'tError_desc = tError_desc & "9h: " & vbCr ' uSearch.'  & vbcr
    'tError_desc = tError_desc & "9i: " & vbCr ' uSearch.'  & vbcr
    'tError_desc = tError_desc & "9j: " & vbCr ' uSearch.'  & vbcr
    'tError_desc = tError_desc & "9k: " & vbCr ' uSearch.'  & vbcr
    'tError_desc = tError_desc & "9l: " & vbCr ' uSearch.Column & vbCr
    'tError_desc = tError_desc & "9m: " & vbCr ' uSearch.Column_str & vbCr
    'tError_desc = tError_desc & "9n: " & vbCr ' uSearch.Row & vbCr
    'tError_desc = tError_desc & "9o: " & vbCr ' uSearch.Row_str & vbCr
    'tError_desc = tError_desc & "9p: " & vbCr ' uSearch.Found & vbCr
    'tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "10a: " & uReg.Email_address & vbCr
    tError_desc = tError_desc & "10b: " & uReg.Trial_30_day & vbCr
    tError_desc = tError_desc & "10c: " & uReg.Trial_30_day_from_date & vbCr
    tError_desc = tError_desc & "10d: " & uReg.Trial_30_day_20_days_notified & vbCr
    tError_desc = tError_desc & "10e: " & uReg.Code & vbCr
    tError_desc = tError_desc & "10f: " & uReg.Code_sent & vbCr
    tError_desc = tError_desc & "10g: " & uReg.Registered & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "11a: " & uActive.Country & vbCr
    tError_desc = tError_desc & "11b: " & uActive.Currency_tla & vbCr
    tError_desc = tError_desc & "11c: " & uActive.Currency_symbol & vbCr
    tError_desc = tError_desc & "11d: " & uActive.Currency_format & vbCr
    tError_desc = tError_desc & "11e: " & uActive.Corporation_name & vbCr
    tError_desc = tError_desc & "11f: " & uActive.Branch_name & vbCr
    tError_desc = tError_desc & vbCr

    tError_desc = tError_desc & "End of report."
    tError_desc = tError_desc & vbCr

    ' uOp.Called_from =  tLocals.Subr 
    Call Error_write_micron(tError_desc)
    If tTable <> uTables.Log And uOp.Called_from <> udtApplication_subroutine_names.Log_write_memory Then
      tLocals.Success = udtBoolean_extended.True_
      Call Log_write(tLocals, 0, tError_desc)
    End If
    uOp.Called_from = ""
  End Sub

  Sub Error_handler_for_text_file(ByRef tLocals_external As udtLocals)
    ' Separated 2020.10.30. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    If tLocals_external.Error_number <> Err.Number Or tLocals_external.Error_description <> Err.Description Or tLocals_external.Error_source <> Err.Source Or _
       uError.Count = 0 Then
      Stop ' Permanent; fix calling sub.
      tLocals_external.Error_number = Err.Number
      tLocals_external.Error_description = Err.Description
      tLocals_external.Error_source = Err.Source
    End If
    tLocals.Success = udtBoolean_extended.Null

    If Err.Number <> 0 Then
      Stop ' Indefinite.
      If tLocals_external.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals_external.Error_number

          Case 0
            Stop ' For debugging.

          Case 53 ' File not found.
            Stop ' Indefinite.
            tLocals.Msg = "File Not Found." & Chr(13) & Chr(13) & "Error number " & Str(Err.Number) & "."

          Case 55 ' File already open.
            FileClose(FreeFile() - 1)
            If tLocals_external.Error_count <= 5 Then
              tLocals_external.Action = udtuStandard_texts.Retry
            Else
              uApp.Terminating = uApp.Terminating + 1
            End If

          Case 75 ' The process cannot access the file 'C:\Users\micha\AppData\Roaming\LivingInfo Systems\Financier\Financier log 2019.txt' because it is being used by another process.
            FileClose(FreeFile() - 1)
            If tLocals_external.Error_count <= 5 Then
              tLocals_external.Action = udtuStandard_texts.Retry
            Else
              uApp.Terminating = uApp.Terminating + 1
              uError.Message = "Error logging is un-available, due to a computer session error. Please restart this machine as soon as possible."
              Call Msgbox_LIS_top(uError.Message, udtAlert_types.Critical, udtDialogue_types.Okay_only, uOp.Subr)
              uOp.Success = False
              Err.Clear()
              tLocals.Error_number = Err.Number
              tLocals.Error_description = Err.Description
              tLocals.Error_source = Err.Source
              uError.Message = ""
            End If

          Case 76 ' Could not find part of the path.
            Err.Clear()
            Call First_run_make_folders(Nothing)
            If uError.Count <= 5 Then
              tLocals_external.Action = udtuStandard_texts.Retry
            Else
              uApp.Terminating = uApp.Terminating + 1
            End If

          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)

        End Select
      End If ' If uApp.Terminating = 0 Then
    End If ' If Err.Number <> 0 Then
  End Sub

  Sub Infinite_loop(ByRef tLocals_external As udtLocals)
    ' Written by Michael John Boyack Hodgson 2012.12.27. Revised 2021.04.17.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = "Infinite loop"
    tLocals.Status = udtuStandard_texts.Initiation
    If udtError.Infinite_loop_limit = 0 Then
      Exit Sub
    End If
    Stop ' Permanent.
    If Err.Number = 0 Or Err.Description = "" Or Err.Source = "" Then
      Stop ' Permanent.
    End If
    If tLocals_external.Error_number <> Err.Number Or tLocals_external.Error_description <> Err.Description Or tLocals_external.Error_source <> Err.Source Or _
       uError.Count = 0 Then
      Stop ' Permanent; fix calling sub.
      tLocals_external.Error_number = Err.Number
      tLocals_external.Error_description = Err.Description
      tLocals_external.Error_source = Err.Source
    End If
    uError.Last_error_sub = tLocals_external.Subr
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    If tLocals_external.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then
      uApp.Terminating = uApp.Terminating + 1
      tLocals.Msg = "Probable infinite loop detected. Terminating processing now." & vbCrLf & "Bummer." & Chr(13) & Chr(13)
      uError.Message = tLocals.Msg
      Call Error_write_mega(tLocals_external)
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Success = udtBoolean_extended.False_
        uApp.Terminating = uApp.Terminating + 1
      End If
      uOp.Success = False

      If uApp.App_name <> "" And uApp.Error_log_loc <> "" Then
        If uApp.Terminating < 5 Then
          tLocals.Action = "Report error"
          With uApp.Email_udt
            If Not .Initialised Then Call .Initialise()
            .Recipient_address = uLIS.Support_email
            .Subject = "Infinite loop detected in " & uApp.App_name & " on " & uMachine.Name & "."
            .Body = ""
            .Attachment_loc = uApp.Error_log_loc
          End With
          If Email_smtp_send_fn() Then
            uError.Reported = True
          Else
            Stop ' Permanent.
            tLocals.Error_count = tLocals.Error_count + 1
            tLocals.Success = udtBoolean_extended.False_
            uApp.Terminating = uApp.Terminating + 1
          End If
        End If
      End If

      tLocals.Action = "Speak"
      mObjSynth.Speak(tLocals.Msg)

      frmMain.Statuslabel1_set(tLocals.Msg)
      frmMain.Close()
      Call Error_clear(tLocals)
    End If

  End Sub

#End Region ' End region, Error handling.

#Region "Application maintenance"

  Sub Version_check(Optional ByVal tTemplate_path As String = "")
    ' Written 2014.06.18. Revised 2021.05.02.
    Dim tLocals As udtLocals = Nothing
    Dim tPass As Boolean = True ' Until failed.
    Dim i As Integer
    Dim tError_message As String = ""
    Dim tSubroutines As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Version_check
    If Not uMachine.Development_machine And Not uMachine.Testing_machine Then
      uOp.Success = True
      Exit Sub
    End If
    uOp.Subr = tLocals.Subr
    tLocals.Status = udtuStandard_texts.Initiation
    ' To do; standardise name:
    If tTemplate_path = "" Then tTemplate_path = uLIS.Development_path & "\Dept 11A-2, Application development\Template\Template"
    If Dir(uLIS.Development_path, FileAttribute.Directory) = "" Then
      Stop ' Permanent.
    End If
    uError.Message = ""
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + mModule.Count + 1
    tLocals.Status = ""

Main:
    tLocals.Action = "Check that Kernel is up to date"
    Call Version_check_sub_01()
    uOp.Subr = tLocals.Subr

    If uMachine.Development_machine And uApp.App_name <> udtApplication.App_Template Then
      tLocals.Action = "Check that each module is up-to-date compared to Template"
      Call Module_list_load()
      uOp.Subr = tLocals.Subr

      For i = 1 To mModule.Count
        uOp.Success = False
        tLocals.Msg = mModule.Item(i) & ".vb"
        Call Version_check_sub_02(tTemplate_path, tLocals.Msg, tPass, tSubroutines)
      Next
      uOp.Subr = tLocals.Subr
      tLocals.Msg = ""

      If tPass Then
        If uApp.App_name = udtApplication.App_Financier_Commercial_Edition Then
          If tTemplate_path = uLIS.Development_path & "\Dept 11A-2, Application development\Template\Template" Then
            tTemplate_path = "C:\LivingInfo Systems\Dept 11A-2, Application development\Financier\Financier\Financier"
            Call Version_check(tTemplate_path)
            uOp.Subr = tLocals.Subr
          End If
        End If
      Else
        tError_message = uApp.English_name & " is out of date."
        If uOp.Called_from = "" Then uOp.Called_from = tLocals.Subr
        Call Speak_it(uOp.Called_from, tError_message, 2)
        uOp.Subr = tLocals.Subr
        uOp.Success = False
        If InStr(tTemplate_path, "Template") > 0 Then
          If InStr(tSubroutines, "Declarations_common.vb") > 0 Or _
             InStr(tSubroutines, "Kernel_common.vb") > 0 Or _
             InStr(tSubroutines, "Database_common.vb") > 0 Or _
             InStr(tSubroutines, "Crypt_rc4.vb") > 0 Or _
             InStr(tSubroutines, "Email_Smtp.vb") > 0 Or _
             InStr(tSubroutines, "Web_common.vb") > 0 Or _
             InStr(tSubroutines, "clsEncrypt.vb") > 0 Or _
             InStr(tSubroutines, "frmDialogue.vb") > 0 Then
            If InStr(tSubroutines, ",") > 1 Then
              tError_message = tError_message & " Please run Template Kernel distribute to update modules " & tSubroutines & "."
            Else
              tError_message = tError_message & " Please run Template Kernel distribute to update module " & tSubroutines & "."
            End If
          Else
            If InStr(tSubroutines, ",") > 1 Then
              tError_message = tError_message & " Please use Winmerge to update modules " & tSubroutines & "."
            Else
              tError_message = tError_message & " Please use Winmerge to update module " & tSubroutines & "."
            End If
          End If
        ElseIf InStr(tTemplate_path, "Financier") > 0 Then
          tError_message = tError_message & " Please update " & uApp.App_name & " compared to " & udtApplication.App_Financier_Personal_Edition & ", for modules " & tSubroutines & "."
        Else
          Stop ' Permanent.
        End If

        uDialogue.Called_from = uOp.Subr
        Call Msgbox_LIS_top(tError_message, udtAlert_types.Critical, udtDialogue_types.Okay_only, tLocals.Subr)
        uOp.Subr = tLocals.Subr
        ' If uDialogue.Active Then GoTo Ender_full
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Status = udtuStandard_texts.Failed
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False

      End If ' If tPass Then
    End If

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    Call frmMain.ToolStripProgressBar1_set()

    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Success = udtBoolean_extended.True_
    End If

    uError.Message = tError_message
    Call Log_write(tLocals, 0)
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub Version_check_after_dialogue_tab()
    ' Revised 2021.03.03.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.Version_check_after_dialogue_tab
    uOp.Called_from = tLocals.Subr

    tLocals.Action = udtuStandard_texts.Main_start
    tLocals.Success = udtBoolean_extended.True_
    Call Log_write(tLocals, 0)

    frmMain.Close()

  End Sub

  Private Sub Version_check_sub_01()
    ' Separated 2015.02.03. Revised 2020.03.04.
    Dim tuApp_revision_original As Date = uApp.App_revision_date
    Dim tChanged As Boolean = False
    uOp.Subr = udtApplication_subroutine_names.Version_check_sub_01

    If cDeclarations_common_version > uApp.App_revision_date Then uApp.App_revision_date = cDeclarations_common_version : tChanged = True
    If cKernel_common_version > uApp.App_revision_date Then uApp.App_revision_date = cKernel_common_version : tChanged = True
    If cDatabase_common_version > uApp.App_revision_date Then uApp.App_revision_date = cDatabase_common_version : tChanged = True
    If cCrypt_rc4_version > uApp.App_revision_date Then uApp.App_revision_date = cCrypt_rc4_version : tChanged = True
    If cEmail_smtp_version > uApp.App_revision_date Then uApp.App_revision_date = cEmail_smtp_version : tChanged = True
    If cFtp_common_version > uApp.App_revision_date Then uApp.App_revision_date = cFtp_common_version : tChanged = True

    ' If any of these trip, check the corresponding WinMerge file:
    If cDeclarations_version > uApp.App_revision_date Then uApp.App_revision_date = cDeclarations_version : tChanged = True
    If cKernel_version > uApp.App_revision_date Then uApp.App_revision_date = cKernel_version : tChanged = True
    If cDatabase_version > uApp.App_revision_date Then uApp.App_revision_date = cDatabase_version : tChanged = True
    If clsEncrypt.cclsEncrypt_version > uApp.App_revision_date Then uApp.App_revision_date = clsEncrypt.cclsEncrypt_version : tChanged = True
    If frmMain.cfrmMain_version > uApp.App_revision_date Then uApp.App_revision_date = frmMain.cfrmMain_version : tChanged = True
    If frmDialogue.cfrmDialogue_version > uApp.App_revision_date Then uApp.App_revision_date = frmDialogue.cfrmDialogue_version : tChanged = True
    Call Version_check_sub_01_local(tChanged)
    uOp.Subr = udtApplication_subroutine_names.Version_check_sub_01

    If tChanged Then uApp.Upgraded = True

    If tuApp_revision_original = uApp.App_revision_date Then
      Call Extract_transform_load_Kernel_vb(False) ' 2018.11.04. Don't update Kernel.vb; delete Kernel.tmp.
    Else
      Call Extract_transform_load_Kernel_vb(True) ' 2018.09.02: Turned back on. Update Kernel.vb.
      If uApp.App_name = udtApplication.App_Template Then Stop ' To do, to re-test:
    End If

    uOp.Subr = ""
  End Sub

  Private Sub Version_check_sub_02(ByRef tTemplate_path As String, ByRef tFile_name As String, ByRef tPass As Boolean, ByRef tSubroutines As String)
    ' Written 2014.06.18. Revised 2021.04.14.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile = Nothing

    Dim tThis_version As Date = udtuStandard_texts.Date_null
    Dim tTarget_date As Date = udtuStandard_texts.Date_null
    tLocals.Subr = udtApplication_subroutine_names.Version_check_sub_02

    With tFile
      .Free_file = FreeFile()
      tLocals.Action = "Check " & tFile_name

      .Open_file = tTemplate_path & "\" & tFile_name
      FileOpen(.Free_file, .Open_file, OpenMode.Input)
      Do
        Input(.Free_file, .Text_read_line)
      Loop Until InStr(.Text_read_line, "Public Const") > 0 And InStr(.Text_read_line, "version") > 0 Or EOF(.Free_file)
      FileClose(.Free_file)
      .Open_file = ""
      .Text_read_line = Mid(.Text_read_line, InStr(.Text_read_line, "#") + 1)
      .Text_read_line = "#" & Left(.Text_read_line, InStr(.Text_read_line, "#"))

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = Date.ParseExact(.Text_read_line, "#M/d/yyyy#", CultureInfo.InvariantCulture)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = Date.ParseExact(.Text_read_line, "#M/dd/yyyy#", CultureInfo.InvariantCulture)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = Date.ParseExact(.Text_read_line, "#MM/d/yyyy#", CultureInfo.InvariantCulture)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = Date.ParseExact(.Text_read_line, "#MM/dd/yyyy#", CultureInfo.InvariantCulture)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = DateAndTime.DateValue(.Text_read_line)
        Catch ex As Exception
          tLocals.Error_description = ex.Message
          ' Conversion from string "#3/28/2021#" to type 'Date' is not valid.
          tLocals.Msg = IsDate(.Text_read_line)
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Try
          tTarget_date = CDate(Replace(.Text_read_line, "#", ""))
        Catch ex As Exception
          tLocals.Error_description = ex.Message
        End Try
      End If

      If tTarget_date = udtuStandard_texts.Date_null Then
        Stop ' Indefinite.
      End If

      Select Case tFile_name
        Case "Declarations_common.vb"
          tThis_version = cDeclarations_common_version
        Case "Kernel_common.vb"
          tThis_version = cKernel_common_version
        Case "Database_common.vb"
          tThis_version = cDatabase_common_version
        Case "Crypt_rc4.vb"
          tThis_version = cCrypt_rc4_version
        Case "Email_smtp.vb", "Email_Smtp.vb"
          tThis_version = cEmail_smtp_version
        Case "Web_common.vb"
          tThis_version = cWeb_common_version
        Case "Declarations.vb"
          tThis_version = cDeclarations_version
        Case "Kernel.vb"
          tThis_version = cKernel_version
        Case "Database_.vb"
          tThis_version = cDatabase_version
        Case "clsEncrypt.vb"
          tThis_version = clsEncrypt.cclsEncrypt_version
        Case "frmMain.vb"
          tThis_version = frmMain.cfrmMain_version
        Case "frmDialogue.vb"
          tThis_version = frmDialogue.cfrmDialogue_version
        Case Else
          Stop
          tLocals.Status = udtuStandard_texts.Failed
      End Select
    End With
    If tTarget_date > tThis_version Then
      tPass = False
      If tSubroutines = "" Then
        tSubroutines = tFile_name
      Else
        tSubroutines = tSubroutines & ", " & tFile_name
      End If
    End If

    Call frmMain.ToolStripProgressBar1_set()
    uOp.Subr = ""
  End Sub

  Sub Module_list_load()
    ' Written by Michael John Boyack Hodgson, 2015.01.25. Revised 2020.06.08.
    If mModule.Count < 12 Then
      mModule.Clear()
      mModule.Add(GetType(Declarations_common).Name)
      mModule.Add(GetType(Kernel_common).Name)
      mModule.Add(GetType(Database_common).Name)
      mModule.Add(GetType(Crypt_rc4).Name)
      mModule.Add(GetType(Email_Smtp).Name)
      mModule.Add(GetType(Web_common).Name)

      mModule.Add(GetType(Declarations).Name)
      mModule.Add(GetType(Kernel).Name)
      mModule.Add(GetType(Database_).Name)
      mModule.Add(GetType(clsEncrypt).Name)
      mModule.Add(frmMain.Name)
      mModule.Add(frmDialogue.Name)
    End If
    Call frmMain.Statuslabel1_set()
  End Sub

  Sub Extract_transform_load_Kernel_vb(ByVal tUpgrade As Boolean)
    ' Separated 2019.05.09. Revised 2021.04.08.
    Dim tLocals As udtLocals = Nothing
    Dim tFile As udtFile

    Dim tApplication_development_folder As String = uLIS.Development_path & "\Dept 11A-2, Application development"
    Dim tKernel_tmp As String = ""
    tLocals.Subr = udtApplication_subroutine_names.Extract_transform_load_Kernel_vb
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    On Error GoTo Error_captured
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    Select Case uApp.App_name
      Case udtApplication.App_Template
        tFile.File_full_name = tApplication_development_folder & "\Template\Template\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\Template\Template\Kernel.tmp"
      Case udtApplication.App_Contacts_merge
        tFile.File_full_name = tApplication_development_folder & "\Contacts merge\Contacts_merge\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\Contacts merge\Contacts_merge\Kernel.tmp"
      Case udtApplication.App_Financier_Personal_Edition
        tFile.File_full_name = tApplication_development_folder & "\Financier\Financier\Financier\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\Financier\Financier\Financier\Kernel.tmp"
      Case udtApplication.App_LivingInfo_Systems_Server
        tFile.File_full_name = tApplication_development_folder & "\LIS server\LIS server\LIS server\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\LIS server\LIS server\LIS server\Kernel.tmp"
      Case udtApplication.App_Seer
        tFile.File_full_name = tApplication_development_folder & "\Seer\Windows application\Seer\Seer\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\Seer\Windows application\Seer\Seer\Kernel.tmp"
      Case udtApplication.App_Financier_Commercial_Edition
        tFile.File_full_name = tApplication_development_folder & "\Financier Commercial Edition\Financier Commercial Edition\Financier Commercial Edition\Kernel.vb"
        tKernel_tmp = tApplication_development_folder & "\Financier Commercial Edition\Financier Commercial Edition\Financier Commercial Edition\Kernel.tmp"
      Case Else
        Stop ' Permanent.
        tLocals.Status = udtuStandard_texts.Failed
    End Select

    If tUpgrade Then
      If tFile.File_full_name <> "" And tKernel_tmp <> "" Then
        ' Modify code of Kernel.vb, to update uApp.App_revision.
        tFile.Open_file = tFile.File_full_name
        FileOpen(1, tFile.File_full_name, OpenMode.Input)
        FileOpen(2, tKernel_tmp, OpenMode.Output)
        Do
          tFile.Text_read_line = LineInput(1)
          If InStr(tFile.Text_read_line, ".App_revision_date = ") = 0 Then PrintLine(2, tFile.Text_read_line)
        Loop Until InStr(tFile.Text_read_line, ".App_revision_date = ") > 0 Or EOF(1)
        tLocals.Msg = Left(tFile.Text_read_line, InStr(tFile.Text_read_line, "#")) & Replace(Format(uApp.App_revision_date, "MM/dd/yyyy"), ".", "/") & "# ' M/d/y." & vbCrLf
        Print(2, tLocals.Msg)
        Do
          tFile.Text_read_line = LineInput(1)
          PrintLine(2, tFile.Text_read_line)
        Loop Until EOF(1)
        FileClose(1, 2)
        tFile.Open_file = ""

        FileCopy(tKernel_tmp, tFile.File_full_name)
        Kill(tKernel_tmp)
      End If
    Else
      If tKernel_tmp <> "" Then
        If Dir(tKernel_tmp, FileAttribute.Normal) <> "" Then Kill(tKernel_tmp)
      End If
    End If
    tLocals.Success = udtBoolean_extended.True_

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Then
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 53, 55, 75, 76
            Call Error_handler_for_text_file(tLocals)
            If tLocals.Action = "Resume" Then tLocals.Action = "" : Resume
          Case Else
            Stop ' Permanent.
            Call Error_display(tLocals)
        End Select
      End If
    End If

Ender_full:
    If tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then tLocals.Status = udtuStandard_texts.Completion
    Call Log_write(tLocals, 0)

    tFile.File_full_name = ""
    If uOp.Called_from = tLocals.Subr Then uOp.Called_from = ""
    Call frmMain.Statuslabel1_set()
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

  Sub App_revision_upgrade(ByRef tApp_revision_last As Date, ByVal tDb As udtDatabase_oledb, ByRef tUpgraded As Boolean)
    ' Revised 2021.03.14.
    Dim tLocals As udtLocals = Nothing
    tLocals.Subr = udtApplication_subroutine_names.App_revision_upgrade
    tLocals.Error_count = 0
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 7
    Call frmMain.Statuslabel1_set()
    tLocals.Status = ""

    tLocals.Action = "Check versions"
    If Not uApp.Environment_dump_emailed Then
      uOp.Called_from = tLocals.Subr
      Call Environment_dump(True)
      uOp.Subr = tLocals.Subr
      ' No error check, because emailing the file can fail.
      uOp.Success = False
    End If

    Select Case uApp.App_build_number
      Case Is > 1
        ' Stop ' Indefinite.
        ' ...
        ' If do something then tUpgraded = True
    End Select

    If tApp_revision_last < uApp.App_revision_date Then
      tLocals.Action = "Upgrade data"

      If tApp_revision_last <= #1/3/2012# Then ' M/d/y.
        tLocals.Action = "Set uUser.Tab_last = frmMain.tabHome.Name"
        If uUser.Tab_last = "tabMain" Then
          uUser.Tab_last = frmMain.tabHome.Name
          tUpgraded = True
        End If
      End If

      If tApp_revision_last < #10/14/2018# Then
        tLocals.Action = "Call " & udtApplication_subroutine_names.Data_fix_currency_sign_to_currency_symbol
        Call Data_fix_currency_sign_to_currency_symbol(tUpgraded)
        uOp.Subr = tLocals.Subr
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        If uDialogue.Active Then GoTo Ender_full
      End If ' If tApp_revision_last < #10/14/2018# Then

      If tApp_revision_last < #10/15/2019# Then
        tLocals.Action = "Call " & udtApplication_subroutine_names.Data_fix_currency_to_currency_
        Call Data_fix_currency_to_currency_(tUpgraded)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        If uDialogue.Active Then GoTo Ender_full
      End If

      If tApp_revision_last < #1/14/2021# Then
        tLocals.Action = "Call Data fix Log.Status"
        Call Data_fix_status(tUpgraded)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        If uDialogue.Active Then GoTo Ender_full
      End If

      If tApp_revision_last < #3/2/2021# Then
        tLocals.Action = "Call Data fix date to date_"
        Call Data_fix_date_to_date_(tUpgraded)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        If uDialogue.Active Then GoTo Ender_full

        tLocals.Action = "Call Data fix time to time_"
        Call Data_fix_time_to_time_(tUpgraded)
        If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
          Stop ' Permanent.
          tLocals.Error_count = tLocals.Error_count + 1
          tLocals.Success = udtBoolean_extended.False_
          GoTo Error_process
        End If
        uOp.Success = False
        If uDialogue.Active Then GoTo Ender_full

      End If

    End If ' If tApp_revision_last < uApp.App_revision Then

    If tDb.Upgrade_table = "All" Then
      tLocals.Action = "Call " & udtApplication_subroutine_names.Diagnostic_all_tables
      Call Diagnostic_all_tables()
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
      uApp.Memory_check_completed_date = Date.Today
      uUser.Preferences_changed = True

    ElseIf tDb.Upgrade_table <> "" Then
      Stop ' Indefinite.
      tLocals.Action = "Call " & udtApplication_subroutine_names.Database_diagnostic_level_05
      Call Database_diagnostic_level_05(tDb)
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      Else
        tUpgraded = True
      End If
      uOp.Success = False

    End If

    tLocals.Action = "Call App_revision upgrade local"
    Call App_revision_upgrade_local(tApp_revision_last, tDb, tUpgraded)
    uOp.Subr = tLocals.Subr
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False
    If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_

    If tUpgraded Then
      uApp.Upgraded = True
      tLocals.Action = "Updated to build " & uApp.App_build_number & ", revision " & uApp.App_revision_date & "."
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
    End If

    If uUser.Preferences_changed And tDb.Upgrade_table = "" Then
      tLocals.Action = "Call " & udtApplication_subroutine_names.Preferences_write
      If uOp.Preferences_loaded Then
        Call Preferences_write()
      Else
        uUser.Preferences_changed = True
        Call Preferences_write_one("App_build", uApp.App_build_number)
        uUser.Preferences_changed = True
        Call Preferences_write_one("App_revision", uApp.App_revision_date)
      End If
      uOp.Subr = tLocals.Subr
    End If

Error_process:
    If tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or _
      uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_

      uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Ender_full:
    If Not uDialogue.Active Then
      If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
        tLocals.Success = udtBoolean_extended.False_
      Else
        tLocals.Status = udtuStandard_texts.Completion
      End If
      If uOp.Event_ <> "" And tLocals.Error_count = 0 And uError.Number = 0 And uApp.Terminating = 0 Then tLocals.Success = udtBoolean_extended.True_
      Call Log_write(tLocals, 0)
      uOp.Subr = tLocals.Subr
    End If

Ender_partial:
    uApp.Progressbar_steps_difference = uApp.Progressbar_steps_total - uApp.Progressbar_step
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False
  End Sub

#End Region ' End region, Application maintenance.

  Sub Sub_new()
    ' Revised 2021.04.28.
    Dim tLocals As udtLocals = Nothing : Stop ' To do: retest using tLocals.
    Dim tDb As udtDatabase_oledb = Nothing
    Dim tError As udtError = Nothing
    Dim tFile As udtFile = Nothing
    Dim tSearch As udtSearch = Nothing

    Dim tCalled_from As String
    ' dim any other variables.

    tLocals.Subr = "New sub-routine"
    tLocals.Subr = System.Reflection.MethodBase.GetCurrentMethod.Name
    If uOp.Event_ = "" Then uOp.Event_ = tLocals.Subr

    tLocals.Status = udtuStandard_texts.Initiation ' For Log_write.
    tLocals.Continue_ = True ' Until set false.
    tLocals.Msg = ""
    tLocals.Error_count = 0 ' Use if there is a chance of an external error.
    tLocals.Success = udtBoolean_extended.False_ ' Use if there is _no_ chance of an external error.
    Call Log_write(tLocals, 0)

    On Error GoTo Error_captured

    uApp.Progressbar_show = True
    uApp.Progressbar_steps_total = 10
    uApp.Progressbar_step = uApp.Progressbar_step + 1
    Call frmMain.Statuslabel1_set()
    tLocals.Status = ""

Main:
    tLocals.Action = udtuStandard_texts.Main_start
    tDb.Database = uMemory.Database ' For example.
    Call Database_instantiate(tDb)
    uOp.Subr = tLocals.Subr ' Needed only if this sub-routine can generate an error; especially one not captured in this sub-routine.
    If uDialogue.Active Then GoTo Ender_full
    If uOp.Event_ = "" Or Not uOp.Success Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      If Not tLocals.Continue_ Then GoTo Error_process ' Or:
      ' GoTo Ender_partial
    Else
      tLocals.Success = udtBoolean_extended.True_
      tLocals.Steps_completed = tLocals.Steps_completed + 1
    End If
    uOp.Result = ""
    uOp.Success = False ' If going to call other sub-routines; else:
    ' uOp.Success = False ' Carry the result forward.

Upgrade:
    tLocals.Action = "Upgrade"
    Select Case tLocals.Action
      Case udtuStandard_texts.Upgrade
    End Select

Error_process:
    If (tLocals.Continue_ = False And _
       (tLocals.Error_count <> 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_)) Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Then
      Stop ' Permanent.
      ' Reduce all error responses down to tLocals.Success = udtBoolean_extended.False_
      tLocals.Success = udtBoolean_extended.False_

      ' If Error_captured exists, the next line should be off, else on.
      ' uError.Count = uError.Count + 1
      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)
    End If

Error_captured:
    tLocals.Error_number = Err.Number
    tLocals.Error_description = Err.Description
    tLocals.Error_source = Err.Source
    If Err.Number <> 0 Or uApp.Terminating > 0 Or uError.Count <> 0 Or uError.Number <> 0 Then
      Stop ' Indefinite.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_

      If tLocals.Error_count >= udtError.Infinite_loop_limit Or uError.Count >= udtError.Infinite_loop_limit Then Call Infinite_loop(tLocals)

      If uApp.Terminating = 0 Then
        Select Case tLocals.Error_number
          Case 0
            tLocals.Error_count = tLocals.Error_count - 1
            Call Error_clear(tLocals)
            Resume ' Or:
            Resume Next
          Case Else
            If tDb.Cmd_str = "" Then
              Stop ' Permanent.
              Call Error_display(tLocals)
            Else
              Call Database_error_handler(tLocals, tDb)
              If tLocals.Action <> "" Or tDb.Upgrade_table <> "" Then
                If tDb.Conn.State = ConnectionState.Open Then tDb.Conn.Close()
                Call Database_diagnostic_00(tDb)
                tLocals.Action = ""
                tLocals.Msg = ""
                Resume
              End If
            End If ' If tDb.Cmd_str = "" Then
        End Select
      End If
    End If

Ender_full:
    ' Section revised 2021.02.16.
    tLocals.Action = "Ender full"

    ' Check all results and reduce down to tLocals.Success = udtBoolean_extended.False_:
    If tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Then
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
    Else
      tLocals.Status = udtuStandard_texts.Completion
    End If

    If tLocals.Action = "" And tLocals.Status <> udtuStandard_texts.Cancelled And tLocals.Status <> udtuStandard_texts.Failed Then
      tLocals.Action = tLocals.Status
    End If

    If tLocals.Error_count > 0 Or tLocals.Status = udtuStandard_texts.Cancelled Or tLocals.Status = udtuStandard_texts.Failed Or tLocals.Success = udtBoolean_extended.False_ Or _
       tLocals.Steps_completed < tLocals.Steps_to_complete Or _
       uOp.Event_ = "" Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      tLocals.Success = udtBoolean_extended.False_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check failed.")
    Else
      tLocals.Success = udtBoolean_extended.True_
      Call frmMain.txtDiagnostic_result_append(tLocals.Subr & " check completion." & vbCrLf)
    End If

    Call Log_write(tLocals, 0)
    If uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 0 Then
      Stop ' Permanent.
      tLocals.Status = udtuStandard_texts.Failed
    End If
    uOp.Success = False

Ender_partial:
    ' Section revised 2021.02.16.
    ' Used to skip usual Ender processing, and where an error handler is not needed:
    tLocals.Action = "Ender partial"
    uOp.Subr = tLocals.Subr
    If uOp.Called_from = tLocals.Subr Then uOp.Called_from = ""
    Call frmMain.Statuslabel1_set() ' No chirp.
    uOp.Subr = ""
    Call frmMain.Statuslabel1_set() ' Chirp.
    If tLocals.Success = udtBoolean_extended.True_ Then uOp.Success = True Else uOp.Success = False : Call frmMain.Statuslabel1_failed(tLocals.Subr)
    If uOp.Event_ = tLocals.Subr Then uOp.Event_ = ""
  End Sub

  Sub End_game()
    ' Revised 2021.03.24.
    Dim tLocals As udtLocals = Nothing
    Dim tCalled_from As String = uOp.Called_from
    tLocals.Subr = udtApplication_subroutine_names.End_game
    If uOp.Called_from = "" Then uOp.Called_from = tLocals.Subr
    tLocals.Status = udtuStandard_texts.Initiation
    Call Log_write(tLocals, 0)
    uOp.Subr = tLocals.Subr
    If tLocals.Error_count > 0 Or uOp.Event_ = "" Or Not uOp.Success Or uError.Number <> 0 Then
      Stop ' Permanent.
      tLocals.Error_count = tLocals.Error_count + 1
      tLocals.Status = udtuStandard_texts.Failed
      tLocals.Success = udtBoolean_extended.False_
      GoTo Error_process
    End If
    uOp.Success = False

    uApp.Shutdown = True
    uOp.Called_from = tCalled_from
    Call frmMain.ToolStripProgressBar1_set()

    uApp.End_game = True
    If uOp.Called_from = "frmMain_FormClosing" Or uOp.Called_from = udtApplication_subroutine_names.frmMain_FormClosed Then
    Else
      uApp.Progressbar_step = 0
      uApp.Progressbar_steps_total = uApp.Progressbar_steps_total + 11
    End If
    uApp.Progressbar_show = True
    Call frmMain.ToolStripProgressBar1_set()

    tLocals.Action = "End game local"
    Call End_game_local()
    If uDialogue.Active And uDialogue.Called_from <> udtApplication_subroutine_names.Version_check Then
      Exit Sub
    End If
    Call frmMain.ToolStripProgressBar1_set()

    If uDialogue.Called_from <> udtApplication_subroutine_names.Version_check Then
      tLocals.Action = "Preferences write"
      uUser.Preferences_changed = True
      Call Preferences_write()
      uOp.Subr = tLocals.Subr
      If uOp.Event_ = "" Or tLocals.Error_count > 0 Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 1 Then
        Stop ' Permanent.
        tLocals.Error_count = tLocals.Error_count + 1
        tLocals.Status = udtuStandard_texts.Failed
        tLocals.Success = udtBoolean_extended.False_
        GoTo Error_process
      End If
      uOp.Success = False
    End If
    Call frmMain.ToolStripProgressBar1_set()

    tLocals.Success = udtBoolean_extended.True_
    tLocals.Action = uApp.App_name & " session end"
    If uApp.Terminating <= 3 Then
      Call Log_write(tLocals, 0)
    End If
    uOp.Subr = tLocals.Subr
    Call frmMain.ToolStripProgressBar1_set()

    tLocals.Action = "Salutation"
    Select Case Left$(TimeString, 2)
      Case Is < 19
        tLocals.Msg = "Good bye, "
      Case Else
        tLocals.Msg = "Good night, "
    End Select
    tLocals.Msg = tLocals.Msg & uUser.First_name & "."
    Call frmMain.Statuslabel1_set(tLocals.Msg)
    If uApp.Terminating <= 3 Then Call Speak_it(tLocals.Subr, tLocals.Msg, 2)
    uOp.Success = False
    Call frmMain.ToolStripProgressBar1_set()

    uOp.Called_from = tLocals.Subr
    Application.Exit()
    If tCalled_from <> udtApplication_subroutine_names.frmMain_FormClosed Then
      On Error Resume Next
      End ' That's all, folks!
    End If

Error_process:
    If uOp.Event_ = "" Or tLocals.Error_count > 0 Or Not uOp.Success Or uError.Number <> 0 Or uApp.Terminating > 1 Then
      Stop ' Permanent.
      tLocals.Success = udtBoolean_extended.False_
    End If

  End Sub

End Module
